@@ -65,8 +65,6 @@ aes_ssse3_cipher_setkey(void *_ctx, const void *userkey, size_t keysize)
 	struct aes_ctx *ctx = _ctx;
 	int ret;
 
-	CHECK_AES_KEYSIZE(keysize);
-
 	if (ctx->enc)
 		ret =
 		    vpaes_set_encrypt_key(userkey, keysize * 8,
@@ -109,6 +107,9 @@ static int aes_setiv(void *_ctx, const void *iv, size_t iv_size)
 {
 	struct aes_ctx *ctx = _ctx;
 
+	if (iv_size != 16)
+		return gnutls_assert_val(GNUTLS_E_INVALID_REQUEST);
+
 	memcpy(ctx->iv, iv, 16);
 	return 0;
 }
