@@ -1,5 +1,6 @@
 /*
  * Copyright (C) 2000-2012 Free Software Foundation, Inc.
+ * Copyright (C) 2017 Red Hat, Inc.
  *
  * Author: Nikos Mavrogiannopoulos
  *
@@ -31,13 +32,13 @@
 #define GNUTLS_FALLBACK_SCSV_MAJOR 0x56
 #define GNUTLS_FALLBACK_SCSV_MINOR 0x00
 
-/* would allow for 256 ciphersuites */
-#define MAX_CIPHERSUITE_SIZE 512
+#define IS_EC(x) (((x)==GNUTLS_PK_ECDSA)||((x)==GNUTLS_PK_ECDH_X25519)||((x)==GNUTLS_PK_EDDSA_ED25519))
 
-#define IS_EC(x) (((x)==GNUTLS_PK_ECDSA)||((x)==GNUTLS_PK_ECDHX))
+#define TLS_SIGN_AID_UNKNOWN {{255, 255}}
+#define HAVE_UNKNOWN_SIGAID(aid) ((aid)->id[0] == 255 && (aid)->id[1] == 255)
 
 /* Functions for version handling. */
-const version_entry_st *version_to_entry(gnutls_protocol_t c);
+const version_entry_st *version_to_entry(gnutls_protocol_t version);
 const version_entry_st *_gnutls_version_lowest(gnutls_session_t session);
 gnutls_protocol_t _gnutls_version_max(gnutls_session_t session);
 int _gnutls_version_priority(gnutls_session_t session,
@@ -48,6 +49,11 @@ gnutls_protocol_t _gnutls_version_get(uint8_t major, uint8_t minor);
 unsigned _gnutls_version_is_too_high(gnutls_session_t session, uint8_t major, uint8_t minor);
 
 /* Functions for feature checks */
+int
+_gnutls_figure_common_ciphersuite(gnutls_session_t session,
+				  const ciphersuite_list_st *peer_clist,
+				  const gnutls_cipher_suite_entry_st **ce);
+
 inline static int
 _gnutls_version_has_selectable_prf(const version_entry_st * ver)
 {
@@ -147,34 +153,32 @@ inline static int _gnutls_mac_get_key_size(const mac_entry_st * e)
 #define _gnutls_digest_get_name _gnutls_mac_get_name
 #define _gnutls_hash_get_algo_len _gnutls_mac_get_algo_len
 
+/* Security against pre-image attacks */
 inline static int _gnutls_digest_is_secure(const mac_entry_st * e)
 {
 	if (unlikely(e == NULL))
 		return 0;
 	else
-		return e->secure;
+		return (e->preimage_insecure==0);
 }
 
 /* Functions for cipher suites. */
+int _gnutls_get_client_ciphersuites(gnutls_session_t session,
+	gnutls_buffer_st * cdata, const version_entry_st *minver,
+	const version_entry_st *maxver, unsigned add_scsv);
+
 int _gnutls_supported_ciphersuites(gnutls_session_t session,
 				   uint8_t * cipher_suites,
 				   unsigned int max_cipher_suite_size);
-int
-_gnutls_remove_unwanted_ciphersuites(gnutls_session_t session,
-			     uint8_t * cipher_suites,
-			     int cipher_suites_size,
-			     gnutls_pk_algorithm_t * pk_algos,
-			     size_t pk_algos_size);
+
+const gnutls_cipher_suite_entry_st
+    *cipher_suite_get(gnutls_kx_algorithm_t kx_algorithm,
+		      gnutls_cipher_algorithm_t cipher_algorithm,
+		      gnutls_mac_algorithm_t mac_algorithm);
 
 const char *_gnutls_cipher_suite_get_name(const uint8_t suite[2]);
-gnutls_mac_algorithm_t _gnutls_cipher_suite_get_prf(const uint8_t
-						    suite[2]);
-const cipher_entry_st *_gnutls_cipher_suite_get_cipher_algo(const uint8_t
-							    suite[2]);
 gnutls_kx_algorithm_t _gnutls_cipher_suite_get_kx_algo(const uint8_t
 						       suite[2]);
-const mac_entry_st *_gnutls_cipher_suite_get_mac_algo(const uint8_t
-						      suite[2]);
 
 int
 _gnutls_cipher_suite_get_id(gnutls_kx_algorithm_t kx_algorithm,
@@ -191,7 +195,7 @@ const cipher_entry_st *cipher_name_to_entry(const char *name);
 inline static cipher_type_t _gnutls_cipher_type(const cipher_entry_st * e)
 {
 	if (unlikely(e == NULL))
-		return 0;
+		return CIPHER_AEAD; /* doesn't matter */
 	return e->type;
 }
 
@@ -270,31 +274,56 @@ inline static int _gnutls_cipher_get_tag_size(const cipher_entry_st * e)
 /* Functions for key exchange. */
 bool _gnutls_kx_needs_dh_params(gnutls_kx_algorithm_t algorithm);
 bool _gnutls_kx_allows_false_start(gnutls_session_t session);
-int _gnutls_kx_cert_pk_params(gnutls_kx_algorithm_t algorithm);
 mod_auth_st *_gnutls_kx_auth_struct(gnutls_kx_algorithm_t algorithm);
 int _gnutls_kx_is_ok(gnutls_kx_algorithm_t algorithm);
 
 int _gnutls_kx_get_id(const char *name);
 
-/* Type to KX mappings. */
-gnutls_kx_algorithm_t _gnutls_map_kx_get_kx(gnutls_credentials_type_t type,
-					    int server);
 gnutls_credentials_type_t _gnutls_map_kx_get_cred(gnutls_kx_algorithm_t
 						  algorithm, int server);
 
 /* KX to PK mapping. */
+unsigned
+_gnutls_kx_supports_pk(gnutls_kx_algorithm_t kx_algorithm,
+		       gnutls_pk_algorithm_t pk_algorithm);
 
-/* DSA + RSA + ECC */
-#define GNUTLS_DISTINCT_PK_ALGORITHMS 3
-gnutls_pk_algorithm_t _gnutls_map_pk_get_pk(gnutls_kx_algorithm_t
-					    kx_algorithm);
+unsigned
+_gnutls_kx_supports_pk_usage(gnutls_kx_algorithm_t kx_algorithm,
+		       gnutls_pk_algorithm_t pk_algorithm,
+		       unsigned key_usage);
 
 enum encipher_type { CIPHER_ENCRYPT = 0, CIPHER_SIGN = 1, CIPHER_IGN };
 
 enum encipher_type _gnutls_kx_encipher_type(gnutls_kx_algorithm_t
-					    algorithm);
+					    kx_algorithm);
 
 /* Functions for sign algorithms. */
+
+typedef enum hash_security_level_t {
+	_SECURE,
+	_INSECURE_FOR_CERTS,
+	_INSECURE
+} hash_security_level_t;
+
+struct gnutls_sign_entry_st {
+	const char *name;
+	const char *oid;
+	gnutls_sign_algorithm_t id;
+	gnutls_pk_algorithm_t pk;
+	gnutls_digest_algorithm_t hash;
+	/* See RFC 5246 HashAlgorithm and SignatureAlgorithm
+	   for values to use in aid struct. */
+	const sign_algorithm_st aid;
+	hash_security_level_t slevel;	/* contains values of hash_security_level_t */
+};
+typedef struct gnutls_sign_entry_st gnutls_sign_entry_st;
+
+const gnutls_sign_entry_st *_gnutls_sign_to_entry(gnutls_sign_algorithm_t sign);
+const gnutls_sign_entry_st *_gnutls_pk_to_sign_entry(gnutls_pk_algorithm_t pk, gnutls_digest_algorithm_t hash);
+const gnutls_sign_entry_st *_gnutls_oid_to_sign_entry(const char *oid);
+
+bool _gnutls_sign_is_secure2(const gnutls_sign_entry_st *se, unsigned int flags);
+
 gnutls_pk_algorithm_t _gnutls_x509_sign_to_pk(gnutls_sign_algorithm_t
 					      sign);
 const char *_gnutls_x509_sign_to_oid(gnutls_pk_algorithm_t,
@@ -304,34 +333,50 @@ gnutls_sign_algorithm_t _gnutls_tls_aid_to_sign(const sign_algorithm_st *
 const sign_algorithm_st *_gnutls_sign_to_tls_aid(gnutls_sign_algorithm_t
 						 sign);
 
-int _gnutls_mac_priority(gnutls_session_t session,
-			 gnutls_mac_algorithm_t algorithm);
-int _gnutls_cipher_priority(gnutls_session_t session,
-			    gnutls_cipher_algorithm_t a);
-int _gnutls_kx_priority(gnutls_session_t session,
-			gnutls_kx_algorithm_t algorithm);
-
 unsigned int _gnutls_pk_bits_to_subgroup_bits(unsigned int pk_bits);
+gnutls_digest_algorithm_t _gnutls_pk_bits_to_sha_hash(unsigned int pk_bits);
+
+gnutls_digest_algorithm_t _gnutls_hash_size_to_sha_hash(unsigned int size);
+
+bool _gnutls_pk_is_not_prehashed(gnutls_pk_algorithm_t algorithm);
 
 /* ECC */
-struct gnutls_ecc_curve_entry_st {
+typedef struct gnutls_ecc_curve_entry_st {
 	const char *name;
 	const char *oid;
 	gnutls_ecc_curve_t id;
 	gnutls_pk_algorithm_t pk;
-	int tls_id;		/* The RFC4492 namedCurve ID */
-	int size;		/* the size in bytes */
-};
-typedef struct gnutls_ecc_curve_entry_st gnutls_ecc_curve_entry_st;
+	unsigned size;		/* the size in bytes */
+	unsigned sig_size;	/* the size of curve signatures in bytes (EdDSA) */
+} gnutls_ecc_curve_entry_st;
 
 const gnutls_ecc_curve_entry_st
     *_gnutls_ecc_curve_get_params(gnutls_ecc_curve_t curve);
-gnutls_ecc_curve_t gnutls_ecc_curve_get_id(const char *name);
-int _gnutls_tls_id_to_ecc_curve(int num);
-int _gnutls_ecc_curve_get_tls_id(gnutls_ecc_curve_t supported_ecc);
-gnutls_ecc_curve_t _gnutls_ecc_bits_to_curve(int bits);
+
+const gnutls_group_entry_st *_gnutls_tls_id_to_group(unsigned num);
+const gnutls_group_entry_st * _gnutls_id_to_group(unsigned id);
+
+gnutls_ecc_curve_t _gnutls_ecc_bits_to_curve(gnutls_pk_algorithm_t pk, int bits);
 #define MAX_ECC_CURVE_SIZE 66
 
+gnutls_pk_algorithm_t _gnutls_oid_to_pk_and_curve(const char *oid, gnutls_ecc_curve_t *curve);
+
+inline static int _curve_is_eddsa(const gnutls_ecc_curve_entry_st * e)
+{
+	size_t ret = 0;
+	if (unlikely(e == NULL))
+		return ret;
+	if (e->pk == GNUTLS_PK_EDDSA_ED25519)
+		return 1;
+	return 0;
+}
+
+inline static int curve_is_eddsa(gnutls_ecc_curve_t id)
+{
+	const gnutls_ecc_curve_entry_st *e = _gnutls_ecc_curve_get_params(id);
+	return _curve_is_eddsa(e);
+}
+
 static inline int _gnutls_kx_is_ecc(gnutls_kx_algorithm_t kx)
 {
 	if (kx == GNUTLS_KX_ECDHE_RSA || kx == GNUTLS_KX_ECDHE_ECDSA ||
@@ -341,6 +386,15 @@ static inline int _gnutls_kx_is_ecc(gnutls_kx_algorithm_t kx)
 	return 0;
 }
 
+static inline int _gnutls_kx_is_dhe(gnutls_kx_algorithm_t kx)
+{
+	if (kx == GNUTLS_KX_DHE_RSA || kx == GNUTLS_KX_DHE_DSS ||
+	    kx == GNUTLS_KX_ANON_DH || kx == GNUTLS_KX_DHE_PSK)
+		return 1;
+
+	return 0;
+}
+
 static inline int _sig_is_ecdsa(gnutls_sign_algorithm_t sig)
 {
 	if (sig == GNUTLS_SIGN_ECDSA_SHA1 || sig == GNUTLS_SIGN_ECDSA_SHA224 ||
@@ -351,4 +405,6 @@ static inline int _sig_is_ecdsa(gnutls_sign_algorithm_t sig)
 	return 0;
 }
 
+bool _gnutls_pk_are_compat(gnutls_pk_algorithm_t pk1, gnutls_pk_algorithm_t pk2);
+
 #endif
