@@ -1,5 +1,6 @@
 /*
  * Copyright (C) 2001-2012 Free Software Foundation, Inc.
+ * Copyright (C) 2017 Red Hat, Inc.
  *
  * Author: Nikos Mavrogiannopoulos
  *
@@ -43,18 +44,9 @@
 #include <x509.h>
 #include <x509/verify-high.h>
 #include <gnutls/abstract.h>
+#include "abstract_int.h"
 #include "debug.h"
 
-#ifdef ENABLE_OPENPGP
-#include "openpgp/openpgp.h"
-
-static gnutls_privkey_t alloc_and_load_pgp_key(const
-					       gnutls_openpgp_privkey_t
-					       key, int deinit);
-static gnutls_pcert_st *alloc_and_load_pgp_certs(gnutls_openpgp_crt_t cert);
-
-#endif
-
 static gnutls_pcert_st *alloc_and_load_x509_certs(gnutls_x509_crt_t *
 						  certs, unsigned);
 static gnutls_privkey_t alloc_and_load_x509_key(gnutls_x509_privkey_t key,
@@ -122,14 +114,6 @@ static int copy_certificate_auth_info(cert_auth_info_t info, gnutls_pcert_st * c
 	info->ncerts = ncerts;
 	info->cert_type = certs[0].type;
 
-#ifdef ENABLE_OPENPGP
-	if (certs[0].type == GNUTLS_CRT_OPENPGP) {
-		if (keyid)
-			memcpy(info->subkey_id, keyid,
-			       GNUTLS_OPENPGP_KEYID_SIZE);
-	}
-#endif
-
 	return 0;
 
  clear:
@@ -218,7 +202,7 @@ find_x509_client_cert(gnutls_session_t session,
 	       int pk_algos_length, int *indx)
 {
 	unsigned size;
-	gnutls_datum_t odn = { NULL, 0 };
+	gnutls_datum_t odn = { NULL, 0 }, asked_dn;
 	uint8_t *data = _data;
 	ssize_t data_size = _data_size;
 	unsigned i, j;
@@ -241,6 +225,10 @@ find_x509_client_cert(gnutls_session_t session,
 		DECR_LENGTH_RET(data_size, size, 0);
 		data += 2;
 
+		asked_dn.data = data;
+		asked_dn.size = size;
+		_gnutls_dn_log("Peer requested CA", &asked_dn);
+
 		for (i = 0; i < cred->ncerts; i++) {
 			for (j = 0; j < cred->certs[i].cert_list_length; j++) {
 				if ((result =
@@ -251,7 +239,7 @@ find_x509_client_cert(gnutls_session_t session,
 					return result;
 				}
 
-				if (odn.size == 0 || odn.size != size)
+				if (odn.size == 0 || odn.size != asked_dn.size)
 					continue;
 
 				/* If the DN matches and
@@ -264,7 +252,7 @@ find_x509_client_cert(gnutls_session_t session,
 								   [0].pubkey,
 								   NULL);
 
-				if ((memcmp(odn.data, data, size) == 0) &&
+				if ((memcmp(odn.data, asked_dn.data, asked_dn.size) == 0) &&
 				    (check_pk_algo_in_list
 				     (pk_algos, pk_algos_length,
 				      cert_pk) == 0)) {
@@ -288,44 +276,6 @@ find_x509_client_cert(gnutls_session_t session,
 
 }
 
-#ifdef ENABLE_OPENPGP
-/* Locates the most appropriate openpgp cert
- */
-static int
-find_openpgp_cert(const gnutls_certificate_credentials_t cred,
-		  gnutls_pk_algorithm_t * pk_algos,
-		  int pk_algos_length, int *indx)
-{
-	unsigned i, j;
-
-	*indx = -1;
-
-	for (i = 0; i < cred->ncerts; i++) {
-		for (j = 0; j < cred->certs[i].cert_list_length; j++) {
-
-			/* If the *_SIGN algorithm matches
-			 * the cert is our cert!
-			 */
-			if ((check_pk_algo_in_list
-			     (pk_algos, pk_algos_length,
-			      gnutls_pubkey_get_pk_algorithm(cred->certs
-							     [i].cert_list
-							     [0].pubkey,
-							     NULL)) == 0)
-			    && (cred->certs[i].cert_list[0].type ==
-				GNUTLS_CRT_OPENPGP)) {
-				*indx = i;
-				break;
-			}
-		}
-		if (*indx != -1)
-			break;
-	}
-
-	return 0;
-}
-#endif
-
 /* Returns the number of issuers in the server's
  * certificate request packet.
  */
@@ -406,6 +356,8 @@ get_issuers(gnutls_session_t session,
 			issuers_dn[i].data = data;
 			issuers_dn[i].size = size;
 
+			_gnutls_dn_log("Peer requested CA", &issuers_dn[i]);
+
 			data += size;
 		}
 	}
@@ -492,19 +444,8 @@ call_get_cert_callback(gnutls_session_t session,
 		local_certs =
 		    alloc_and_load_x509_certs(st2.cert.x509, st2.ncerts);
 	} else {		/* PGP */
-		if (st2.ncerts > 1) {
-			gnutls_assert();
-			ret = GNUTLS_E_INVALID_REQUEST;
-			goto cleanup;
-		}
-#ifdef ENABLE_OPENPGP
-		{
-			local_certs = alloc_and_load_pgp_certs(st2.cert.pgp);
-		}
-#else
 		ret = GNUTLS_E_UNIMPLEMENTED_FEATURE;
 		goto cleanup;
-#endif
 	}
 
 	if (local_certs == NULL) {
@@ -514,19 +455,6 @@ call_get_cert_callback(gnutls_session_t session,
 	}
 
 	switch (st2.key_type) {
-	case GNUTLS_PRIVKEY_OPENPGP:
-#ifdef ENABLE_OPENPGP
-		if (st2.key.pgp != NULL) {
-			local_key =
-			    alloc_and_load_pgp_key(st2.key.pgp, st2.deinit_all);
-			if (local_key == NULL) {
-				gnutls_assert();
-				ret = GNUTLS_E_INTERNAL_ERROR;
-				goto cleanup;
-			}
-		}
-#endif
-		break;
 	case GNUTLS_PRIVKEY_PKCS11:
 #ifdef ENABLE_PKCS11
 		if (st2.key.pkcs11 != NULL) {
@@ -574,12 +502,6 @@ call_get_cert_callback(gnutls_session_t session,
 			}
 			gnutls_free(st2.cert.x509);
 		}
-	} else {
-#ifdef ENABLE_OPENPGP
-		if (st2.deinit_all) {
-			gnutls_openpgp_crt_deinit(st2.cert.pgp);
-		}
-#endif
 	}
 
 	return ret;
@@ -617,9 +539,7 @@ select_client_cert(gnutls_session_t session,
 
 		/* use a callback to get certificate 
 		 */
-		if (session->security_parameters.cert_type != GNUTLS_CRT_X509)
-			issuers_dn_length = 0;
-		else {
+		if (session->security_parameters.cert_type == GNUTLS_CRT_X509) {
 			issuers_dn_length =
 			    get_issuers_num(session, data, data_size);
 			if (issuers_dn_length < 0) {
@@ -645,6 +565,8 @@ select_client_cert(gnutls_session_t session,
 					goto cleanup;
 				}
 			}
+		} else {
+			issuers_dn_length = 0;
 		}
 
 		result =
@@ -656,20 +578,13 @@ select_client_cert(gnutls_session_t session,
 	} else {
 		/* If we have no callbacks, try to guess.
 		 */
-		result = 0;
-
-		if (session->security_parameters.cert_type == GNUTLS_CRT_X509)
+		if (session->security_parameters.cert_type == GNUTLS_CRT_X509) {
 			result =
 			    find_x509_client_cert(session, cred, _data, _data_size,
 					   pk_algos, pk_algos_length, &indx);
-#ifdef ENABLE_OPENPGP
-		else if (session->security_parameters.cert_type ==
-			 GNUTLS_CRT_OPENPGP)
-			result =
-			    find_openpgp_cert(cred, pk_algos,
-					      pk_algos_length, &indx);
-#endif
-
+		} else {
+			result = GNUTLS_E_UNIMPLEMENTED_FEATURE;
+		}
 		if (result < 0) {
 			gnutls_assert();
 			return result;
@@ -681,7 +596,7 @@ select_client_cert(gnutls_session_t session,
 						   cert_list[0],
 						   cred->certs[indx].
 						   cert_list_length,
-						   cred->pkey[indx], 0,
+						   cred->certs[indx].pkey, 0,
 						   NULL, NULL);
 		} else {
 			_gnutls_selected_certs_set(session, NULL, 0, NULL, 0,
@@ -749,171 +664,12 @@ static int gen_x509_crt(gnutls_session_t session, gnutls_buffer_st * data)
 	return data->length;
 }
 
-enum PGPKeyDescriptorType
-    { PGP_EMPTY_KEY = 1, PGP_KEY_SUBKEY, PGP_KEY_FINGERPRINT_SUBKEY };
-
-#ifdef ENABLE_OPENPGP
-static int
-gen_openpgp_certificate(gnutls_session_t session, gnutls_buffer_st * data)
-{
-	int ret;
-	gnutls_pcert_st *apr_cert_list;
-	gnutls_privkey_t apr_pkey;
-	int apr_cert_list_length;
-	unsigned int subkey;
-	uint8_t type;
-	uint8_t fpr[GNUTLS_OPENPGP_V4_FINGERPRINT_SIZE];
-	char buf[2 * GNUTLS_OPENPGP_KEYID_SIZE + 1];
-	size_t fpr_size;
-
-	/* find the appropriate certificate */
-	if ((ret =
-	     _gnutls_get_selected_cert(session, &apr_cert_list,
-				       &apr_cert_list_length, &apr_pkey)) < 0) {
-		gnutls_assert();
-		return ret;
-	}
-
-	ret = 3 + 1 + 3;
-
-	if (apr_cert_list_length > 0) {
-		fpr_size = sizeof(fpr);
-		ret =
-		    gnutls_pubkey_get_openpgp_key_id(apr_cert_list[0].pubkey, 0,
-						     fpr, &fpr_size, &subkey);
-		if (ret < 0)
-			return gnutls_assert_val(ret);
-
-		ret += 1 + fpr_size;	/* for the keyid */
-		_gnutls_handshake_log("Sending PGP key ID %s (%s)\n",
-				      _gnutls_bin2hex(fpr,
-						      GNUTLS_OPENPGP_KEYID_SIZE,
-						      buf, sizeof(buf),
-						      NULL),
-				      subkey ? "subkey" : "master");
-
-		ret += apr_cert_list[0].cert.size;
-	}
-
-	ret = _gnutls_buffer_append_prefix(data, 24, ret - 3);
-	if (ret < 0)
-		return gnutls_assert_val(ret);
-
-	if (apr_cert_list_length > 0) {
-		type = PGP_KEY_SUBKEY;
-
-		ret = _gnutls_buffer_append_data(data, &type, 1);
-		if (ret < 0)
-			return gnutls_assert_val(ret);
-
-		ret = _gnutls_buffer_append_data_prefix(data, 8, fpr, fpr_size);
-		if (ret < 0)
-			return gnutls_assert_val(ret);
-
-		ret =
-		    _gnutls_buffer_append_data_prefix(data, 24,
-						      apr_cert_list[0].
-						      cert.data,
-						      apr_cert_list[0].
-						      cert.size);
-		if (ret < 0)
-			return gnutls_assert_val(ret);
-	} else {		/* empty - no certificate */
-
-		type = PGP_EMPTY_KEY;
-
-		ret = _gnutls_buffer_append_data(data, &type, 1);
-		if (ret < 0)
-			return gnutls_assert_val(ret);
-
-		ret = _gnutls_buffer_append_prefix(data, 24, 0);
-		if (ret < 0)
-			return gnutls_assert_val(ret);
-	}
-
-	return data->length;
-}
-
-static int
-gen_openpgp_certificate_fpr(gnutls_session_t session, gnutls_buffer_st * data)
-{
-	int ret, packet_size;
-	uint8_t type, fpr[GNUTLS_OPENPGP_V4_FINGERPRINT_SIZE];
-	uint8_t id[GNUTLS_OPENPGP_KEYID_SIZE];
-	unsigned int subkey;
-	size_t fpr_size, id_size;
-	gnutls_pcert_st *apr_cert_list;
-	gnutls_privkey_t apr_pkey;
-	int apr_cert_list_length;
-
-	/* find the appropriate certificate */
-	if ((ret =
-	     _gnutls_get_selected_cert(session, &apr_cert_list,
-				       &apr_cert_list_length, &apr_pkey)) < 0) {
-		gnutls_assert();
-		return ret;
-	}
-
-	if (apr_cert_list_length <= 0)
-		return gen_openpgp_certificate(session, data);
-
-	id_size = sizeof(id);
-	ret =
-	    gnutls_pubkey_get_openpgp_key_id(apr_cert_list[0].pubkey, 0,
-					     id, &id_size, &subkey);
-	if (ret < 0)
-		return gnutls_assert_val(ret);
-
-	fpr_size = sizeof(fpr);
-	ret =
-	    gnutls_pubkey_get_openpgp_key_id(apr_cert_list[0].pubkey,
-					     GNUTLS_PUBKEY_GET_OPENPGP_FINGERPRINT,
-					     fpr, &fpr_size, NULL);
-	if (ret < 0)
-		return gnutls_assert_val(ret);
-
-	packet_size = 3 + 1;
-	packet_size += 1 + fpr_size;	/* for the keyid */
-
-	/* Only v4 fingerprints are sent 
-	 */
-	packet_size += 20 + 1;
-
-	ret = _gnutls_buffer_append_prefix(data, 24, packet_size - 3);
-	if (ret < 0)
-		return gnutls_assert_val(ret);
-
-	type = PGP_KEY_FINGERPRINT_SUBKEY;
-	ret = _gnutls_buffer_append_data(data, &type, 1);
-	if (ret < 0)
-		return gnutls_assert_val(ret);
-
-	ret = _gnutls_buffer_append_data_prefix(data, 8, id, id_size);
-	if (ret < 0)
-		return gnutls_assert_val(ret);
-
-	ret = _gnutls_buffer_append_data_prefix(data, 8, fpr, fpr_size);
-	if (ret < 0)
-		return gnutls_assert_val(ret);
-
-	return data->length;
-}
-#endif
-
 int
 _gnutls_gen_cert_client_crt(gnutls_session_t session, gnutls_buffer_st * data)
 {
 	switch (session->security_parameters.cert_type) {
-#ifdef ENABLE_OPENPGP
-	case GNUTLS_CRT_OPENPGP:
-		if (_gnutls_openpgp_send_fingerprint(session) == 0)
-			return gen_openpgp_certificate(session, data);
-		else
-			return gen_openpgp_certificate_fpr(session, data);
-#endif
 	case GNUTLS_CRT_X509:
 		return gen_x509_crt(session, data);
-
 	default:
 		gnutls_assert();
 		return GNUTLS_E_INTERNAL_ERROR;
@@ -924,10 +680,6 @@ int
 _gnutls_gen_cert_server_crt(gnutls_session_t session, gnutls_buffer_st * data)
 {
 	switch (session->security_parameters.cert_type) {
-#ifdef ENABLE_OPENPGP
-	case GNUTLS_CRT_OPENPGP:
-		return gen_openpgp_certificate(session, data);
-#endif
 	case GNUTLS_CRT_X509:
 		return gen_x509_crt(session, data);
 	default:
@@ -940,7 +692,6 @@ static
 int check_pk_compat(gnutls_session_t session, gnutls_pubkey_t pubkey)
 {
 	unsigned cert_pk;
-	unsigned req_cert_pk;
 	unsigned kx;
 
 	if (session->security_parameters.entity != GNUTLS_CLIENT)
@@ -952,15 +703,10 @@ int check_pk_compat(gnutls_session_t session, gnutls_pubkey_t pubkey)
 		return GNUTLS_E_CERTIFICATE_ERROR;
 	}
 
-	kx = _gnutls_cipher_suite_get_kx_algo(session->
-					      security_parameters.cipher_suite);
+	kx = session->security_parameters.cs->kx_algorithm;
 
-	req_cert_pk = _gnutls_kx_cert_pk_params(kx);
-
-	if (req_cert_pk == GNUTLS_PK_UNKNOWN)	/* doesn't matter */
-		return 0;
-
-	if (req_cert_pk != cert_pk) {
+	if (_gnutls_map_kx_get_cred(kx, 1) == GNUTLS_CRD_CERTIFICATE &&
+	    !_gnutls_kx_supports_pk(kx, cert_pk)) {
 		gnutls_assert();
 		return GNUTLS_E_CERTIFICATE_ERROR;
 	}
@@ -1011,10 +757,14 @@ _gnutls_proc_x509_server_crt(gnutls_session_t session,
 	size = _gnutls_read_uint24(p);
 	p += 3;
 
+	/* ensure no discrepancy in data */
+	if (size != dsize)
+		return gnutls_assert_val(GNUTLS_E_UNEXPECTED_PACKET_LENGTH);
+
 	/* some implementations send 0B 00 00 06 00 00 03 00 00 00
 	 * instead of just 0B 00 00 03 00 00 00 as an empty certificate message.
 	 */
-	if (size == 0 || size == 3) {
+	if (size == 0 || (size == 3 && memcmp(p, "\x00\x00\x00", 3) == 0)) {
 		gnutls_assert();
 		/* no certificate was sent */
 		return GNUTLS_E_NO_CERTIFICATE_FOUND;
@@ -1073,6 +823,7 @@ _gnutls_proc_x509_server_crt(gnutls_session_t session,
 		if (ret < 0) {
 			gnutls_assert();
 			peer_certificate_list_size = j;
+			ret = GNUTLS_E_CERTIFICATE_ERROR;
 			goto cleanup;
 		}
 
@@ -1103,199 +854,6 @@ _gnutls_proc_x509_server_crt(gnutls_session_t session,
 
 }
 
-#ifdef ENABLE_OPENPGP
-static int
-_gnutls_proc_openpgp_server_crt(gnutls_session_t session,
-				uint8_t * data, size_t data_size)
-{
-	int size, ret, len;
-	uint8_t *p = data;
-	cert_auth_info_t info;
-	gnutls_certificate_credentials_t cred;
-	ssize_t dsize = data_size;
-	int key_type;
-	gnutls_pcert_st *peer_certificate_list = NULL;
-	gnutls_datum_t tmp, akey = { NULL, 0 };
-	unsigned int compat = 0;
-	uint8_t subkey_id[GNUTLS_OPENPGP_KEYID_SIZE];
-
-	cred = (gnutls_certificate_credentials_t)
-	    _gnutls_get_cred(session, GNUTLS_CRD_CERTIFICATE);
-	if (cred == NULL) {
-		gnutls_assert();
-		return GNUTLS_E_INSUFFICIENT_CREDENTIALS;
-	}
-
-	if ((ret =
-	     _gnutls_auth_info_set(session, GNUTLS_CRD_CERTIFICATE,
-				   sizeof(cert_auth_info_st), 1)) < 0) {
-		gnutls_assert();
-		return ret;
-	}
-
-	info = _gnutls_get_auth_info(session, GNUTLS_CRD_CERTIFICATE);
-
-	if (data == NULL || data_size == 0) {
-		gnutls_assert();
-		return GNUTLS_E_NO_CERTIFICATE_FOUND;
-	}
-
-	DECR_LEN(dsize, 3);
-	size = _gnutls_read_uint24(p);
-	p += 3;
-
-	if (size == 0) {
-		gnutls_assert();
-		/* no certificate was sent */
-		return GNUTLS_E_NO_CERTIFICATE_FOUND;
-	}
-
-	/* Read PGPKeyDescriptor */
-	DECR_LEN(dsize, 1);
-	key_type = *p;
-	p++;
-
-	/* Try to read the keyid if present */
-	if (key_type == PGP_KEY_FINGERPRINT_SUBKEY
-	    || key_type == PGP_KEY_SUBKEY) {
-		/* check size */
-		if (*p != GNUTLS_OPENPGP_KEYID_SIZE) {
-			gnutls_assert();
-			return GNUTLS_E_UNSUPPORTED_CERTIFICATE_TYPE;
-		}
-
-		DECR_LEN(dsize, 1);
-		p++;
-
-		DECR_LEN(dsize, GNUTLS_OPENPGP_KEYID_SIZE);
-		memcpy(subkey_id, p, GNUTLS_OPENPGP_KEYID_SIZE);
-		p += GNUTLS_OPENPGP_KEYID_SIZE;
-	}
-
-	if (key_type == PGP_KEY_FINGERPRINT_SUBKEY) {
-		DECR_LEN(dsize, 1);
-		len = (uint8_t) * p;
-		p++;
-
-		if (len != 20) {
-			gnutls_assert();
-			return GNUTLS_E_OPENPGP_FINGERPRINT_UNSUPPORTED;
-		}
-
-		DECR_LEN(dsize, 20);
-
-		/* request the actual key from our database, or
-		 * a key server or anything.
-		 */
-		if ((ret =
-		     _gnutls_openpgp_request_key(session, &akey, cred, p,
-						 20)) < 0) {
-			gnutls_assert();
-			return ret;
-		}
-		tmp = akey;
-	} else if (key_type == PGP_KEY_SUBKEY) {	/* the whole key */
-
-		/* Read the actual certificate */
-		DECR_LEN(dsize, 3);
-		len = _gnutls_read_uint24(p);
-		p += 3;
-
-		if (len == 0) {
-			gnutls_assert();
-			/* no certificate was sent */
-			return
-			    gnutls_assert_val
-			    (GNUTLS_E_UNEXPECTED_PACKET_LENGTH);
-		}
-
-		DECR_LEN(dsize, len);
-
-		tmp.size = len;
-		tmp.data = p;
-
-	} else if (key_type == PGP_EMPTY_KEY) {	/* the whole key */
-
-		/* Read the actual certificate */
-		DECR_LEN(dsize, 3);
-		len = _gnutls_read_uint24(p);
-		p += 3;
-
-		if (len == 0)	/* PGP_EMPTY_KEY */
-			return GNUTLS_E_NO_CERTIFICATE_FOUND;
-		/* Uncomment to remove compatibility with RFC5081.
-		   else
-		   return gnutls_assert_val(GNUTLS_E_UNEXPECTED_PACKET_LENGTH); */
-
-		DECR_LEN(dsize, len);
-
-		tmp.size = len;
-		tmp.data = p;
-
-		compat = 1;
-	} else {
-		gnutls_assert();
-		return GNUTLS_E_UNSUPPORTED_CERTIFICATE_TYPE;
-	}
-
-	if (dsize != 0)
-		return gnutls_assert_val(GNUTLS_E_UNEXPECTED_PACKET_LENGTH);
-
-	/* ok we now have the peer's key in tmp datum
-	 */
-	peer_certificate_list = gnutls_calloc(1, sizeof(gnutls_pcert_st));
-	if (peer_certificate_list == NULL) {
-		gnutls_assert();
-		ret = GNUTLS_E_MEMORY_ERROR;
-		goto cleanup;
-	}
-
-	ret =
-	    gnutls_pcert_import_openpgp_raw(&peer_certificate_list[0],
-					    &tmp,
-					    GNUTLS_OPENPGP_FMT_RAW,
-					    (compat ==
-					     0) ? subkey_id : NULL, 0);
-	if (ret < 0) {
-		gnutls_assert();
-		goto cleanup;
-	}
-
-	if (compat != 0) {
-		size_t t = sizeof(subkey_id);
-		gnutls_pubkey_get_openpgp_key_id(peer_certificate_list
-						 [0].pubkey, 0, subkey_id, &t,
-						 NULL);
-	}
-
-	ret = check_pk_compat(session, peer_certificate_list[0].pubkey);
-	if (ret < 0) {
-		gnutls_assert();
-		goto cleanup;
-	}
-
-	ret =
-	    copy_certificate_auth_info(info,
-				       peer_certificate_list, 1, subkey_id);
-	if (ret < 0) {
-		gnutls_assert();
-		goto cleanup;
-	}
-
-	ret = 0;
-
- cleanup:
-
-	_gnutls_free_datum(&akey);
-	if (peer_certificate_list != NULL) {
-		gnutls_pcert_deinit(&peer_certificate_list[0]);
-		gnutls_free(peer_certificate_list);
-	}
-	return ret;
-
-}
-#endif
-
 int _gnutls_proc_crt(gnutls_session_t session, uint8_t * data, size_t data_size)
 {
 	int ret;
@@ -1310,11 +868,6 @@ int _gnutls_proc_crt(gnutls_session_t session, uint8_t * data, size_t data_size)
 	}
 
 	switch (session->security_parameters.cert_type) {
-#ifdef ENABLE_OPENPGP
-	case GNUTLS_CRT_OPENPGP:
-		ret = _gnutls_proc_openpgp_server_crt(session, data, data_size);
-		break;
-#endif
 	case GNUTLS_CRT_X509:
 		ret = _gnutls_proc_x509_server_crt(session, data, data_size);
 		break;
@@ -1419,12 +972,6 @@ _gnutls_proc_cert_cert_req(gnutls_session_t session, uint8_t * data,
 	size = _gnutls_read_uint16(p);
 	p += 2;
 
-	if (session->security_parameters.cert_type == GNUTLS_CRT_OPENPGP
-	    && size != 0) {
-		gnutls_assert();
-		return GNUTLS_E_UNEXPECTED_PACKET_LENGTH;
-	}
-
 	DECR_LEN_FINAL(dsize, size);
 
 	/* We should reply with a certificate message, 
@@ -1490,8 +1037,8 @@ _gnutls_gen_cert_client_crt_vrfy(gnutls_session_t session,
 		if (aid == NULL)
 			return gnutls_assert_val(GNUTLS_E_UNKNOWN_ALGORITHM);
 
-		p[0] = aid->hash_algorithm;
-		p[1] = aid->sign_algorithm;
+		p[0] = aid->id[0];
+		p[1] = aid->id[1];
 		ret = _gnutls_buffer_append_data(data, p, 2);
 		if (ret < 0) {
 			gnutls_assert();
@@ -1526,6 +1073,8 @@ _gnutls_proc_cert_client_crt_vrfy(gnutls_session_t session,
 	gnutls_pcert_st peer_cert;
 	gnutls_sign_algorithm_t sign_algo = GNUTLS_SIGN_UNKNOWN;
 	const version_entry_st *ver = get_version(session);
+	gnutls_certificate_credentials_t cred;
+	unsigned vflags;
 
 	if (unlikely(info == NULL || info->ncerts == 0 || ver == NULL)) {
 		gnutls_assert();
@@ -1533,12 +1082,21 @@ _gnutls_proc_cert_client_crt_vrfy(gnutls_session_t session,
 		return GNUTLS_E_INTERNAL_ERROR;
 	}
 
+	cred = (gnutls_certificate_credentials_t)
+	    _gnutls_get_cred(session, GNUTLS_CRD_CERTIFICATE);
+	if (cred == NULL) {
+		gnutls_assert();
+		return GNUTLS_E_INSUFFICIENT_CREDENTIALS;
+	}
+
+	vflags = cred->verify_flags | session->internals.additional_verify_flags;
+
 	if (_gnutls_version_has_selectable_sighash(ver)) {
 		sign_algorithm_st aid;
 
 		DECR_LEN(dsize, 2);
-		aid.hash_algorithm = pdata[0];
-		aid.sign_algorithm = pdata[1];
+		aid.id[0] = pdata[0];
+		aid.id[1] = pdata[1];
 
 		sign_algo = _gnutls_tls_aid_to_sign(&aid);
 		if (sign_algo == GNUTLS_SIGN_UNKNOWN) {
@@ -1571,7 +1129,7 @@ _gnutls_proc_cert_client_crt_vrfy(gnutls_session_t session,
 	}
 
 	if ((ret =
-	     _gnutls_handshake_verify_crt_vrfy(session, &peer_cert, &sig,
+	     _gnutls_handshake_verify_crt_vrfy(session, vflags, &peer_cert, &sig,
 					       sign_algo)) < 0) {
 		gnutls_assert();
 		gnutls_pcert_deinit(&peer_cert);
@@ -1616,19 +1174,12 @@ _gnutls_gen_cert_server_cert_req(gnutls_session_t session,
 		return gnutls_assert_val(ret);
 
 	if (_gnutls_version_has_selectable_sighash(ver)) {
-		uint8_t p[MAX_SIGN_ALGO_SIZE];
-
 		ret =
-		    _gnutls_sign_algorithm_write_params(session, p,
-							MAX_SIGN_ALGO_SIZE);
+		    _gnutls_sign_algorithm_write_params(session, data);
 		if (ret < 0) {
 			gnutls_assert();
 			return ret;
 		}
-
-		ret = _gnutls_buffer_append_data(data, p, ret);
-		if (ret < 0)
-			return gnutls_assert_val(ret);
 	}
 
 	if (session->security_parameters.cert_type == GNUTLS_CRT_X509 &&
@@ -1765,73 +1316,6 @@ alloc_and_load_x509_key(gnutls_x509_privkey_t key, int deinit)
 	return local_key;
 }
 
-/* converts the given pgp certificate to gnutls_cert* and allocates
- * space for them.
- */
-#ifdef ENABLE_OPENPGP
-static gnutls_pcert_st *alloc_and_load_pgp_certs(gnutls_openpgp_crt_t cert)
-{
-	gnutls_pcert_st *local_certs;
-	int ret = 0;
-
-	if (cert == NULL)
-		return NULL;
-
-	local_certs = gnutls_malloc(sizeof(gnutls_pcert_st));
-	if (local_certs == NULL) {
-		gnutls_assert();
-		return NULL;
-	}
-
-	ret = gnutls_pcert_import_openpgp(local_certs, cert, 0);
-	if (ret < 0) {
-		gnutls_assert();
-		return NULL;
-	}
-
-	if (ret < 0) {
-		gnutls_assert();
-		gnutls_pcert_deinit(local_certs);
-		gnutls_free(local_certs);
-		return NULL;
-	}
-
-	return local_certs;
-}
-
-/* converts the given raw key to gnutls_privkey* and allocates
- * space for it.
- */
-static gnutls_privkey_t
-alloc_and_load_pgp_key(gnutls_openpgp_privkey_t key, int deinit)
-{
-	gnutls_privkey_t local_key;
-	int ret = 0;
-
-	if (key == NULL)
-		return NULL;
-
-	ret = gnutls_privkey_init(&local_key);
-	if (ret < 0) {
-		gnutls_assert();
-		return NULL;
-	}
-
-	ret =
-	    gnutls_privkey_import_openpgp(local_key, key,
-					  deinit ?
-					  GNUTLS_PRIVKEY_IMPORT_AUTO_RELEASE
-					  : 0);
-	if (ret < 0) {
-		gnutls_assert();
-		gnutls_privkey_deinit(local_key);
-		return NULL;
-	}
-
-	return local_key;
-}
-#endif
-
 #ifdef ENABLE_PKCS11
 
 /* converts the given raw key to gnutls_privkey* and allocates
@@ -1931,6 +1415,66 @@ static void get_server_name(gnutls_session_t session, uint8_t * name,
 	return;
 }
 
+static
+unsigned pubkey_is_compat_with_cs(gnutls_session_t session,
+				     gnutls_pubkey_t pubkey,
+				     gnutls_certificate_type_t cert_type,
+				     const gnutls_cipher_suite_entry_st *cs)
+{
+	unsigned pk = pubkey->params.algo;
+	unsigned key_usage;
+
+	if (session->security_parameters.cert_type != cert_type) {
+		return 0;
+	}
+
+	if (unlikely(session->internals.priorities->allow_server_key_usage_violation)) {
+		key_usage = 0;
+	} else {
+		key_usage = pubkey->key_usage;
+	}
+
+	if (!_gnutls_kx_supports_pk_usage(cs->kx_algorithm, pk, key_usage)) {
+		return 0;
+	}
+
+	return 1;
+}
+
+/* Selects a signature algorithm (if required by the ciphersuite and TLS
+ * version), appropriate for the certificate. If none can be selected
+ * returns an error.
+ */
+static
+int select_sign_algorithm(gnutls_session_t session,
+			  gnutls_pcert_st * cert,
+			  gnutls_privkey_t pkey,
+			  const gnutls_cipher_suite_entry_st *cs)
+{
+	gnutls_sign_algorithm_t algo;
+	const version_entry_st *ver = get_version(session);
+
+	if (_gnutls_kx_encipher_type(cs->kx_algorithm) != CIPHER_SIGN)
+		return 0;
+
+	if (!_gnutls_version_has_selectable_sighash(ver)) {
+		/* For SSL3.0 and TLS1.0 we lie as we cannot express md5-sha1 as
+		 * signature algorithm. */
+		algo = gnutls_pk_to_sign(cert->pubkey->params.algo, GNUTLS_DIG_SHA1);
+		gnutls_sign_algorithm_set_server(session, algo);
+		return 0;
+	}
+
+	algo = _gnutls_session_get_sign_algo(session, cert, pkey, 0);
+	if (algo == GNUTLS_SIGN_UNKNOWN)
+		return gnutls_assert_val(GNUTLS_E_INCOMPATIBLE_SIG_WITH_KEY);
+
+	gnutls_sign_algorithm_set_server(session, algo);
+	_gnutls_handshake_log("Selected signature algorithm: %s\n", gnutls_sign_algorithm_get_name(algo));
+
+	return 0;
+}
+
 /* finds the most appropriate certificate in the cert list.
  * The 'appropriate' is defined by the user.
  *
@@ -1942,9 +1486,7 @@ static void get_server_name(gnutls_session_t session, uint8_t * name,
  *
  */
 int
-_gnutls_server_select_cert(gnutls_session_t session,
-			   gnutls_pk_algorithm_t * pk_algos,
-			   size_t pk_algos_size)
+_gnutls_server_select_cert(gnutls_session_t session, const gnutls_cipher_suite_entry_st *cs)
 {
 	unsigned i, j;
 	int idx, ret;
@@ -1954,22 +1496,50 @@ _gnutls_server_select_cert(gnutls_session_t session,
 	cred = (gnutls_certificate_credentials_t)
 	    _gnutls_get_cred(session, GNUTLS_CRD_CERTIFICATE);
 	if (cred == NULL) {
-		gnutls_assert();
-		return GNUTLS_E_INSUFFICIENT_CREDENTIALS;
+		gnutls_assert(); /* we don't need to select a cert */
+		return 0;
 	}
 
+	/* When a callback is set, we call it once to get the
+	 * certificate and then check its compatibility with
+	 * the ciphersuites.
+	 */
+
 	/* If the callback which retrieves certificate has been set,
-	 * use it and leave.
+	 * use it and leave. We make sure that this is called once.
 	 */
 	if (cred->get_cert_callback
 	    || cred->get_cert_callback2) {
-		ret = call_get_cert_callback(session, NULL, 0, NULL, 0);
-		if (ret < 0)
+
+		if (session->internals.selected_cert_list_length == 0) {
+			ret = call_get_cert_callback(session, NULL, 0, NULL, 0);
+			if (ret < 0)
+				return gnutls_assert_val(ret);
+
+			_gnutls_debug_log("Selected (%s) cert\n",
+					  gnutls_pk_get_name(session->internals.selected_cert_list[0].pubkey->params.algo));
+		}
+
+		if (!pubkey_is_compat_with_cs(session,
+					     session->internals.selected_cert_list[0].pubkey,
+					     session->internals.selected_cert_list[0].type,
+					     cs)) {
+			return gnutls_assert_val(GNUTLS_E_INSUFFICIENT_CREDENTIALS);
+		}
+
+		ret = select_sign_algorithm(session,
+					    &session->internals.selected_cert_list[0],
+					    session->internals.selected_key,
+					    cs);
+		if (ret < 0) {
 			return gnutls_assert_val(ret);
-		return ret;
+		}
+
+		return 0;
 	}
 
-	/* Otherwise... */
+	/* Otherwise... we check the compatibility of the ciphersuite
+	 * with all the certificates available. */
 
 	get_server_name(session, (unsigned char *)server_name,
 			sizeof(server_name));
@@ -1982,70 +1552,73 @@ _gnutls_server_select_cert(gnutls_session_t session,
 	 */
 
 	if (server_name[0] != 0) {
-		for (i = 0; i < cred->ncerts; i++) {
+		for (j = 0; j < cred->ncerts; j++) {
+			i = cred->sorted_cert_idx[j];
+
 			if (cred->certs[i].names != NULL
 			    && _gnutls_str_array_match(cred->certs[i].names,
 						       server_name) != 0) {
 				/* if requested algorithms are also compatible select it */
-				gnutls_pk_algorithm_t pk =
-				    gnutls_pubkey_get_pk_algorithm(cred->certs
-								   [i].cert_list
-								   [0].pubkey,
-								   NULL);
 
-				_gnutls_handshake_log
-				    ("HSK[%p]: Requested server name: '%s', ctype: %s (%d)\n",
-				     session, server_name,
-				     gnutls_certificate_type_get_name
-				     (session->security_parameters.cert_type),
-				     session->security_parameters.cert_type);
-
-				if (session->security_parameters.cert_type ==
-				    cred->certs[i].cert_list[0].type) {
-					for (j = 0; j < pk_algos_size; j++)
-						if (pk_algos[j] == pk) {
-							idx = i;
-							goto finished;
-						}
+				if (!pubkey_is_compat_with_cs(session,
+							     cred->certs[i].cert_list[0].pubkey,
+							     cred->certs[i].cert_list[0].type,
+							     cs)) {
+					continue;
 				}
+
+				ret = select_sign_algorithm(session,
+							    &cred->certs[i].cert_list[0],
+							    cred->certs[i].pkey,
+							    cs);
+				if (ret >= 0) {
+					idx = i;
+					_gnutls_debug_log("Selected (%s) cert based on ciphersuite %x.%x: %s\n",
+						  gnutls_pk_get_name(cred->certs[i].cert_list[0].pubkey->params.algo),
+						  (unsigned)cs->id[0],
+						  (unsigned)cs->id[1],
+						  cs->name);
+					/* found */
+					goto finished;
+				}
+
+
 			}
 		}
 	}
 
-	for (j = 0; j < pk_algos_size; j++) {
+	/* no name match */
+	for (j = 0; j < cred->ncerts; j++) {
+		i = cred->sorted_cert_idx[j];
+
 		_gnutls_handshake_log
-		    ("HSK[%p]: Requested PK algorithm: %s (%d) -- ctype: %s (%d)\n",
-		     session, gnutls_pk_get_name(pk_algos[j]), pk_algos[j],
-		     gnutls_certificate_type_get_name
-		     (session->security_parameters.cert_type),
-		     session->security_parameters.cert_type);
+		    ("HSK[%p]: checking compat of %s with certificate[%d] (%s/%s)\n",
+		     session, cs->name, i,
+		     gnutls_pk_get_name(cred->certs[i].cert_list[0].pubkey->params.algo),
+		     gnutls_certificate_type_get_name(cred->certs
+						      [i].cert_list
+						      [0].type));
 
-		for (i = 0; i < cred->ncerts; i++) {
-			gnutls_pk_algorithm_t pk =
-			    gnutls_pubkey_get_pk_algorithm(cred->certs[i].
-							   cert_list[0].pubkey,
-							   NULL);
-			/* find one compatible certificate
-			 */
-			_gnutls_handshake_log
-			    ("HSK[%p]: certificate[%d] PK algorithm: %s (%d) - ctype: %s (%d)\n",
-			     session, i, gnutls_pk_get_name(pk), pk,
-			     gnutls_certificate_type_get_name(cred->certs
-							      [i].cert_list
-							      [0].type),
-			     cred->certs[i].cert_list[0].type);
+		if (!pubkey_is_compat_with_cs(session,
+					     cred->certs[i].cert_list[0].pubkey,
+					     cred->certs[i].cert_list[0].type,
+					     cs)) {
+			continue;
+		}
 
-			if (pk_algos[j] == pk) {
-				/* if cert type matches
-				 */
-	  /* *INDENT-OFF* */
-	  if (session->security_parameters.cert_type == cred->certs[i].cert_list[0].type)
-	    {
-	      idx = i;
-	      goto finished;
-	    }
-	  /* *INDENT-ON* */
-			}
+		ret = select_sign_algorithm(session,
+					    &cred->certs[i].cert_list[0],
+					    cred->certs[i].pkey,
+					    cs);
+		if (ret >= 0) {
+			idx = i;
+			_gnutls_debug_log("Selected (%s) cert based on ciphersuite %x.%x: %s\n",
+					  gnutls_pk_get_name(cred->certs[i].cert_list[0].pubkey->params.algo),
+					  (unsigned)cs->id[0],
+					  (unsigned)cs->id[1],
+					  cs->name);
+			/* found */
+			goto finished;
 		}
 	}
 
@@ -2057,7 +1630,7 @@ _gnutls_server_select_cert(gnutls_session_t session,
 		_gnutls_selected_certs_set(session,
 					   &cred->certs[idx].cert_list[0],
 					   cred->certs[idx].cert_list_length,
-					   cred->pkey[idx], 0,
+					   cred->certs[idx].pkey, 0,
 					   cred->certs[idx].ocsp_func,
 					   cred->certs[idx].ocsp_func_ptr);
 	} else {
@@ -2134,8 +1707,8 @@ int _gnutls_gen_dhe_signature(gnutls_session_t session,
 			goto cleanup;
 		}
 
-		p[0] = aid->hash_algorithm;
-		p[1] = aid->sign_algorithm;
+		p[0] = aid->id[0];
+		p[1] = aid->id[1];
 
 		ret = _gnutls_buffer_append_data(data, p, 2);
 		if (ret < 0) {
@@ -2170,6 +1743,8 @@ _gnutls_proc_dhe_signature(gnutls_session_t session, uint8_t * data,
 	gnutls_pcert_st peer_cert;
 	gnutls_sign_algorithm_t sign_algo = GNUTLS_SIGN_UNKNOWN;
 	const version_entry_st *ver = get_version(session);
+	gnutls_certificate_credentials_t cred;
+	unsigned vflags;
 
 	if (unlikely(info == NULL || info->ncerts == 0 || ver == NULL)) {
 		gnutls_assert();
@@ -2177,19 +1752,27 @@ _gnutls_proc_dhe_signature(gnutls_session_t session, uint8_t * data,
 		return GNUTLS_E_INTERNAL_ERROR;
 	}
 
+	cred = (gnutls_certificate_credentials_t)
+	    _gnutls_get_cred(session, GNUTLS_CRD_CERTIFICATE);
+	if (cred == NULL) {
+		gnutls_assert();
+		return GNUTLS_E_INSUFFICIENT_CREDENTIALS;
+	}
+
+	vflags = cred->verify_flags | session->internals.additional_verify_flags;
+
 	/* VERIFY SIGNATURE */
 	if (_gnutls_version_has_selectable_sighash(ver)) {
 		sign_algorithm_st aid;
 
 		DECR_LEN(data_size, 1);
-		aid.hash_algorithm = *data++;
+		aid.id[0] = *data++;
 		DECR_LEN(data_size, 1);
-		aid.sign_algorithm = *data++;
+		aid.id[1] = *data++;
 		sign_algo = _gnutls_tls_aid_to_sign(&aid);
 		if (sign_algo == GNUTLS_SIGN_UNKNOWN) {
 			_gnutls_debug_log("unknown signature %d.%d\n",
-					  aid.sign_algorithm,
-					  aid.hash_algorithm);
+					  aid.id[0], aid.id[1]);
 			gnutls_assert();
 			return GNUTLS_E_UNSUPPORTED_SIGNATURE_ALGORITHM;
 		}
@@ -2211,7 +1794,7 @@ _gnutls_proc_dhe_signature(gnutls_session_t session, uint8_t * data,
 	}
 
 	ret =
-	    _gnutls_handshake_verify_data(session, &peer_cert, vparams,
+	    _gnutls_handshake_verify_data(session, vflags, &peer_cert, vparams,
 					  &signature, sign_algo);
 
 	gnutls_pcert_deinit(&peer_cert);
