@@ -1,5 +1,6 @@
 /*
  * Copyright (C) 2002-2012 Free Software Foundation, Inc.
+ * Copyright (C) 2016-2017 Red Hat, Inc.
  *
  * Author: Nikos Mavrogiannopoulos
  *
@@ -25,7 +26,6 @@
 #include "auth.h"
 #include <auth/dh_common.h>
 #include <x509/x509_int.h>
-#include <openpgp/openpgp_int.h>
 #include <gnutls/abstract.h>
 #include <gnutls/compat.h>
 #include <str_array.h>
@@ -38,6 +38,9 @@ typedef struct {
 	gnutls_status_request_ocsp_func ocsp_func;
 	void *ocsp_func_ptr; /* corresponding OCSP response function + ptr */
 	char *ocsp_response_file; /* corresponding OCSP response file */
+
+	/* the private key corresponding to certificate */
+	gnutls_privkey_t pkey;
 } certs_st;
 
 /* This structure may be complex, but it's the only way to
@@ -45,6 +48,9 @@ typedef struct {
  */
 typedef struct gnutls_certificate_credentials_st {
 	gnutls_dh_params_t dh_params;
+	unsigned deinit_dh_params; /* if the internal values are set */
+	gnutls_sec_param_t dh_sec_param; /* used in RFC7919 negotiation */
+
 	/* this callback is used to retrieve the DH or RSA
 	 * parameters.
 	 */
@@ -53,16 +59,12 @@ typedef struct gnutls_certificate_credentials_st {
 	certs_st *certs;
 	unsigned ncerts;	/* the number of certs */
 
-	gnutls_privkey_t *pkey;
-	/* private keys. It contains ncerts private
-	 * keys. pkey[i] corresponds to certificate in
-	 * cert_list[i][0].
+	/* contains sorted index values for certs. Sorted in a way
+	 * that RSA-PSS keys always take precedence over plain RSA keys
+	 * to ensure that we use only RSA-PSS keys if present for RSA-PSS
+	 * operations. We keep indexes to certs structures above.
 	 */
-
-#ifdef ENABLE_OPENPGP
-	/* OpenPGP specific stuff */
-	gnutls_openpgp_keyring_t keyring;
-#endif
+	unsigned int *sorted_cert_idx;
 
 	/* X509 specific stuff */
 	gnutls_x509_trust_list_t tlist;
@@ -109,9 +111,6 @@ typedef struct cert_auth_info_st {
 	unsigned int ncerts;	/* holds the size of the list above */
 
 	gnutls_certificate_type_t cert_type;
-#ifdef ENABLE_OPENPGP
-	uint8_t subkey_id[GNUTLS_OPENPGP_KEYID_SIZE];
-#endif
 } *cert_auth_info_t;
 
 typedef struct cert_auth_info_st cert_auth_info_st;
@@ -131,20 +130,14 @@ int _gnutls_get_selected_cert(gnutls_session_t session,
 			      int *apr_cert_list_length,
 			      gnutls_privkey_t * apr_pkey);
 
-int _gnutls_server_select_cert(struct gnutls_session_int *,
-			       gnutls_pk_algorithm_t *, size_t);
+int
+_gnutls_server_select_cert(gnutls_session_t session, const gnutls_cipher_suite_entry_st *cs);
 void _gnutls_selected_certs_deinit(gnutls_session_t session);
 
 int _gnutls_get_auth_info_pcert(gnutls_pcert_st * gcert,
 				gnutls_certificate_type_t type,
 				cert_auth_info_t info);
 
-int certificate_credential_append_crt_list(gnutls_certificate_credentials_t
-					   res, gnutls_str_array_t names,
-					   gnutls_pcert_st * crt, int nr);
-int certificate_credentials_append_pkey(gnutls_certificate_credentials_t
-					res, gnutls_privkey_t pkey);
-
 int _gnutls_selected_cert_supported_kx(struct gnutls_session_int *session,
 				       gnutls_kx_algorithm_t * alg,
 				       int *alg_size);
