@@ -1,5 +1,6 @@
 /*
  * Copyright (C) 2011-2012 Free Software Foundation, Inc.
+ * Copyright (C) 2017 Red Hat, Inc.
  *
  * Author: Nikos Mavrogiannopoulos
  *
@@ -98,7 +99,7 @@ static int calc_ecdh_key(gnutls_session_t session,
 	pub.params[ECC_Y] = session->key.ecdh_y;
 	pub.raw_pub.data = session->key.ecdhx.data;
 	pub.raw_pub.size = session->key.ecdhx.size;
-	pub.flags = ecurve->id;
+	pub.curve = ecurve->id;
 
 	ret =
 	    _gnutls_pk_derive(ecurve->pk, &tmp_dh_key,
@@ -136,15 +137,19 @@ static int calc_ecdh_key(gnutls_session_t session,
 
 int _gnutls_proc_ecdh_common_client_kx(gnutls_session_t session,
 				       uint8_t * data, size_t _data_size,
-				       gnutls_ecc_curve_t curve,
+				       const struct gnutls_group_entry_st *group,
 				       gnutls_datum_t * psk_key)
 {
 	ssize_t data_size = _data_size;
 	int ret, i = 0;
-	int point_size;
-	const gnutls_ecc_curve_entry_st *ecurve = _gnutls_ecc_curve_get_params(curve);
+	unsigned point_size;
+	const gnutls_ecc_curve_entry_st *ecurve;
 
-	if (curve == GNUTLS_ECC_CURVE_INVALID || ecurve == NULL)
+	if (group == NULL)
+		return gnutls_assert_val(GNUTLS_E_ECC_NO_SUPPORTED_CURVES);
+
+	 ecurve = _gnutls_ecc_curve_get_params(group->curve);
+	if (ecurve == NULL)
 		return gnutls_assert_val(GNUTLS_E_ECC_NO_SUPPORTED_CURVES);
 
 	DECR_LEN(data_size, 1);
@@ -155,14 +160,14 @@ int _gnutls_proc_ecdh_common_client_kx(gnutls_session_t session,
 
 	if (ecurve->pk == GNUTLS_PK_EC) {
 		ret =
-		    _gnutls_ecc_ansi_x963_import(&data[i], point_size,
+		    _gnutls_ecc_ansi_x962_import(&data[i], point_size,
 					 &session->key.ecdh_x,
 					 &session->key.ecdh_y);
 		if (ret < 0) {
 			gnutls_assert();
 			goto cleanup;
 		}
-	} else if (ecurve->pk == GNUTLS_PK_ECDHX) {
+	} else if (ecurve->pk == GNUTLS_PK_ECDH_X25519) {
 		if (ecurve->size != point_size)
 			return gnutls_assert_val(GNUTLS_E_RECEIVED_ILLEGAL_PARAMETER);
 
@@ -211,7 +216,7 @@ proc_ecdhe_client_kx(gnutls_session_t session,
 
 	return _gnutls_proc_ecdh_common_client_kx(session, data,
 						  _data_size,
-						  _gnutls_session_ecc_curve_get
+						  get_group
 						  (session), NULL);
 }
 
@@ -229,10 +234,14 @@ _gnutls_gen_ecdh_common_client_kx_int(gnutls_session_t session,
 {
 	int ret;
 	gnutls_datum_t out;
-	int curve = _gnutls_session_ecc_curve_get(session);
-	const gnutls_ecc_curve_entry_st *ecurve = _gnutls_ecc_curve_get_params(curve);
+	const gnutls_group_entry_st *group = get_group(session);
+	const gnutls_ecc_curve_entry_st *ecurve;
 	int pk;
 
+	if (group == NULL)
+		return gnutls_assert_val(GNUTLS_E_ECC_NO_SUPPORTED_CURVES);
+
+	 ecurve = _gnutls_ecc_curve_get_params(group->curve);
 	if (ecurve == NULL)
 		return gnutls_assert_val(GNUTLS_E_ECC_NO_SUPPORTED_CURVES);
 
@@ -240,15 +249,14 @@ _gnutls_gen_ecdh_common_client_kx_int(gnutls_session_t session,
 
 	/* generate temporal key */
 	ret =
-	    _gnutls_pk_generate_keys(pk, curve,
-				     &session->key.ecdh_params);
+	    _gnutls_pk_generate_keys(pk, ecurve->id,
+				     &session->key.ecdh_params, 1);
 	if (ret < 0)
 		return gnutls_assert_val(ret);
 
-	ret = GNUTLS_E_RECEIVED_ILLEGAL_PARAMETER;
 	if (pk == GNUTLS_PK_EC) {
 		ret =
-		    _gnutls_ecc_ansi_x963_export(curve,
+		    _gnutls_ecc_ansi_x962_export(ecurve->id,
 						 session->key.ecdh_params.
 						 params[ECC_X] /* x */ ,
 						 session->key.ecdh_params.
@@ -268,7 +276,7 @@ _gnutls_gen_ecdh_common_client_kx_int(gnutls_session_t session,
 			gnutls_assert();
 			goto cleanup;
 		}
-	} else if (pk == GNUTLS_PK_ECDHX) {
+	} else if (pk == GNUTLS_PK_ECDH_X25519) {
 		ret =
 		    _gnutls_buffer_append_data_prefix(data, 8,
 					session->key.ecdh_params.raw_pub.data,
@@ -315,8 +323,9 @@ int
 _gnutls_proc_ecdh_common_server_kx(gnutls_session_t session,
 				   uint8_t * data, size_t _data_size)
 {
-	int i, ret, point_size;
-	gnutls_ecc_curve_t curve;
+	int i, ret;
+	unsigned point_size;
+	const gnutls_group_entry_st *group;
 	ssize_t data_size = _data_size;
 	const gnutls_ecc_curve_entry_st *ecurve;
 
@@ -331,27 +340,27 @@ _gnutls_proc_ecdh_common_server_kx(gnutls_session_t session,
 		return gnutls_assert_val(GNUTLS_E_ECC_NO_SUPPORTED_CURVES);
 
 	DECR_LEN(data_size, 2);
-	curve = _gnutls_tls_id_to_ecc_curve(_gnutls_read_uint16(&data[i]));
 
-	if (curve == GNUTLS_ECC_CURVE_INVALID) {
-		_gnutls_debug_log("received curve %u.%u\n", (unsigned)data[i], (unsigned)data[i+1]);
+	group = _gnutls_tls_id_to_group(_gnutls_read_uint16(&data[i]));
+	if (group == NULL || group->curve == 0) {
+		_gnutls_debug_log("received unknown curve %u.%u\n", (unsigned)data[i], (unsigned)data[i+1]);
+		return gnutls_assert_val(GNUTLS_E_RECEIVED_ILLEGAL_PARAMETER);
 	} else {
-		_gnutls_debug_log("received curve %s\n", gnutls_ecc_curve_get_name(curve));
+		_gnutls_debug_log("received curve %s\n", group->name);
 	}
 
 	i += 2;
 
-	ret = _gnutls_session_supports_ecc_curve(session, curve);
+	ret = _gnutls_session_supports_group(session, group->id);
 	if (ret < 0)
 		return gnutls_assert_val(ret);
 
-	ecurve = _gnutls_ecc_curve_get_params(curve);
+	ecurve = _gnutls_ecc_curve_get_params(group->curve);
 	if (ecurve == NULL) {
-		gnutls_assert();
 		return gnutls_assert_val(GNUTLS_E_RECEIVED_ILLEGAL_PARAMETER);
 	}
 
-	_gnutls_session_ecc_curve_set(session, curve);
+	_gnutls_session_group_set(session, group);
 
 	DECR_LEN(data_size, 1);
 	point_size = data[i];
@@ -361,13 +370,13 @@ _gnutls_proc_ecdh_common_server_kx(gnutls_session_t session,
 
 	if (ecurve->pk == GNUTLS_PK_EC) {
 		ret =
-		    _gnutls_ecc_ansi_x963_import(&data[i], point_size,
+		    _gnutls_ecc_ansi_x962_import(&data[i], point_size,
 						 &session->key.ecdh_x,
 						 &session->key.ecdh_y);
 		if (ret < 0)
 			return gnutls_assert_val(ret);
 
-	} else if (ecurve->pk == GNUTLS_PK_ECDHX) {
+	} else if (ecurve->pk == GNUTLS_PK_ECDH_X25519) {
 		if (ecurve->size != point_size)
 			return gnutls_assert_val(GNUTLS_E_RECEIVED_ILLEGAL_PARAMETER);
 
@@ -393,13 +402,13 @@ _gnutls_proc_ecdh_common_server_kx(gnutls_session_t session,
  * be inserted */
 int _gnutls_ecdh_common_print_server_kx(gnutls_session_t session,
 					gnutls_buffer_st * data,
-					gnutls_ecc_curve_t curve)
+					const gnutls_group_entry_st *group)
 {
 	uint8_t p;
-	int ret, pk;
+	int ret;
 	gnutls_datum_t out;
 
-	if (curve == GNUTLS_ECC_CURVE_INVALID)
+	if (group == NULL || group->curve == 0)
 		return gnutls_assert_val(GNUTLS_E_ECC_NO_SUPPORTED_CURVES);
 
 	/* just in case we are resuming a session */
@@ -416,24 +425,21 @@ int _gnutls_ecdh_common_print_server_kx(gnutls_session_t session,
 
 	ret =
 	    _gnutls_buffer_append_prefix(data, 16,
-					 _gnutls_ecc_curve_get_tls_id
-					 (curve));
+					 group->tls_id);
 	if (ret < 0)
 		return gnutls_assert_val(ret);
 
-	pk = gnutls_ecc_curve_get_pk(curve);
 
 	/* generate temporal key */
 	ret =
-	    _gnutls_pk_generate_keys(pk, curve,
-				     &session->key.ecdh_params);
+	    _gnutls_pk_generate_keys(group->pk, group->curve,
+				     &session->key.ecdh_params, 1);
 	if (ret < 0)
 		return gnutls_assert_val(ret);
 
-	ret = GNUTLS_E_RECEIVED_ILLEGAL_PARAMETER;
-	if (pk == GNUTLS_PK_EC) {
+	if (group->pk == GNUTLS_PK_EC) {
 		ret =
-		    _gnutls_ecc_ansi_x963_export(curve,
+		    _gnutls_ecc_ansi_x962_export(group->curve,
 						 session->key.ecdh_params.
 						 params[ECC_X] /* x */ ,
 						 session->key.ecdh_params.
@@ -449,15 +455,18 @@ int _gnutls_ecdh_common_print_server_kx(gnutls_session_t session,
 		if (ret < 0)
 			return gnutls_assert_val(ret);
 
-	} else if (pk == GNUTLS_PK_ECDHX) {
+	} else if (group->pk == GNUTLS_PK_ECDH_X25519) {
 		ret =
 			_gnutls_buffer_append_data_prefix(data, 8,
 					session->key.ecdh_params.raw_pub.data,
 					session->key.ecdh_params.raw_pub.size);
 		if (ret < 0)
 			return gnutls_assert_val(ret);
+	} else {
+		return gnutls_assert_val(GNUTLS_E_ECC_NO_SUPPORTED_CURVES);
 	}
 
+
 	return data->length;
 }
 
@@ -483,7 +492,7 @@ gen_ecdhe_server_kx(gnutls_session_t session, gnutls_buffer_st * data)
 
 	ret =
 	    _gnutls_ecdh_common_print_server_kx(session, data,
-						_gnutls_session_ecc_curve_get
+						get_group
 						(session));
 	if (ret < 0) {
 		gnutls_assert();
