@@ -224,6 +224,8 @@ int errno_to_gerr(int err, unsigned dtls)
 			return GNUTLS_E_LARGE_PACKET;
 		else
 			return GNUTLS_E_PUSH_ERROR;
+	case ECONNRESET:
+		return GNUTLS_E_PREMATURE_TERMINATION;
 	default:
 		gnutls_assert();
 		return GNUTLS_E_PUSH_ERROR;
@@ -540,8 +542,8 @@ _gnutls_io_read_buffered(gnutls_session_t session, size_t total,
 	size_t recvdata, readsize;
 
 	if (total > max_record_recv_size(session) || total == 0) {
-		gnutls_assert();	/* internal error */
-		return GNUTLS_E_INVALID_REQUEST;
+		gnutls_assert();
+		return GNUTLS_E_RECORD_OVERFLOW;
 	}
 
 	/* calculate the actual size, ie. get the minimum of the
@@ -953,13 +955,12 @@ parse_handshake_header(gnutls_session_t session, mbuffer_st * bufel,
 		return
 		    gnutls_assert_val(GNUTLS_E_UNEXPECTED_PACKET_LENGTH);
 
-	if (hsk->length > 0 && (hsk->start_offset > hsk->end_offset ||
+	if (hsk->length > 0 && (hsk->start_offset >= hsk->end_offset ||
 				hsk->end_offset - hsk->start_offset >=
 				data_size
-				|| hsk->end_offset >= hsk->length)) {
+				|| hsk->end_offset >= hsk->length))
 		return
 		    gnutls_assert_val(GNUTLS_E_UNEXPECTED_PACKET_LENGTH);
-	}
 	else if (hsk->length == 0 && hsk->end_offset != 0
 		 && hsk->start_offset != 0)
 		return
@@ -1077,12 +1078,12 @@ static int merge_handshake_packet(gnutls_session_t session,
 inline static int cmp_hsk_types(gnutls_handshake_description_t expected,
 				gnutls_handshake_description_t recvd)
 {
+	if ((expected != GNUTLS_HANDSHAKE_CLIENT_HELLO
 #ifdef ENABLE_SSL2
-	if (expected == GNUTLS_HANDSHAKE_CLIENT_HELLO
-	     && recvd == GNUTLS_HANDSHAKE_CLIENT_HELLO_V2)
-		return 1;
+	     || recvd != GNUTLS_HANDSHAKE_CLIENT_HELLO_V2
 #endif
-	if (expected != recvd)
+	     )
+	    && (expected != recvd))
 		return 0;
 
 	return 1;
