@@ -1,6 +1,6 @@
 /*
  * Copyright (C) 2001-2016 Free Software Foundation, Inc.
- * Copyright (C) 2015-2016 Red Hat, Inc.
+ * Copyright (C) 2015-2017 Red Hat, Inc.
  *
  * Author: Nikos Mavrogiannopoulos
  *
@@ -40,9 +40,6 @@
 #include <str_array.h>
 #include <x509/verify-high.h>
 #include "x509/x509_int.h"
-#ifdef ENABLE_OPENPGP
-#include "openpgp/openpgp.h"
-#endif
 #include "dh.h"
 
 /**
@@ -65,17 +62,13 @@ void gnutls_certificate_free_keys(gnutls_certificate_credentials_t sc)
 		gnutls_free(sc->certs[i].cert_list);
 		gnutls_free(sc->certs[i].ocsp_response_file);
 		_gnutls_str_array_clear(&sc->certs[i].names);
+		gnutls_privkey_deinit(sc->certs[i].pkey);
 	}
 
 	gnutls_free(sc->certs);
+	gnutls_free(sc->sorted_cert_idx);
 	sc->certs = NULL;
-
-	for (i = 0; i < sc->ncerts; i++) {
-		gnutls_privkey_deinit(sc->pkey[i]);
-	}
-
-	gnutls_free(sc->pkey);
-	sc->pkey = NULL;
+	sc->sorted_cert_idx = NULL;
 
 	sc->ncerts = 0;
 }
@@ -202,9 +195,7 @@ gnutls_certificate_free_credentials(gnutls_certificate_credentials_t sc)
 	gnutls_x509_trust_list_deinit(sc->tlist, 1);
 	gnutls_certificate_free_keys(sc);
 	memset(sc->pin_tmp, 0, sizeof(sc->pin_tmp));
-#ifdef ENABLE_OPENPGP
-	gnutls_openpgp_keyring_deinit(sc->keyring);
-#endif
+
 	if (sc->deinit_dh_params) {
 		gnutls_dh_params_deinit(sc->dh_params);
 	}
@@ -245,112 +236,6 @@ gnutls_certificate_allocate_credentials(gnutls_certificate_credentials_t *
 	return 0;
 }
 
-/* Returns 0 if it's ok to use the gnutls_kx_algorithm_t with this 
- * certificate (uses the KeyUsage field). 
- */
-static int
-check_key_usage(const gnutls_pcert_st * cert,
-		gnutls_kx_algorithm_t alg)
-{
-	unsigned int key_usage = 0;
-	int encipher_type;
-
-	if (cert == NULL || alg == GNUTLS_KX_UNKNOWN) {
-		gnutls_assert();
-		return GNUTLS_E_INTERNAL_ERROR;
-	}
-
-	if (_gnutls_map_kx_get_cred(alg, 1) == GNUTLS_CRD_CERTIFICATE ||
-	    _gnutls_map_kx_get_cred(alg, 0) == GNUTLS_CRD_CERTIFICATE) {
-
-		gnutls_pubkey_get_key_usage(cert->pubkey, &key_usage);
-
-		encipher_type = _gnutls_kx_encipher_type(alg);
-
-		if (key_usage != 0 && encipher_type != CIPHER_IGN) {
-			/* If key_usage has been set in the certificate
-			 */
-
-			if (encipher_type == CIPHER_ENCRYPT) {
-				/* If the key exchange method requires an encipher
-				 * type algorithm, and key's usage does not permit
-				 * encipherment, then fail.
-				 */
-				if (!(key_usage & GNUTLS_KEY_KEY_ENCIPHERMENT)) {
-					gnutls_assert();
-					return
-					    GNUTLS_E_KEY_USAGE_VIOLATION;
-				}
-			}
-
-			if (encipher_type == CIPHER_SIGN) {
-				/* The same as above, but for sign only keys
-				 */
-				if (!(key_usage & GNUTLS_KEY_DIGITAL_SIGNATURE)) {
-					gnutls_assert();
-					return
-					    GNUTLS_E_KEY_USAGE_VIOLATION;
-				}
-			}
-		}
-	}
-	return 0;
-}
-
-
-/* returns the KX algorithms that are supported by a
- * certificate. (Eg a certificate with RSA params, supports
- * GNUTLS_KX_RSA algorithm).
- * This function also uses the KeyUsage field of the certificate
- * extensions in order to disable unneeded algorithms.
- */
-int
-_gnutls_selected_cert_supported_kx(gnutls_session_t session,
-				   gnutls_kx_algorithm_t * alg,
-				   int *alg_size)
-{
-	unsigned kx;
-	gnutls_pk_algorithm_t pk, cert_pk;
-	gnutls_pcert_st *cert;
-	int i;
-
-	if (session->internals.selected_cert_list_length == 0) {
-		*alg_size = 0;
-		return 0;
-	}
-
-	cert = &session->internals.selected_cert_list[0];
-	cert_pk = gnutls_pubkey_get_pk_algorithm(cert->pubkey, NULL);
-	i = 0;
-
-	for (kx = 0; kx < MAX_ALGOS; kx++) {
-		pk = _gnutls_map_kx_get_pk(kx);
-		if (pk == cert_pk) {
-			/* then check key usage */
-			if (check_key_usage(cert, kx) == 0 ||
-			    unlikely(session->internals.priorities.allow_server_key_usage_violation != 0)) {
-				alg[i] = kx;
-				i++;
-
-				if (i > *alg_size)
-					return
-					    gnutls_assert_val
-					    (GNUTLS_E_INTERNAL_ERROR);
-			}
-		}
-	}
-
-	if (i == 0) {
-		gnutls_assert();
-		return GNUTLS_E_INVALID_REQUEST;
-	}
-
-	*alg_size = i;
-
-	return 0;
-}
-
-
 /**
  * gnutls_certificate_server_set_request:
  * @session: is a #gnutls_session_t type.
@@ -565,80 +450,13 @@ _gnutls_x509_get_raw_crt_expiration_time(const gnutls_datum_t * cert)
 	return result;
 }
 
-#ifdef ENABLE_OPENPGP
-/*-
- * _gnutls_openpgp_crt_verify_peers - return the peer's certificate status
- * @session: is a gnutls session
- *
- * This function will try to verify the peer's certificate and return its status (TRUSTED, INVALID etc.).
- * Returns a negative error code in case of an error, or GNUTLS_E_NO_CERTIFICATE_FOUND if no certificate was sent.
- -*/
-static int
-_gnutls_openpgp_crt_verify_peers(gnutls_session_t session,
-				 gnutls_x509_subject_alt_name_t type,
-				 const char *hostname,
-				 unsigned int *status)
-{
-	cert_auth_info_t info;
-	gnutls_certificate_credentials_t cred;
-	int peer_certificate_list_size, ret;
-	unsigned int verify_flags;
-
-	CHECK_AUTH(GNUTLS_CRD_CERTIFICATE, GNUTLS_E_INVALID_REQUEST);
-
-	info = _gnutls_get_auth_info(session, GNUTLS_CRD_CERTIFICATE);
-	if (info == NULL)
-		return GNUTLS_E_INVALID_REQUEST;
-
-	cred = (gnutls_certificate_credentials_t)
-	    _gnutls_get_cred(session, GNUTLS_CRD_CERTIFICATE);
-	if (cred == NULL) {
-		gnutls_assert();
-		return GNUTLS_E_INSUFFICIENT_CREDENTIALS;
-	}
-
-	if (info->raw_certificate_list == NULL || info->ncerts == 0) {
-		gnutls_assert();
-		return GNUTLS_E_NO_CERTIFICATE_FOUND;
-	}
-
-	verify_flags = cred->verify_flags | session->internals.additional_verify_flags;
-
-	/* generate a list of gnutls_certs based on the auth info
-	 * raw certs.
-	 */
-	peer_certificate_list_size = info->ncerts;
-
-	if (peer_certificate_list_size != 1) {
-		gnutls_assert();
-		return GNUTLS_E_INTERNAL_ERROR;
-	}
-
-	/* Verify certificate 
-	 */
-	ret =
-	    _gnutls_openpgp_verify_key(cred, type, hostname,
-				       &info->raw_certificate_list[0],
-				       peer_certificate_list_size,
-				       verify_flags,
-				       status);
-
-	if (ret < 0) {
-		gnutls_assert();
-		return ret;
-	}
-
-	return 0;
-}
-#endif
-
 /**
  * gnutls_certificate_verify_peers2:
  * @session: is a gnutls session
  * @status: is the output of the verification
  *
  * This function will verify the peer's certificate and store
- * the status in the @status variable as a bitwise or'd gnutls_certificate_status_t
+ * the status in the @status variable as a bitwise OR of gnutls_certificate_status_t
  * values or zero if the certificate is trusted. Note that value in @status
  * is set only when the return value of this function is success (i.e, failure 
  * to trust a certificate does not imply a negative return value).
@@ -674,7 +492,7 @@ gnutls_certificate_verify_peers2(gnutls_session_t session,
  * @status: is the output of the verification
  *
  * This function will verify the peer's certificate and store the
- * status in the @status variable as a bitwise or'd gnutls_certificate_status_t
+ * the status in the @status variable as a bitwise OR of gnutls_certificate_status_t
  * values or zero if the certificate is trusted. Note that value in @status
  * is set only when the return value of this function is success (i.e, failure 
  * to trust a certificate does not imply a negative return value).
@@ -717,7 +535,7 @@ gnutls_typed_vdata_st data;
  * @status: is the output of the verification
  *
  * This function will verify the peer's certificate and store the
- * status in the @status variable as a bitwise or'd gnutls_certificate_status_t
+ * the status in the @status variable as a bitwise OR of gnutls_certificate_status_t
  * values or zero if the certificate is trusted. Note that value in @status
  * is set only when the return value of this function is success (i.e, failure 
  * to trust a certificate does not imply a negative return value).
@@ -749,8 +567,6 @@ gnutls_certificate_verify_peers(gnutls_session_t session,
 				unsigned int *status)
 {
 	cert_auth_info_t info;
-	const char *hostname = NULL;
-	unsigned i, type = 0;
 
 	CHECK_AUTH(GNUTLS_CRD_CERTIFICATE, GNUTLS_E_INVALID_REQUEST);
 
@@ -767,24 +583,6 @@ gnutls_certificate_verify_peers(gnutls_session_t session,
 	case GNUTLS_CRT_X509:
 		return _gnutls_x509_cert_verify_peers(session, data, elements,
 						      status);
-#ifdef ENABLE_OPENPGP
-	case GNUTLS_CRT_OPENPGP:
-		for (i=0;i<elements;i++) {
-			if (data[i].type == GNUTLS_DT_DNS_HOSTNAME) {
-				hostname = (char*)data[i].data;
-				type = GNUTLS_SAN_DNSNAME;
-				break;
-			} else if (data[i].type == GNUTLS_DT_RFC822NAME) {
-				hostname = (char*)data[i].data;
-				type = GNUTLS_SAN_RFC822NAME;
-				break;
-			}
-		}
-		return _gnutls_openpgp_crt_verify_peers(session,
-							type,
-							hostname,
-							status);
-#endif
 	default:
 		return GNUTLS_E_INVALID_REQUEST;
 	}
@@ -822,12 +620,6 @@ time_t gnutls_certificate_expiration_time_peers(gnutls_session_t session)
 		    _gnutls_x509_get_raw_crt_expiration_time(&info->
 							     raw_certificate_list
 							     [0]);
-#ifdef ENABLE_OPENPGP
-	case GNUTLS_CRT_OPENPGP:
-		return
-		    _gnutls_openpgp_get_raw_key_expiration_time
-		    (&info->raw_certificate_list[0]);
-#endif
 	default:
 		return (time_t) - 1;
 	}
@@ -838,7 +630,6 @@ time_t gnutls_certificate_expiration_time_peers(gnutls_session_t session)
  * @session: is a gnutls session
  *
  * This function will return the peer's certificate activation time.
- * This is the creation time for openpgp keys.
  *
  * Returns: (time_t)-1 on error.
  *
@@ -866,13 +657,6 @@ time_t gnutls_certificate_activation_time_peers(gnutls_session_t session)
 		    _gnutls_x509_get_raw_crt_activation_time(&info->
 							     raw_certificate_list
 							     [0]);
-#ifdef ENABLE_OPENPGP
-	case GNUTLS_CRT_OPENPGP:
-		return
-		    _gnutls_openpgp_get_raw_key_creation_time(&info->
-							      raw_certificate_list
-							      [0]);
-#endif
 	default:
 		return (time_t) - 1;
 	}
@@ -887,6 +671,7 @@ int _gnutls_check_key_cert_match(gnutls_certificate_credentials_t res)
 	gnutls_datum_t test = {(void*)TEST_TEXT, sizeof(TEST_TEXT)-1};
 	gnutls_datum_t sig = {NULL, 0};
 	int pk, pk2, ret;
+	unsigned sign_algo;
 
 	if (res->flags & GNUTLS_CERTIFICATE_SKIP_KEY_CERT_MATCH)
 		return 0;
@@ -895,19 +680,31 @@ int _gnutls_check_key_cert_match(gnutls_certificate_credentials_t res)
 	    gnutls_pubkey_get_pk_algorithm(res->certs[res->ncerts - 1].
 					   cert_list[0].pubkey, NULL);
 	pk2 =
-	    gnutls_privkey_get_pk_algorithm(res->pkey[res->ncerts - 1],
+	    gnutls_privkey_get_pk_algorithm(res->certs[res->ncerts - 1].pkey,
 					    NULL);
 
-	if (pk2 != pk) {
+	if (GNUTLS_PK_IS_RSA(pk) && GNUTLS_PK_IS_RSA(pk2)) {
+		if (pk2 == GNUTLS_PK_RSA_PSS && pk == GNUTLS_PK_RSA) {
+			_gnutls_debug_log("you cannot mix an RSA-PSS key with an RSA certificate\n");
+			return GNUTLS_E_CERTIFICATE_KEY_MISMATCH;
+		}
+
+		if (pk2 == GNUTLS_PK_RSA_PSS || pk == GNUTLS_PK_RSA_PSS)
+			pk = GNUTLS_PK_RSA_PSS;
+	} else if (pk2 != pk) {
 		gnutls_assert();
+		_gnutls_debug_log("key is %s, certificate is %s\n", gnutls_pk_get_name(pk2),
+			gnutls_pk_get_name(pk));
 		return GNUTLS_E_CERTIFICATE_KEY_MISMATCH;
 	}
 
+	sign_algo = gnutls_pk_to_sign(pk, GNUTLS_DIG_SHA256);
+
 	/* now check if keys really match. We use the sign/verify approach
 	 * because we cannot always obtain the parameters from the abstract
 	 * keys (e.g. PKCS #11). */
-	ret = gnutls_privkey_sign_data(res->pkey[res->ncerts - 1],
-		GNUTLS_DIG_SHA256, 0, &test, &sig);
+	ret = gnutls_privkey_sign_data2(res->certs[res->ncerts - 1].pkey,
+		sign_algo, 0, &test, &sig);
 	if (ret < 0) {
 		/* for some reason we couldn't sign that. That shouldn't have
 		 * happened, but since it did, report the issue and do not
@@ -917,8 +714,8 @@ int _gnutls_check_key_cert_match(gnutls_certificate_credentials_t res)
 	}
 
 	ret = gnutls_pubkey_verify_data2(res->certs[res->ncerts - 1].cert_list[0].pubkey,
-		gnutls_pk_to_sign(pk, GNUTLS_DIG_SHA256),
-		GNUTLS_VERIFY_ALLOW_BROKEN, &test, &sig);
+					 sign_algo,
+					 GNUTLS_VERIFY_ALLOW_BROKEN, &test, &sig);
 
 	gnutls_free(sig.data);
 
@@ -995,16 +792,6 @@ gnutls_certificate_verification_status_print(unsigned int status,
 			_gnutls_buffer_append_str(&str,
 						  _
 						  ("The certificate issuer is not a CA. "));
-	} else if (type == GNUTLS_CRT_OPENPGP) {
-		if (status & GNUTLS_CERT_SIGNER_NOT_FOUND)
-			_gnutls_buffer_append_str(&str,
-						  _
-						  ("Could not find a signer of the certificate. "));
-
-		if (status & GNUTLS_CERT_REVOKED)
-			_gnutls_buffer_append_str(&str,
-						  _
-						  ("The certificate is revoked. "));
 	}
 
 	if (status & GNUTLS_CERT_INSECURE_ALGORITHM)
@@ -1052,6 +839,11 @@ gnutls_certificate_verification_status_print(unsigned int status,
 					  _
 					  ("The received OCSP status response is invalid. "));
 
+	if (status & GNUTLS_CERT_UNKNOWN_CRIT_EXTENSIONS)
+		_gnutls_buffer_append_str(&str,
+					  _
+					  ("The certificate contains an unknown critical extension. "));
+
 	return _gnutls_buffer_to_datum(&str, out, 1);
 }
 
@@ -1079,6 +871,7 @@ gnutls_certificate_set_dh_params(gnutls_certificate_credentials_t res,
 	}
 
 	res->dh_params = dh_params;
+	res->dh_sec_param = gnutls_pk_bits_to_sec_param(GNUTLS_PK_DH, _gnutls_mpi_get_nbits(dh_params->params[0]));
 }
 
 
@@ -1101,19 +894,7 @@ int
 gnutls_certificate_set_known_dh_params(gnutls_certificate_credentials_t res,
 				       gnutls_sec_param_t sec_param)
 {
-	int ret;
-
-	if (res->deinit_dh_params) {
-		res->deinit_dh_params = 0;
-		gnutls_dh_params_deinit(res->dh_params);
-		res->dh_params = NULL;
-	}
-
-	ret = _gnutls_set_cred_dh_params(&res->dh_params, sec_param);
-	if (ret < 0)
-		return gnutls_assert_val(ret);
-
-	res->deinit_dh_params = 1;
+	res->dh_sec_param = sec_param;
 
 	return 0;
 }
