@@ -1,5 +1,6 @@
 /*
  * Copyright (C) 2001-2012 Free Software Foundation, Inc.
+ * Copyright (C) 2017 Red Hat, Inc.
  *
  * Author: Nikos Mavrogiannopoulos
  *
@@ -34,9 +35,7 @@
 #include <state.h>
 #include <extensions.h>
 #include <buffers.h>
-
-static int
-_gnutls_set_kx(gnutls_session_t session, gnutls_kx_algorithm_t algo);
+#include "dtls.h"
 
 static const char keyexp[] = "key expansion";
 static const int keyexp_length = sizeof(keyexp) - 1;
@@ -218,28 +217,19 @@ _gnutls_init_record_state(record_parameters_st * params,
 	if (ret < 0 && params->cipher->id != GNUTLS_CIPHER_NULL)
 		return gnutls_assert_val(ret);
 
-	ret =
-	    _gnutls_comp_init(&state->compression_state,
-			      params->compression_algorithm,
-			      read /*1==decompress */ );
-
-	if (ret < 0)
-		return gnutls_assert_val(ret);
-
 	return 0;
 }
 
 int
-_gnutls_epoch_set_cipher_suite(gnutls_session_t session,
-			       int epoch_rel, const uint8_t suite[2])
+_gnutls_set_cipher_suite2(gnutls_session_t session,
+			  const gnutls_cipher_suite_entry_st *cs)
 {
 	const cipher_entry_st *cipher_algo;
 	const mac_entry_st *mac_algo;
 	record_parameters_st *params;
-	const gnutls_cipher_suite_entry_st *cs;
 	int ret;
 
-	ret = _gnutls_epoch_get(session, epoch_rel, &params);
+	ret = _gnutls_epoch_get(session, EPOCH_NEXT, &params);
 	if (ret < 0)
 		return gnutls_assert_val(ret);
 
@@ -247,10 +237,6 @@ _gnutls_epoch_set_cipher_suite(gnutls_session_t session,
 	    || params->cipher != NULL || params->mac != NULL)
 		return gnutls_assert_val(GNUTLS_E_INTERNAL_ERROR);
 
-	cs = ciphersuite_to_entry(suite);
-	if (cs == NULL)
-		return gnutls_assert_val(GNUTLS_E_INTERNAL_ERROR);
-
 	cipher_algo = cipher_to_entry(cs->block_algorithm);
 	mac_algo = mac_to_entry(cs->mac_algorithm);
 
@@ -258,44 +244,22 @@ _gnutls_epoch_set_cipher_suite(gnutls_session_t session,
 	    || _gnutls_mac_is_ok(mac_algo) == 0)
 		return gnutls_assert_val(GNUTLS_E_UNWANTED_ALGORITHM);
 
-	if (_gnutls_cipher_priority(session, cipher_algo->id) < 0)
-		return gnutls_assert_val(GNUTLS_E_UNWANTED_ALGORITHM);
-
-	if (_gnutls_mac_priority(session, mac_algo->id) < 0)
-		return gnutls_assert_val(GNUTLS_E_UNWANTED_ALGORITHM);
+	if (_gnutls_version_has_selectable_prf(get_version(session))) {
+		if (cs->prf == GNUTLS_MAC_UNKNOWN ||
+		    _gnutls_mac_is_ok(mac_to_entry(cs->prf)) == 0)
+			return gnutls_assert_val(GNUTLS_E_UNWANTED_ALGORITHM);
+		session->security_parameters.prf_mac = cs->prf;
+	} else {
+		session->security_parameters.prf_mac = GNUTLS_MAC_MD5_SHA1;
+	}
 
+	session->security_parameters.cs = cs;
 	params->cipher = cipher_algo;
 	params->mac = mac_algo;
 
 	return 0;
 }
 
-int
-_gnutls_epoch_set_compression(gnutls_session_t session,
-			      int epoch_rel,
-			      gnutls_compression_method_t comp_algo)
-{
-	record_parameters_st *params;
-	int ret;
-
-	ret = _gnutls_epoch_get(session, epoch_rel, &params);
-	if (ret < 0)
-		return gnutls_assert_val(ret);
-
-	if (params->initialized
-	    || params->compression_algorithm != GNUTLS_COMP_UNKNOWN)
-		return gnutls_assert_val(GNUTLS_E_INTERNAL_ERROR);
-
-	if (_gnutls_compression_is_ok(comp_algo) != 0)
-		return
-		    gnutls_assert_val
-		    (GNUTLS_E_UNKNOWN_COMPRESSION_ALGORITHM);
-
-	params->compression_algorithm = comp_algo;
-
-	return 0;
-}
-
 void
 _gnutls_epoch_set_null_algos(gnutls_session_t session,
 			     record_parameters_st * params)
@@ -310,7 +274,6 @@ _gnutls_epoch_set_null_algos(gnutls_session_t session,
 
 	params->cipher = cipher_to_entry(GNUTLS_CIPHER_NULL);
 	params->mac = mac_to_entry(GNUTLS_MAC_NULL);
-	params->compression_algorithm = GNUTLS_COMP_NULL;
 	params->initialized = 1;
 }
 
@@ -319,7 +282,6 @@ int _gnutls_epoch_set_keys(gnutls_session_t session, uint16_t epoch)
 	int hash_size;
 	int IV_size;
 	int key_size;
-	gnutls_compression_method_t comp_algo;
 	record_parameters_st *params;
 	int ret;
 	const version_entry_st *ver = get_version(session);
@@ -337,23 +299,10 @@ int _gnutls_epoch_set_keys(gnutls_session_t session, uint16_t epoch)
 	_gnutls_record_log
 	    ("REC[%p]: Initializing epoch #%u\n", session, params->epoch);
 
-	comp_algo = params->compression_algorithm;
-
 	if (_gnutls_cipher_is_ok(params->cipher) == 0
 	    || _gnutls_mac_is_ok(params->mac) == 0)
 		return gnutls_assert_val(GNUTLS_E_UNWANTED_ALGORITHM);
 
-	if (_gnutls_cipher_priority(session, params->cipher->id) < 0)
-		return gnutls_assert_val(GNUTLS_E_UNWANTED_ALGORITHM);
-
-	if (_gnutls_mac_priority(session, params->mac->id) < 0)
-		return gnutls_assert_val(GNUTLS_E_UNWANTED_ALGORITHM);
-
-	if (_gnutls_compression_is_ok(comp_algo) != 0)
-		return
-		    gnutls_assert_val
-		    (GNUTLS_E_UNKNOWN_COMPRESSION_ALGORITHM);
-
 	if (!_gnutls_version_has_explicit_iv(ver) &&
 	    _gnutls_cipher_type(params->cipher) == CIPHER_BLOCK) {
 		IV_size = _gnutls_cipher_get_iv_size(params->cipher);
@@ -378,7 +327,12 @@ int _gnutls_epoch_set_keys(gnutls_session_t session, uint16_t epoch)
 	if (ret < 0)
 		return gnutls_assert_val(ret);
 
-	params->record_sw_size = 0;
+	session->internals.max_recv_size = _gnutls_record_overhead(params->cipher, params->mac, params->etm, 0);
+	session->internals.max_recv_size += session->security_parameters.max_record_recv_size + RECORD_HEADER_SIZE(session);
+	if (session->internals.allow_large_records != 0)
+		session->internals.max_recv_size += EXTRA_COMP_SIZE;
+
+	_dtls_reset_window(params);
 
 	_gnutls_record_log("REC[%p]: Epoch #%u ready\n", session,
 			   params->epoch);
@@ -389,15 +343,13 @@ int _gnutls_epoch_set_keys(gnutls_session_t session, uint16_t epoch)
 
 
 #define CPY_COMMON dst->entity = src->entity; \
-	dst->kx_algorithm = src->kx_algorithm; \
-	memcpy( dst->cipher_suite, src->cipher_suite, 2); \
+	dst->cs = src->cs; \
 	memcpy( dst->master_secret, src->master_secret, GNUTLS_MASTER_SIZE); \
 	memcpy( dst->client_random, src->client_random, GNUTLS_RANDOM_SIZE); \
 	memcpy( dst->server_random, src->server_random, GNUTLS_RANDOM_SIZE); \
 	memcpy( dst->session_id, src->session_id, GNUTLS_MAX_SESSION_ID_SIZE); \
 	dst->session_id_size = src->session_id_size; \
 	dst->cert_type = src->cert_type; \
-	dst->compression_method = src->compression_method; \
 	dst->timestamp = src->timestamp; \
 	dst->ext_master_secret = src->ext_master_secret; \
 	dst->etm = src->etm; \
@@ -432,18 +384,6 @@ int _gnutls_connection_state_init(gnutls_session_t session)
 	return 0;
 }
 
-int _gnutls_epoch_get_compression(gnutls_session_t session, int epoch)
-{
-	record_parameters_st *params;
-	int ret;
-
-	ret = _gnutls_epoch_get(session, epoch, &params);
-	if (ret < 0)
-		return GNUTLS_COMP_UNKNOWN;
-
-	return params->compression_algorithm;
-}
-
 /* Initializes the read connection session
  * (read encrypted data)
  */
@@ -456,14 +396,8 @@ int _gnutls_read_connection_state_init(gnutls_session_t session)
 	/* Update internals from CipherSuite selected.
 	 * If we are resuming just copy the connection session
 	 */
-	if (session->internals.resumed == RESUME_FALSE) {
-		ret = _gnutls_set_kx(session,
-				     _gnutls_cipher_suite_get_kx_algo
-				     (session->security_parameters.
-				      cipher_suite));
-		if (ret < 0)
-			return ret;
-	} else if (session->security_parameters.entity == GNUTLS_CLIENT)
+	if (session->internals.resumed != RESUME_FALSE &&
+	    session->security_parameters.entity == GNUTLS_CLIENT)
 		_gnutls_set_resumed_parameters(session);
 
 	ret = _gnutls_epoch_set_keys(session, epoch_next);
@@ -472,8 +406,7 @@ int _gnutls_read_connection_state_init(gnutls_session_t session)
 
 	_gnutls_handshake_log("HSK[%p]: Cipher Suite: %s\n",
 			      session,
-			      _gnutls_cipher_suite_get_name
-			      (session->security_parameters.cipher_suite));
+			      session->security_parameters.cs->name);
 
 	session->security_parameters.epoch_read = epoch_next;
 
@@ -494,14 +427,8 @@ int _gnutls_write_connection_state_init(gnutls_session_t session)
 /* Update internals from CipherSuite selected.
  * If we are resuming just copy the connection session
  */
-	if (session->internals.resumed == RESUME_FALSE) {
-		ret = _gnutls_set_kx(session,
-				     _gnutls_cipher_suite_get_kx_algo
-				     (session->security_parameters.
-				      cipher_suite));
-		if (ret < 0)
-			return ret;
-	} else if (session->security_parameters.entity == GNUTLS_SERVER)
+	if (session->internals.resumed != RESUME_FALSE &&
+	    session->security_parameters.entity == GNUTLS_SERVER)
 		_gnutls_set_resumed_parameters(session);
 
 	ret = _gnutls_epoch_set_keys(session, epoch_next);
@@ -509,8 +436,7 @@ int _gnutls_write_connection_state_init(gnutls_session_t session)
 		return gnutls_assert_val(ret);
 
 	_gnutls_handshake_log("HSK[%p]: Cipher Suite: %s\n", session,
-			      _gnutls_cipher_suite_get_name
-			      (session->security_parameters.cipher_suite));
+			      session->security_parameters.cs->name);
 
 	_gnutls_handshake_log
 	    ("HSK[%p]: Initializing internal [write] cipher sessions\n",
@@ -521,23 +447,6 @@ int _gnutls_write_connection_state_init(gnutls_session_t session)
 	return 0;
 }
 
-/* Sets the specified kx algorithm into pending session
- */
-static int
-_gnutls_set_kx(gnutls_session_t session, gnutls_kx_algorithm_t algo)
-{
-
-	if (_gnutls_kx_is_ok(algo) == 0) {
-		session->security_parameters.kx_algorithm = algo;
-	} else
-		return gnutls_assert_val(GNUTLS_E_INTERNAL_ERROR);
-
-	if (_gnutls_kx_priority(session, algo) < 0)
-		return gnutls_assert_val(GNUTLS_E_UNWANTED_ALGORITHM);
-
-	return 0;
-}
-
 static inline int
 epoch_resolve(gnutls_session_t session,
 	      unsigned int epoch_rel, uint16_t * epoch_out)
@@ -628,7 +537,6 @@ _gnutls_epoch_alloc(gnutls_session_t session, uint16_t epoch,
 	(*slot)->epoch = epoch;
 	(*slot)->cipher = NULL;
 	(*slot)->mac = NULL;
-	(*slot)->compression_algorithm = GNUTLS_COMP_UNKNOWN;
 
 	if (IS_DTLS(session))
 		_gnutls_write_uint16(epoch,
@@ -722,9 +630,6 @@ static inline void free_record_state(record_state_st * state, int d)
 	_gnutls_free_datum(&state->key);
 
 	_gnutls_auth_cipher_deinit(&state->cipher_state);
-
-	if (state->compression_state.handle != NULL)
-		_gnutls_comp_deinit(&state->compression_state, d);
 }
 
 void
