@@ -1,5 +1,6 @@
 /*
- * Copyright (C) 2002-2012 Free Software Foundation, Inc.
+ * Copyright (C) 2002-2016 Free Software Foundation, Inc.
+ * Copyright (C) 2015-2017 Red Hat, Inc.
  *
  * Author: Nikos Mavrogiannopoulos
  *
@@ -42,11 +43,11 @@ static int _gnutls_signature_algorithm_send_params(gnutls_session_t
 						   session,
 						   gnutls_buffer_st *
 						   extdata);
-static void signature_algorithms_deinit_data(extension_priv_data_t priv);
-static int signature_algorithms_pack(extension_priv_data_t epriv,
+static void signature_algorithms_deinit_data(gnutls_ext_priv_data_t priv);
+static int signature_algorithms_pack(gnutls_ext_priv_data_t epriv,
 				     gnutls_buffer_st * ps);
 static int signature_algorithms_unpack(gnutls_buffer_st * ps,
-				       extension_priv_data_t * _priv);
+				       gnutls_ext_priv_data_t * _priv);
 
 const extension_entry_st ext_mod_sig = {
 	.name = "Signature Algorithms",
@@ -58,11 +59,12 @@ const extension_entry_st ext_mod_sig = {
 	.pack_func = signature_algorithms_pack,
 	.unpack_func = signature_algorithms_unpack,
 	.deinit_func = signature_algorithms_deinit_data,
+	.cannot_be_overriden = 1
 };
 
 typedef struct {
 	/* TLS 1.2 signature algorithms */
-	gnutls_sign_algorithm_t sign_algorithms[MAX_SIGNATURE_ALGORITHMS];
+	gnutls_sign_algorithm_t sign_algorithms[MAX_ALGOS];
 	uint16_t sign_algorithms_size;
 } sig_ext_st;
 
@@ -71,48 +73,40 @@ typedef struct {
  */
 int
 _gnutls_sign_algorithm_write_params(gnutls_session_t session,
-				    uint8_t * data, size_t max_data_size)
+				    gnutls_buffer_st * extdata)
 {
-	uint8_t *p = data, *len_p;
-	unsigned int len, i, j;
+	uint8_t *p;
+	unsigned int len, i;
 	const sign_algorithm_st *aid;
+	uint8_t buffer[MAX_ALGOS*2];
 
-	if (max_data_size <
-	    (session->internals.priorities.sign_algo.algorithms * 2) + 2) {
-		gnutls_assert();
-		return GNUTLS_E_SHORT_MEMORY_BUFFER;
-	}
-
+	p = buffer;
 	len = 0;
-	len_p = p;
 
-	p += 2;
+	for (i=0;i<session->internals.priorities->sigalg.size;i++) {
+		aid = &session->internals.priorities->sigalg.entry[i]->aid;
 
-	for (i = j = 0;
-	     j < session->internals.priorities.sign_algo.algorithms;
-	     i += 2, j++) {
-		aid =
-		    _gnutls_sign_to_tls_aid(session->internals.
-					    priorities.sign_algo.
-					    priority[j]);
-
-		if (aid == NULL)
+		if (HAVE_UNKNOWN_SIGAID(aid))
 			continue;
 
 		_gnutls_handshake_log
 		    ("EXT[%p]: sent signature algo (%d.%d) %s\n", session,
-		     aid->hash_algorithm, aid->sign_algorithm,
-		     gnutls_sign_get_name(session->internals.priorities.
-					  sign_algo.priority[j]));
-		*p = aid->hash_algorithm;
-		p++;
-		*p = aid->sign_algorithm;
-		p++;
+		     (int)aid->id[0], (int)aid->id[1],
+		     session->internals.priorities->sigalg.entry[i]->name);
+
 		len += 2;
+		if (unlikely(len >= sizeof(buffer))) {
+			len -= 2;
+			break;
+		}
+
+		*p = aid->id[0];
+		p++;
+		*p = aid->id[1];
+		p++;
 	}
 
-	_gnutls_write_uint16(len, len_p);
-	return len + 2;
+	return _gnutls_buffer_append_data_prefix(extdata, 16, buffer, len);
 }
 
 
@@ -125,9 +119,9 @@ _gnutls_sign_algorithm_parse_data(gnutls_session_t session,
 {
 	unsigned int sig, i;
 	sig_ext_st *priv;
-	extension_priv_data_t epriv;
+	gnutls_ext_priv_data_t epriv;
 
-	if (data_size % 2 != 0)
+	if (data_size == 0 || data_size % 2 != 0)
 		return gnutls_assert_val(GNUTLS_E_UNEXPECTED_PACKET_LENGTH);
 
 	priv = gnutls_calloc(1, sizeof(*priv));
@@ -139,23 +133,23 @@ _gnutls_sign_algorithm_parse_data(gnutls_session_t session,
 	for (i = 0; i < data_size; i += 2) {
 		sign_algorithm_st aid;
 
-		aid.hash_algorithm = data[i];
-		aid.sign_algorithm = data[i + 1];
+		aid.id[0] = data[i];
+		aid.id[1] = data[i + 1];
 
 		sig = _gnutls_tls_aid_to_sign(&aid);
 
 		_gnutls_handshake_log
 		    ("EXT[%p]: rcvd signature algo (%d.%d) %s\n", session,
-		     aid.hash_algorithm, aid.sign_algorithm,
+		     aid.id[0], aid.id[1],
 		     gnutls_sign_get_name(sig));
 
 		if (sig != GNUTLS_SIGN_UNKNOWN) {
+			if (priv->sign_algorithms_size ==
+			    MAX_ALGOS)
+				break;
 			priv->sign_algorithms[priv->
 					      sign_algorithms_size++] =
 			    sig;
-			if (priv->sign_algorithms_size ==
-			    MAX_SIGNATURE_ALGORITHMS)
-				break;
 		}
 	}
 
@@ -195,13 +189,16 @@ _gnutls_signature_algorithm_recv_params(gnutls_session_t session,
 	} else {
 		/* SERVER SIDE - we must check if the sent cert type is the right one
 		 */
-		if (data_size > 2) {
+		if (data_size >= 2) {
 			uint16_t len;
 
 			DECR_LEN(data_size, 2);
 			len = _gnutls_read_uint16(data);
 			DECR_LEN(data_size, len);
 
+			if (data_size > 0)
+				return gnutls_assert_val(GNUTLS_E_UNEXPECTED_PACKET_LENGTH);
+
 			ret =
 			    _gnutls_sign_algorithm_parse_data(session,
 							      data + 2,
@@ -210,6 +207,8 @@ _gnutls_signature_algorithm_recv_params(gnutls_session_t session,
 				gnutls_assert();
 				return ret;
 			}
+		} else {
+			return gnutls_assert_val(GNUTLS_E_UNEXPECTED_PACKET_LENGTH);
 		}
 	}
 
@@ -232,16 +231,9 @@ _gnutls_signature_algorithm_send_params(gnutls_session_t session,
 	/* this function sends the client extension data */
 	if (session->security_parameters.entity == GNUTLS_CLIENT
 	    && _gnutls_version_has_selectable_sighash(ver)) {
-		if (session->internals.priorities.sign_algo.algorithms > 0) {
-			uint8_t p[MAX_SIGN_ALGO_SIZE];
-
+		if (session->internals.priorities->sigalg.size > 0) {
 			ret =
-			    _gnutls_sign_algorithm_write_params(session, p,
-								sizeof(p));
-			if (ret < 0)
-				return gnutls_assert_val(ret);
-
-			ret = _gnutls_buffer_append_data(extdata, p, ret);
+			    _gnutls_sign_algorithm_write_params(session, extdata);
 			if (ret < 0)
 				return gnutls_assert_val(ret);
 
@@ -263,13 +255,15 @@ _gnutls_signature_algorithm_send_params(gnutls_session_t session,
  */
 gnutls_sign_algorithm_t
 _gnutls_session_get_sign_algo(gnutls_session_t session,
-			      gnutls_pcert_st * cert, unsigned client_cert)
+			      gnutls_pcert_st * cert,
+			      gnutls_privkey_t privkey,
+			      unsigned client_cert)
 {
 	unsigned i;
 	int ret;
 	const version_entry_st *ver = get_version(session);
 	sig_ext_st *priv;
-	extension_priv_data_t epriv;
+	gnutls_ext_priv_data_t epriv;
 	unsigned int cert_algo;
 
 	if (unlikely(ver == NULL))
@@ -283,10 +277,8 @@ _gnutls_session_get_sign_algo(gnutls_session_t session,
 					 &epriv);
 	priv = epriv;
 
-	if (ret < 0 || !_gnutls_version_has_selectable_sighash(ver)
-	    || priv->sign_algorithms_size == 0)
+	if (ret < 0 || !_gnutls_version_has_selectable_sighash(ver)) {
 		/* none set, allow SHA-1 only */
-	{
 		ret = gnutls_pk_to_sign(cert_algo, GNUTLS_DIG_SHA1);
 
 		if (!client_cert && _gnutls_session_sign_algo_enabled(session, ret) < 0)
@@ -295,8 +287,10 @@ _gnutls_session_get_sign_algo(gnutls_session_t session,
 	}
 
 	for (i = 0; i < priv->sign_algorithms_size; i++) {
-		if (gnutls_sign_get_pk_algorithm(priv->sign_algorithms[i])
-		    == cert_algo) {
+		if (_gnutls_privkey_compatible_with_sig(privkey, priv->sign_algorithms[i]) == 0)
+			continue;
+
+		if (gnutls_sign_supports_pk_algorithm(priv->sign_algorithms[i], cert_algo) != 0) {
 			if (_gnutls_pubkey_compatible_with_sig
 			    (session, cert->pubkey, ver,
 			     priv->sign_algorithms[i]) < 0)
@@ -332,9 +326,9 @@ _gnutls_session_sign_algo_enabled(gnutls_session_t session,
 		return 0;
 	}
 
-	for (i = 0; i < session->internals.priorities.sign_algo.algorithms;
+	for (i = 0; i < session->internals.priorities->sigalg.size;
 	     i++) {
-		if (session->internals.priorities.sign_algo.priority[i] ==
+		if (session->internals.priorities->sigalg.entry[i]->id ==
 		    sig) {
 			return 0;	/* ok */
 		}
@@ -343,13 +337,13 @@ _gnutls_session_sign_algo_enabled(gnutls_session_t session,
 	return GNUTLS_E_UNSUPPORTED_SIGNATURE_ALGORITHM;
 }
 
-static void signature_algorithms_deinit_data(extension_priv_data_t priv)
+static void signature_algorithms_deinit_data(gnutls_ext_priv_data_t priv)
 {
 	gnutls_free(priv);
 }
 
 static int
-signature_algorithms_pack(extension_priv_data_t epriv,
+signature_algorithms_pack(gnutls_ext_priv_data_t epriv,
 			  gnutls_buffer_st * ps)
 {
 	sig_ext_st *priv = epriv;
@@ -364,11 +358,11 @@ signature_algorithms_pack(extension_priv_data_t epriv,
 
 static int
 signature_algorithms_unpack(gnutls_buffer_st * ps,
-			    extension_priv_data_t * _priv)
+			    gnutls_ext_priv_data_t * _priv)
 {
 	sig_ext_st *priv;
 	int i, ret;
-	extension_priv_data_t epriv;
+	gnutls_ext_priv_data_t epriv;
 
 	priv = gnutls_calloc(1, sizeof(*priv));
 	if (priv == NULL) {
@@ -421,7 +415,7 @@ gnutls_sign_algorithm_get_requested(gnutls_session_t session,
 {
 	const version_entry_st *ver = get_version(session);
 	sig_ext_st *priv;
-	extension_priv_data_t epriv;
+	gnutls_ext_priv_data_t epriv;
 	int ret;
 
 	if (unlikely(ver == NULL))
