@@ -38,9 +38,6 @@
 #ifdef ENABLE_OCSP
 
 typedef struct {
-	gnutls_datum_t *responder_id;
-	size_t responder_id_size;
-	gnutls_datum_t request_extensions;
 	gnutls_datum_t response;
 
 	unsigned int expect_cstatus;
@@ -72,48 +69,17 @@ static int
 client_send(gnutls_session_t session,
 	    gnutls_buffer_st * extdata, status_request_ext_st * priv)
 {
-	int ret_len = 1 + 2;
+	const uint8_t data[5] = "\x01\x00\x00\x00\x00";
+	const int len = 5;
 	int ret;
-	size_t i;
 
-	ret = _gnutls_buffer_append_prefix(extdata, 8, 1);
+	/* We do not support setting either ResponderID or Extensions */
+
+	ret = _gnutls_buffer_append_data(extdata, data, len);
 	if (ret < 0)
 		return gnutls_assert_val(ret);
 
-	ret =
-	    _gnutls_buffer_append_prefix(extdata, 16,
-					 priv->responder_id_size);
-	if (ret < 0)
-		return gnutls_assert_val(ret);
-
-	for (i = 0; i < priv->responder_id_size; i++) {
-		if (priv->responder_id[i].size <= 0)
-			return gnutls_assert_val(GNUTLS_E_INVALID_REQUEST);
-
-		ret = _gnutls_buffer_append_data_prefix(extdata, 16,
-							priv->
-							responder_id[i].
-							data,
-							priv->
-							responder_id[i].
-							size);
-		if (ret < 0)
-			return gnutls_assert_val(ret);
-
-		ret_len += 2 + priv->responder_id[i].size;
-	}
-
-	ret = _gnutls_buffer_append_data_prefix(extdata, 16,
-						priv->request_extensions.
-						data,
-						priv->request_extensions.
-						size);
-	if (ret < 0)
-		return gnutls_assert_val(ret);
-
-	ret_len += 2 + priv->request_extensions.size;
-
-	return ret_len;
+	return len;
 }
 
 static int
@@ -231,7 +197,7 @@ static int
 _gnutls_status_request_send_params(gnutls_session_t session,
 				   gnutls_buffer_st * extdata)
 {
-	extension_priv_data_t epriv;
+	gnutls_ext_priv_data_t epriv;
 	status_request_ext_st *priv;
 	int ret;
 
@@ -267,7 +233,7 @@ static int
 _gnutls_status_request_recv_params(gnutls_session_t session,
 				   const uint8_t * data, size_t size)
 {
-	extension_priv_data_t epriv;
+	gnutls_ext_priv_data_t epriv;
 	status_request_ext_st *priv;
 	int ret;
 
@@ -288,21 +254,18 @@ _gnutls_status_request_recv_params(gnutls_session_t session,
 /**
  * gnutls_ocsp_status_request_enable_client:
  * @session: is a #gnutls_session_t type.
- * @responder_id: array with #gnutls_datum_t with DER data of responder id
- * @responder_id_size: number of members in @responder_id array
- * @extensions: a #gnutls_datum_t with DER encoded OCSP extensions
+ * @responder_id: ignored, must be %NULL
+ * @responder_id_size: ignored, must be zero
+ * @extensions: ignored, must be %NULL
  *
  * This function is to be used by clients to request OCSP response
  * from the server, using the "status_request" TLS extension.  Only
  * OCSP status type is supported.
  *
- * The @responder_id array, its containing elements as well as
- * the data of @extensions, must be allocated using gnutls_malloc(). They
- * will be deinitialized on session cleanup.
- *
- * Due to the difficult semantics of the @responder_id and @extensions
- * parameters, it is recommended to only call this function with these
- * parameters set to %NULL.
+ * Previous versions of GnuTLS supported setting @responder_id and
+ * @extensions fields, but due to the difficult semantics of the
+ * parameter usage, and other issues, this support was removed
+ * since 3.6.0 and these parameters must be set to %NULL.
  *
  * Returns: On success, %GNUTLS_E_SUCCESS (0) is returned,
  *   otherwise a negative error code is returned.
@@ -316,7 +279,7 @@ gnutls_ocsp_status_request_enable_client(gnutls_session_t session,
 					 gnutls_datum_t * extensions)
 {
 	status_request_ext_st *priv;
-	extension_priv_data_t epriv;
+	gnutls_ext_priv_data_t epriv;
 
 	if (session->security_parameters.entity == GNUTLS_SERVER)
 		return gnutls_assert_val(GNUTLS_E_INVALID_REQUEST);
@@ -325,13 +288,6 @@ gnutls_ocsp_status_request_enable_client(gnutls_session_t session,
 	if (priv == NULL)
 		return gnutls_assert_val(GNUTLS_E_MEMORY_ERROR);
 
-	priv->responder_id = responder_id;
-	priv->responder_id_size = responder_id_size;
-	if (extensions) {
-		priv->request_extensions.data = extensions->data;
-		priv->request_extensions.size = extensions->size;
-	}
-
 	_gnutls_ext_set_session_data(session,
 				     GNUTLS_EXTENSION_STATUS_REQUEST,
 				     epriv);
@@ -359,7 +315,7 @@ gnutls_ocsp_status_request_get(gnutls_session_t session,
 			       gnutls_datum_t * response)
 {
 	status_request_ext_st *priv;
-	extension_priv_data_t epriv;
+	gnutls_ext_priv_data_t epriv;
 	int ret;
 
 	if (session->security_parameters.entity == GNUTLS_SERVER)
@@ -517,28 +473,19 @@ gnutls_certificate_set_ocsp_status_request_file
 	return 0;
 }
 
-static void _gnutls_status_request_deinit_data(extension_priv_data_t epriv)
+static void _gnutls_status_request_deinit_data(gnutls_ext_priv_data_t epriv)
 {
 	status_request_ext_st *priv = epriv;
-	unsigned i;
 
 	if (priv == NULL)
 		return;
 
-	if (priv->responder_id != NULL) {
-		for (i = 0; i < priv->responder_id_size; i++)
-			gnutls_free(priv->responder_id[i].data);
-
-		gnutls_free(priv->responder_id);
-	}
-
-	gnutls_free(priv->request_extensions.data);
 	gnutls_free(priv->response.data);
 	gnutls_free(priv);
 }
 
 static int
-_gnutls_status_request_pack(extension_priv_data_t epriv,
+_gnutls_status_request_pack(gnutls_ext_priv_data_t epriv,
 			    gnutls_buffer_st * ps)
 {
 	status_request_ext_st *priv = epriv;
@@ -552,7 +499,7 @@ _gnutls_status_request_pack(extension_priv_data_t epriv,
 
 static int
 _gnutls_status_request_unpack(gnutls_buffer_st * ps,
-			      extension_priv_data_t * epriv)
+			      gnutls_ext_priv_data_t * epriv)
 {
 	status_request_ext_st *priv;
 	int ret;
@@ -582,7 +529,8 @@ const extension_entry_st ext_mod_status_request = {
 	.send_func = _gnutls_status_request_send_params,
 	.pack_func = _gnutls_status_request_pack,
 	.unpack_func = _gnutls_status_request_unpack,
-	.deinit_func = _gnutls_status_request_deinit_data
+	.deinit_func = _gnutls_status_request_deinit_data,
+	.cannot_be_overriden = 1
 };
 
 /* Functions to be called from handshake */
@@ -594,7 +542,7 @@ _gnutls_send_server_certificate_status(gnutls_session_t session, int again)
 	uint8_t *data;
 	int data_size = 0;
 	int ret;
-	extension_priv_data_t epriv;
+	gnutls_ext_priv_data_t epriv;
 	status_request_ext_st *priv;
 
 	if (again == 0) {
@@ -638,7 +586,7 @@ int _gnutls_recv_server_certificate_status(gnutls_session_t session)
 	gnutls_buffer_st buf;
 	int ret;
 	status_request_ext_st *priv = NULL;
-	extension_priv_data_t epriv;
+	gnutls_ext_priv_data_t epriv;
 
 	ret =
 	    _gnutls_ext_get_session_data(session,
