@@ -1,6 +1,6 @@
 /*
  * Copyright (C) 2000-2016 Free Software Foundation, Inc.
- * Copyright (C) 2015-2016 Red Hat, Inc.
+ * Copyright (C) 2015-2017 Red Hat, Inc.
  *
  * Author: Nikos Mavrogiannopoulos
  *
@@ -148,6 +148,10 @@ typedef struct {
 #define DEFAULT_EXPIRE_TIME 3600
 #define DEFAULT_HANDSHAKE_TIMEOUT_MS 40*1000
 
+/* The EC group to be used when the extension
+ * supported groups/curves is not present */
+#define DEFAULT_EC_GROUP GNUTLS_GROUP_SECP256R1
+
 typedef enum transport_t {
 	GNUTLS_STREAM,
 	GNUTLS_DGRAM
@@ -171,7 +175,7 @@ typedef enum record_flush_t {
  * when receiving we use a different way as there are implementations that
  * store more data than allowed.
  */
-#define MAX_RECORD_SEND_OVERHEAD(session) (MAX_CIPHER_BLOCK_SIZE/*iv*/+MAX_PAD_SIZE+((gnutls_compression_get(session)!=GNUTLS_COMP_NULL)?(EXTRA_COMP_SIZE):(0))+MAX_HASH_SIZE/*MAC*/)
+#define MAX_RECORD_SEND_OVERHEAD(session) (MAX_CIPHER_BLOCK_SIZE/*iv*/+MAX_PAD_SIZE+MAX_HASH_SIZE/*MAC*/)
 #define MAX_RECORD_SEND_SIZE(session) (IS_DTLS(session)? \
 	(MIN((size_t)gnutls_dtls_get_mtu(session), (size_t)session->security_parameters.max_record_send_size+MAX_RECORD_SEND_OVERHEAD(session))): \
 	((size_t)session->security_parameters.max_record_send_size+MAX_RECORD_SEND_OVERHEAD(session)))
@@ -300,6 +304,8 @@ typedef enum content_type_t {
 #define GNUTLS_PK_ANY (gnutls_pk_algorithm_t)-1
 #define GNUTLS_PK_NONE (gnutls_pk_algorithm_t)-2
 
+#define GNUTLS_PK_IS_RSA(pk) ((pk) == GNUTLS_PK_RSA || (pk) == GNUTLS_PK_RSA_PSS)
+
 /* Message buffers (mbuffers) structures */
 
 /* this is actually the maximum number of distinct handshake
@@ -469,6 +475,18 @@ typedef struct gnutls_cipher_suite_entry_st {
 	gnutls_mac_algorithm_t prf;
 } gnutls_cipher_suite_entry_st;
 
+
+typedef struct gnutls_group_entry_st {
+	const char *name;
+	gnutls_group_t id;
+	const gnutls_datum_t *prime;
+	const gnutls_datum_t *generator;
+	const unsigned *q_bits;
+	gnutls_ecc_curve_t curve;
+	gnutls_pk_algorithm_t pk;
+	unsigned tls_id;		/* The RFC4492 namedCurve ID or TLS 1.3 group ID */
+} gnutls_group_entry_st;
+
 /* This structure is used both for MACs and digests
  */
 typedef struct mac_entry_st {
@@ -480,8 +498,8 @@ typedef struct mac_entry_st {
 	unsigned key_size;
 	unsigned nonce_size;
 	unsigned placeholder;	/* if set, then not a real MAC */
-	unsigned secure;	/* must be set to zero if this hash is known to be broken */
 	unsigned block_size;	/* internal block size for HMAC */
+	unsigned preimage_insecure; /* if this algorithm should not be trusted for pre-image attacks */
 } mac_entry_st;
 
 typedef struct {
@@ -508,11 +526,9 @@ typedef struct {
 
 #include <hash_int.h>
 #include <cipher_int.h>
-#include <compress.h>
 
 typedef struct {
-	uint8_t hash_algorithm;
-	uint8_t sign_algorithm;	/* pk algorithm actually */
+	uint8_t id[2]; /* used to be (in TLS 1.2) hash algorithm , PK algorithm */
 } sign_algorithm_st;
 
 /* This structure holds parameters got from TLS extension
@@ -520,9 +536,6 @@ typedef struct {
  * structures also - see SRP).
  */
 
-#define MAX_SIGNATURE_ALGORITHMS 16
-#define MAX_SIGN_ALGO_SIZE (2 + MAX_SIGNATURE_ALGORITHMS * 2)
-
 #define MAX_VERIFY_DATA_SIZE 36	/* in SSL 3.0, 12 in TLS 1.0 */
 
 /* auth_info_t structures now MAY contain malloced 
@@ -541,7 +554,6 @@ typedef struct {
  */
 typedef struct {
 	unsigned int entity;	/* GNUTLS_SERVER or GNUTLS_CLIENT */
-	gnutls_kx_algorithm_t kx_algorithm;
 
 	/* The epoch used to read and write */
 	uint16_t epoch_read;
@@ -557,8 +569,11 @@ typedef struct {
 	 * moved here from internals in order to be restored
 	 * on resume;
 	 */
-	uint8_t cipher_suite[2];
-	gnutls_compression_method_t compression_method;
+	const struct gnutls_cipher_suite_entry_st *cs;
+
+	/* This is kept outside the ciphersuite entry as on certain
+	 * TLS versions we need a separate PRF MAC, i.e., MD5_SHA1. */
+	gnutls_mac_algorithm_t prf_mac;
 	uint8_t master_secret[GNUTLS_MASTER_SIZE];
 	uint8_t client_random[GNUTLS_RANDOM_SIZE];
 	uint8_t server_random[GNUTLS_RANDOM_SIZE];
@@ -573,10 +588,16 @@ typedef struct {
 	uint16_t max_record_recv_size;
 	/* holds the negotiated certificate type */
 	gnutls_certificate_type_t cert_type;
-	gnutls_ecc_curve_t ecc_curve;	/* holds the first supported ECC curve requested by client */
+
+	/* The selected (after server hello EC or DH group */
+	const gnutls_group_entry_st *grp;
+
+	/* Holds the signature algorithm that will be used in this session,
+	 * selected by the server at the time of Ciphersuite/certificate
+	 * selection - see select_sign_algorithm() */
+	gnutls_sign_algorithm_t server_sign_algo;
 
 	/* Holds the signature algorithm used in this session - If any */
-	gnutls_sign_algorithm_t server_sign_algo;
 	gnutls_sign_algorithm_t client_sign_algo;
 
 	/* Whether the master secret negotiation will be according to
@@ -603,7 +624,6 @@ struct record_state_st {
 	gnutls_datum_t IV;
 	gnutls_datum_t key;
 	auth_cipher_hd_st cipher_state;
-	comp_hd_st compression_state;
 	gnutls_uint64 sequence_number;
 };
 
@@ -620,8 +640,6 @@ struct record_parameters_st {
 	uint16_t epoch;
 	int initialized;
 
-	gnutls_compression_method_t compression_algorithm;
-
 	const cipher_entry_st *cipher;
 	bool etm;
 	const mac_entry_st *mac;
@@ -653,44 +671,81 @@ typedef enum {
 	SR_SAFE
 } safe_renegotiation_t;
 
+#define MAX_CIPHERSUITE_SIZE 256
+
+typedef struct ciphersuite_list_st {
+	const gnutls_cipher_suite_entry_st *entry[MAX_CIPHERSUITE_SIZE];
+	unsigned int size;
+} ciphersuite_list_st;
+
+typedef struct group_list_st {
+	const gnutls_group_entry_st *entry[MAX_ALGOS];
+	unsigned int size;
+	bool have_ffdhe;
+} group_list_st;
+
+typedef struct sign_algo_list_st {
+	const struct gnutls_sign_entry_st *entry[MAX_ALGOS];
+	unsigned int size;
+} sign_algo_list_st;
+
+#include "atomic.h"
+
 /* For the external api */
 struct gnutls_priority_st {
-	priority_st cipher;
-	priority_st mac;
-	priority_st kx;
-	priority_st compression;
 	priority_st protocol;
 	priority_st cert_type;
-	priority_st sign_algo;
-	priority_st supported_ecc;
+
+	/* The following are not necessary to be stored in
+	 * the structure; however they are required by the
+	 * external APIs: gnutls_priority_*_list() */
+	priority_st _cipher;
+	priority_st _mac;
+	priority_st _kx;
+	priority_st _sign_algo;
+	priority_st _supported_ecc;
+
+	/* the supported groups */
+	group_list_st groups;
+
+	/* the supported signature algorithms */
+	sign_algo_list_st sigalg;
+
+	/* the supported ciphersuites */
+	ciphersuite_list_st cs;
 
 	/* to disable record padding */
 	bool no_extensions;
-	bool no_ext_master_secret;
-	bool allow_large_records;
-	unsigned int dumbfw;
+
+
 	safe_renegotiation_t sr;
 	bool min_record_version;
 	bool server_precedence;
-	bool allow_key_usage_violation;
 	bool allow_server_key_usage_violation; /* for test suite purposes only */
-	bool allow_wrong_pms;
 	bool no_tickets;
-	bool no_etm;
 	bool have_cbc;
-	/* Whether stateless compression will be used */
-	bool stateless_compression;
 	unsigned int additional_verify_flags;
 
+	/* TLS_FALLBACK_SCSV */
+	bool fallback;
+
 	/* The session's expected security level.
 	 * Will be used to determine the minimum DH bits,
 	 * (or the acceptable certificate security level).
 	 */
 	gnutls_sec_param_t level;
-	unsigned int dh_prime_bits;	/* old (deprecated) variable */
 
-	/* TLS_FALLBACK_SCSV */
-	bool fallback;
+	/* these should be accessed from
+	 * session->internals.VAR names */
+	bool _allow_large_records;
+	bool _no_etm;
+	bool _no_ext_master_secret;
+	bool _allow_key_usage_violation;
+	bool _allow_wrong_pms;
+	bool _dumbfw;
+	unsigned int _dh_prime_bits;	/* old (deprecated) variable */
+
+	DEF_ATOMIC_INT(usage_cnt);
 };
 
 /* Allow around 50KB of length-hiding padding
@@ -706,6 +761,14 @@ struct gnutls_priority_st {
 	      (x)->allow_wrong_pms = 1; \
 	      (x)->dumbfw = 1
 
+#define ENABLE_PRIO_COMPAT(x) \
+	      (x)->_allow_large_records = 1; \
+	      (x)->_no_etm = 1; \
+	      (x)->_no_ext_master_secret = 1; \
+	      (x)->_allow_key_usage_violation = 1; \
+	      (x)->_allow_wrong_pms = 1; \
+	      (x)->_dumbfw = 1
+
 /* DH and RSA parameters types.
  */
 typedef struct gnutls_dh_params_int {
@@ -716,11 +779,6 @@ typedef struct gnutls_dh_params_int {
 				 */
 } dh_params_st;
 
-typedef struct {
-	gnutls_dh_params_t dh_params;
-	int free_dh_params;
-} internal_params_st;
-
 /* DTLS session state
  */
 typedef struct {
@@ -802,7 +860,18 @@ typedef struct {
 	int last_handshake_out;
 
 	/* priorities */
-	struct gnutls_priority_st priorities;
+	struct gnutls_priority_st *priorities;
+
+	/* variables directly set when setting the priorities above, or
+	 * when overriding them */
+	bool allow_large_records;
+	bool no_etm;
+	bool no_ext_master_secret;
+	bool allow_key_usage_violation;
+	bool allow_wrong_pms;
+	bool dumbfw;
+	unsigned int dh_prime_bits;	/* old (deprecated) variable */
+
 
 	/* resumed session */
 	bool resumed;	/* RESUME_TRUE or FALSE - if we are resuming a session */
@@ -897,16 +966,6 @@ typedef struct {
 	gnutls_status_request_ocsp_func selected_ocsp_func;
 	void *selected_ocsp_func_ptr;
 
-	/* In case of a client holds the extensions we sent to the peer;
-	 * otherwise the extensions we received from the client.
-	 */
-	uint16_t extensions_sent[MAX_EXT_TYPES];
-	uint16_t extensions_sent_size;
-
-	/* is 0 if we are to send the whole PGP key, or non zero
-	 * if the fingerprint is to be sent.
-	 */
-	bool pgp_fingerprint;
 
 	/* This holds the default version that our first
 	 * record packet will have. */
@@ -915,21 +974,11 @@ typedef struct {
 
 	void *user_ptr;
 
-	bool enable_private;	/* non zero to
-					 * enable cipher suites
-					 * which have 0xFF status.
-					 */
-
 	/* Holds 0 if the last called function was interrupted while
 	 * receiving, and non zero otherwise.
 	 */
 	bool direction;
 
-	/* This callback will be used (if set) to receive an
-	 * openpgp key. (if the peer sends a fingerprint)
-	 */
-	gnutls_openpgp_recv_key_func openpgp_recv_key_func;
-
 	/* If non zero the server will not advertise the CA's he
 	 * trusts (do not send an RDN sequence).
 	 */
@@ -941,12 +990,6 @@ typedef struct {
 	 */
 	uint8_t rsa_pms_version[2];
 
-	/* Here we cache the DH or RSA parameters got from the
-	 * credentials structure, or from a callback. That is to
-	 * minimize external calls.
-	 */
-	internal_params_st params;
-
 	/* To avoid using global variables, and especially on Windows where
 	 * the application may use a different errno variable than GnuTLS,
 	 * it is possible to use gnutls_transport_set_errno to set a
@@ -965,17 +1008,6 @@ typedef struct {
 	/* A handshake process has been completed */
 	bool initial_negotiation_completed;
 
-	struct {
-		uint16_t type;
-		gnutls_ext_priv_data_t priv;
-		bool set;
-	} extension_int_data[MAX_EXT_TYPES];
-
-	struct {
-		uint16_t type;
-		gnutls_ext_priv_data_t priv;
-		bool set;
-	} resumed_extension_int_data[MAX_EXT_TYPES];
 	/* The type of transport protocol; stream or datagram */
 	transport_t transport;
 
@@ -1062,12 +1094,39 @@ typedef struct {
 
 	struct extension_entry_st *rexts;
 	unsigned rexts_size;
+
+	struct {
+		uint16_t type;
+		gnutls_ext_priv_data_t priv;
+		gnutls_ext_priv_data_t resumed_priv;
+		uint8_t set;
+		uint8_t resumed_set;
+	} ext_data[MAX_EXT_TYPES];
+
+	/* In case of a client holds the extensions we sent to the peer;
+	 * otherwise the extensions we received from the client.
+	 */
+	const struct extension_entry_st *used_exts[MAX_EXT_TYPES];
+	unsigned used_exts_size;
+
+	/* this is not the negotiated max_record_recv_size, but the actual maximum
+	 * receive size */
+	unsigned max_recv_size;
+
+	/* whether the peer has advertized at least an FFDHE group */
+	bool have_ffdhe;
+	bool used_ffdhe; /* whether ffdhe was actually negotiated and used */
+
+	/* candidate groups to be selected for security params groups */
+	const gnutls_group_entry_st *cand_ec_group;
+	const gnutls_group_entry_st *cand_dh_group;
+
 	/* If you add anything here, check _gnutls_handshake_internal_state_clear().
 	 */
 } internals_st;
 
 /* Maximum number of epochs we keep around. */
-#define MAX_EPOCH_INDEX 16
+#define MAX_EPOCH_INDEX 4
 
 struct gnutls_session_int {
 	security_parameters_st security_parameters;
