@@ -1,6 +1,6 @@
 /*
  * Copyright (C) 2000-2016 Free Software Foundation, Inc.
- * Copyright (C) 2015-2016 Red Hat, Inc.
+ * Copyright (C) 2015-2017 Red Hat, Inc.
  *
  * Author: Nikos Mavrogiannopoulos
  *
@@ -29,7 +29,6 @@
 #include "dh.h"
 #include "debug.h"
 #include "algorithms.h"
-#include "compress.h"
 #include "cipher.h"
 #include "buffers.h"
 #include "mbuffers.h"
@@ -55,22 +54,17 @@
 #include <random.h>
 #include <dtls.h>
 
-#ifdef HANDSHAKE_DEBUG
-#define ERR(x, y) _gnutls_handshake_log("HSK[%p]: %s (%d)\n", session, x,y)
-#else
-#define ERR(x, y)
-#endif
-
 #define TRUE 1
 #define FALSE 0
 
-static int server_select_comp_method(gnutls_session_t session,
+static int check_if_null_comp_present(gnutls_session_t session,
 					     uint8_t * data, int datalen);
 static int handshake_client(gnutls_session_t session);
 static int handshake_server(gnutls_session_t session);
 
 static int
-recv_hello(gnutls_session_t session, uint8_t * data, int datalen);
+read_server_hello(gnutls_session_t session,
+		  uint8_t * data, int datalen);
 
 static int
 recv_handshake_final(gnutls_session_t session, int init);
@@ -85,6 +79,7 @@ handshake_hash_buffer_empty(gnutls_session_t session)
 
 	_gnutls_buffers_log("BUF[HSK]: Emptied buffer\n");
 
+	session->internals.used_exts_size = 0;
 	session->internals.handshake_hash_buffer_prev_len = 0;
 	session->internals.handshake_hash_buffer.length = 0;
 	return;
@@ -135,24 +130,10 @@ static int resume_copy_required_values(gnutls_session_t session)
 	 * That is because the client must see these in our
 	 * hello message.
 	 */
-	memcpy(session->security_parameters.cipher_suite,
-	       session->internals.resumed_security_parameters.cipher_suite,
-	       2);
-	session->security_parameters.compression_method =
-	    session->internals.resumed_security_parameters.
-	    compression_method;
-
-	ret = _gnutls_epoch_set_cipher_suite(session, EPOCH_NEXT,
-					     session->internals.
-					     resumed_security_parameters.
-					     cipher_suite);
-	if (ret < 0)
-		return gnutls_assert_val(ret);
-
-	ret = _gnutls_epoch_set_compression(session, EPOCH_NEXT,
-					    session->internals.
-					    resumed_security_parameters.
-					    compression_method);
+	ret = _gnutls_set_cipher_suite2(session,
+				       session->internals.
+				       resumed_security_parameters.
+				       cs);
 	if (ret < 0)
 		return gnutls_assert_val(ret);
 
@@ -191,17 +172,23 @@ static int resume_copy_required_values(gnutls_session_t session)
  */
 static int create_tls_random(uint8_t * dst)
 {
-	uint32_t tim;
 	int ret;
 
-	/* Use weak random numbers for the most of the
+	/* Use nonce rng level for the most of the
 	 * buffer except for the first 4 that are the
 	 * system's time.
 	 */
 
+#if defined(FUZZING_BUILD_MODE_UNSAFE_FOR_PRODUCTION)
+	/* When fuzzying avoid timing dependencies */
+	memset(dst, 1, 4);
+#else
+	uint32_t tim;
+
 	tim = gnutls_time(NULL);
 	/* generate server random value */
 	_gnutls_write_uint32(tim, dst);
+#endif
 
 	ret =
 	    gnutls_rnd(GNUTLS_RND_NONCE, &dst[3], GNUTLS_RANDOM_SIZE - 3);
@@ -341,48 +328,25 @@ _gnutls_finished(gnutls_session_t session, int type, void *ret,
 		 int sending)
 {
 	const int siz = TLS_MSG_LEN;
-	uint8_t concat[MAX_HASH_SIZE + 16 /*MD5 */ ];
+	uint8_t concat[MAX_HASH_SIZE];
 	size_t hash_len;
 	const char *mesg;
-	int rc, len;
+	int rc, len, algorithm;
 
 	if (sending)
 		len = session->internals.handshake_hash_buffer.length;
 	else
 		len = session->internals.handshake_hash_buffer_prev_len;
 
-	if (!_gnutls_version_has_selectable_prf(get_version(session))) {
-		rc = _gnutls_hash_fast(GNUTLS_DIG_SHA1,
-				       session->internals.
-				       handshake_hash_buffer.data, len,
-				       &concat[16]);
-		if (rc < 0)
-			return gnutls_assert_val(rc);
+	algorithm = session->security_parameters.prf_mac;
+	rc = _gnutls_hash_fast(algorithm,
+			       session->internals.
+			       handshake_hash_buffer.data, len,
+			       concat);
+	if (rc < 0)
+		return gnutls_assert_val(rc);
 
-		rc = _gnutls_hash_fast(GNUTLS_DIG_MD5,
-				       session->internals.
-				       handshake_hash_buffer.data, len,
-				       concat);
-		if (rc < 0)
-			return gnutls_assert_val(rc);
-
-		hash_len = 20 + 16;
-	} else {
-		int algorithm =
-		    _gnutls_cipher_suite_get_prf(session->
-						 security_parameters.
-						 cipher_suite);
-
-		rc = _gnutls_hash_fast(algorithm,
-				       session->internals.
-				       handshake_hash_buffer.data, len,
-				       concat);
-		if (rc < 0)
-			return gnutls_assert_val(rc);
-
-		hash_len =
-		    _gnutls_hash_get_algo_len(mac_to_entry(algorithm));
-	}
+	hash_len = _gnutls_hash_get_algo_len(mac_to_entry(algorithm));
 
 	if (type == GNUTLS_SERVER) {
 		mesg = SERVER_MSG;
@@ -478,7 +442,8 @@ read_client_hello(gnutls_session_t session, uint8_t * data,
 {
 	uint8_t session_id_len;
 	int pos = 0, ret;
-	uint16_t suite_size, comp_size, ext_size;
+	uint16_t suite_size, comp_size;
+	int ext_size;
 	gnutls_protocol_t adv_version;
 	int neg_version, sret = 0;
 	int len = datalen;
@@ -548,6 +513,9 @@ read_client_hello(gnutls_session_t session, uint8_t * data,
 	suite_size = _gnutls_read_uint16(&data[pos]);
 	pos += 2;
 
+	if (suite_size == 0 || (suite_size % 2) != 0)
+		return gnutls_assert_val(GNUTLS_E_UNEXPECTED_PACKET_LENGTH);
+
 	suite_ptr = &data[pos];
 	DECR_LEN(len, suite_size);
 	pos += suite_size;
@@ -555,6 +523,9 @@ read_client_hello(gnutls_session_t session, uint8_t * data,
 	DECR_LEN(len, 1);
 	comp_size = data[pos++]; /* the number of compression methods */
 
+	if (comp_size == 0)
+		return gnutls_assert_val(GNUTLS_E_UNEXPECTED_PACKET_LENGTH);
+
 	comp_ptr = &data[pos];
 	DECR_LEN(len, comp_size);
 	pos += comp_size;
@@ -583,10 +554,6 @@ read_client_hello(gnutls_session_t session, uint8_t * data,
 		session->internals.resumption_requested = 1;
 
 	if (ret == 0) {		/* resumed using default TLS resumption! */
-		ret = _gnutls_server_select_suite(session, suite_ptr, suite_size, 1);
-		if (ret < 0)
-			return gnutls_assert_val(ret);
-
 		ret = resume_copy_required_values(session);
 		if (ret < 0)
 			return gnutls_assert_val(ret);
@@ -646,10 +613,6 @@ read_client_hello(gnutls_session_t session, uint8_t * data,
 		    max_record_send_size =
 		    session->security_parameters.max_record_send_size;
 
-		ret = _gnutls_server_select_suite(session, suite_ptr, suite_size, 1);
-		if (ret < 0)
-			return gnutls_assert_val(ret);
-
 		ret = resume_copy_required_values(session);
 		if (ret < 0)
 			return gnutls_assert_val(ret);
@@ -659,7 +622,7 @@ read_client_hello(gnutls_session_t session, uint8_t * data,
 
 	/* select an appropriate cipher suite (as well as certificate)
 	 */
-	ret = _gnutls_server_select_suite(session, suite_ptr, suite_size, 0);
+	ret = _gnutls_server_select_suite(session, suite_ptr, suite_size);
 	if (ret < 0) {
 		gnutls_assert();
 		return ret;
@@ -667,7 +630,7 @@ read_client_hello(gnutls_session_t session, uint8_t * data,
 
 	/* select appropriate compression method */
 	ret =
-	    server_select_comp_method(session, comp_ptr,
+	    check_if_null_comp_present(session, comp_ptr,
 					      comp_size);
 	if (ret < 0) {
 		gnutls_assert();
@@ -791,7 +754,6 @@ static int _gnutls_recv_finished(gnutls_session_t session)
 	    _gnutls_recv_handshake(session, GNUTLS_HANDSHAKE_FINISHED,
 				   0, &buf);
 	if (ret < 0) {
-		ERR("recv finished int", ret);
 		gnutls_assert();
 		return ret;
 	}
@@ -830,11 +792,18 @@ static int _gnutls_recv_finished(gnutls_session_t session)
 		goto cleanup;
 	}
 
+#if defined(FUZZING_BUILD_MODE_UNSAFE_FOR_PRODUCTION)
+	/* When fuzzying allow to proceed without verifying the handshake
+	 * consistency */
+# warning This is unsafe for production builds
+
+#else
 	if (memcmp(vrfy, data, data_size) != 0) {
 		gnutls_assert();
 		ret = GNUTLS_E_ERROR_IN_FINISHED_PACKET;
 		goto cleanup;
 	}
+#endif
 
 	ret = _gnutls_ext_sr_finished(session, data, data_size, 1);
 	if (ret < 0) {
@@ -862,77 +831,25 @@ static int _gnutls_recv_finished(gnutls_session_t session)
 	return ret;
 }
 
-/* returns PK_RSA if the given cipher suite list only supports,
- * RSA algorithms, PK_DSA if DSS, and PK_ANY for both or PK_NONE for none.
- */
-static int
-server_find_pk_algos_in_ciphersuites(const uint8_t *
-				     data, unsigned int datalen,
-				     gnutls_pk_algorithm_t * algos,
-				     size_t * algos_size)
-{
-	unsigned int j, x;
-	gnutls_kx_algorithm_t kx;
-	gnutls_pk_algorithm_t pk;
-	unsigned found;
-	unsigned int max = *algos_size;
-
-	if (datalen % 2 != 0) {
-		gnutls_assert();
-		return GNUTLS_E_UNEXPECTED_PACKET_LENGTH;
-	}
-
-	*algos_size = 0;
-	for (j = 0; j < datalen; j += 2) {
-		kx = _gnutls_cipher_suite_get_kx_algo(&data[j]);
-		if (_gnutls_map_kx_get_cred(kx, 1) ==
-		    GNUTLS_CRD_CERTIFICATE) {
-			pk = _gnutls_map_kx_get_pk(kx);
-			found = 0;
-			for (x = 0; x < *algos_size; x++) {
-				if (algos[x] == pk) {
-					found = 1;
-					break;
-				}
-			}
-
-			if (found == 0) {
-				algos[(*algos_size)++] =
-				    _gnutls_map_kx_get_pk(kx);
-				if ((*algos_size) >= max)
-					return 0;
-			}
-		}
-	}
-
-	return 0;
-}
-
 /* This selects the best supported ciphersuite from the given ones. Then
  * it adds the suite to the session and performs some checks.
- *
- * When @scsv_only is non-zero only the available SCSVs are parsed
- * and acted upon.
  */
 int
 _gnutls_server_select_suite(gnutls_session_t session, uint8_t * data,
-			    unsigned int datalen, unsigned scsv_only)
+			    unsigned int datalen)
 {
 	int ret;
-	unsigned int i, j, cipher_suites_size;
-	size_t pk_algos_size;
-	uint8_t cipher_suites[MAX_CIPHERSUITE_SIZE];
+	unsigned int i;
+	ciphersuite_list_st peer_clist;
+	const gnutls_cipher_suite_entry_st *selected;
 	int retval;
-	gnutls_pk_algorithm_t pk_algos[MAX_ALGOS];	/* will hold the pk algorithms
-							 * supported by the peer.
-							 */
+
+	peer_clist.size = 0;
 
 	for (i = 0; i < datalen; i += 2) {
-		/* we support the TLS renegotiation SCSV, even if we are
-		 * not under SSL 3.0, because openssl sends this SCSV
-		 * on resumption unconditionally. */
+#ifdef ENABLE_SSL3 /* No need to support certain SCSV's without SSL 3.0 */
 		/* TLS_RENEGO_PROTECTION_REQUEST = { 0x00, 0xff } */
-		if (session->internals.priorities.sr != SR_DISABLED &&
+		if (session->internals.priorities->sr != SR_DISABLED &&
 		    data[i] == GNUTLS_RENEGO_PROTECTION_REQUEST_MAJOR &&
 		    data[i + 1] == GNUTLS_RENEGO_PROTECTION_REQUEST_MINOR) {
 			_gnutls_handshake_log
@@ -944,6 +861,7 @@ _gnutls_server_select_suite(gnutls_session_t session, uint8_t * data,
 				return retval;
 			}
 		}
+#endif
 
 		/* TLS_FALLBACK_SCSV */
 		if (data[i] == GNUTLS_FALLBACK_SCSV_MAJOR &&
@@ -955,131 +873,32 @@ _gnutls_server_select_suite(gnutls_session_t session, uint8_t * data,
 
 			if (gnutls_protocol_get_version(session) != max)
 				return gnutls_assert_val(GNUTLS_E_INAPPROPRIATE_FALLBACK);
+		} else {
+			if (peer_clist.size < MAX_CIPHERSUITE_SIZE) {
+				peer_clist.entry[peer_clist.size] = ciphersuite_to_entry(&data[i]);
+				if (peer_clist.entry[peer_clist.size] != NULL)
+					peer_clist.size++;
+			}
 		}
 	}
 
-	if (scsv_only)
-		return 0;
-
-	pk_algos_size = MAX_ALGOS;
-	ret =
-	    server_find_pk_algos_in_ciphersuites(data, datalen, pk_algos,
-						 &pk_algos_size);
-	if (ret < 0)
+	ret = _gnutls_figure_common_ciphersuite(session, &peer_clist, &selected);
+	if (ret < 0) {
 		return gnutls_assert_val(ret);
-
-	ret =
-	    _gnutls_supported_ciphersuites(session, cipher_suites,
-					   sizeof(cipher_suites));
-	if (ret < 0)
-		return gnutls_assert_val(ret);
-
-	cipher_suites_size = ret;
-
-	/* Here we remove any ciphersuite that does not conform
-	 * the certificate requested, or to the
-	 * authentication requested (e.g. SRP).
-	 */
-	ret =
-	    _gnutls_remove_unwanted_ciphersuites(session, cipher_suites,
-						 cipher_suites_size,
-						 pk_algos, pk_algos_size);
-	if (ret <= 0) {
-		gnutls_assert();
-		if (ret < 0)
-			return ret;
-		else
-			return GNUTLS_E_UNKNOWN_CIPHER_SUITE;
 	}
 
-	cipher_suites_size = ret;
-
-	/* Data length should be zero mod 2 since
-	 * every ciphersuite is 2 bytes. (this check is needed
-	 * see below).
-	 */
-	if (datalen % 2 != 0) {
-		gnutls_assert();
-		return GNUTLS_E_UNEXPECTED_PACKET_LENGTH;
-	}
-
-	memset(session->security_parameters.cipher_suite, 0, 2);
-
-	retval = GNUTLS_E_UNKNOWN_CIPHER_SUITE;
-
 	_gnutls_handshake_log
-	    ("HSK[%p]: Requested cipher suites[size: %d]: \n", session,
-	     (int) datalen);
+		    ("HSK[%p]: Selected cipher suite: %s\n", session, selected->name);
 
-	if (session->internals.priorities.server_precedence == 0) {
-		for (j = 0; j < datalen; j += 2) {
-			_gnutls_handshake_log("\t0x%.2x, 0x%.2x %s\n",
-					      data[j], data[j + 1],
-					      _gnutls_cipher_suite_get_name
-					      (&data[j]));
-			for (i = 0; i < cipher_suites_size; i += 2) {
-				if (memcmp(&cipher_suites[i], &data[j], 2)
-				    == 0) {
-					_gnutls_handshake_log
-					    ("HSK[%p]: Selected cipher suite: %s\n",
-					     session,
-					     _gnutls_cipher_suite_get_name
-					     (&data[j]));
-					memcpy(session->
-					       security_parameters.
-					       cipher_suite,
-					       &cipher_suites[i], 2);
-					_gnutls_epoch_set_cipher_suite
-					    (session, EPOCH_NEXT,
-					     session->security_parameters.
-					     cipher_suite);
-
-
-					retval = 0;
-					goto finish;
-				}
-			}
-		}
-	} else {		/* server selects */
-
-		for (i = 0; i < cipher_suites_size; i += 2) {
-			for (j = 0; j < datalen; j += 2) {
-				if (memcmp(&cipher_suites[i], &data[j], 2)
-				    == 0) {
-					_gnutls_handshake_log
-					    ("HSK[%p]: Selected cipher suite: %s\n",
-					     session,
-					     _gnutls_cipher_suite_get_name
-					     (&data[j]));
-					memcpy(session->
-					       security_parameters.
-					       cipher_suite,
-					       &cipher_suites[i], 2);
-					_gnutls_epoch_set_cipher_suite
-					    (session, EPOCH_NEXT,
-					     session->security_parameters.
-					     cipher_suite);
-
-
-					retval = 0;
-					goto finish;
-				}
-			}
-		}
-	}
-      finish:
-
-	if (retval != 0) {
+	ret = _gnutls_set_cipher_suite2(session, selected);
+	if (ret < 0) {
 		gnutls_assert();
-		return retval;
+		return ret;
 	}
 
 	/* check if the credentials (username, public key etc.) are ok
 	 */
-	if (_gnutls_get_kx_cred
-	    (session,
-	     _gnutls_cipher_suite_get_kx_algo(session->security_parameters.
-					      cipher_suite)) == NULL) {
+	if (_gnutls_get_kx_cred(session, selected->kx_algorithm) == NULL) {
 		gnutls_assert();
 		return GNUTLS_E_INSUFFICIENT_CREDENTIALS;
 	}
@@ -1089,12 +908,8 @@ _gnutls_server_select_suite(gnutls_session_t session, uint8_t * data,
 	 * according to the KX algorithm. This is needed since all the
 	 * handshake functions are read from there;
 	 */
-	session->internals.auth_struct =
-	    _gnutls_kx_auth_struct(_gnutls_cipher_suite_get_kx_algo
-				   (session->security_parameters.
-				    cipher_suite));
+	session->internals.auth_struct = _gnutls_kx_auth_struct(selected->kx_algorithm);
 	if (session->internals.auth_struct == NULL) {
-
 		_gnutls_handshake_log
 		    ("HSK[%p]: Cannot find the appropriate handler for the KX algorithm\n",
 		     session);
@@ -1107,71 +922,20 @@ _gnutls_server_select_suite(gnutls_session_t session, uint8_t * data,
 }
 
 
-/* This selects the best supported compression method from the ones provided 
+/* This checks whether the null compression method is present.
  */
 static int
-server_select_comp_method(gnutls_session_t session,
+check_if_null_comp_present(gnutls_session_t session,
 			  uint8_t * data, int datalen)
 {
-	int x, i, j;
-	uint8_t comps[MAX_ALGOS];
+	int j;
 
-	x = _gnutls_supported_compression_methods(session, comps,
-						  MAX_ALGOS);
-	if (x < 0) {
-		gnutls_assert();
-		return x;
+	for (j = 0; j < datalen; j++) {
+		if (data[j] == 0)
+			return 0;
 	}
 
-	if (session->internals.priorities.server_precedence == 0) {
-		for (j = 0; j < datalen; j++) {
-			for (i = 0; i < x; i++) {
-				if (comps[i] == data[j]) {
-					gnutls_compression_method_t method
-					    =
-					    _gnutls_compression_get_id
-					    (comps[i]);
-
-					_gnutls_epoch_set_compression
-					    (session, EPOCH_NEXT, method);
-					session->security_parameters.
-					    compression_method = method;
-
-					_gnutls_handshake_log
-					    ("HSK[%p]: Selected Compression Method: %s\n",
-					     session,
-					     gnutls_compression_get_name
-					     (method));
-					return 0;
-				}
-			}
-		}
-	} else {
-		for (i = 0; i < x; i++) {
-			for (j = 0; j < datalen; j++) {
-				if (comps[i] == data[j]) {
-					gnutls_compression_method_t method
-					    =
-					    _gnutls_compression_get_id
-					    (comps[i]);
-
-					_gnutls_epoch_set_compression
-					    (session, EPOCH_NEXT, method);
-					session->security_parameters.
-					    compression_method = method;
-
-					_gnutls_handshake_log
-					    ("HSK[%p]: Selected Compression Method: %s\n",
-					     session,
-					     gnutls_compression_get_name
-					     (method));
-					return 0;
-				}
-			}
-		}
-	}
-
-	/* we were not able to find a compatible compression
+	/* we were not able to find a the NULL compression
 	 * algorithm
 	 */
 	gnutls_assert();
@@ -1494,7 +1258,6 @@ _gnutls_recv_handshake(gnutls_session_t session,
 	switch (hsk.htype) {
 	case GNUTLS_HANDSHAKE_CLIENT_HELLO_V2:
 	case GNUTLS_HANDSHAKE_CLIENT_HELLO:
-	case GNUTLS_HANDSHAKE_SERVER_HELLO:
 #ifdef ENABLE_SSL2
 		if (hsk.htype == GNUTLS_HANDSHAKE_CLIENT_HELLO_V2)
 			ret =
@@ -1503,8 +1266,18 @@ _gnutls_recv_handshake(gnutls_session_t session,
 							 hsk.data.length);
 		else
 #endif
-			ret =
-			    recv_hello(session, hsk.data.data,
+			ret = read_client_hello(session, hsk.data.data,
+						hsk.data.length);
+
+		if (ret < 0) {
+			gnutls_assert();
+			goto cleanup;
+		}
+
+		break;
+
+	case GNUTLS_HANDSHAKE_SERVER_HELLO:
+		ret = read_server_hello(session, hsk.data.data,
 					hsk.data.length);
 
 		if (ret < 0) {
@@ -1580,55 +1353,41 @@ _gnutls_recv_handshake(gnutls_session_t session,
 static int
 set_client_ciphersuite(gnutls_session_t session, uint8_t suite[2])
 {
-	uint8_t z;
-	uint8_t cipher_suites[MAX_CIPHERSUITE_SIZE];
-	int cipher_suite_size;
-	int i;
+	unsigned j;
+	int ret;
+	const gnutls_cipher_suite_entry_st *selected = NULL;
 
-	z = 1;
-	cipher_suite_size =
-	    _gnutls_supported_ciphersuites(session, cipher_suites,
-					   sizeof(cipher_suites));
-	if (cipher_suite_size < 0) {
-		gnutls_assert();
-		return cipher_suite_size;
-	}
-
-	for (i = 0; i < cipher_suite_size; i += 2) {
-		if (memcmp(&cipher_suites[i], suite, 2) == 0) {
-			z = 0;
+	for (j = 0; j < session->internals.priorities->cs.size; j++) {
+		if (suite[0] == session->internals.priorities->cs.entry[j]->id[0] &&
+		    suite[1] == session->internals.priorities->cs.entry[j]->id[1]) {
+			selected = session->internals.priorities->cs.entry[j];
 			break;
 		}
 	}
 
-	if (z != 0) {
+	if (!selected) {
 		gnutls_assert();
 		_gnutls_handshake_log
-		    ("HSK[%p]: unsupported cipher suite %.2X.%.2X\n",
+		    ("HSK[%p]: unsupported cipher suite %.2X.%.2X was negotiated\n",
 		     session, (unsigned int) suite[0],
 		     (unsigned int) suite[1]);
 		return GNUTLS_E_UNKNOWN_CIPHER_SUITE;
 	}
 
-	memcpy(session->security_parameters.cipher_suite, suite, 2);
-	_gnutls_epoch_set_cipher_suite(session, EPOCH_NEXT,
-				       session->security_parameters.
-				       cipher_suite);
+	ret = _gnutls_set_cipher_suite2(session, selected);
+	if (ret < 0)
+		return gnutls_assert_val(ret);
 
 	_gnutls_handshake_log("HSK[%p]: Selected cipher suite: %s\n",
 			      session,
-			      _gnutls_cipher_suite_get_name
-			      (session->security_parameters.cipher_suite));
-
+			      selected->name);
 
 	/* check if the credentials (username, public key etc.) are ok.
 	 * Actually checks if they exist.
 	 */
 	if (!session->internals.premaster_set &&
 	    _gnutls_get_kx_cred
-	    (session,
-	     _gnutls_cipher_suite_get_kx_algo
-	     (session->security_parameters.cipher_suite)) == NULL) {
+	    (session, selected->kx_algorithm) == NULL) {
 		gnutls_assert();
 		return GNUTLS_E_INSUFFICIENT_CREDENTIALS;
 	}
@@ -1639,9 +1398,7 @@ set_client_ciphersuite(gnutls_session_t session, uint8_t suite[2])
 	 * handshake functions are read from there;
 	 */
 	session->internals.auth_struct =
-	    _gnutls_kx_auth_struct(_gnutls_cipher_suite_get_kx_algo
-				   (session->security_parameters.
-				    cipher_suite));
+	    _gnutls_kx_auth_struct(selected->kx_algorithm);
 
 	if (session->internals.auth_struct == NULL) {
 
@@ -1652,48 +1409,6 @@ set_client_ciphersuite(gnutls_session_t session, uint8_t suite[2])
 		return GNUTLS_E_INTERNAL_ERROR;
 	}
 
-
-	return 0;
-}
-
-/* This function sets the given comp method to the session.
- */
-static int
-set_client_comp_method(gnutls_session_t session,
-			       uint8_t comp_method)
-{
-	int comp_methods_num;
-	uint8_t compression_methods[MAX_ALGOS];
-	int id = _gnutls_compression_get_id(comp_method);
-	int i;
-
-	_gnutls_handshake_log
-	    ("HSK[%p]: Selected compression method: %s (%d)\n", session,
-	     gnutls_compression_get_name(id), (int) comp_method);
-
-	comp_methods_num = _gnutls_supported_compression_methods(session,
-								 compression_methods,
-								 MAX_ALGOS);
-	if (comp_methods_num < 0) {
-		gnutls_assert();
-		return comp_methods_num;
-	}
-
-	for (i = 0; i < comp_methods_num; i++) {
-		if (compression_methods[i] == comp_method) {
-			comp_methods_num = 0;
-			break;
-		}
-	}
-
-	if (comp_methods_num != 0) {
-		gnutls_assert();
-		return GNUTLS_E_UNKNOWN_COMPRESSION_ALGORITHM;
-	}
-
-	session->security_parameters.compression_method = id;
-	_gnutls_epoch_set_compression(session, EPOCH_NEXT, id);
-
 	return 0;
 }
 
@@ -1706,6 +1421,7 @@ client_check_if_resuming(gnutls_session_t session,
 				 uint8_t * session_id, int session_id_len)
 {
 	char buf[2 * GNUTLS_MAX_SESSION_ID_SIZE + 1];
+	int ret;
 
 	_gnutls_handshake_log("HSK[%p]: SessionID length: %d\n", session,
 			      session_id_len);
@@ -1731,24 +1447,20 @@ client_check_if_resuming(gnutls_session_t session,
 		       session->security_parameters.client_random,
 		       GNUTLS_RANDOM_SIZE);
 
-		memcpy(session->security_parameters.cipher_suite,
-			session->internals.resumed_security_parameters.cipher_suite, 2);
-		session->security_parameters.compression_method =
-			session->internals.resumed_security_parameters.compression_method;
-
-		_gnutls_epoch_set_cipher_suite
-		    (session, EPOCH_NEXT,
+		ret = _gnutls_set_cipher_suite2
+		    (session,
 		     session->internals.resumed_security_parameters.
-		     cipher_suite);
-		_gnutls_epoch_set_compression(session, EPOCH_NEXT,
-					      session->internals.
-					      resumed_security_parameters.
-					      compression_method);
+		     cs);
+		if (ret < 0) {
+			gnutls_assert();
+			goto no_resume;
+		}
 
 		session->internals.resumed = RESUME_TRUE;	/* we are resuming */
 
 		return 0;
 	} else {
+no_resume:
 		/* keep the new session id */
 		session->internals.resumed = RESUME_FALSE;	/* we are not resuming */
 		session->security_parameters.session_id_size =
@@ -1769,7 +1481,7 @@ client_check_if_resuming(gnutls_session_t session,
  */
 static int
 read_server_hello(gnutls_session_t session,
-			  uint8_t * data, int datalen)
+		  uint8_t * data, int datalen)
 {
 	uint8_t session_id_len = 0;
 	int pos = 0;
@@ -1851,12 +1563,7 @@ read_server_hello(gnutls_session_t session,
 	/* move to compression 
 	 */
 	DECR_LEN(len, 1);
-
-	ret = set_client_comp_method(session, data[pos++]);
-	if (ret < 0) {
-		gnutls_assert();
-		return GNUTLS_E_UNKNOWN_COMPRESSION_ALGORITHM;
-	}
+	pos++;
 
 	/* Parse extensions.
 	 */
@@ -1871,103 +1578,19 @@ read_server_hello(gnutls_session_t session,
 	return ret;
 }
 
-#define RESERVED_CIPHERSUITES 4
-/* This function copies the appropriate ciphersuites to a locally allocated buffer
- * Needed in client hello messages. Returns the new data length. If add_scsv is
- * true, add the special safe renegotiation CS.
- */
-static int
-copy_ciphersuites(gnutls_session_t session,
-		  gnutls_buffer_st * cdata, int add_scsv)
-{
-	int ret;
-	uint8_t cipher_suites[MAX_CIPHERSUITE_SIZE + RESERVED_CIPHERSUITES]; /* allow space for SCSV */
-	int cipher_suites_size;
-	size_t init_length = cdata->length;
-
-	ret =
-	    _gnutls_supported_ciphersuites(session, cipher_suites,
-					   sizeof(cipher_suites) - RESERVED_CIPHERSUITES);
-	if (ret < 0)
-		return gnutls_assert_val(ret);
-
-	/* Here we remove any ciphersuite that does not conform
-	 * the certificate requested, or to the
-	 * authentication requested (eg SRP).
-	 */
-	ret =
-	    _gnutls_remove_unwanted_ciphersuites(session, cipher_suites,
-						 ret, NULL, 0);
-	if (ret < 0)
-		return gnutls_assert_val(ret);
-
-	/* If no cipher suites were enabled.
-	 */
-	if (ret == 0)
-		return
-		    gnutls_assert_val(GNUTLS_E_INSUFFICIENT_CREDENTIALS);
-
-	cipher_suites_size = ret;
-#ifdef ENABLE_SSL3
-	if (add_scsv) {
-		cipher_suites[cipher_suites_size] = 0x00;
-		cipher_suites[cipher_suites_size + 1] = 0xff;
-		cipher_suites_size += 2;
-
-		ret = _gnutls_ext_sr_send_cs(session);
-		if (ret < 0)
-			return gnutls_assert_val(ret);
-	}
-#endif
-
-	if (session->internals.priorities.fallback) {
-		cipher_suites[cipher_suites_size] =
-			GNUTLS_FALLBACK_SCSV_MAJOR;
-		cipher_suites[cipher_suites_size + 1] =
-			GNUTLS_FALLBACK_SCSV_MINOR;
-		cipher_suites_size += 2;
-	}
-
-	ret =
-	    _gnutls_buffer_append_data_prefix(cdata, 16, cipher_suites,
-					      cipher_suites_size);
-	if (ret < 0)
-		return gnutls_assert_val(ret);
-
-	ret = cdata->length - init_length;
-
-	return ret;
-}
-
-
 /* This function copies the appropriate compression methods, to a locally allocated buffer 
  * Needed in hello messages. Returns the new data length.
  */
 static int
-copy_comp_methods(gnutls_session_t session,
+append_null_comp(gnutls_session_t session,
 			  gnutls_buffer_st * cdata)
 {
-	int ret;
-	uint8_t compression_methods[MAX_ALGOS], comp_num;
+	uint8_t compression_methods[2] = {0x01, 0x00};
 	size_t init_length = cdata->length;
+	int ret;
 
 	ret =
-	    _gnutls_supported_compression_methods(session,
-						  compression_methods,
-						  MAX_ALGOS);
-	if (ret < 0)
-		return gnutls_assert_val(ret);
-
-	comp_num = ret;
-
-	/* put the number of compression methods */
-	ret = _gnutls_buffer_append_prefix(cdata, 8, comp_num);
-	if (ret < 0)
-		return gnutls_assert_val(ret);
-
-	ret =
-	    _gnutls_buffer_append_data(cdata, compression_methods,
-				       comp_num);
+	    _gnutls_buffer_append_data(cdata, compression_methods, 2);
 	if (ret < 0)
 		return gnutls_assert_val(ret);
 
@@ -1983,15 +1606,17 @@ static int send_client_hello(gnutls_session_t session, int again)
 	mbuffer_st *bufel = NULL;
 	int type;
 	int ret = 0;
-	const version_entry_st *hver;
+	const version_entry_st *hver, *min_ver;
 	uint8_t tver[2];
 	gnutls_buffer_st extdata;
 	int rehandshake = 0;
+	unsigned add_sr_scsv = 0;
 	uint8_t session_id_len =
 	    session->internals.resumed_security_parameters.session_id_size;
 
 	_gnutls_buffer_init(&extdata);
 
+
 	/* note that rehandshake is different than resuming
 	 */
 	if (session->security_parameters.session_id_size)
@@ -2036,6 +1661,13 @@ static int send_client_hello(gnutls_session_t session, int again)
 		_gnutls_handshake_log("HSK[%p]: Adv. version: %u.%u\n", session,
 				      (unsigned)tver[0], (unsigned)tver[1]);
 
+		min_ver = _gnutls_version_lowest(session);
+		if (min_ver == NULL) {
+			gnutls_assert();
+			ret = GNUTLS_E_NO_PRIORITIES_WERE_SET;
+			goto cleanup;
+		}
+
 		/* Set the version we advertized as maximum 
 		 * (RSA uses it).
 		 */
@@ -2043,22 +1675,18 @@ static int send_client_hello(gnutls_session_t session, int again)
 		if (_gnutls_set_current_version(session, hver->id) < 0)
 			return gnutls_assert_val(GNUTLS_E_UNSUPPORTED_VERSION_PACKET);
 
-		if (session->internals.priorities.min_record_version != 0) {
+		if (session->internals.priorities->min_record_version != 0) {
 			/* Advertize the lowest supported (SSL 3.0) record packet
 			 * version in record packets during the handshake.
 			 * That is to avoid confusing implementations
 			 * that do not support TLS 1.2 and don't know
 			 * how 3,3 version of record packets look like.
 			 */
-			const version_entry_st *v = _gnutls_version_lowest(session);
-
-			if (v == NULL) {
-				gnutls_assert();
-				return GNUTLS_E_NO_PRIORITIES_WERE_SET;
-			} else {
-				_gnutls_record_set_default_version(session,
-								   v->major, v->minor);
-			}
+			_gnutls_record_set_default_version(session,
+							   min_ver->major, min_ver->minor);
+		} else {
+			_gnutls_record_set_default_version(session,
+							   hver->major, hver->minor);
 		}
 
 		/* In order to know when this session was initiated.
@@ -2115,19 +1743,11 @@ static int send_client_hello(gnutls_session_t session, int again)
 		if (!session->internals.initial_negotiation_completed &&
 		    session->security_parameters.entity == GNUTLS_CLIENT &&
 		    (hver->id == GNUTLS_SSL3 &&
-		     session->internals.priorities.no_extensions != 0)) {
-			ret =
-			    copy_ciphersuites(session, &extdata,
-					      TRUE);
-			if (session->security_parameters.entity == GNUTLS_CLIENT)
-				_gnutls_extension_list_add(session,
-						   GNUTLS_EXTENSION_SAFE_RENEGOTIATION);
-		} else
+		     session->internals.priorities->no_extensions != 0)) {
+			add_sr_scsv = 1;
+		}
 #endif
-			ret =
-			    copy_ciphersuites(session, &extdata,
-					      FALSE);
-
+		ret = _gnutls_get_client_ciphersuites(session, &extdata, min_ver, hver, add_sr_scsv);
 		if (ret < 0) {
 			gnutls_assert();
 			goto cleanup;
@@ -2135,7 +1755,7 @@ static int send_client_hello(gnutls_session_t session, int again)
 
 		/* Copy the compression methods.
 		 */
-		ret = copy_comp_methods(session, &extdata);
+		ret = append_null_comp(session, &extdata);
 		if (ret < 0) {
 			gnutls_assert();
 			goto cleanup;
@@ -2143,7 +1763,7 @@ static int send_client_hello(gnutls_session_t session, int again)
 
 		/* Generate and copy TLS extensions.
 		 */
-		if (session->internals.priorities.no_extensions == 0) {
+		if (session->internals.priorities->no_extensions == 0) {
 			if (_gnutls_version_has_extensions(hver)) {
 				type = GNUTLS_EXT_ANY;
 			} else {
@@ -2196,7 +1816,6 @@ static int send_server_hello(gnutls_session_t session, int again)
 	gnutls_buffer_st extdata;
 	int pos = 0;
 	int datalen, ret = 0;
-	uint8_t comp;
 	uint8_t session_id_len =
 	    session->security_parameters.session_id_size;
 	char buf[2 * GNUTLS_MAX_SESSION_ID_SIZE + 1];
@@ -2254,14 +1873,10 @@ static int send_server_hello(gnutls_session_t session, int again)
 						      sizeof(buf), NULL));
 
 		memcpy(&data[pos],
-		       session->security_parameters.cipher_suite, 2);
+		       session->security_parameters.cs->id, 2);
 		pos += 2;
 
-		comp =
-		    _gnutls_compression_get_num(session->
-						security_parameters.
-						compression_method);
-		data[pos++] = comp;
+		data[pos++] = 0x00;
 
 		if (extdata.length > 0) {
 			memcpy(&data[pos], extdata.data, extdata.length);
@@ -2277,47 +1892,6 @@ static int send_server_hello(gnutls_session_t session, int again)
 	return ret;
 }
 
-static int send_hello(gnutls_session_t session, int again)
-{
-	int ret;
-
-	if (session->security_parameters.entity == GNUTLS_CLIENT) {
-		ret = send_client_hello(session, again);
-
-	} else {		/* SERVER */
-		ret = send_server_hello(session, again);
-	}
-
-	return ret;
-}
-
-/* RECEIVE A HELLO MESSAGE. This should be called from gnutls_recv_handshake_int only if a
- * hello message is expected. It uses the security_parameters.cipher_suite
- * and internals.compression_method.
- */
-static int
-recv_hello(gnutls_session_t session, uint8_t * data, int datalen)
-{
-	int ret;
-
-	if (session->security_parameters.entity == GNUTLS_CLIENT) {
-		ret = read_server_hello(session, data, datalen);
-		if (ret < 0) {
-			gnutls_assert();
-			return ret;
-		}
-	} else {		/* Server side reading a client hello */
-
-		ret = read_client_hello(session, data, datalen);
-		if (ret < 0) {
-			gnutls_assert();
-			return ret;
-		}
-	}
-
-	return 0;
-}
-
 static int
 recv_hello_verify_request(gnutls_session_t session,
 				  uint8_t * data, int datalen)
@@ -2588,10 +2162,11 @@ int gnutls_handshake(gnutls_session_t session)
 
 	if (STATE == STATE0) {
 		/* first call */
-		if (session->internals.priorities.protocol.algorithms == 0)
+		if (session->internals.priorities == NULL ||
+		    session->internals.priorities->cs.size == 0)
 			return gnutls_assert_val(GNUTLS_E_NO_PRIORITIES_WERE_SET);
 
-		session->internals.extensions_sent_size = 0;
+		session->internals.used_exts_size = 0;
 		session->internals.crt_requested = 0;
 		session->internals.handshake_in_progress = 1;
 		session->internals.vc_status = -1;
@@ -2714,7 +2289,6 @@ gnutls_handshake_set_timeout(gnutls_session_t session, unsigned int ms)
 			} \
 		} \
 		gnutls_assert(); \
-		ERR( str, ret); \
 		/* do not allow non-fatal errors at this point */ \
 		if (gnutls_error_is_fatal(ret) == 0) ret = gnutls_assert_val(GNUTLS_E_INTERNAL_ERROR); \
 		session_invalidate(session); \
@@ -2794,25 +2368,10 @@ static int handshake_client(gnutls_session_t session)
 {
 	int ret = 0;
 
-#ifdef HANDSHAKE_DEBUG
-	char buf[64];
-
-	if (session->internals.resumed_security_parameters.
-	    session_id_size > 0)
-		_gnutls_handshake_log("HSK[%p]: Ask to resume: %s\n",
-				      session,
-				      _gnutls_bin2hex(session->internals.
-						      resumed_security_parameters.
-						      session_id,
-						      session->internals.
-						      resumed_security_parameters.
-						      session_id_size, buf,
-						      sizeof(buf), NULL));
-#endif
 	switch (STATE) {
 	case STATE0:
 	case STATE1:
-		ret = send_hello(session, AGAIN(STATE1));
+		ret = send_client_hello(session, AGAIN(STATE1));
 		STATE = STATE1;
 		IMED_RET("send hello", ret, 1);
 		/* fall through */
@@ -3074,7 +2633,6 @@ static int send_handshake_final(gnutls_session_t session, int init)
 		FINAL_STATE = STATE0;
 
 		if (ret < 0) {
-			ERR("send ChangeCipherSpec", ret);
 			gnutls_assert();
 			return ret;
 		}
@@ -3100,7 +2658,6 @@ static int send_handshake_final(gnutls_session_t session, int init)
 		ret = _gnutls_send_finished(session, FAGAIN(STATE2));
 		FINAL_STATE = STATE2;
 		if (ret < 0) {
-			ERR("send Finished", ret);
 			gnutls_assert();
 			return ret;
 		}
@@ -3158,7 +2715,6 @@ static int recv_handshake_final(gnutls_session_t session, int init)
 		    _gnutls_recv_int(session, GNUTLS_CHANGE_CIPHER_SPEC,
 				     ccs, ccs_len, NULL, tleft);
 		if (ret <= 0) {
-			ERR("recv ChangeCipherSpec", ret);
 			gnutls_assert();
 			return (ret<0)?ret:GNUTLS_E_UNEXPECTED_PACKET;
 		}
@@ -3194,7 +2750,6 @@ static int recv_handshake_final(gnutls_session_t session, int init)
 
 		ret = _gnutls_recv_finished(session);
 		if (ret < 0) {
-			ERR("recv finished", ret);
 			gnutls_assert();
 			return ret;
 		}
@@ -3239,7 +2794,7 @@ static int handshake_server(gnutls_session_t session)
 		IMED_RET("recv hello", ret, 0);
 		/* fall through */
 	case STATE3:
-		ret = send_hello(session, AGAIN(STATE3));
+		ret = send_server_hello(session, AGAIN(STATE3));
 		STATE = STATE3;
 		IMED_RET("send hello", ret, 1);
 		/* fall through */
@@ -3502,8 +3057,6 @@ int _gnutls_handshake_get_session_hash(gnutls_session_t session, gnutls_datum_t
 	int ret;
 	const mac_entry_st *me;
 	uint8_t concat[2*MAX_HASH_SIZE];
-	digest_hd_st td_md5;
-	digest_hd_st td_sha;
 
 	if (unlikely(ver == NULL))
 		return gnutls_assert_val(GNUTLS_E_INTERNAL_ERROR);
@@ -3514,50 +3067,18 @@ int _gnutls_handshake_get_session_hash(gnutls_session_t session, gnutls_datum_t
 			return gnutls_assert_val(GNUTLS_E_INTERNAL_ERROR);
 	}
 
-	if (_gnutls_version_has_selectable_prf(ver)) { /* TLS 1.2+ */
-		gnutls_mac_algorithm_t prf;
+	me = mac_to_entry(session->security_parameters.prf_mac);
+	if (me == NULL)
+		return gnutls_assert_val(GNUTLS_E_INTERNAL_ERROR);
 
-		prf = _gnutls_cipher_suite_get_prf(session->security_parameters.cipher_suite);
-		if (prf == GNUTLS_MAC_UNKNOWN)
-			return gnutls_assert_val(GNUTLS_E_UNKNOWN_PK_ALGORITHM);
+	ret =
+	    _gnutls_hash_fast((gnutls_digest_algorithm_t)me->id,
+			      session->internals.handshake_hash_buffer.
+			      data,
+			      session->internals.handshake_hash_buffer_client_kx_len,
+			      concat);
+	if (ret < 0)
+		return gnutls_assert_val(ret);
 
-		me = mac_to_entry(prf);
-
-		ret =
-		    _gnutls_hash_fast((gnutls_digest_algorithm_t)me->id,
-				      session->internals.handshake_hash_buffer.
-				      data,
-				      session->internals.handshake_hash_buffer_client_kx_len,
-				      concat);
-		if (ret < 0)
-			return gnutls_assert_val(ret);
-
-		return _gnutls_set_datum(shash, concat, me->output_size);
-	} else {
-		ret = _gnutls_hash_init(&td_sha, hash_to_entry(GNUTLS_DIG_SHA1));
-		if (ret < 0) {
-			gnutls_assert();
-			return ret;
-		}
-
-		_gnutls_hash(&td_sha,
-			     session->internals.handshake_hash_buffer.data,
-			     session->internals.handshake_hash_buffer_client_kx_len);
-
-		_gnutls_hash_deinit(&td_sha, &concat[16]);
-
-		ret =
-		    _gnutls_hash_init(&td_md5,
-				      hash_to_entry(GNUTLS_DIG_MD5));
-		if (ret < 0)
-			return gnutls_assert_val(ret);
-
-		_gnutls_hash(&td_md5,
-			     session->internals.handshake_hash_buffer.data,
-			     session->internals.handshake_hash_buffer_client_kx_len);
-
-		_gnutls_hash_deinit(&td_md5, concat);
-
-		return _gnutls_set_datum(shash, concat, 36);
-	}
+	return _gnutls_set_datum(shash, concat, me->output_size);
 }
