@@ -1,5 +1,6 @@
 /*
  * Copyright (C) 2001-2014 Free Software Foundation, Inc.
+ * Copyright (C) 2017 Red Hat, Inc.
  *
  * Author: Nikos Mavrogiannopoulos
  *
@@ -37,7 +38,34 @@
 #include <random.h>
 #include <gnutls/crypto.h>
 
-/* encodes the Dss-Sig-Value structure
+/**
+ * gnutls_encode_rs_value:
+ * @sig_value: will hold a Dss-Sig-Value DER encoded structure
+ * @r: must contain the r value
+ * @s: must contain the s value
+ *
+ * This function will encode the provided r and s values, 
+ * into a Dss-Sig-Value structure, used for DSA and ECDSA
+ * signatures.
+ *
+ * The output value should be deallocated using gnutls_free().
+ *
+ * Returns: On success, %GNUTLS_E_SUCCESS (0) is returned, otherwise
+ *   an error code is returned.
+ *
+ * Since: 3.6.0
+ *
+ **/
+int
+gnutls_encode_rs_value(gnutls_datum_t * sig_value,
+			const gnutls_datum_t * r,
+			const gnutls_datum_t * s)
+{
+	return _gnutls_encode_ber_rs_raw(sig_value, r, s);
+}
+
+/* same as gnutls_encode_rs_value(), but kept since it used
+ * to be exported for FIPS140 CAVS testing.
  */
 int
 _gnutls_encode_ber_rs_raw(gnutls_datum_t * sig_value,
@@ -180,7 +208,7 @@ _gnutls_decode_ber_rs(const gnutls_datum_t * sig_value, bigint_t * r,
 	result = _gnutls_x509_read_int(sig, "s", s);
 	if (result < 0) {
 		gnutls_assert();
-		_gnutls_mpi_release(s);
+		_gnutls_mpi_release(r);
 		asn1_delete_structure(&sig);
 		return result;
 	}
@@ -190,6 +218,35 @@ _gnutls_decode_ber_rs(const gnutls_datum_t * sig_value, bigint_t * r,
 	return 0;
 }
 
+/**
+ * gnutls_decode_rs_value:
+ * @sig_value: holds a Dss-Sig-Value DER or BER encoded structure
+ * @r: will contain the r value
+ * @s: will contain the s value
+ *
+ * This function will decode the provided @sig_value, 
+ * into @r and @s elements. The Dss-Sig-Value is used for DSA and ECDSA
+ * signatures.
+ *
+ * The output values may be padded with a zero byte to prevent them
+ * from being interpreted as negative values. The value
+ * should be deallocated using gnutls_free().
+ *
+ * Returns: On success, %GNUTLS_E_SUCCESS (0) is returned, otherwise
+ *   an error code is returned.
+ *
+ * Since: 3.6.0
+ *
+ **/
+int gnutls_decode_rs_value(const gnutls_datum_t * sig_value, gnutls_datum_t *r,
+			   gnutls_datum_t *s)
+{
+	return _gnutls_decode_ber_rs_raw(sig_value, r, s);
+}
+
+/* same as gnutls_decode_rs_value(), but kept since it used
+ * to be exported for FIPS140 CAVS testing.
+ */
 int
 _gnutls_decode_ber_rs_raw(const gnutls_datum_t * sig_value, gnutls_datum_t *r,
 			  gnutls_datum_t *s)
@@ -242,12 +299,14 @@ int _gnutls_pk_params_copy(gnutls_pk_params_st * dst,
 	unsigned int i, j;
 	dst->params_nr = 0;
 
-	if (src == NULL || src->params_nr == 0) {
+	if (src == NULL || (src->params_nr == 0 && src->raw_pub.size == 0)) {
 		gnutls_assert();
 		return GNUTLS_E_INVALID_REQUEST;
 	}
 
-	dst->flags = src->flags;
+	dst->pkflags = src->pkflags;
+	dst->curve = src->curve;
+	dst->qbits = src->qbits;
 	dst->algo = src->algo;
 
 	for (i = 0; i < src->params_nr; i++) {
@@ -275,6 +334,8 @@ int _gnutls_pk_params_copy(gnutls_pk_params_st * dst,
 	}
 	dst->palgo = src->palgo;
 
+	memcpy(&dst->spki, &src->spki, sizeof(gnutls_x509_spki_st));
+
 	return 0;
 
 fail:
@@ -317,6 +378,24 @@ void gnutls_pk_params_clear(gnutls_pk_params_st * p)
 	}
 }
 
+unsigned
+_gnutls_find_rsa_pss_salt_size(unsigned bits, const mac_entry_st *me,
+			       unsigned salt_size)
+{
+	unsigned max_salt_size, digest_size;
+
+	digest_size = _gnutls_hash_get_algo_len(me);
+	max_salt_size = (bits + 7) / 8 - digest_size - 2;
+
+	if (salt_size < digest_size)
+		salt_size = digest_size;
+
+	if (salt_size > max_salt_size)
+		salt_size = max_salt_size;
+
+	return salt_size;
+}
+
 /* Writes the digest information and the digest in a DER encoded
  * structure. The digest info is allocated and stored into the info structure.
  */
@@ -530,22 +609,27 @@ _gnutls_params_get_rsa_raw(const gnutls_pk_params_st* params,
 				    gnutls_datum_t * d, gnutls_datum_t * p,
 				    gnutls_datum_t * q, gnutls_datum_t * u,
 				    gnutls_datum_t * e1,
-				    gnutls_datum_t * e2)
+				    gnutls_datum_t * e2,
+				    unsigned int flags)
 {
 	int ret;
+	mpi_dprint_func dprint = _gnutls_mpi_dprint_lz;
+
+	if (flags & GNUTLS_EXPORT_FLAG_NO_LZ)
+		dprint = _gnutls_mpi_dprint;
 
 	if (params == NULL) {
 		gnutls_assert();
 		return GNUTLS_E_INVALID_REQUEST;
 	}
 
-	if (params->algo != GNUTLS_PK_RSA) {
+	if (!GNUTLS_PK_IS_RSA(params->algo)) {
 		gnutls_assert();
 		return GNUTLS_E_INVALID_REQUEST;
 	}
 
 	if (m) {
-		ret = _gnutls_mpi_dprint_lz(params->params[0], m);
+		ret = dprint(params->params[0], m);
 		if (ret < 0) {
 			gnutls_assert();
 			goto error;
@@ -554,7 +638,7 @@ _gnutls_params_get_rsa_raw(const gnutls_pk_params_st* params,
 
 	/* E */
 	if (e) {
-		ret = _gnutls_mpi_dprint_lz(params->params[1], e);
+		ret = dprint(params->params[1], e);
 		if (ret < 0) {
 			gnutls_assert();
 			goto error;
@@ -563,7 +647,7 @@ _gnutls_params_get_rsa_raw(const gnutls_pk_params_st* params,
 
 	/* D */
 	if (d && params->params[2]) {
-		ret = _gnutls_mpi_dprint_lz(params->params[2], d);
+		ret = dprint(params->params[2], d);
 		if (ret < 0) {
 			gnutls_assert();
 			goto error;
@@ -575,7 +659,7 @@ _gnutls_params_get_rsa_raw(const gnutls_pk_params_st* params,
 
 	/* P */
 	if (p && params->params[3]) {
-		ret = _gnutls_mpi_dprint_lz(params->params[3], p);
+		ret = dprint(params->params[3], p);
 		if (ret < 0) {
 			gnutls_assert();
 			goto error;
@@ -587,7 +671,7 @@ _gnutls_params_get_rsa_raw(const gnutls_pk_params_st* params,
 
 	/* Q */
 	if (q && params->params[4]) {
-		ret = _gnutls_mpi_dprint_lz(params->params[4], q);
+		ret = dprint(params->params[4], q);
 		if (ret < 0) {
 			gnutls_assert();
 			goto error;
@@ -599,7 +683,7 @@ _gnutls_params_get_rsa_raw(const gnutls_pk_params_st* params,
 
 	/* U */
 	if (u && params->params[5]) {
-		ret = _gnutls_mpi_dprint_lz(params->params[5], u);
+		ret = dprint(params->params[5], u);
 		if (ret < 0) {
 			gnutls_assert();
 			goto error;
@@ -611,7 +695,7 @@ _gnutls_params_get_rsa_raw(const gnutls_pk_params_st* params,
 
 	/* E1 */
 	if (e1 && params->params[6]) {
-		ret = _gnutls_mpi_dprint_lz(params->params[6], e1);
+		ret = dprint(params->params[6], e1);
 		if (ret < 0) {
 			gnutls_assert();
 			goto error;
@@ -623,7 +707,7 @@ _gnutls_params_get_rsa_raw(const gnutls_pk_params_st* params,
 
 	/* E2 */
 	if (e2 && params->params[7]) {
-		ret = _gnutls_mpi_dprint_lz(params->params[7], e2);
+		ret = dprint(params->params[7], e2);
 		if (ret < 0) {
 			gnutls_assert();
 			goto error;
@@ -651,9 +735,13 @@ int
 _gnutls_params_get_dsa_raw(const gnutls_pk_params_st* params,
 			     gnutls_datum_t * p, gnutls_datum_t * q,
 			     gnutls_datum_t * g, gnutls_datum_t * y,
-			     gnutls_datum_t * x)
+			     gnutls_datum_t * x, unsigned int flags)
 {
 	int ret;
+	mpi_dprint_func dprint = _gnutls_mpi_dprint_lz;
+
+	if (flags & GNUTLS_EXPORT_FLAG_NO_LZ)
+		dprint = _gnutls_mpi_dprint;
 
 	if (params == NULL) {
 		gnutls_assert();
@@ -667,7 +755,7 @@ _gnutls_params_get_dsa_raw(const gnutls_pk_params_st* params,
 
 	/* P */
 	if (p) {
-		ret = _gnutls_mpi_dprint_lz(params->params[0], p);
+		ret = dprint(params->params[0], p);
 		if (ret < 0) {
 			gnutls_assert();
 			return ret;
@@ -676,7 +764,7 @@ _gnutls_params_get_dsa_raw(const gnutls_pk_params_st* params,
 
 	/* Q */
 	if (q) {
-		ret = _gnutls_mpi_dprint_lz(params->params[1], q);
+		ret = dprint(params->params[1], q);
 		if (ret < 0) {
 			gnutls_assert();
 			_gnutls_free_datum(p);
@@ -687,7 +775,7 @@ _gnutls_params_get_dsa_raw(const gnutls_pk_params_st* params,
 
 	/* G */
 	if (g) {
-		ret = _gnutls_mpi_dprint_lz(params->params[2], g);
+		ret = dprint(params->params[2], g);
 		if (ret < 0) {
 			gnutls_assert();
 			_gnutls_free_datum(p);
@@ -699,7 +787,7 @@ _gnutls_params_get_dsa_raw(const gnutls_pk_params_st* params,
 
 	/* Y */
 	if (y) {
-		ret = _gnutls_mpi_dprint_lz(params->params[3], y);
+		ret = dprint(params->params[3], y);
 		if (ret < 0) {
 			gnutls_assert();
 			_gnutls_free_datum(p);
@@ -711,7 +799,7 @@ _gnutls_params_get_dsa_raw(const gnutls_pk_params_st* params,
 
 	/* X */
 	if (x) {
-		ret = _gnutls_mpi_dprint_lz(params->params[4], x);
+		ret = dprint(params->params[4], x);
 		if (ret < 0) {
 			gnutls_assert();
 			_gnutls_free_datum(y);
@@ -729,9 +817,14 @@ int _gnutls_params_get_ecc_raw(const gnutls_pk_params_st* params,
 				       gnutls_ecc_curve_t * curve,
 				       gnutls_datum_t * x,
 				       gnutls_datum_t * y,
-				       gnutls_datum_t * k)
+				       gnutls_datum_t * k,
+				       unsigned int flags)
 {
 	int ret;
+	mpi_dprint_func dprint = _gnutls_mpi_dprint_lz;
+
+	if (flags & GNUTLS_EXPORT_FLAG_NO_LZ)
+		dprint = _gnutls_mpi_dprint;
 
 	if (params == NULL) {
 		gnutls_assert();
@@ -739,11 +832,36 @@ int _gnutls_params_get_ecc_raw(const gnutls_pk_params_st* params,
 	}
 
 	if (curve)
-		*curve = params->flags;
+		*curve = params->curve;
+
+	if (curve_is_eddsa(params->curve)) {
+		if (x) {
+			ret = _gnutls_set_datum(x, params->raw_pub.data, params->raw_pub.size);
+			if (ret < 0) {
+				return gnutls_assert_val(ret);
+			}
+		}
+
+		if (y) {
+			y->data = NULL;
+			y->size = 0;
+		}
+
+		if (k) {
+			ret = _gnutls_set_datum(k, params->raw_priv.data, params->raw_priv.size);
+			if (ret < 0) {
+				_gnutls_free_datum(x);
+				return gnutls_assert_val(ret);
+			}
+		}
+
+		return 0;
+	}
+
 
 	/* X */
 	if (x) {
-		ret = _gnutls_mpi_dprint_lz(params->params[ECC_X], x);
+		ret = dprint(params->params[ECC_X], x);
 		if (ret < 0) {
 			gnutls_assert();
 			return ret;
@@ -752,7 +870,7 @@ int _gnutls_params_get_ecc_raw(const gnutls_pk_params_st* params,
 
 	/* Y */
 	if (y) {
-		ret = _gnutls_mpi_dprint_lz(params->params[ECC_Y], y);
+		ret = dprint(params->params[ECC_Y], y);
 		if (ret < 0) {
 			gnutls_assert();
 			_gnutls_free_datum(x);
@@ -763,7 +881,7 @@ int _gnutls_params_get_ecc_raw(const gnutls_pk_params_st* params,
 
 	/* K */
 	if (k) {
-		ret = _gnutls_mpi_dprint_lz(params->params[ECC_K], k);
+		ret = dprint(params->params[ECC_K], k);
 		if (ret < 0) {
 			gnutls_assert();
 			_gnutls_free_datum(x);
@@ -833,8 +951,10 @@ pk_prepare_hash(gnutls_pk_algorithm_t pk,
 
 		_gnutls_free_datum(&old_digest);
 		break;
+	case GNUTLS_PK_RSA_PSS:
 	case GNUTLS_PK_DSA:
-	case GNUTLS_PK_EC:
+	case GNUTLS_PK_ECDSA:
+	case GNUTLS_PK_EDDSA_ED25519:
 		break;
 	default:
 		gnutls_assert();
