@@ -61,23 +61,28 @@ struct gnutls_pkcs11_obj_st {
 	struct pin_info_st pin;
 };
 
+struct gnutls_pkcs11_privkey_st {
+	gnutls_pk_algorithm_t pk_algorithm;
+	unsigned int rsa_pss_ok; /* if it is an RSA key, it can do RSA-PSS */
+
+	unsigned int flags;
+	struct p11_kit_uri *uinfo;
+	char *url;
+
+	struct pkcs11_session_info sinfo;
+	ck_object_handle_t ref;	/* the key in the session */
+	unsigned reauth; /* whether we need to login on each operation */
+
+	void *mutex; /* lock for operations requiring co-ordination */
+
+	struct pin_info_st pin;
+};
+
 /* This must be called on every function that uses a PKCS #11 function
  * directly. It can be provided a callback function to run when a reinitialization
  * occurs. */
 typedef int (*pkcs11_reinit_function)(void *priv);
-
-typedef enum init_level_t {
-	PROV_UNINITIALIZED = 0,
-	PROV_INIT_MANUAL,
-	PROV_INIT_MANUAL_TRUSTED,
-	PROV_INIT_TRUSTED,
-	PROV_INIT_ALL
-} init_level_t;
-
-/* See _gnutls_pkcs11_check_init() for possible Transitions.
- */
-
-int _gnutls_pkcs11_check_init(init_level_t req_level, void *priv, pkcs11_reinit_function cb);
+int _gnutls_pkcs11_check_init(void *priv, pkcs11_reinit_function cb);
 
 #define FIX_KEY_USAGE(pk, usage) \
 	if (usage == 0) { \
@@ -88,26 +93,15 @@ int _gnutls_pkcs11_check_init(init_level_t req_level, void *priv, pkcs11_reinit_
 	}
 
 #define PKCS11_CHECK_INIT \
-	ret = _gnutls_pkcs11_check_init(PROV_INIT_ALL, NULL, NULL); \
+	ret = _gnutls_pkcs11_check_init(NULL, NULL); \
 	if (ret < 0) \
 		return gnutls_assert_val(ret)
 
 #define PKCS11_CHECK_INIT_RET(x) \
-	ret = _gnutls_pkcs11_check_init(PROV_INIT_ALL, NULL, NULL); \
+	ret = _gnutls_pkcs11_check_init(NULL, NULL); \
 	if (ret < 0) \
 		return gnutls_assert_val(x)
 
-#define PKCS11_CHECK_INIT_FLAGS(f) \
-	ret = _gnutls_pkcs11_check_init((f & GNUTLS_PKCS11_OBJ_FLAG_PRESENT_IN_TRUSTED_MODULE)?PROV_INIT_TRUSTED:PROV_INIT_ALL, NULL, NULL); \
-	if (ret < 0) \
-		return gnutls_assert_val(ret)
-
-#define PKCS11_CHECK_INIT_FLAGS_RET(f, x) \
-	ret = _gnutls_pkcs11_check_init((f & GNUTLS_PKCS11_OBJ_FLAG_PRESENT_IN_TRUSTED_MODULE)?PROV_INIT_TRUSTED:PROV_INIT_ALL, NULL, NULL); \
-	if (ret < 0) \
-		return gnutls_assert_val(x)
-
-
 /* thus function is called for every token in the traverse_tokens
  * function. Once everything is traversed it is called with NULL tinfo.
  * It should return 0 if found what it was looking for.
@@ -179,9 +173,11 @@ int pkcs11_token_matches_info(struct p11_kit_uri *info,
 unsigned int pkcs11_obj_flags_to_int(unsigned int flags);
 
 int
-_gnutls_pkcs11_privkey_sign_hash(gnutls_pkcs11_privkey_t key,
-				 const gnutls_datum_t * hash,
-				 gnutls_datum_t * signature);
+_gnutls_pkcs11_privkey_sign(gnutls_pkcs11_privkey_t key,
+			    const gnutls_sign_entry_st *se,
+			    const gnutls_datum_t * hash,
+			    gnutls_datum_t * signature,
+			    gnutls_x509_spki_st *spki_params);
 
 int
 _gnutls_pkcs11_privkey_decrypt_data(gnutls_pkcs11_privkey_t key,
@@ -198,8 +194,12 @@ static inline int pk_to_mech(gnutls_pk_algorithm_t pk)
 		return CKM_DSA;
 	else if (pk == GNUTLS_PK_EC)
 		return CKM_ECDSA;
-	else
+	else if (pk == GNUTLS_PK_RSA)
 		return CKM_RSA_PKCS;
+	else if (pk == GNUTLS_PK_RSA_PSS)
+		return CKM_RSA_PKCS_PSS;
+	else
+		return -1;
 }
 
 static inline int pk_to_key_type(gnutls_pk_algorithm_t pk)
@@ -208,8 +208,10 @@ static inline int pk_to_key_type(gnutls_pk_algorithm_t pk)
 		return CKK_DSA;
 	else if (pk == GNUTLS_PK_EC)
 		return CKK_ECDSA;
-	else
+	else if (pk == GNUTLS_PK_RSA_PSS || pk == GNUTLS_PK_RSA)
 		return CKK_RSA;
+	else
+		return -1;
 }
 
 static inline gnutls_pk_algorithm_t key_type_to_pk(ck_key_type_t m)
@@ -232,9 +234,12 @@ static inline int pk_to_genmech(gnutls_pk_algorithm_t pk, ck_key_type_t *type)
 	} else if (pk == GNUTLS_PK_EC) {
 		*type = CKK_ECDSA;
 		return CKM_ECDSA_KEY_PAIR_GEN;
-	} else {
+	} else if (pk == GNUTLS_PK_RSA_PSS || pk == GNUTLS_PK_RSA) {
 		*type = CKK_RSA;
 		return CKM_RSA_PKCS_KEY_PAIR_GEN;
+	} else {
+		*type = -1;
+		return -1;
 	}
 }
 
@@ -319,6 +324,12 @@ pkcs11_get_mechanism_list(struct ck_function_list *module,
 			  ck_mechanism_type_t * mechanism_list,
 			  unsigned long *count);
 
+ck_rv_t
+pkcs11_get_mechanism_info(struct ck_function_list *module,
+			  ck_slot_id_t slot_id,
+			  ck_mechanism_type_t mechanism,
+			  struct ck_mechanism_info *ptr);
+
 ck_rv_t
 pkcs11_sign_init(struct ck_function_list *module,
 		 ck_session_handle_t sess,
