@@ -27,13 +27,14 @@
 #include <pkcs11_int.h>
 #include "pkcs11x.h"
 #include <x509/common.h>
+#include "pk.h"
 
 static const ck_bool_t tval = 1;
 static const ck_bool_t fval = 0;
 
 #define MAX_ASIZE 24
 
-static void mark_flags(unsigned flags, struct ck_attribute *a, unsigned *a_val)
+static void mark_flags(unsigned flags, struct ck_attribute *a, unsigned *a_val, unsigned trusted)
 {
 	static const unsigned long category = 2;
 
@@ -44,6 +45,17 @@ static void mark_flags(unsigned flags, struct ck_attribute *a, unsigned *a_val)
 		(*a_val)++;
 	}
 
+	if (flags & GNUTLS_PKCS11_OBJ_FLAG_MARK_DISTRUSTED) {
+		if (trusted) {
+			a[*a_val].type = CKA_X_DISTRUSTED;
+			a[*a_val].value = (void *) &tval;
+			a[*a_val].value_len = sizeof(tval);
+			(*a_val)++;
+		} else {
+			_gnutls_debug_log("p11: ignoring the distrusted flag as it is not valid on non-p11-kit-trust modules\n");
+		}
+	}
+
 	if (flags & GNUTLS_PKCS11_OBJ_FLAG_MARK_TRUSTED) {
 		a[*a_val].type = CKA_TRUSTED;
 		a[*a_val].value = (void *) &tval;
@@ -97,6 +109,7 @@ gnutls_pkcs11_copy_x509_crt2(const char *token_url,
 	struct p11_kit_uri *info = NULL;
 	ck_rv_t rv;
 	size_t der_size, id_size, serial_size;
+	gnutls_datum_t serial_der = {NULL, 0};
 	uint8_t *der = NULL;
 	uint8_t serial[128];
 	uint8_t id[20];
@@ -109,8 +122,6 @@ gnutls_pkcs11_copy_x509_crt2(const char *token_url,
 	
 	PKCS11_CHECK_INIT;
 
-	memset(&sinfo, 0, sizeof(sinfo));
-
 	ret = pkcs11_url_to_info(token_url, &info, 0);
 	if (ret < 0) {
 		gnutls_assert();
@@ -161,12 +172,12 @@ gnutls_pkcs11_copy_x509_crt2(const char *token_url,
 		id_size = sizeof(id);
 		ret = gnutls_x509_crt_get_subject_key_id(crt, id, &id_size, NULL);
 		if (ret < 0) {
-		        id_size = sizeof(id);
+			id_size = sizeof(id);
 			ret = gnutls_x509_crt_get_key_id(crt, 0, id, &id_size);
 			if (ret < 0) {
 			  gnutls_assert();
 			  goto cleanup;
-	                }
+			}
 		}
 
 		a[1].value = id;
@@ -201,10 +212,13 @@ gnutls_pkcs11_copy_x509_crt2(const char *token_url,
 
 	serial_size = sizeof(serial);
 	if (gnutls_x509_crt_get_serial(crt, serial, &serial_size) >= 0) {
-		a[a_val].type = CKA_SERIAL_NUMBER;
-		a[a_val].value = (void *) serial;
-		a[a_val].value_len = serial_size;
-		a_val++;
+		ret = _gnutls_x509_ext_gen_number(serial, serial_size, &serial_der);
+		if (ret >= 0) {
+			a[a_val].type = CKA_SERIAL_NUMBER;
+			a[a_val].value = (void *) serial_der.data;
+			a[a_val].value_len = serial_der.size;
+			a_val++;
+		}
 	}
 
 	if (label) {
@@ -214,7 +228,7 @@ gnutls_pkcs11_copy_x509_crt2(const char *token_url,
 		a_val++;
 	}
 
-	mark_flags(flags, a, &a_val);
+	mark_flags(flags, a, &a_val, sinfo.trusted);
 
 	rv = pkcs11_create_object(sinfo.module, sinfo.pks, a, a_val, &ctx);
 	if (rv != CKR_OK) {
@@ -231,6 +245,7 @@ gnutls_pkcs11_copy_x509_crt2(const char *token_url,
 
       cleanup:
 	gnutls_free(der);
+	gnutls_free(serial_der.data);
 	pkcs11_close_session(&sinfo);
 	return ret;
 
@@ -265,15 +280,20 @@ static int add_pubkey(gnutls_pubkey_t pubkey, struct ck_attribute *a, unsigned *
 	pk = gnutls_pubkey_get_pk_algorithm(pubkey, NULL);
 
 	switch (pk) {
+	case GNUTLS_PK_RSA_PSS:
 	case GNUTLS_PK_RSA: {
 		gnutls_datum_t m, e;
 
-		ret = gnutls_pubkey_export_rsa_raw(pubkey, &m, &e);
+		/* PKCS#11 defines integers as unsigned having most significant byte
+		 * first, e.g., 32768 = 0x80 0x00. This is interpreted literraly by
+		 * some HSMs which do not accept an integer with a leading zero */
+		ret = gnutls_pubkey_export_rsa_raw2(pubkey, &m, &e, GNUTLS_EXPORT_FLAG_NO_LZ);
 		if (ret < 0) {
 			gnutls_assert();
 			return ret;
 		}
 
+
 		a[*a_val].type = CKA_MODULUS;
 		a[*a_val].value = m.data;
 		a[*a_val].value_len = m.size;
@@ -288,7 +308,7 @@ static int add_pubkey(gnutls_pubkey_t pubkey, struct ck_attribute *a, unsigned *
 	case GNUTLS_PK_DSA: {
 		gnutls_datum_t p, q, g, y;
 
-		ret = gnutls_pubkey_export_dsa_raw(pubkey, &p, &q, &g, &y);
+		ret = gnutls_pubkey_export_dsa_raw2(pubkey, &p, &q, &g, &y, GNUTLS_EXPORT_FLAG_NO_LZ);
 		if (ret < 0) {
 			gnutls_assert();
 			return ret;
@@ -383,8 +403,6 @@ gnutls_pkcs11_copy_pubkey(const char *token_url,
 	
 	PKCS11_CHECK_INIT;
 
-	memset(&sinfo, 0, sizeof(sinfo));
-
 	ret = pkcs11_url_to_info(token_url, &info, 0);
 	if (ret < 0) {
 		gnutls_assert();
@@ -451,7 +469,7 @@ gnutls_pkcs11_copy_pubkey(const char *token_url,
 	}
 	a_val++;
 
-	mark_flags(flags, a, &a_val);
+	mark_flags(flags, a, &a_val, sinfo.trusted);
 
 	a[a_val].type = CKA_VERIFY;
 	if (key_usage & GNUTLS_KEY_DIGITAL_SIGNATURE) {
@@ -489,7 +507,7 @@ gnutls_pkcs11_copy_pubkey(const char *token_url,
 	ret = 0;
 
       cleanup:
-      	clean_pubkey(a, a_val);
+	clean_pubkey(a, a_val);
 	pkcs11_close_session(&sinfo);
 	return ret;
 
@@ -533,8 +551,6 @@ gnutls_pkcs11_copy_attached_extension(const char *token_url,
 	
 	PKCS11_CHECK_INIT;
 
-	memset(&sinfo, 0, sizeof(sinfo));
-
 	ret = pkcs11_url_to_info(token_url, &info, 0);
 	if (ret < 0) {
 		gnutls_assert();
@@ -641,8 +657,6 @@ gnutls_pkcs11_copy_x509_privkey2(const char *token_url,
 
 	PKCS11_CHECK_INIT;
 
-	memset(&sinfo, 0, sizeof(sinfo));
-
 	memset(&p, 0, sizeof(p));
 	memset(&q, 0, sizeof(q));
 	memset(&g, 0, sizeof(g));
@@ -782,14 +796,12 @@ gnutls_pkcs11_copy_x509_privkey2(const char *token_url,
 
 	switch (pk) {
 	case GNUTLS_PK_RSA:
+	case GNUTLS_PK_RSA_PSS:
 		{
 
-			ret =
-			    gnutls_x509_privkey_export_rsa_raw2(key, &m,
-								&e, &d, &p,
-								&q, &u,
-								&exp1,
-								&exp2);
+			ret = _gnutls_params_get_rsa_raw(&key->params, &m, &e, &d, &p,
+							 &q, &u, &exp1, &exp2,
+							 GNUTLS_EXPORT_FLAG_NO_LZ);
 			if (ret < 0) {
 				gnutls_assert();
 				goto cleanup;
@@ -841,9 +853,8 @@ gnutls_pkcs11_copy_x509_privkey2(const char *token_url,
 		}
 	case GNUTLS_PK_DSA:
 		{
-			ret =
-			    gnutls_x509_privkey_export_dsa_raw(key, &p, &q,
-							       &g, &y, &x);
+			ret = _gnutls_params_get_dsa_raw(&key->params, &p, &q, &g, &y, &x,
+							 GNUTLS_EXPORT_FLAG_NO_LZ);
 			if (ret < 0) {
 				gnutls_assert();
 				goto cleanup;
@@ -876,7 +887,7 @@ gnutls_pkcs11_copy_x509_privkey2(const char *token_url,
 	case GNUTLS_PK_EC:
 		{
 			ret =
-			    _gnutls_x509_write_ecc_params(key->params.flags,
+			    _gnutls_x509_write_ecc_params(key->params.curve,
 							  &p);
 			if (ret < 0) {
 				gnutls_assert();
@@ -884,8 +895,8 @@ gnutls_pkcs11_copy_x509_privkey2(const char *token_url,
 			}
 
 			ret =
-			    _gnutls_mpi_dprint_lz(key->params.
-						  params[ECC_K], &x);
+			    _gnutls_mpi_dprint(key->params.
+						params[ECC_K], &x);
 			if (ret < 0) {
 				gnutls_assert();
 				goto cleanup;
@@ -928,6 +939,7 @@ gnutls_pkcs11_copy_x509_privkey2(const char *token_url,
 
       cleanup:
 	switch (pk) {
+	case GNUTLS_PK_RSA_PSS:
 	case GNUTLS_PK_RSA:
 		{
 			gnutls_free(m.data);
@@ -975,8 +987,8 @@ struct delete_data_st {
 
 static int
 delete_obj_url_cb(struct ck_function_list *module, struct pkcs11_session_info *sinfo,
-	          struct ck_token_info *tinfo,
-	          struct ck_info *lib_info, void *input)
+		  struct ck_token_info *tinfo,
+		  struct ck_info *lib_info, void *input)
 {
 	struct delete_data_st *find_data = input;
 	struct ck_attribute a[4];
@@ -1152,7 +1164,7 @@ gnutls_pkcs11_token_init(const char *token_url,
 		return ret;
 	}
 
-	ret = pkcs11_find_slot(&module, &slot, info, NULL, NULL);
+	ret = pkcs11_find_slot(&module, &slot, info, NULL, NULL, NULL);
 	p11_kit_uri_free(info);
 
 	if (ret < 0) {
@@ -1204,8 +1216,6 @@ gnutls_pkcs11_token_set_pin(const char *token_url,
 
 	PKCS11_CHECK_INIT;
 
-	memset(&sinfo, 0, sizeof(sinfo));
-
 	ret = pkcs11_url_to_info(token_url, &info, 0);
 	if (ret < 0) {
 		gnutls_assert();
@@ -1281,8 +1291,6 @@ gnutls_pkcs11_token_get_random(const char *token_url,
 
 	PKCS11_CHECK_INIT;
 
-	memset(&sinfo, 0, sizeof(sinfo));
-
 	ret = pkcs11_url_to_info(token_url, &info, 0);
 	if (ret < 0) {
 		gnutls_assert();
