@@ -2,6 +2,7 @@
  * GnuTLS PKCS#11 support
  * Copyright (C) 2010-2014 Free Software Foundation, Inc.
  * Copyright (C) 2012-2015 Nikos Mavrogiannopoulos
+ * Copyright (C) 2016-2017 Red Hat, Inc.
  * 
  * Author: Nikos Mavrogiannopoulos
  *
@@ -29,20 +30,20 @@
 #include <gnutls/abstract.h>
 #include <pk.h>
 #include <x509_int.h>
-#include <openpgp/openpgp_int.h>
-#include <openpgp/openpgp.h>
 #include <tls-sig.h>
 #include <algorithms.h>
 #include <fips.h>
 #include <system-keys.h>
 #include "urls.h"
+#include "pkcs11_int.h"
 #include <abstract_int.h>
 
 static int
-_gnutls_privkey_sign_raw_data(gnutls_privkey_t key,
-			     unsigned flags,
-			     const gnutls_datum_t * data,
-			     gnutls_datum_t * signature);
+privkey_sign_prehashed(gnutls_privkey_t signer,
+		       const gnutls_sign_entry_st *se,
+		       const gnutls_datum_t * hash_data,
+		       gnutls_datum_t * signature,
+		       gnutls_x509_spki_st * params, unsigned flags);
 
 /**
  * gnutls_privkey_get_type:
@@ -123,21 +124,16 @@ int gnutls_privkey_verify_seed(gnutls_privkey_t key, gnutls_digest_algorithm_t d
 int gnutls_privkey_get_pk_algorithm(gnutls_privkey_t key, unsigned int *bits)
 {
 	switch (key->type) {
-#ifdef ENABLE_OPENPGP
-	case GNUTLS_PRIVKEY_OPENPGP:
-		return gnutls_openpgp_privkey_get_pk_algorithm(key->key.openpgp,
-							       bits);
-#endif
 #ifdef ENABLE_PKCS11
 	case GNUTLS_PRIVKEY_PKCS11:
 		return gnutls_pkcs11_privkey_get_pk_algorithm(key->key.pkcs11,
 							      bits);
 #endif
 	case GNUTLS_PRIVKEY_X509:
-		if (bits)
-			*bits =
-			    _gnutls_mpi_get_nbits(key->key.x509->
-						  params.params[0]);
+		if (bits) {
+			*bits = pubkey_to_bits(&key->key.x509->params);
+		}
+
 		return gnutls_x509_privkey_get_pk_algorithm(key->key.x509);
 	case GNUTLS_PRIVKEY_EXT:
 		if (bits)
@@ -157,9 +153,13 @@ privkey_to_pubkey(gnutls_pk_algorithm_t pk,
 	int ret;
 
 	pub->algo = priv->algo;
-	pub->flags = priv->flags;
+	pub->pkflags = priv->pkflags;
+	pub->curve = priv->curve;
+	pub->qbits = priv->qbits;
+	memcpy(&pub->spki, &priv->spki, sizeof(gnutls_x509_spki_st));
 
 	switch (pk) {
+	case GNUTLS_PK_RSA_PSS:
 	case GNUTLS_PK_RSA:
 		pub->params[0] = _gnutls_mpi_copy(priv->params[0]);
 		pub->params[1] = _gnutls_mpi_copy(priv->params[1]);
@@ -189,7 +189,7 @@ privkey_to_pubkey(gnutls_pk_algorithm_t pk,
 		}
 
 		break;
-	case GNUTLS_PK_EC:
+	case GNUTLS_PK_ECDSA:
 		pub->params[ECC_X] = _gnutls_mpi_copy(priv->params[ECC_X]);
 		pub->params[ECC_Y] = _gnutls_mpi_copy(priv->params[ECC_Y]);
 
@@ -201,6 +201,12 @@ privkey_to_pubkey(gnutls_pk_algorithm_t pk,
 			goto cleanup;
 		}
 
+		break;
+	case GNUTLS_PK_EDDSA_ED25519:
+		ret = _gnutls_set_datum(&pub->raw_pub, priv->raw_pub.data, priv->raw_pub.size);
+		if (ret < 0)
+			return gnutls_assert_val(ret);
+
 		break;
 	default:
 		gnutls_assert();
@@ -221,33 +227,6 @@ _gnutls_privkey_get_mpis(gnutls_privkey_t key, gnutls_pk_params_st * params)
 	int ret;
 
 	switch (key->type) {
-#ifdef ENABLE_OPENPGP
-	case GNUTLS_PRIVKEY_OPENPGP:
-		{
-			uint32_t kid[2];
-			uint8_t keyid[GNUTLS_OPENPGP_KEYID_SIZE];
-
-			ret =
-			    gnutls_openpgp_privkey_get_preferred_key_id
-			    (key->key.openpgp, keyid);
-			if (ret == 0) {
-				KEYID_IMPORT(kid, keyid);
-				ret =
-				    _gnutls_openpgp_privkey_get_mpis
-				    (key->key.openpgp, kid, params);
-			} else
-				ret =
-				    _gnutls_openpgp_privkey_get_mpis
-				    (key->key.openpgp, NULL, params);
-
-			if (ret < 0) {
-				gnutls_assert();
-				return ret;
-			}
-		}
-
-		break;
-#endif
 	case GNUTLS_PRIVKEY_X509:
 		ret = _gnutls_pk_params_copy(params, &key->key.x509->params);
 		break;
@@ -296,6 +275,91 @@ _gnutls_privkey_get_public_mpis(gnutls_privkey_t key,
 	return ret;
 }
 
+/* This function retrieves default sign parameters from KEY. */
+int
+_gnutls_privkey_get_spki_params(gnutls_privkey_t key,
+				gnutls_x509_spki_st * params)
+{
+	switch (key->type) {
+#ifdef ENABLE_PKCS11
+	case GNUTLS_PRIVKEY_PKCS11:
+		break;
+#endif
+	case GNUTLS_PRIVKEY_EXT:
+		break;
+	case GNUTLS_PRIVKEY_X509:
+		_gnutls_x509_privkey_get_spki_params(key->key.x509, params);
+		return 0;
+	default:
+		gnutls_assert();
+		return GNUTLS_E_INVALID_REQUEST;
+	}
+
+	memset(params, 0, sizeof(gnutls_x509_spki_st));
+
+	return 0;
+}
+
+/* This function fills in PARAMS with the necessary parameters to sign
+ * with PK and DIG. PARAMS must be initialized with
+ * _gnutls_privkey_get_spki_params in advance.
+ *
+ * After calling this function the params structure will
+ * be initialized even if the original SubjectPublicKeyInfo was empty.
+ */
+int
+_gnutls_privkey_update_spki_params(gnutls_privkey_t key,
+				 gnutls_pk_algorithm_t pk,
+				 gnutls_digest_algorithm_t dig,
+				 unsigned flags,
+				 gnutls_x509_spki_st *params)
+{
+	unsigned salt_size = 0;
+	unsigned bits = 0;
+	gnutls_pk_algorithm_t key_pk;
+
+	if (flags & GNUTLS_PRIVKEY_SIGN_FLAG_RSA_PSS) {
+		if (!GNUTLS_PK_IS_RSA(pk))
+			return gnutls_assert_val(GNUTLS_E_INVALID_REQUEST);
+		pk = GNUTLS_PK_RSA_PSS;
+	}
+
+	key_pk = gnutls_privkey_get_pk_algorithm(key, &bits);
+	if ((key_pk != pk) &&
+	      !(key_pk == GNUTLS_PK_RSA && pk == GNUTLS_PK_RSA_PSS)) {
+		gnutls_assert();
+		return GNUTLS_E_CONSTRAINT_ERROR;
+	}
+
+	if (pk == GNUTLS_PK_RSA_PSS) {
+		const mac_entry_st *me;
+
+		me = hash_to_entry(dig);
+		if (unlikely(me == NULL))
+			return gnutls_assert_val(GNUTLS_E_INVALID_REQUEST);
+
+		if (params->pk == GNUTLS_PK_RSA)
+			salt_size = 0;
+		else if (params->pk == GNUTLS_PK_RSA_PSS) {
+			if (params->rsa_pss_dig != GNUTLS_DIG_UNKNOWN && dig != params->rsa_pss_dig) {
+				return gnutls_assert_val(GNUTLS_E_CONSTRAINT_ERROR);
+			}
+
+			salt_size = params->salt_size;
+		}
+
+		if (flags & GNUTLS_PRIVKEY_FLAG_REPRODUCIBLE)
+			params->salt_size = 0;
+		else
+			params->salt_size = _gnutls_find_rsa_pss_salt_size(bits, me, salt_size);
+		params->rsa_pss_dig = dig;
+	}
+
+	params->pk = pk;
+
+	return 0;
+}
+
 /**
  * gnutls_privkey_init:
  * @key: A pointer to the type to be initialized
@@ -345,11 +409,6 @@ void gnutls_privkey_deinit(gnutls_privkey_t key)
 	if (key->flags & GNUTLS_PRIVKEY_IMPORT_AUTO_RELEASE
 	    || key->flags & GNUTLS_PRIVKEY_IMPORT_COPY)
 		switch (key->type) {
-#ifdef ENABLE_OPENPGP
-		case GNUTLS_PRIVKEY_OPENPGP:
-			gnutls_openpgp_privkey_deinit(key->key.openpgp);
-			break;
-#endif
 #ifdef ENABLE_PKCS11
 		case GNUTLS_PRIVKEY_PKCS11:
 			gnutls_pkcs11_privkey_deinit(key->key.pkcs11);
@@ -575,6 +634,9 @@ gnutls_privkey_import_ext(gnutls_privkey_t pkey,
 					  decrypt_func, NULL, flags);
 }
 
+#define PK_IS_OK_FOR_EXT2(pk) \
+	((pk == GNUTLS_PK_RSA) || (pk == GNUTLS_PK_ECDSA) || (pk == GNUTLS_PK_DSA))
+
 /**
  * gnutls_privkey_import_ext2:
  * @pkey: The private key
@@ -619,6 +681,9 @@ gnutls_privkey_import_ext2(gnutls_privkey_t pkey,
 		return ret;
 	}
 
+	if (!PK_IS_OK_FOR_EXT2(pk))
+		return gnutls_assert_val(GNUTLS_E_INVALID_REQUEST);
+
 	if (sign_fn == NULL && decrypt_fn == NULL)
 		return gnutls_assert_val(GNUTLS_E_INVALID_REQUEST);
 
@@ -699,6 +764,96 @@ gnutls_privkey_import_ext3(gnutls_privkey_t pkey,
 
 	pkey->pk_algorithm = pkey->key.ext.info_func(pkey, GNUTLS_PRIVKEY_INFO_PK_ALGO, pkey->key.ext.userdata);
 
+	if (!PK_IS_OK_FOR_EXT2(pkey->pk_algorithm))
+		return gnutls_assert_val(GNUTLS_E_INVALID_REQUEST);
+
+	/* Ensure gnutls_privkey_deinit() calls the deinit_func */
+	if (deinit_fn)
+		pkey->flags |= GNUTLS_PRIVKEY_IMPORT_AUTO_RELEASE;
+
+	return 0;
+}
+
+/**
+ * gnutls_privkey_import_ext4:
+ * @pkey: The private key
+ * @userdata: private data to be provided to the callbacks
+ * @sign_data_fn: callback for signature operations (may be %NULL)
+ * @sign_hash_fn: callback for signature operations (may be %NULL)
+ * @decrypt_fn: callback for decryption operations (may be %NULL)
+ * @deinit_fn: a deinitialization function
+ * @info_fn: returns info about the public key algorithm (should not be %NULL)
+ * @flags: Flags for the import
+ *
+ * This function will associate the given callbacks with the
+ * #gnutls_privkey_t type. At least one of the callbacks
+ * must be non-null. If a deinitialization function is provided
+ * then flags is assumed to contain %GNUTLS_PRIVKEY_IMPORT_AUTO_RELEASE.
+ *
+ * Note that in contrast with the signing function of
+ * gnutls_privkey_import_ext3(), the signing functions provided to this
+ * function take explicitly the signature algorithm as parameter and
+ * different functions are provided to sign the data and hashes.
+ *
+ * The @sign_hash_fn is to be called to sign pre-hashed data. The input
+ * to the callback is the output of the hash (such as SHA256) corresponding
+ * to the signature algorithm. For RSA PKCS#1 signatures, the signature
+ * algorithm can be set to %GNUTLS_SIGN_RSA_RAW, and in that case the data
+ * should be handled as if they were an RSA PKCS#1 DigestInfo structure.
+ *
+ * The @sign_data_fn is to be called to sign data. The input data will be
+ * he data to be signed (and hashed), with the provided signature
+ * algorithm. This function is to be used for signature algorithms like
+ * Ed25519 which cannot take pre-hashed data as input.
+ *
+ * When both @sign_data_fn and @sign_hash_fn functions are provided they
+ * must be able to operate on all the supported signature algorithms,
+ * unless prohibited by the type of the algorithm (e.g., as with Ed25519).
+ *
+ * The @info_fn must provide information on the signature algorithms supported by
+ * this private key, and should support the flags %GNUTLS_PRIVKEY_INFO_PK_ALGO and
+ * %GNUTLS_PRIVKEY_INFO_HAVE_SIGN_ALGO. It must return -1 on unknown flags.
+ *
+ * Returns: On success, %GNUTLS_E_SUCCESS (0) is returned, otherwise a
+ *   negative error value.
+ *
+ * Since: 3.6.0
+ **/
+int
+gnutls_privkey_import_ext4(gnutls_privkey_t pkey,
+                           void *userdata,
+                           gnutls_privkey_sign_data_func sign_data_fn,
+                           gnutls_privkey_sign_hash_func sign_hash_fn,
+                           gnutls_privkey_decrypt_func decrypt_fn,
+                           gnutls_privkey_deinit_func deinit_fn,
+                           gnutls_privkey_info_func info_fn,
+                           unsigned int flags)
+{
+	int ret;
+
+	ret = check_if_clean(pkey);
+	if (ret < 0) {
+		gnutls_assert();
+		return ret;
+	}
+
+	if (sign_data_fn == NULL && sign_hash_fn == NULL && decrypt_fn == NULL)
+		return gnutls_assert_val(GNUTLS_E_INVALID_REQUEST);
+
+	if (info_fn == NULL)
+		return gnutls_assert_val(GNUTLS_E_INVALID_REQUEST);
+
+	pkey->key.ext.sign_data_func = sign_data_fn;
+	pkey->key.ext.sign_hash_func = sign_hash_fn;
+	pkey->key.ext.decrypt_func = decrypt_fn;
+	pkey->key.ext.deinit_func = deinit_fn;
+	pkey->key.ext.info_func = info_fn;
+	pkey->key.ext.userdata = userdata;
+	pkey->type = GNUTLS_PRIVKEY_EXT;
+	pkey->flags = flags;
+
+	pkey->pk_algorithm = pkey->key.ext.info_func(pkey, GNUTLS_PRIVKEY_INFO_PK_ALGO, pkey->key.ext.userdata);
+
 	/* Ensure gnutls_privkey_deinit() calls the deinit_func */
 	if (deinit_fn)
 		pkey->flags |= GNUTLS_PRIVKEY_IMPORT_AUTO_RELEASE;
@@ -764,8 +919,9 @@ gnutls_privkey_import_x509(gnutls_privkey_t pkey,
  * @key: Location for the key to be exported.
  *
  * Converts the given abstract private key to a #gnutls_x509_privkey_t
- * type. The key must be of type %GNUTLS_PRIVKEY_X509. The key returned
- * in @key must be deinitialized with gnutls_x509_privkey_deinit().
+ * type. The abstract key must be of type %GNUTLS_PRIVKEY_X509. The input
+ * @key must not be initialized. The key returned in @key should be deinitialized
+ * using gnutls_x509_privkey_deinit().
  *
  * Returns: On success, %GNUTLS_E_SUCCESS (0) is returned, otherwise a
  *   negative error value.
@@ -883,177 +1039,6 @@ gnutls_privkey_generate2(gnutls_privkey_t pkey,
 	return 0;
 }
 
-#ifdef ENABLE_OPENPGP
-/**
- * gnutls_privkey_import_openpgp:
- * @pkey: The private key
- * @key: The private key to be imported
- * @flags: Flags for the import
- *
- * This function will import the given private key to the abstract
- * #gnutls_privkey_t type.
- *
- * The #gnutls_openpgp_privkey_t object must not be deallocated
- * during the lifetime of this structure. The subkey set as
- * preferred will be used, or the master key otherwise.
- *
- * @flags might be zero or one of %GNUTLS_PRIVKEY_IMPORT_AUTO_RELEASE
- * and %GNUTLS_PRIVKEY_IMPORT_COPY.
- *
- * Returns: On success, %GNUTLS_E_SUCCESS (0) is returned, otherwise a
- *   negative error value.
- *
- * Since: 2.12.0
- **/
-int
-gnutls_privkey_import_openpgp(gnutls_privkey_t pkey,
-			      gnutls_openpgp_privkey_t key, unsigned int flags)
-{
-	int ret, idx;
-	uint8_t keyid[GNUTLS_OPENPGP_KEYID_SIZE];
-
-	ret = check_if_clean(pkey);
-	if (ret < 0) {
-		gnutls_assert();
-		return ret;
-	}
-
-	if (flags & GNUTLS_PRIVKEY_IMPORT_COPY) {
-		ret = gnutls_openpgp_privkey_init(&pkey->key.openpgp);
-		if (ret < 0)
-			return gnutls_assert_val(ret);
-
-		ret = _gnutls_openpgp_privkey_cpy(pkey->key.openpgp, key);
-		if (ret < 0) {
-			gnutls_openpgp_privkey_deinit(pkey->key.openpgp);
-			return gnutls_assert_val(ret);
-		}
-	} else
-		pkey->key.openpgp = key;
-
-	pkey->type = GNUTLS_PRIVKEY_OPENPGP;
-
-	ret = gnutls_openpgp_privkey_get_preferred_key_id(key, keyid);
-	if (ret == GNUTLS_E_OPENPGP_PREFERRED_KEY_ERROR) {
-		pkey->pk_algorithm =
-		    gnutls_openpgp_privkey_get_pk_algorithm(key, NULL);
-	} else {
-		if (ret < 0)
-			return gnutls_assert_val(ret);
-
-		idx = gnutls_openpgp_privkey_get_subkey_idx(key, keyid);
-
-		pkey->pk_algorithm =
-		    gnutls_openpgp_privkey_get_subkey_pk_algorithm(key,
-								   idx, NULL);
-	}
-
-	pkey->flags = flags;
-
-	return 0;
-}
-
-/**
- * gnutls_privkey_import_openpgp_raw:
- * @pkey: The private key
- * @data: The private key data to be imported
- * @format: The format of the private key
- * @keyid: The key id to use (optional)
- * @password: A password (optional)
- *
- * This function will import the given private key to the abstract
- * #gnutls_privkey_t type. 
- *
- * Returns: On success, %GNUTLS_E_SUCCESS (0) is returned, otherwise a
- *   negative error value.
- *
- * Since: 3.1.0
- **/
-int gnutls_privkey_import_openpgp_raw(gnutls_privkey_t pkey,
-				      const gnutls_datum_t * data,
-				      gnutls_openpgp_crt_fmt_t format,
-				      const gnutls_openpgp_keyid_t keyid,
-				      const char *password)
-{
-	gnutls_openpgp_privkey_t xpriv;
-	int ret;
-
-	ret = gnutls_openpgp_privkey_init(&xpriv);
-	if (ret < 0)
-		return gnutls_assert_val(ret);
-
-	ret = gnutls_openpgp_privkey_import(xpriv, data, format, password, 0);
-	if (ret < 0) {
-		gnutls_assert();
-		goto cleanup;
-	}
-
-	if (keyid) {
-		ret = gnutls_openpgp_privkey_set_preferred_key_id(xpriv, keyid);
-		if (ret < 0) {
-			gnutls_assert();
-			goto cleanup;
-		}
-	}
-
-	ret =
-	    gnutls_privkey_import_openpgp(pkey, xpriv,
-					  GNUTLS_PRIVKEY_IMPORT_AUTO_RELEASE);
-	if (ret < 0) {
-		gnutls_assert();
-		goto cleanup;
-	}
-
-	return 0;
-
- cleanup:
-	gnutls_openpgp_privkey_deinit(xpriv);
-
-	return ret;
-}
-
-/**
- * gnutls_privkey_export_openpgp:
- * @pkey: The private key
- * @key: Location for the key to be exported.
- *
- * Converts the given abstract private key to a #gnutls_openpgp_privkey_t
- * type. The key must be of type %GNUTLS_PRIVKEY_OPENPGP. The key
- * returned in @key must be deinitialized with
- * gnutls_openpgp_privkey_deinit().
- *
- * Returns: On success, %GNUTLS_E_SUCCESS (0) is returned, otherwise a
- *   negative error value.
- *
- * Since: 3.4.0
- */
-int
-gnutls_privkey_export_openpgp(gnutls_privkey_t pkey,
-			      gnutls_openpgp_privkey_t *key)
-{
-	int ret;
-
-	if (pkey->type != GNUTLS_PRIVKEY_OPENPGP) {
-		gnutls_assert();
-		return GNUTLS_E_INVALID_REQUEST;
-	}
-
-	ret = gnutls_openpgp_privkey_init(key);
-	if (ret < 0)
-		return gnutls_assert_val(ret);
-
-	ret = _gnutls_openpgp_privkey_cpy(*key, pkey->key.openpgp);
-	if (ret < 0) {
-		gnutls_openpgp_privkey_deinit(*key);
-		*key = NULL;
-
-		return gnutls_assert_val(ret);
-	}
-
-	return 0;
-}
-#endif
-
 /**
  * gnutls_privkey_sign_data:
  * @signer: Holds the key
@@ -1083,25 +1068,173 @@ gnutls_privkey_sign_data(gnutls_privkey_t signer,
 			 gnutls_datum_t * signature)
 {
 	int ret;
-	gnutls_datum_t digest;
-	const mac_entry_st *me = hash_to_entry(hash);
+	gnutls_x509_spki_st params;
 
 	if (flags & GNUTLS_PRIVKEY_SIGN_FLAG_TLS1_RSA)
 		return gnutls_assert_val(GNUTLS_E_INVALID_REQUEST);
 
-	ret = pk_hash_data(signer->pk_algorithm, me, NULL, data, &digest);
+	ret = _gnutls_privkey_get_spki_params(signer, &params);
 	if (ret < 0) {
 		gnutls_assert();
 		return ret;
 	}
 
-	ret = pk_prepare_hash(signer->pk_algorithm, me, &digest);
+	ret = _gnutls_privkey_update_spki_params(signer, signer->pk_algorithm,
+					         hash, flags, &params);
+	if (ret < 0) {
+		gnutls_assert();
+		return ret;
+	}
+
+	return privkey_sign_and_hash_data(signer, _gnutls_pk_to_sign_entry(params.pk, hash), data, signature, &params);
+}
+
+/**
+ * gnutls_privkey_sign_data2:
+ * @signer: Holds the key
+ * @algo: The signature algorithm used
+ * @flags: Zero or one of %gnutls_privkey_flags_t
+ * @data: holds the data to be signed
+ * @signature: will contain the signature allocated with gnutls_malloc()
+ *
+ * This function will sign the given data using the specified signature
+ * algorithm. This function is an enhancement of gnutls_privkey_sign_data(),
+ * as it allows utilizing a alternative signature algorithm where possible
+ * (e.g, use an RSA key with RSA-PSS).
+ *
+ * Returns: On success, %GNUTLS_E_SUCCESS (0) is returned, otherwise a
+ * negative error value.
+ *
+ * Since: 3.6.0
+ **/
+int
+gnutls_privkey_sign_data2(gnutls_privkey_t signer,
+			  gnutls_sign_algorithm_t algo,
+			  unsigned int flags,
+			  const gnutls_datum_t * data,
+			  gnutls_datum_t * signature)
+{
+	int ret;
+	gnutls_x509_spki_st params;
+	const gnutls_sign_entry_st *e;
+
+	if (flags & GNUTLS_PRIVKEY_SIGN_FLAG_TLS1_RSA)
+		return gnutls_assert_val(GNUTLS_E_INVALID_REQUEST);
+
+	e = _gnutls_sign_to_entry(algo);
+	if (e == NULL)
+		return gnutls_assert_val(GNUTLS_E_INVALID_REQUEST);
+
+	ret = _gnutls_privkey_get_spki_params(signer, &params);
+	if (ret < 0) {
+		gnutls_assert();
+		return ret;
+	}
+
+	ret = _gnutls_privkey_update_spki_params(signer, e->pk, e->hash,
+					         flags, &params);
+	if (ret < 0) {
+		gnutls_assert();
+		return ret;
+	}
+
+	return privkey_sign_and_hash_data(signer, _gnutls_sign_to_entry(algo), data, signature, &params);
+}
+
+/**
+ * gnutls_privkey_sign_hash2:
+ * @signer: Holds the signer's key
+ * @algo: The signature algorithm used
+ * @flags: Zero or one of %gnutls_privkey_flags_t
+ * @hash_data: holds the data to be signed
+ * @signature: will contain newly allocated signature
+ *
+ * This function will sign the given hashed data using a signature algorithm
+ * supported by the private key. Signature algorithms are always used
+ * together with a hash functions.  Different hash functions may be
+ * used for the RSA algorithm, but only SHA-XXX for the DSA keys.
+ *
+ * You may use gnutls_pubkey_get_preferred_hash_algorithm() to determine
+ * the hash algorithm.
+ *
+ * The flags may be %GNUTLS_PRIVKEY_SIGN_FLAG_TLS1_RSA or %GNUTLS_PRIVKEY_SIGN_FLAG_RSA_PSS.
+ * In the former case this function will ignore @hash_algo and perform a raw PKCS1 signature,
+ * and in the latter an RSA-PSS signature will be generated.
+ *
+ * Note that, not all algorithm support signing already hashed data. When
+ * signing with Ed25519, gnutls_privkey_sign_data() should be used.
+ *
+ * Returns: On success, %GNUTLS_E_SUCCESS (0) is returned, otherwise a
+ *   negative error value.
+ *
+ * Since: 3.6.0
+ **/
+int
+gnutls_privkey_sign_hash2(gnutls_privkey_t signer,
+			  gnutls_sign_algorithm_t algo,
+			  unsigned int flags,
+			  const gnutls_datum_t * hash_data,
+			  gnutls_datum_t * signature)
+{
+	int ret;
+	gnutls_x509_spki_st params;
+	const gnutls_sign_entry_st *se;
+
+	se = _gnutls_sign_to_entry(algo);
+	if (unlikely(se == NULL))
+		return gnutls_assert_val(GNUTLS_E_INVALID_REQUEST);
+
+	ret = _gnutls_privkey_get_spki_params(signer, &params);
+	if (ret < 0) {
+		gnutls_assert();
+		return ret;
+	}
+
+	ret = _gnutls_privkey_update_spki_params(signer, se->pk, se->hash,
+					         flags, &params);
+	if (ret < 0) {
+		gnutls_assert();
+		return ret;
+	}
+
+	return privkey_sign_prehashed(signer, se, hash_data, signature, &params, flags);
+}
+
+int
+privkey_sign_and_hash_data(gnutls_privkey_t signer,
+			   const gnutls_sign_entry_st *se,
+			   const gnutls_datum_t * data,
+			   gnutls_datum_t * signature,
+			   gnutls_x509_spki_st * params)
+{
+	int ret;
+	gnutls_datum_t digest;
+	const mac_entry_st *me;
+
+	if (unlikely(se == NULL))
+		return gnutls_assert_val(GNUTLS_E_INVALID_REQUEST);
+
+	if (_gnutls_pk_is_not_prehashed(se->pk)) {
+		return privkey_sign_raw_data(signer, se, data, signature, params);
+	}
+
+	me = hash_to_entry(se->hash);
+	if (me == NULL)
+		return gnutls_assert_val(GNUTLS_E_INVALID_REQUEST);
+
+	ret = pk_hash_data(se->pk, me, NULL, data, &digest);
+	if (ret < 0) {
+		gnutls_assert();
+		return ret;
+	}
+
+	ret = pk_prepare_hash(se->pk, me, &digest);
 	if (ret < 0) {
 		gnutls_assert();
 		goto cleanup;
 	}
 
-	ret = _gnutls_privkey_sign_raw_data(signer, flags, &digest, signature);
+	ret = privkey_sign_raw_data(signer, se, &digest, signature, params);
 	_gnutls_free_datum(&digest);
 
 	if (ret < 0) {
@@ -1116,6 +1249,7 @@ gnutls_privkey_sign_data(gnutls_privkey_t signer,
 	return ret;
 }
 
+
 /**
  * gnutls_privkey_sign_hash:
  * @signer: Holds the signer's key
@@ -1132,8 +1266,12 @@ gnutls_privkey_sign_data(gnutls_privkey_t signer,
  * You may use gnutls_pubkey_get_preferred_hash_algorithm() to determine
  * the hash algorithm.
  *
- * Note that if %GNUTLS_PRIVKEY_SIGN_FLAG_TLS1_RSA flag is specified this function
- * will ignore @hash_algo and perform a raw PKCS1 signature.
+ * The flags may be %GNUTLS_PRIVKEY_SIGN_FLAG_TLS1_RSA or %GNUTLS_PRIVKEY_SIGN_FLAG_RSA_PSS.
+ * In the former case this function will ignore @hash_algo and perform a raw PKCS1 signature,
+ * and in the latter an RSA-PSS signature will be generated.
+ *
+ * Note that, not all algorithm support signing already hashed data. When
+ * signing with Ed25519, gnutls_privkey_sign_data() should be used.
  *
  * Returns: On success, %GNUTLS_E_SUCCESS (0) is returned, otherwise a
  *   negative error value.
@@ -1146,13 +1284,70 @@ gnutls_privkey_sign_hash(gnutls_privkey_t signer,
 			 unsigned int flags,
 			 const gnutls_datum_t * hash_data,
 			 gnutls_datum_t * signature)
+{
+	int ret;
+	gnutls_x509_spki_st params;
+	const gnutls_sign_entry_st *se;
+
+	ret = _gnutls_privkey_get_spki_params(signer, &params);
+	if (ret < 0) {
+		gnutls_assert();
+		return ret;
+	}
+
+	ret = _gnutls_privkey_update_spki_params(signer, signer->pk_algorithm,
+					       hash_algo, flags, &params);
+	if (ret < 0) {
+		gnutls_assert();
+		return ret;
+	}
+
+	/* legacy callers of this API could use a hash algorithm of 0 (unknown)
+	 * to indicate raw hashing. As we now always want to know the signing
+	 * algorithm involved, we try discovering the hash algorithm. */
+	if (hash_algo == 0 && (params.pk == GNUTLS_PK_DSA || params.pk == GNUTLS_PK_ECDSA)) {
+		hash_algo = _gnutls_hash_size_to_sha_hash(hash_data->size);
+	}
+
+	if (params.pk == GNUTLS_PK_RSA && (flags & GNUTLS_PRIVKEY_SIGN_FLAG_TLS1_RSA)) {
+		/* the corresponding signature algorithm is SIGN_RSA_RAW,
+		 * irrespective of hash algorithm. */
+		se = _gnutls_sign_to_entry(GNUTLS_SIGN_RSA_RAW);
+	} else {
+		se = _gnutls_pk_to_sign_entry(params.pk, hash_algo);
+	}
+
+	if (unlikely(se == NULL))
+		return gnutls_assert_val(GNUTLS_E_INVALID_REQUEST);
+
+	return privkey_sign_prehashed(signer, se,
+				      hash_data, signature, &params, flags);
+}
+
+static int
+privkey_sign_prehashed(gnutls_privkey_t signer,
+		       const gnutls_sign_entry_st *se,
+		       const gnutls_datum_t * hash_data,
+		       gnutls_datum_t * signature,
+		       gnutls_x509_spki_st * params,
+		       unsigned flags)
 {
 	int ret;
 	gnutls_datum_t digest;
 
-	if (flags & GNUTLS_PRIVKEY_SIGN_FLAG_TLS1_RSA)
-		return _gnutls_privkey_sign_raw_data(signer, flags,
-						    hash_data, signature);
+	if (unlikely(se == NULL))
+		return gnutls_assert_val(GNUTLS_E_INVALID_REQUEST);
+
+	if (se->id == GNUTLS_SIGN_RSA_RAW) {
+		return privkey_sign_raw_data(signer,
+					     se,
+					     hash_data, signature,
+					     params);
+	}
+
+	if (_gnutls_pk_is_not_prehashed(signer->pk_algorithm)) {
+		return gnutls_assert_val(GNUTLS_E_INVALID_REQUEST);
+	}
 
 	digest.data = gnutls_malloc(hash_data->size);
 	if (digest.data == NULL) {
@@ -1162,15 +1357,16 @@ gnutls_privkey_sign_hash(gnutls_privkey_t signer,
 	digest.size = hash_data->size;
 	memcpy(digest.data, hash_data->data, digest.size);
 
-	ret =
-	    pk_prepare_hash(signer->pk_algorithm, hash_to_entry(hash_algo),
-			    &digest);
+	ret = pk_prepare_hash(se->pk, hash_to_entry(se->hash), &digest);
 	if (ret < 0) {
 		gnutls_assert();
 		goto cleanup;
 	}
 
-	ret = _gnutls_privkey_sign_raw_data(signer, flags, &digest, signature);
+	ret = privkey_sign_raw_data(signer,
+				    se,
+				    &digest, signature,
+				    params);
 	if (ret < 0) {
 		gnutls_assert();
 		goto cleanup;
@@ -1184,17 +1380,18 @@ gnutls_privkey_sign_hash(gnutls_privkey_t signer,
 }
 
 /*-
- * gnutls_privkey_sign_raw_data:
+ * privkey_sign_raw_data:
  * @key: Holds the key
- * @flags: should be zero
  * @data: holds the data to be signed
  * @signature: will contain the signature allocated with gnutls_malloc()
+ * @params: holds the signing parameters
  *
  * This function will sign the given data using a signature algorithm
  * supported by the private key. Note that this is a low-level function
  * and does not apply any preprocessing or hash on the signed data. 
  * For example on an RSA key the input @data should be of the DigestInfo
- * PKCS #1 1.5 format. Use it only if you know what are you doing.
+ * PKCS #1 1.5 format, on RSA-PSS, DSA or ECDSA the input should be a hash output
+ * and on Ed25519 the raw data to be signed.
  *
  * Note this function is equivalent to using the %GNUTLS_PRIVKEY_SIGN_FLAG_TLS1_RSA
  * flag with gnutls_privkey_sign_hash().
@@ -1204,31 +1401,59 @@ gnutls_privkey_sign_hash(gnutls_privkey_t signer,
  *
  * Since: 3.1.10
  -*/
-static int
-_gnutls_privkey_sign_raw_data(gnutls_privkey_t key,
-			     unsigned flags,
-			     const gnutls_datum_t * data,
-			     gnutls_datum_t * signature)
+int
+privkey_sign_raw_data(gnutls_privkey_t key,
+		      const gnutls_sign_entry_st *se,
+		      const gnutls_datum_t * data,
+		      gnutls_datum_t * signature,
+		      gnutls_x509_spki_st * params)
 {
+	if (unlikely(se == NULL))
+		return gnutls_assert_val(GNUTLS_E_INVALID_REQUEST);
+
 	switch (key->type) {
-#ifdef ENABLE_OPENPGP
-	case GNUTLS_PRIVKEY_OPENPGP:
-		return gnutls_openpgp_privkey_sign_hash(key->key.openpgp,
-							data, signature);
-#endif
 #ifdef ENABLE_PKCS11
 	case GNUTLS_PRIVKEY_PKCS11:
-		return _gnutls_pkcs11_privkey_sign_hash(key->key.pkcs11,
-							data, signature);
+		return _gnutls_pkcs11_privkey_sign(key->key.pkcs11, se,
+						   data, signature,
+						   params);
 #endif
 	case GNUTLS_PRIVKEY_X509:
-		return _gnutls_pk_sign(key->key.x509->pk_algorithm,
-				       signature, data, &key->key.x509->params);
+		return _gnutls_pk_sign(se->pk, signature, data,
+				       &key->key.x509->params, params);
 	case GNUTLS_PRIVKEY_EXT:
-		if (key->key.ext.sign_func == NULL)
+		if (unlikely(key->key.ext.sign_data_func == NULL &&
+			key->key.ext.sign_hash_func == NULL &&
+			key->key.ext.sign_func == NULL))
 			return gnutls_assert_val(GNUTLS_E_INVALID_REQUEST);
-		return key->key.ext.sign_func(key, key->key.ext.userdata,
-					      data, signature);
+
+		if (_gnutls_pk_is_not_prehashed(se->pk)) {
+			if (!key->key.ext.sign_data_func)
+				return gnutls_assert_val(GNUTLS_E_INVALID_REQUEST);
+
+			return key->key.ext.sign_data_func(key, se->id,
+							   key->key.ext.userdata,
+							   0,
+							   data, signature);
+		} else if (key->key.ext.sign_hash_func) {
+			unsigned int flags = 0;
+
+			if (se->pk == GNUTLS_PK_RSA) {
+				se = _gnutls_sign_to_entry(GNUTLS_SIGN_RSA_RAW);
+			}
+
+			/* se may not be set here if we are doing legacy RSA */
+			return key->key.ext.sign_hash_func(key, se->id,
+							   key->key.ext.userdata,
+							   flags,
+							   data, signature);
+		} else {
+			if (!PK_IS_OK_FOR_EXT2(se->pk))
+				return gnutls_assert_val(GNUTLS_E_INVALID_REQUEST);
+
+			return key->key.ext.sign_func(key, key->key.ext.userdata,
+						      data, signature);
+		}
 	default:
 		gnutls_assert();
 		return GNUTLS_E_INVALID_REQUEST;
@@ -1257,12 +1482,6 @@ gnutls_privkey_decrypt_data(gnutls_privkey_t key,
 			    gnutls_datum_t * plaintext)
 {
 	switch (key->type) {
-#ifdef ENABLE_OPENPGP
-	case GNUTLS_PRIVKEY_OPENPGP:
-		return _gnutls_openpgp_privkey_decrypt_data(key->key.openpgp,
-							    flags, ciphertext,
-							    plaintext);
-#endif
 	case GNUTLS_PRIVKEY_X509:
 		return _gnutls_pk_decrypt(key->pk_algorithm, plaintext,
 					  ciphertext, &key->key.x509->params);
@@ -1508,24 +1727,111 @@ int gnutls_privkey_verify_params(gnutls_privkey_t key)
 	return 0;
 }
 
-/*-
- * _gnutls_privkey_get_preferred_sign_algo:
- * @key: should contain a #gnutls_privkey_t type
+/**
+ * gnutls_privkey_get_spki:
+ * @privkey: a public key of type #gnutls_privkey_t
+ * @spki: a SubjectPublicKeyInfo structure of type #gnutls_privkey_spki_t
+ * @flags: must be zero
  *
- * This function returns the preferred signature algorithm for this
- * private key.
+ * This function will return the public key information if available.
+ * The provided @spki must be initialized.
  *
  * Returns: On success, %GNUTLS_E_SUCCESS (0) is returned, otherwise a
  *   negative error value.
  *
- * Since: 3.4.0
- -*/
-gnutls_sign_algorithm_t
-_gnutls_privkey_get_preferred_sign_algo(gnutls_privkey_t key)
+ * Since: 3.6.0
+ **/
+int
+gnutls_privkey_get_spki(gnutls_privkey_t privkey, gnutls_x509_spki_t spki, unsigned int flags)
 {
-	if (key->type == GNUTLS_PRIVKEY_EXT) {
-		if (key->key.ext.info_func)
-			return key->key.ext.info_func(key, GNUTLS_PRIVKEY_INFO_SIGN_ALGO, key->key.ext.userdata);
+	if (privkey == NULL || privkey->type != GNUTLS_PRIVKEY_X509) {
+		gnutls_assert();
+		return GNUTLS_E_REQUESTED_DATA_NOT_AVAILABLE;
 	}
-	return key->preferred_sign_algo;
+
+	if (privkey->key.x509->params.spki.pk == GNUTLS_PK_UNKNOWN)
+		return gnutls_assert_val(GNUTLS_E_REQUESTED_DATA_NOT_AVAILABLE);
+
+	memcpy(spki, &privkey->key.x509->params.spki, sizeof(gnutls_x509_spki_st));
+
+	return 0;
+}
+
+/**
+ * gnutls_privkey_set_spki:
+ * @privkey: a public key of type #gnutls_privkey_t
+ * @spki: a SubjectPublicKeyInfo structure of type #gnutls_privkey_spki_t
+ * @flags: must be zero
+ *
+ * This function will set the public key information.
+ * The provided @spki must be initialized.
+ *
+ * Returns: On success, %GNUTLS_E_SUCCESS (0) is returned, otherwise a
+ *   negative error value.
+ *
+ * Since: 3.6.0
+ **/
+int
+gnutls_privkey_set_spki(gnutls_privkey_t privkey, const gnutls_x509_spki_t spki, unsigned int flags)
+{
+	if (privkey == NULL || privkey->type != GNUTLS_PRIVKEY_X509) {
+		gnutls_assert();
+		return GNUTLS_E_REQUESTED_DATA_NOT_AVAILABLE;
+	}
+
+	return gnutls_x509_privkey_set_spki(privkey->key.x509, spki, flags);
+}
+
+/* Checks whether the public key given is compatible with the
+ * signature algorithm used. The session is only used for audit logging, and
+ * it may be null.
+ */
+unsigned _gnutls_privkey_compatible_with_sig(gnutls_privkey_t privkey,
+					     gnutls_sign_algorithm_t sign)
+{
+	const gnutls_sign_entry_st *se;
+
+	se = _gnutls_sign_to_entry(sign);
+	if (unlikely(se == NULL))
+		return gnutls_assert_val(0);
+
+	/* Prevent RSA-PSS private keys from negotiating an RSA signature,
+	 * and RSA keys which cannot do RSA-PSS (e.g., smart card) from
+	 * negotiating RSA-PSS sig.
+	 */
+	if (privkey->pk_algorithm == GNUTLS_PK_RSA_PSS && se->pk != GNUTLS_PK_RSA_PSS) {
+			return 0;
+	}
+
+	if (privkey->type == GNUTLS_PRIVKEY_EXT) {
+		if (privkey->key.ext.info_func) {
+			int ret;
+
+			ret = privkey->key.ext.info_func(privkey,
+						     GNUTLS_SIGN_ALGO_TO_FLAGS(sign)|GNUTLS_PRIVKEY_INFO_HAVE_SIGN_ALGO,
+						     privkey->key.ext.userdata);
+			if (ret != -1)
+				return ret;
+
+			/* use the old flag */
+			ret = privkey->key.ext.info_func(privkey, GNUTLS_PRIVKEY_INFO_SIGN_ALGO,
+						     privkey->key.ext.userdata);
+			if (ret == (int)sign)
+				return 1;
+		}
+
+		/* This key type is very limited on what it can handle */
+		if (!PK_IS_OK_FOR_EXT2(se->pk))
+			return gnutls_assert_val(0);
+	}
+#ifdef ENABLE_PKCS11
+	else if (privkey->type == GNUTLS_PRIVKEY_PKCS11) {
+		if (privkey->pk_algorithm == GNUTLS_PK_RSA && se->pk == GNUTLS_PK_RSA_PSS) {
+			if (!privkey->key.pkcs11->rsa_pss_ok)
+				return 0;
+		}
+	}
+#endif
+
+	return 1;
 }
