@@ -1,5 +1,6 @@
 /*
  * Copyright (C) 2000-2012 Free Software Foundation, Inc.
+ * Copyright (C) 2017 Red Hat, Inc.
  *
  * Author: Nikos Mavrogiannopoulos
  *
@@ -274,7 +275,7 @@ _gnutls_session_unpack(gnutls_session_t session,
 /* Format: 
  *      1 byte the credentials type
  *      4 bytes the size of the whole structure
- *        DH stuff
+ *	DH stuff
  *      2 bytes the size of secret key in bits
  *      4 bytes the size of the prime
  *      x bytes the prime
@@ -282,12 +283,12 @@ _gnutls_session_unpack(gnutls_session_t session,
  *      x bytes the generator
  *      4 bytes the size of the public key
  *      x bytes the public key
- *        RSA stuff
+ *	RSA stuff
  *      4 bytes the size of the modulus
  *      x bytes the modulus
  *      4 bytes the size of the exponent
  *      x bytes the exponent
- *        CERTIFICATES
+ *	CERTIFICATES
  *      4 bytes the length of the certificate list
  *      4 bytes the size of first certificate
  *      x bytes the certificate
@@ -691,11 +692,9 @@ unpack_psk_auth_info(gnutls_session_t session, gnutls_buffer_st * ps)
  *      1 byte the key exchange algorithm used
  *      1 byte the read cipher algorithm
  *      1 byte the read mac algorithm
- *      1 byte the read compression algorithm
  *
  *      1 byte the write cipher algorithm
  *      1 byte the write mac algorithm
- *      1 byte the write compression algorithm
  *
  *      1 byte the certificate type
  *      1 byte the protocol version
@@ -712,8 +711,8 @@ unpack_psk_auth_info(gnutls_session_t session, gnutls_buffer_st * ps)
  *
  *      4 bytes the new record padding flag
  *      4 bytes the ECC curve
- *            -------------------
- *                MAX: 169 bytes
+ *	    -------------------
+ *		MAX: 169 bytes
  *
  */
 static int
@@ -745,10 +744,7 @@ pack_security_parameters(gnutls_session_t session, gnutls_buffer_st * ps)
 
 
 	BUFFER_APPEND_NUM(ps, session->security_parameters.entity);
-	BUFFER_APPEND_NUM(ps, session->security_parameters.kx_algorithm);
-	BUFFER_APPEND(ps, session->security_parameters.cipher_suite, 2);
-	BUFFER_APPEND_NUM(ps,
-			  session->security_parameters.compression_method);
+	BUFFER_APPEND(ps, session->security_parameters.cs->id, 2);
 	BUFFER_APPEND_NUM(ps, session->security_parameters.cert_type);
 	BUFFER_APPEND_NUM(ps, session->security_parameters.pversion->id);
 
@@ -770,7 +766,12 @@ pack_security_parameters(gnutls_session_t session, gnutls_buffer_st * ps)
 	BUFFER_APPEND_NUM(ps,
 			  session->security_parameters.
 			  max_record_recv_size);
-	BUFFER_APPEND_NUM(ps, session->security_parameters.ecc_curve);
+
+	if (session->security_parameters.grp) {
+		BUFFER_APPEND_NUM(ps, session->security_parameters.grp->id);
+	} else {
+		BUFFER_APPEND_NUM(ps, 0);
+	}
 
 	BUFFER_APPEND_NUM(ps,
 			  session->security_parameters.server_sign_algo);
@@ -794,6 +795,7 @@ unpack_security_parameters(gnutls_session_t session, gnutls_buffer_st * ps)
 	int ret;
 	unsigned version;
 	time_t timestamp;
+	uint8_t cs[2];
 
 	BUFFER_POP_NUM(ps, pack_size);
 
@@ -812,15 +814,11 @@ unpack_security_parameters(gnutls_session_t session, gnutls_buffer_st * ps)
 	BUFFER_POP_NUM(ps,
 		       session->internals.resumed_security_parameters.
 		       entity);
-	BUFFER_POP_NUM(ps,
-		       session->internals.resumed_security_parameters.
-		       kx_algorithm);
-	BUFFER_POP(ps,
-		   session->internals.resumed_security_parameters.
-		   cipher_suite, 2);
-	BUFFER_POP_NUM(ps,
-		       session->internals.resumed_security_parameters.
-		       compression_method);
+	BUFFER_POP(ps, cs, 2);
+	session->internals.resumed_security_parameters.cs = ciphersuite_to_entry(cs);
+	if (session->internals.resumed_security_parameters.cs == NULL)
+		return gnutls_assert_val(GNUTLS_E_INVALID_REQUEST);
+
 	BUFFER_POP_NUM(ps,
 		       session->internals.resumed_security_parameters.
 		       cert_type);
@@ -858,9 +856,10 @@ unpack_security_parameters(gnutls_session_t session, gnutls_buffer_st * ps)
 		       session->internals.resumed_security_parameters.
 		       max_record_recv_size);
 
-	BUFFER_POP_NUM(ps,
-		       session->internals.resumed_security_parameters.
-		       ecc_curve);
+	BUFFER_POP_NUM(ps, ret);
+	session->internals.resumed_security_parameters.grp = _gnutls_id_to_group(ret);
+	/* it can be null */
+
 	BUFFER_POP_NUM(ps,
 		       session->internals.resumed_security_parameters.
 		       server_sign_algo);
@@ -904,7 +903,7 @@ unpack_security_parameters(gnutls_session_t session, gnutls_buffer_st * ps)
  * @kx: the key exchange method
  * @cipher: the cipher
  * @mac: the MAC algorithm
- * @comp: the compression method
+ * @comp: the compression method (ignored)
  * @master: the master key to use
  * @session_id: the session identifier
  *
@@ -927,23 +926,22 @@ gnutls_session_set_premaster(gnutls_session_t session, unsigned int entity,
 			     const gnutls_datum_t * session_id)
 {
 	int ret;
+	uint8_t cs[2];
 
 	memset(&session->internals.resumed_security_parameters, 0,
 	       sizeof(session->internals.resumed_security_parameters));
 
 	session->internals.resumed_security_parameters.entity = entity;
-	session->internals.resumed_security_parameters.kx_algorithm = kx;
 
 	ret =
-	    _gnutls_cipher_suite_get_id(kx, cipher, mac,
-					session->internals.
-					resumed_security_parameters.
-					cipher_suite);
+	    _gnutls_cipher_suite_get_id(kx, cipher, mac, cs);
 	if (ret < 0)
 		return gnutls_assert_val(ret);
 
-	session->internals.resumed_security_parameters.compression_method =
-	    comp;
+	session->internals.resumed_security_parameters.cs = ciphersuite_to_entry(cs);
+	if (cs == NULL)
+		return gnutls_assert_val(GNUTLS_E_INVALID_REQUEST);
+
 	session->internals.resumed_security_parameters.cert_type =
 	    DEFAULT_CERT_TYPE;
 	session->internals.resumed_security_parameters.pversion =
@@ -975,8 +973,7 @@ gnutls_session_set_premaster(gnutls_session_t session, unsigned int entity,
 	session->internals.resumed_security_parameters.timestamp =
 	    gnutls_time(0);
 
-	session->internals.resumed_security_parameters.ecc_curve =
-	    GNUTLS_ECC_CURVE_INVALID;
+	session->internals.resumed_security_parameters.grp = 0;
 
 	session->internals.premaster_set = 1;
 
