@@ -58,25 +58,6 @@ void
 _gnutls_rsa_pms_set_version(gnutls_session_t session,
 			    unsigned char major, unsigned char minor);
 
-void
-_gnutls_session_cert_type_set(gnutls_session_t session,
-			      gnutls_certificate_type_t ct)
-{
-	_gnutls_handshake_log
-	    ("HSK[%p]: Selected certificate type %s (%d)\n", session,
-	     gnutls_certificate_type_get_name(ct), ct);
-	session->security_parameters.cert_type = ct;
-}
-
-void
-_gnutls_session_ecc_curve_set(gnutls_session_t session,
-			      gnutls_ecc_curve_t c)
-{
-	_gnutls_handshake_log("HSK[%p]: Selected ECC curve %s (%d)\n",
-			      session, gnutls_ecc_curve_get_name(c), c);
-	session->security_parameters.ecc_curve = c;
-}
-
 /**
  * gnutls_cipher_get:
  * @session: is a #gnutls_session_t type.
@@ -126,12 +107,9 @@ gnutls_certificate_type_get(gnutls_session_t session)
  **/
 gnutls_kx_algorithm_t gnutls_kx_get(gnutls_session_t session)
 {
-	if (session->internals.handshake_in_progress) {
-		/* This allows early call during handshake */
-		return _gnutls_cipher_suite_get_kx_algo(session->security_parameters.cipher_suite);
-	} else {
-		return session->security_parameters.kx_algorithm;
-	}
+	if (session->security_parameters.cs == 0)
+		return 0;
+	return session->security_parameters.cs->kx_algorithm;
 }
 
 /**
@@ -168,66 +146,7 @@ gnutls_mac_algorithm_t gnutls_mac_get(gnutls_session_t session)
 gnutls_compression_method_t
 gnutls_compression_get(gnutls_session_t session)
 {
-	record_parameters_st *record_params;
-	int ret;
-
-	ret =
-	    _gnutls_epoch_get(session, EPOCH_READ_CURRENT, &record_params);
-	if (ret < 0)
-		return gnutls_assert_val(GNUTLS_COMP_NULL);
-
-	return record_params->compression_algorithm;
-}
-
-/* Check if the given certificate type is supported.
- * This means that it is enabled by the priority functions,
- * and a matching certificate exists.
- */
-int
-_gnutls_session_cert_type_supported(gnutls_session_t session,
-				    gnutls_certificate_type_t cert_type)
-{
-	unsigned i;
-	unsigned cert_found = 0;
-	gnutls_certificate_credentials_t cred;
-
-	if (session->security_parameters.entity == GNUTLS_SERVER) {
-		cred = (gnutls_certificate_credentials_t)
-		    _gnutls_get_cred(session, GNUTLS_CRD_CERTIFICATE);
-
-		if (cred == NULL)
-			return GNUTLS_E_UNSUPPORTED_CERTIFICATE_TYPE;
-
-		if (cred->get_cert_callback == NULL && cred->get_cert_callback2 == NULL) {
-			for (i = 0; i < cred->ncerts; i++) {
-				if (cred->certs[i].cert_list[0].type ==
-				    cert_type) {
-					cert_found = 1;
-					break;
-				}
-			}
-
-			if (cert_found == 0)
-				/* no certificate is of that type.
-				 */
-				return
-				    GNUTLS_E_UNSUPPORTED_CERTIFICATE_TYPE;
-		}
-	}
-
-	if (session->internals.priorities.cert_type.algorithms == 0
-	    && cert_type == DEFAULT_CERT_TYPE)
-		return 0;
-
-	for (i = 0; i < session->internals.priorities.cert_type.algorithms;
-	     i++) {
-		if (session->internals.priorities.cert_type.priority[i] ==
-		    cert_type) {
-			return 0;	/* ok */
-		}
-	}
-
-	return GNUTLS_E_UNSUPPORTED_CERTIFICATE_TYPE;
+	return GNUTLS_COMP_NULL;
 }
 
 static void deinit_keys(gnutls_session_t session)
@@ -256,28 +175,9 @@ static void deinit_keys(gnutls_session_t session)
 	_gnutls_free_temp_key_datum(&session->key.key);
 }
 
-/* this function deinitializes all the internal parameters stored
- * in a session struct.
- */
-inline static void deinit_internal_params(gnutls_session_t session)
-{
-#if defined(ENABLE_DHE) || defined(ENABLE_ANON)
-	if (session->internals.params.free_dh_params)
-		gnutls_dh_params_deinit(session->internals.params.
-					dh_params);
-#endif
-
-	_gnutls_handshake_hash_buffers_clear(session);
-
-	memset(&session->internals.params, 0,
-	       sizeof(session->internals.params));
-}
-
-/* This function will clear all the variables in internals
- * structure within the session, which depend on the current handshake.
- * This is used to allow further handshakes.
- */
-static void _gnutls_handshake_internal_state_init(gnutls_session_t session)
+/* An internal version of _gnutls_handshake_internal_state_clear(),
+ * it will not attempt to deallocate, only initialize */
+static void handshake_internal_state_clear1(gnutls_session_t session)
 {
 	/* by default no selected certificate */
 	session->internals.adv_version_major = 0;
@@ -295,13 +195,21 @@ static void _gnutls_handshake_internal_state_init(gnutls_session_t session)
 	session->internals.handshake_suspicious_loops = 0;
 	session->internals.dtls.hsk_read_seq = 0;
 	session->internals.dtls.hsk_write_seq = 0;
+
+	session->internals.have_ffdhe = 0;
+	session->internals.cand_ec_group = 0;
+	session->internals.cand_dh_group = 0;
 }
 
+/* This function will clear all the variables in internals
+ * structure within the session, which depend on the current handshake.
+ * This is used to allow further handshakes.
+ */
 void _gnutls_handshake_internal_state_clear(gnutls_session_t session)
 {
-	_gnutls_handshake_internal_state_init(session);
+	handshake_internal_state_clear1(session);
 
-	deinit_internal_params(session);
+	_gnutls_handshake_hash_buffers_clear(session);
 	deinit_keys(session);
 
 	_gnutls_epoch_gc(session);
@@ -395,14 +303,7 @@ int gnutls_init(gnutls_session_t * session, unsigned int flags)
 	 * as NULL or 0. This is why calloc is used.
 	 */
 
-	_gnutls_handshake_internal_state_init(*session);
-
-	(*session)->internals.extensions_sent_size = 0;
-
-	/* emulate old gnutls behavior for old applications that do not use the priority_*
-	 * functions.
-	 */
-	(*session)->internals.priorities.sr = SR_PARTIAL;
+	handshake_internal_state_clear1(*session);
 
 #ifdef HAVE_WRITEV
 #ifdef MSG_NOSIGNAL
@@ -503,6 +404,9 @@ void gnutls_deinit(gnutls_session_t session)
 	gnutls_credentials_clear(session);
 	_gnutls_selected_certs_deinit(session);
 
+	/* we rely on priorities' internal reference counting */
+	gnutls_priority_deinit(session->internals.priorities);
+
 	gnutls_free(session);
 }
 
@@ -601,7 +505,7 @@ int _gnutls_dh_set_secret_bits(gnutls_session_t session, unsigned bits)
 /* Sets the prime and the generator in the auth info structure.
  */
 int
-_gnutls_dh_set_group(gnutls_session_t session, bigint_t gen,
+_gnutls_dh_save_group(gnutls_session_t session, bigint_t gen,
 		     bigint_t prime)
 {
 	dh_info_st *dh;
@@ -669,25 +573,6 @@ _gnutls_dh_set_group(gnutls_session_t session, bigint_t gen,
 	return 0;
 }
 
-#ifdef ENABLE_OPENPGP
-/**
- * gnutls_openpgp_send_cert:
- * @session: a #gnutls_session_t type.
- * @status: is one of GNUTLS_OPENPGP_CERT, or GNUTLS_OPENPGP_CERT_FINGERPRINT
- *
- * This function will order gnutls to send the key fingerprint
- * instead of the key in the initial handshake procedure. This should
- * be used with care and only when there is indication or knowledge
- * that the server can obtain the client's key.
- **/
-void
-gnutls_openpgp_send_cert(gnutls_session_t session,
-			 gnutls_openpgp_crt_status_t status)
-{
-	session->internals.pgp_fingerprint = status;
-}
-#endif
-
 /**
  * gnutls_certificate_send_x509_rdn_sequence:
  * @session: a #gnutls_session_t type.
@@ -709,13 +594,6 @@ gnutls_certificate_send_x509_rdn_sequence(gnutls_session_t session,
 	session->internals.ignore_rdn_sequence = status;
 }
 
-#ifdef ENABLE_OPENPGP
-int _gnutls_openpgp_send_fingerprint(gnutls_session_t session)
-{
-	return session->internals.pgp_fingerprint;
-}
-#endif
-
 /*-
  * _gnutls_record_set_default_version - Used to set the default version for the first record packet
  * @session: is a #gnutls_session_t type.
@@ -773,7 +651,8 @@ void
 gnutls_handshake_set_private_extensions(gnutls_session_t session,
 					int allow)
 {
-	session->internals.enable_private = allow;
+	/* we have no private extensions */
+	return;
 }
 
 
@@ -837,8 +716,7 @@ int _gnutls_session_is_psk(gnutls_session_t session)
 {
 	gnutls_kx_algorithm_t kx;
 
-	kx = _gnutls_cipher_suite_get_kx_algo(session->security_parameters.
-					      cipher_suite);
+	kx = session->security_parameters.cs->kx_algorithm;
 	if (kx == GNUTLS_KX_PSK || kx == GNUTLS_KX_DHE_PSK
 	    || kx == GNUTLS_KX_RSA_PSK)
 		return 1;
@@ -860,8 +738,7 @@ int _gnutls_session_is_ecc(gnutls_session_t session)
 	/* We get the key exchange algorithm through the ciphersuite because
 	 * the negotiated key exchange might not have been set yet.
 	 */
-	kx = _gnutls_cipher_suite_get_kx_algo(session->security_parameters.
-					      cipher_suite);
+	kx = session->security_parameters.cs->kx_algorithm;
 
 	return _gnutls_kx_is_ecc(kx);
 }
@@ -1014,10 +891,12 @@ gnutls_handshake_set_post_client_hello_function(gnutls_session_t session,
  *
  * Note that this function must be called after any call to gnutls_priority
  * functions.
+ *
+ * Since: 2.1.4
  **/
 void gnutls_session_enable_compatibility_mode(gnutls_session_t session)
 {
-	ENABLE_COMPAT(&session->internals.priorities);
+	ENABLE_COMPAT(&session->internals);
 }
 
 /**
@@ -1061,7 +940,7 @@ gnutls_session_channel_binding(gnutls_session_t session,
  * gnutls_ecc_curve_get:
  * @session: is a #gnutls_session_t type.
  *
- * Returns the currently used elliptic curve. Only valid
+ * Returns the currently used elliptic curve for key exchange. Only valid
  * when using an elliptic curve ciphersuite.
  *
  * Returns: the currently used curve, a #gnutls_ecc_curve_t
@@ -1071,7 +950,34 @@ gnutls_session_channel_binding(gnutls_session_t session,
  **/
 gnutls_ecc_curve_t gnutls_ecc_curve_get(gnutls_session_t session)
 {
-	return _gnutls_session_ecc_curve_get(session);
+	const gnutls_group_entry_st *e;
+
+	e = get_group(session);
+	if (e == NULL || e->curve == 0)
+		return 0;
+	return e->curve;
+}
+
+/**
+ * gnutls_group_get:
+ * @session: is a #gnutls_session_t type.
+ *
+ * Returns the currently used group for key exchange. Only valid
+ * when using an elliptic curve or DH ciphersuite.
+ *
+ * Returns: the currently used group, a #gnutls_group_t
+ *   type.
+ *
+ * Since: 3.6.0
+ **/
+gnutls_group_t gnutls_group_get(gnutls_session_t session)
+{
+	const gnutls_group_entry_st *e;
+
+	e = get_group(session);
+	if (e == NULL)
+		return 0;
+	return e->id;
 }
 
 /**
@@ -1360,6 +1266,8 @@ unsigned gnutls_session_get_flags(gnutls_session_t session)
 		flags |= GNUTLS_SFLAGS_HB_PEER_SEND;
 	if (session->internals.false_start_used)
 		flags |= GNUTLS_SFLAGS_FALSE_START;
+	if (session->internals.used_ffdhe)
+		flags |= GNUTLS_SFLAGS_RFC7919;
 
 	return flags;
 }
