@@ -27,6 +27,7 @@
 #include <stdarg.h>
 #include <c-ctype.h>
 #include <intprops.h>
+#include <nettle/base64.h>
 #include "vasprintf.h"
 #include "extras/hex.h"
 
@@ -66,20 +67,6 @@ void _gnutls_str_cpy(char *dest, size_t dest_tot_size, const char *src)
 	}
 }
 
-void
-_gnutls_mem_cpy(char *dest, size_t dest_tot_size, const char *src,
-		size_t src_size)
-{
-
-	if (dest_tot_size >= src_size) {
-		memcpy(dest, src, src_size);
-	} else {
-		if (dest_tot_size > 0) {
-			memcpy(dest, src, dest_tot_size);
-		}
-	}
-}
-
 void _gnutls_buffer_init(gnutls_buffer_st * str)
 {
 	str->data = str->allocd = NULL;
@@ -110,6 +97,7 @@ void _gnutls_buffer_clear(gnutls_buffer_st * str)
 
 static void align_allocd_with_data(gnutls_buffer_st * dest)
 {
+	assert(dest->allocd != NULL);
 	if (dest->length)
 		memmove(dest->allocd, dest->data, dest->length);
 	dest->data = dest->allocd;
@@ -254,14 +242,15 @@ int _gnutls_buffer_to_datum(gnutls_buffer_st * str, gnutls_datum_t * data, unsig
 	if (str->length == 0) {
 		data->data = NULL;
 		data->size = 0;
-		_gnutls_buffer_clear(str);
-		return 0;
+		ret = 0;
+		goto fail;
 	}
 
 	if (is_str) {
 		ret = _gnutls_buffer_append_data(str, "\x00", 1);
 		if (ret < 0) {
-			return gnutls_assert_val(ret);
+			gnutls_assert();
+			goto fail;
 		}
 	}
 
@@ -269,8 +258,8 @@ int _gnutls_buffer_to_datum(gnutls_buffer_st * str, gnutls_datum_t * data, unsig
 		data->data = gnutls_malloc(str->length);
 		if (data->data == NULL) {
 			gnutls_assert();
-			_gnutls_buffer_clear(str);
-			return GNUTLS_E_MEMORY_ERROR;
+			ret = GNUTLS_E_MEMORY_ERROR;
+			goto fail;
 		}
 		memcpy(data->data, str->data, str->length);
 		data->size = str->length;
@@ -286,6 +275,9 @@ int _gnutls_buffer_to_datum(gnutls_buffer_st * str, gnutls_datum_t * data, unsig
 	}
 
 	return 0;
+ fail:
+	_gnutls_buffer_clear(str);
+	return ret;
 }
 
 /* returns data from a string in a constant buffer.
@@ -340,6 +332,8 @@ _gnutls_buffer_insert_data(gnutls_buffer_st * dest, int pos,
 	if (ret < 0)
 		return ret;
 
+	assert(dest->data != NULL);
+
 	memmove(&dest->data[pos + str_size], &dest->data[pos],
 		orig_length - pos);
 
@@ -718,7 +712,7 @@ _gnutls_hostname_compare(const char *certname,
 	unsigned i;
 
 	for (i=0;i<certnamesize;i++) {
-		if (c_isascii(certname[i]) == 0)
+		if (c_isprint(certname[i]) == 0)
 			return hostname_compare_raw(certname, certnamesize, hostname);
 	}
 
@@ -837,20 +831,19 @@ _gnutls_buffer_append_data_prefix(gnutls_buffer_st * buf,
 				  int pfx_size, const void *data,
 				  size_t data_size)
 {
-	int ret = 0, ret1;
+	int ret;
 
-	ret1 = _gnutls_buffer_append_prefix(buf, pfx_size, data_size);
-	if (ret1 < 0)
-		return gnutls_assert_val(ret1);
+	ret = _gnutls_buffer_append_prefix(buf, pfx_size, data_size);
+	if (ret < 0)
+		return gnutls_assert_val(ret);
 
 	if (data_size > 0) {
 		ret = _gnutls_buffer_append_data(buf, data, data_size);
-
 		if (ret < 0)
 			return gnutls_assert_val(ret);
 	}
 
-	return ret + ret1;
+	return 0;
 }
 
 int _gnutls_buffer_append_mpi(gnutls_buffer_st * buf, int pfx_size,
@@ -876,25 +869,6 @@ int _gnutls_buffer_append_mpi(gnutls_buffer_st * buf, int pfx_size,
 	return ret;
 }
 
-int
-_gnutls_buffer_pop_data_prefix(gnutls_buffer_st * buf, void *data,
-			       size_t * data_size)
-{
-	size_t size;
-	int ret;
-
-	ret = _gnutls_buffer_pop_prefix(buf, &size, 1);
-	if (ret < 0) {
-		gnutls_assert();
-		return ret;
-	}
-
-	if (size > 0)
-		_gnutls_buffer_pop_data(buf, data, data_size);
-
-	return 0;
-}
-
 void
 _gnutls_buffer_hexprint(gnutls_buffer_st * str,
 			const void *_data, size_t len)
@@ -911,6 +885,26 @@ _gnutls_buffer_hexprint(gnutls_buffer_st * str,
 	}
 }
 
+int
+_gnutls_buffer_base64print(gnutls_buffer_st * str,
+			   const void *_data, size_t len)
+{
+	const unsigned char *data = _data;
+	unsigned b64len = BASE64_ENCODE_RAW_LENGTH(len);
+	int ret;
+
+	ret = _gnutls_buffer_resize(str, str->length+b64len+1);
+	if (ret < 0) {
+		return gnutls_assert_val(ret);
+	}
+
+	base64_encode_raw(&str->data[str->length], len, data);
+	str->length += b64len;
+	str->data[str->length] = 0;
+
+	return 0;
+}
+
 void
 _gnutls_buffer_hexdump(gnutls_buffer_st * str, const void *_data,
 		       size_t len, const char *spc)
