@@ -43,8 +43,6 @@ struct gnutls_tdb_int {
 
 static int raw_pubkey_to_base64(const gnutls_datum_t * raw,
 				gnutls_datum_t * b64);
-static int pgp_crt_to_raw_pubkey(const gnutls_datum_t * cert,
-				 gnutls_datum_t * rpubkey);
 static int verify_pubkey(const char *file, const char *host,
 			 const char *service, const gnutls_datum_t * skey);
 
@@ -116,8 +114,7 @@ gnutls_verify_stored_pubkey(const char *db_name,
 	int ret;
 	char local_file[MAX_FILENAME];
 
-	if (cert_type != GNUTLS_CRT_X509
-	    && cert_type != GNUTLS_CRT_OPENPGP)
+	if (cert_type != GNUTLS_CRT_X509)
 		return
 		    gnutls_assert_val
 		    (GNUTLS_E_UNSUPPORTED_CERTIFICATE_TYPE);
@@ -132,11 +129,7 @@ gnutls_verify_stored_pubkey(const char *db_name,
 	if (tdb == NULL)
 		tdb = &default_tdb;
 
-	if (cert_type == GNUTLS_CRT_X509)
-		ret = x509_raw_crt_to_raw_pubkey(cert, &pubkey);
-	else
-		ret = pgp_crt_to_raw_pubkey(cert, &pubkey);
-
+	ret = x509_raw_crt_to_raw_pubkey(cert, &pubkey);
 	if (ret < 0) {
 		gnutls_assert();
 		goto cleanup;
@@ -383,74 +376,6 @@ static int raw_pubkey_to_base64(const gnutls_datum_t * raw,
 	return 0;
 }
 
-static int pgp_crt_to_raw_pubkey(const gnutls_datum_t * cert,
-				 gnutls_datum_t * rpubkey)
-{
-#ifdef ENABLE_OPENPGP
-	gnutls_openpgp_crt_t crt = NULL;
-	gnutls_pubkey_t pubkey = NULL;
-	size_t size;
-	int ret;
-
-	ret = gnutls_openpgp_crt_init(&crt);
-	if (ret < 0)
-		return gnutls_assert_val(ret);
-
-	ret = gnutls_pubkey_init(&pubkey);
-	if (ret < 0) {
-		gnutls_assert();
-		goto cleanup;
-	}
-
-	ret = gnutls_openpgp_crt_import(crt, cert, GNUTLS_OPENPGP_FMT_RAW);
-	if (ret < 0) {
-		gnutls_assert();
-		goto cleanup;
-	}
-
-	ret = gnutls_pubkey_import_openpgp(pubkey, crt, 0);
-	if (ret < 0) {
-		gnutls_assert();
-		goto cleanup;
-	}
-
-	size = 0;
-	ret =
-	    gnutls_pubkey_export(pubkey, GNUTLS_X509_FMT_DER, NULL, &size);
-	if (ret < 0 && ret != GNUTLS_E_SHORT_MEMORY_BUFFER) {
-		gnutls_assert();
-		goto cleanup;
-	}
-
-	rpubkey->data = gnutls_malloc(size);
-	if (rpubkey->data == NULL) {
-		ret = GNUTLS_E_MEMORY_ERROR;
-		gnutls_assert();
-		goto cleanup;
-	}
-
-	ret =
-	    gnutls_pubkey_export(pubkey, GNUTLS_X509_FMT_DER,
-				 rpubkey->data, &size);
-	if (ret < 0) {
-		gnutls_free(rpubkey->data);
-		gnutls_assert();
-		goto cleanup;
-	}
-
-	rpubkey->size = size;
-	ret = 0;
-
-      cleanup:
-	gnutls_openpgp_crt_deinit(crt);
-	gnutls_pubkey_deinit(pubkey);
-
-	return ret;
-#else
-	return GNUTLS_E_UNIMPLEMENTED_FEATURE;
-#endif
-}
-
 static
 int store_pubkey(const char *db_name, const char *host,
 		 const char *service, time_t expiration,
@@ -543,6 +468,13 @@ int store_commitment(const char *db_name, const char *host,
  * the storage of entries. If it is NULL then the
  * default file backend will be used.
  *
+ * Unless an alternative @tdb is provided, the storage format is a textual format
+ * consisting of a line for each host with fields separated by '|'. The contents of
+ * the fields are a format-identifier which is set to 'g0', the hostname that the
+ * rest of the data applies to, the numeric port or host name, the expiration
+ * time in seconds since the epoch (0 for no expiration), and a base64
+ * encoding of the raw (DER) public key information (SPKI) of the peer.
+ *
  * Returns: On success, %GNUTLS_E_SUCCESS (0) is returned, otherwise a
  *   negative error value.
  *
@@ -561,8 +493,7 @@ gnutls_store_pubkey(const char *db_name,
 	int ret;
 	char local_file[MAX_FILENAME];
 
-	if (cert_type != GNUTLS_CRT_X509
-	    && cert_type != GNUTLS_CRT_OPENPGP)
+	if (cert_type != GNUTLS_CRT_X509)
 		return
 		    gnutls_assert_val
 		    (GNUTLS_E_UNSUPPORTED_CERTIFICATE_TYPE);
@@ -586,10 +517,7 @@ gnutls_store_pubkey(const char *db_name,
 	if (tdb == NULL)
 		tdb = &default_tdb;
 
-	if (cert_type == GNUTLS_CRT_X509)
-		ret = x509_raw_crt_to_raw_pubkey(cert, &pubkey);
-	else
-		ret = pgp_crt_to_raw_pubkey(cert, &pubkey);
+	ret = x509_raw_crt_to_raw_pubkey(cert, &pubkey);
 	if (ret < 0) {
 		gnutls_assert();
 		goto cleanup;
@@ -616,7 +544,7 @@ gnutls_store_pubkey(const char *db_name,
  * @hash_algo: The hash algorithm type
  * @hash: The raw hash
  * @expiration: The expiration time (use 0 to disable expiration)
- * @flags: should be 0.
+ * @flags: should be 0 or %GNUTLS_SCOMMIT_FLAG_ALLOW_BROKEN.
  *
  * This function will store the provided hash commitment to 
  * the list of stored public keys. The key with the given
@@ -646,9 +574,12 @@ gnutls_store_commitment(const char *db_name,
 	char local_file[MAX_FILENAME];
 	const mac_entry_st *me = hash_to_entry(hash_algo);
 
-	if (me == NULL || _gnutls_digest_is_secure(me) == 0)
+	if (me == NULL)
 		return gnutls_assert_val(GNUTLS_E_ILLEGAL_PARAMETER);
 
+	if (!(flags & GNUTLS_SCOMMIT_FLAG_ALLOW_BROKEN) && _gnutls_digest_is_secure(me) == 0)
+		return gnutls_assert_val(GNUTLS_E_INSUFFICIENT_SECURITY);
+
 	if (_gnutls_hash_get_algo_len(me) != hash->size)
 		return gnutls_assert_val(GNUTLS_E_INVALID_REQUEST);
 
