@@ -1,6 +1,7 @@
 /*
  * Copyright (C) 2003-2016 Free Software Foundation, Inc.
  * Copyright (C) 2012-2016 Nikos Mavrogiannopoulos
+ * Copyright (C) 2016-2017 Red Hat, Inc.
  *
  * Author: Nikos Mavrogiannopoulos
  *
@@ -36,6 +37,8 @@
 #include <gnutls/x509-ext.h>
 #include "x509_int.h"
 #include <libtasn1.h>
+#include <pk.h>
+#include "attributes.h"
 
 /**
  * gnutls_x509_crq_init:
@@ -167,15 +170,18 @@ gnutls_x509_crq_import(gnutls_x509_crq_t crq,
  * enumeration that is the signature algorithm that has been used to
  * sign this certificate request.
  *
- * Returns: a #gnutls_sign_algorithm_t value, or a negative error code on
- *   error.
+ * Since 3.6.0 this function never returns a negative error code.
+ * Error cases and unknown/unsupported signature algorithms are
+ * mapped to %GNUTLS_SIGN_UNKNOWN.
+ *
+ * Returns: a #gnutls_sign_algorithm_t value
  *
  * Since: 3.4.0
  **/
 int gnutls_x509_crq_get_signature_algorithm(gnutls_x509_crq_t crq)
 {
-	return _gnutls_x509_get_signature_algorithm(crq->crq,
-						    "signatureAlgorithm.algorithm");
+	return map_errs_to_zero(_gnutls_x509_get_signature_algorithm(crq->crq,
+						    "signatureAlgorithm"));
 }
 
 /**
@@ -413,126 +419,6 @@ gnutls_x509_crq_get_dn_oid(gnutls_x509_crq_t crq,
 				       indx, oid, sizeof_oid);
 }
 
-/* Parses an Attribute list in the asn1_struct, and searches for the
- * given OID. The index indicates the attribute value to be returned.
- *
- * If raw==0 only printable data are returned, or
- * GNUTLS_E_X509_UNSUPPORTED_ATTRIBUTE.
- *
- * asn1_attr_name must be a string in the form
- * "certificationRequestInfo.attributes"
- *
- */
-static int
-parse_attribute(ASN1_TYPE asn1_struct,
-		const char *attr_name, const char *given_oid, unsigned indx,
-		int raw, gnutls_datum_t * out)
-{
-	int k1, result;
-	char tmpbuffer1[MAX_NAME_SIZE];
-	char tmpbuffer3[MAX_NAME_SIZE];
-	char value[200];
-	gnutls_datum_t td;
-	char oid[MAX_OID_SIZE];
-	int len;
-
-	k1 = 0;
-	do {
-
-		k1++;
-		/* create a string like "attribute.?1"
-		 */
-		if (attr_name[0] != 0)
-			snprintf(tmpbuffer1, sizeof(tmpbuffer1), "%s.?%u",
-				 attr_name, k1);
-		else
-			snprintf(tmpbuffer1, sizeof(tmpbuffer1), "?%u",
-				 k1);
-
-		len = sizeof(value) - 1;
-		result =
-		    asn1_read_value(asn1_struct, tmpbuffer1, value, &len);
-
-		if (result == ASN1_ELEMENT_NOT_FOUND) {
-			gnutls_assert();
-			break;
-		}
-
-		if (result != ASN1_VALUE_NOT_FOUND) {
-			gnutls_assert();
-			result = _gnutls_asn2err(result);
-			goto cleanup;
-		}
-
-		/* Move to the attibute type and values
-		 */
-		/* Read the OID
-		 */
-		_gnutls_str_cpy(tmpbuffer3, sizeof(tmpbuffer3),
-				tmpbuffer1);
-		_gnutls_str_cat(tmpbuffer3, sizeof(tmpbuffer3), ".type");
-
-		len = sizeof(oid) - 1;
-		result =
-		    asn1_read_value(asn1_struct, tmpbuffer3, oid, &len);
-
-		if (result == ASN1_ELEMENT_NOT_FOUND)
-			break;
-		else if (result != ASN1_SUCCESS) {
-			gnutls_assert();
-			result = _gnutls_asn2err(result);
-			goto cleanup;
-		}
-
-		if (strcmp(oid, given_oid) == 0) {	/* Found the OID */
-
-			/* Read the Value
-			 */
-			snprintf(tmpbuffer3, sizeof(tmpbuffer3),
-				 "%s.values.?%u", tmpbuffer1, indx + 1);
-
-			len = sizeof(value) - 1;
-			result =
-			    _gnutls_x509_read_value(asn1_struct,
-						    tmpbuffer3, &td);
-
-			if (result != ASN1_SUCCESS) {
-				gnutls_assert();
-				result = _gnutls_asn2err(result);
-				goto cleanup;
-			}
-
-			if (raw == 0) {
-				result =
-				    _gnutls_x509_dn_to_string
-				    (oid, td.data, td.size, out);
-
-				_gnutls_free_datum(&td);
-
-				if (result < 0) {
-					gnutls_assert();
-					goto cleanup;
-				}
-				return 0;
-			} else {	/* raw!=0 */
-				out->data = td.data;
-				out->size = td.size;
-
-				return 0;
-			}
-		}
-
-	}
-	while (1);
-
-	gnutls_assert();
-
-	result = GNUTLS_E_REQUESTED_DATA_NOT_AVAILABLE;
-
-      cleanup:
-	return result;
-}
-
 /**
  * gnutls_x509_crq_get_challenge_password:
  * @crq: should contain a #gnutls_x509_crq_t type
@@ -559,7 +445,7 @@ gnutls_x509_crq_get_challenge_password(gnutls_x509_crq_t crq,
 	}
 
 	ret =
-	    parse_attribute(crq->crq,
+	    _x509_parse_attribute(crq->crq,
 			    "certificationRequestInfo.attributes",
 			    "1.2.840.113549.1.9.7", 0, 0, &td);
 	if (ret < 0)
@@ -568,149 +454,6 @@ gnutls_x509_crq_get_challenge_password(gnutls_x509_crq_t crq,
 	return _gnutls_strdatum_to_buf(&td, pass, pass_size);
 }
 
-/* This function will attempt to set the requested attribute in
- * the given X509v3 certificate.
- *
- * Critical will be either 0 or 1.
- */
-static int
-add_attribute(ASN1_TYPE asn, const char *root, const char *attribute_id,
-	      const gnutls_datum_t * ext_data)
-{
-	int result;
-	char name[MAX_NAME_SIZE];
-
-	snprintf(name, sizeof(name), "%s", root);
-
-	/* Add a new attribute in the list.
-	 */
-	result = asn1_write_value(asn, name, "NEW", 1);
-	if (result != ASN1_SUCCESS) {
-		gnutls_assert();
-		return _gnutls_asn2err(result);
-	}
-
-	snprintf(name, sizeof(name), "%s.?LAST.type", root);
-
-	result = asn1_write_value(asn, name, attribute_id, 1);
-	if (result != ASN1_SUCCESS) {
-		gnutls_assert();
-		return _gnutls_asn2err(result);
-	}
-
-	snprintf(name, sizeof(name), "%s.?LAST.values", root);
-
-	result = asn1_write_value(asn, name, "NEW", 1);
-	if (result != ASN1_SUCCESS) {
-		gnutls_assert();
-		return _gnutls_asn2err(result);
-	}
-
-	snprintf(name, sizeof(name), "%s.?LAST.values.?LAST", root);
-
-	result = _gnutls_x509_write_value(asn, name, ext_data);
-	if (result < 0) {
-		gnutls_assert();
-		return result;
-	}
-
-	return 0;
-}
-
-/* Overwrite the given attribute (using the index)
- * index here starts from one.
- */
-static int
-overwrite_attribute(ASN1_TYPE asn, const char *root, unsigned indx,
-		    const gnutls_datum_t * ext_data)
-{
-	char name[MAX_NAME_SIZE], name2[MAX_NAME_SIZE];
-	int result;
-
-	snprintf(name, sizeof(name), "%s.?%u", root, indx);
-
-	_gnutls_str_cpy(name2, sizeof(name2), name);
-	_gnutls_str_cat(name2, sizeof(name2), ".values.?LAST");
-
-	result = _gnutls_x509_write_value(asn, name2, ext_data);
-	if (result < 0) {
-		gnutls_assert();
-		return result;
-	}
-
-
-	return 0;
-}
-
-static int
-set_attribute(ASN1_TYPE asn, const char *root,
-	      const char *ext_id, const gnutls_datum_t * ext_data)
-{
-	int result;
-	int k, len;
-	char name[MAX_NAME_SIZE], name2[MAX_NAME_SIZE];
-	char extnID[MAX_OID_SIZE];
-
-	/* Find the index of the given attribute.
-	 */
-	k = 0;
-	do {
-		k++;
-
-		snprintf(name, sizeof(name), "%s.?%u", root, k);
-
-		len = sizeof(extnID) - 1;
-		result = asn1_read_value(asn, name, extnID, &len);
-
-		/* move to next
-		 */
-
-		if (result == ASN1_ELEMENT_NOT_FOUND) {
-			break;
-		}
-
-		do {
-
-			_gnutls_str_cpy(name2, sizeof(name2), name);
-			_gnutls_str_cat(name2, sizeof(name2), ".type");
-
-			len = sizeof(extnID) - 1;
-			result = asn1_read_value(asn, name2, extnID, &len);
-
-			if (result == ASN1_ELEMENT_NOT_FOUND) {
-				gnutls_assert();
-				break;
-			} else if (result != ASN1_SUCCESS) {
-				gnutls_assert();
-				return _gnutls_asn2err(result);
-			}
-
-			/* Handle Extension
-			 */
-			if (strcmp(extnID, ext_id) == 0) {
-				/* attribute was found
-				 */
-				return overwrite_attribute(asn, root, k,
-							   ext_data);
-			}
-
-
-		}
-		while (0);
-	}
-	while (1);
-
-	if (result == ASN1_ELEMENT_NOT_FOUND) {
-		return add_attribute(asn, root, ext_id, ext_data);
-	} else {
-		gnutls_assert();
-		return _gnutls_asn2err(result);
-	}
-
-
-	return 0;
-}
-
 /**
  * gnutls_x509_crq_set_attribute_by_oid:
  * @crq: should contain a #gnutls_x509_crq_t type
@@ -743,7 +486,7 @@ gnutls_x509_crq_set_attribute_by_oid(gnutls_x509_crq_t crq,
 		return GNUTLS_E_INVALID_REQUEST;
 	}
 
-	return set_attribute(crq->crq,
+	return _x509_set_attribute(crq->crq,
 			     "certificationRequestInfo.attributes", oid,
 			     &data);
 }
@@ -781,7 +524,7 @@ gnutls_x509_crq_get_attribute_by_oid(gnutls_x509_crq_t crq,
 	}
 
 	ret =
-	    parse_attribute(crq->crq,
+	    _x509_parse_attribute(crq->crq,
 			    "certificationRequestInfo.attributes", oid,
 			    indx, 1, &td);
 	if (ret < 0)
@@ -919,7 +662,7 @@ gnutls_x509_crq_set_key(gnutls_x509_crq_t crq, gnutls_x509_privkey_t key)
 	result = _gnutls_x509_encode_and_copy_PKI_params
 	    (crq->crq,
 	     "certificationRequestInfo.subjectPKInfo",
-	     key->pk_algorithm, &key->params);
+	     &key->params);
 
 	if (result < 0) {
 		gnutls_assert();
@@ -1037,11 +780,12 @@ gnutls_x509_crq_set_key_rsa_raw(gnutls_x509_crq_t crq,
 	}
 
 	temp_params.params_nr = RSA_PUBLIC_PARAMS;
+	temp_params.algo = GNUTLS_PK_RSA;
 
 	result = _gnutls_x509_encode_and_copy_PKI_params
 	    (crq->crq,
 	     "certificationRequestInfo.subjectPKInfo",
-	     GNUTLS_PK_RSA, &temp_params);
+	     &temp_params);
 
 	if (result < 0) {
 		gnutls_assert();
@@ -1099,6 +843,8 @@ gnutls_x509_crq_set_challenge_password(gnutls_x509_crq_t crq,
 		password = (char*)out.data;
 	}
 
+	assert(password != NULL);
+
 	result = _gnutls_x509_encode_and_write_attribute
 	    ("1.2.840.113549.1.9.7", crq->crq,
 	     "certificationRequestInfo.attributes.?LAST", password,
@@ -1284,14 +1030,58 @@ gnutls_x509_crq_get_pk_algorithm(gnutls_x509_crq_t crq, unsigned int *bits)
 	}
 
 	result = _gnutls_x509_get_pk_algorithm
-	    (crq->crq, "certificationRequestInfo.subjectPKInfo", bits);
+	    (crq->crq, "certificationRequestInfo.subjectPKInfo", NULL, bits);
 	if (result < 0) {
 		gnutls_assert();
+		return result;
 	}
 
 	return result;
 }
 
+/**
+ * gnutls_x509_crq_get_spki;
+ * @crq: should contain a #gnutls_x509_crq_t type
+ * @spki: a SubjectPublicKeyInfo structure of type #gnutls_x509_spki_t
+ * @flags: must be zero
+ *
+ * This function will return the public key information of a PKCS#10
+ * certificate request. The provided @spki must be initialized.
+ *
+ * Returns: Zero on success, or a negative error code on error.
+ **/
+int
+gnutls_x509_crq_get_spki(gnutls_x509_crq_t crq,
+			 gnutls_x509_spki_t spki,
+			 unsigned int flags)
+{
+	int result;
+	gnutls_x509_spki_st params;
+
+	if (crq == NULL) {
+		gnutls_assert();
+		return GNUTLS_E_INVALID_REQUEST;
+	}
+
+	memset(&params, 0, sizeof(params));
+
+	spki->pk = gnutls_x509_crq_get_pk_algorithm(crq, NULL);
+
+	result = _gnutls_x509_crq_read_spki_params(crq, &params);
+	if (result < 0) {
+		gnutls_assert();
+		return result;
+	}
+
+	if (params.pk == GNUTLS_PK_UNKNOWN)
+		return gnutls_assert_val(GNUTLS_E_REQUESTED_DATA_NOT_AVAILABLE);
+
+	spki->rsa_pss_dig = params.rsa_pss_dig;
+	spki->salt_size = params.salt_size;
+
+	return 0;
+}
+
 /**
  * gnutls_x509_crq_get_signature_oid:
  * @crq: should contain a #gnutls_x509_crq_t type
@@ -2714,7 +2504,7 @@ gnutls_x509_crq_get_key_id(gnutls_x509_crq_t crq, unsigned int flags,
 			   unsigned char *output_data,
 			   size_t * output_data_size)
 {
-	int pk, ret = 0;
+	int ret = 0;
 	gnutls_pk_params_st params;
 
 	if (crq == NULL) {
@@ -2722,12 +2512,6 @@ gnutls_x509_crq_get_key_id(gnutls_x509_crq_t crq, unsigned int flags,
 		return GNUTLS_E_INVALID_REQUEST;
 	}
 
-	pk = gnutls_x509_crq_get_pk_algorithm(crq, NULL);
-	if (pk < 0) {
-		gnutls_assert();
-		return pk;
-	}
-
 	ret = _gnutls_x509_crq_get_mpis(crq, &params);
 	if (ret < 0) {
 		gnutls_assert();
@@ -2735,7 +2519,7 @@ gnutls_x509_crq_get_key_id(gnutls_x509_crq_t crq, unsigned int flags,
 	}
 
 	ret =
-	    _gnutls_get_key_id(pk, &params, output_data, output_data_size, flags);
+	    _gnutls_get_key_id(&params, output_data, output_data_size, flags);
 
 	gnutls_pk_params_release(&params);
 
@@ -2772,6 +2556,9 @@ gnutls_x509_crq_privkey_sign(gnutls_x509_crq_t crq, gnutls_privkey_t key,
 	int result;
 	gnutls_datum_t signature;
 	gnutls_datum_t tbs;
+	gnutls_pk_algorithm_t pk;
+	gnutls_x509_spki_st params;
+	const gnutls_sign_entry_st *se;
 
 	if (crq == NULL) {
 		gnutls_assert();
@@ -2788,6 +2575,19 @@ gnutls_x509_crq_privkey_sign(gnutls_x509_crq_t crq, gnutls_privkey_t key,
 		}
 	}
 
+	result = _gnutls_privkey_get_spki_params(key, &params);
+	if (result < 0) {
+		gnutls_assert();
+		return result;
+	}
+
+	pk = gnutls_privkey_get_pk_algorithm(key, NULL);
+	result = _gnutls_privkey_update_spki_params(key, pk, dig, 0, &params);
+	if (result < 0) {
+		gnutls_assert();
+		return result;
+	}
+
 	/* Step 1. Self sign the request.
 	 */
 	result =
@@ -2799,7 +2599,12 @@ gnutls_x509_crq_privkey_sign(gnutls_x509_crq_t crq, gnutls_privkey_t key,
 		return result;
 	}
 
-	result = gnutls_privkey_sign_data(key, dig, 0, &tbs, &signature);
+	se = _gnutls_pk_to_sign_entry(params.pk, dig);
+	if (se == NULL)
+		return gnutls_assert_val(GNUTLS_E_INVALID_REQUEST);
+
+	result = privkey_sign_and_hash_data(key, se,
+					    &tbs, &signature, &params);
 	gnutls_free(tbs.data);
 
 	if (result < 0) {
@@ -2823,9 +2628,8 @@ gnutls_x509_crq_privkey_sign(gnutls_x509_crq_t crq, gnutls_privkey_t key,
 	/* Step 3. Write the signatureAlgorithm field.
 	 */
 	result =
-	    _gnutls_x509_write_sig_params(crq->crq, "signatureAlgorithm",
-					  gnutls_privkey_get_pk_algorithm
-					  (key, NULL), dig, 0);
+	    _gnutls_x509_write_sign_params(crq->crq, "signatureAlgorithm",
+					   se, &params);
 	if (result < 0) {
 		gnutls_assert();
 		return result;
@@ -2853,7 +2657,8 @@ int gnutls_x509_crq_verify(gnutls_x509_crq_t crq, unsigned int flags)
 	gnutls_datum_t data = { NULL, 0 };
 	gnutls_datum_t signature = { NULL, 0 };
 	gnutls_pk_params_st params;
-	gnutls_digest_algorithm_t algo;
+	gnutls_x509_spki_st sign_params;
+	const gnutls_sign_entry_st *se;
 	int ret;
 
 	gnutls_pk_params_init(&params);
@@ -2869,13 +2674,18 @@ int gnutls_x509_crq_verify(gnutls_x509_crq_t crq, unsigned int flags)
 
 	ret =
 	    _gnutls_x509_get_signature_algorithm(crq->crq,
-						 "signatureAlgorithm.algorithm");
+						 "signatureAlgorithm");
 	if (ret < 0) {
 		gnutls_assert();
 		goto cleanup;
 	}
 
-	algo = gnutls_sign_get_hash_algorithm(ret);
+	se = _gnutls_sign_to_entry(ret);
+	if (se == NULL) {
+		gnutls_assert();
+		ret = GNUTLS_E_UNSUPPORTED_SIGNATURE_ALGORITHM;
+		goto cleanup;
+	}
 
 	ret =
 	    _gnutls_x509_get_signature(crq->crq, "signature", &signature);
@@ -2890,10 +2700,17 @@ int gnutls_x509_crq_verify(gnutls_x509_crq_t crq, unsigned int flags)
 		goto cleanup;
 	}
 
+	ret = _gnutls_x509_read_sign_params(crq->crq,
+					    "signatureAlgorithm",
+					    &sign_params);
+	if (ret < 0) {
+		gnutls_assert();
+		goto cleanup;
+	}
+
 	ret =
-	    pubkey_verify_data(gnutls_x509_crq_get_pk_algorithm(crq, NULL),
-			       hash_to_entry(algo), &data, &signature,
-			       &params);
+	    pubkey_verify_data(se, hash_to_entry(se->hash), &data, &signature,
+			       &params, &sign_params, flags);
 	if (ret < 0) {
 		gnutls_assert();
 		goto cleanup;
@@ -2973,7 +2790,7 @@ gnutls_x509_crq_set_private_key_usage_period(gnutls_x509_crq_t crq,
 
 /**
  * gnutls_x509_crq_get_tlsfeatures:
- * @crt: A X.509 certificate request
+ * @crq: An X.509 certificate request
  * @features: If the function succeeds, the
  *   features will be stored in this variable.
  * @flags: zero or %GNUTLS_EXT_FLAG_APPEND
@@ -3035,7 +2852,7 @@ int gnutls_x509_crq_get_tlsfeatures(gnutls_x509_crq_t crq,
 
 /**
  * gnutls_x509_crq_set_tlsfeatures:
- * @crt: A X.509 certificate request
+ * @crq: An X.509 certificate request
  * @features: If the function succeeds, the
  *   features will be added to the certificate
  *   request.
@@ -3049,7 +2866,7 @@ int gnutls_x509_crq_get_tlsfeatures(gnutls_x509_crq_t crq,
  * Since: 3.5.1
  **/
 int gnutls_x509_crq_set_tlsfeatures(gnutls_x509_crq_t crq,
-								   gnutls_x509_tlsfeatures_t features)
+				    gnutls_x509_tlsfeatures_t features)
 {
 	int ret;
 	gnutls_datum_t der;
@@ -3118,3 +2935,118 @@ gnutls_x509_crq_set_extension_by_oid(gnutls_x509_crq_t crq,
 	return 0;
 
 }
+
+/**
+ * gnutls_x509_crq_set_spki:
+ * @crq: a certificate request of type #gnutls_x509_crq_t
+ * @spki: a SubjectPublicKeyInfo structure of type #gnutls_x509_spki_t
+ * @flags: must be zero
+ *
+ * This function will set the certificate request's subject public key
+ * information explicitly. This is intended to be used in the cases
+ * where a single public key (e.g., RSA) can be used for multiple
+ * signature algorithms (RSA PKCS1-1.5, and RSA-PSS).
+ *
+ * To export the public key (i.e., the SubjectPublicKeyInfo part), check
+ * gnutls_pubkey_import_x509().
+ *
+ * Returns: On success, %GNUTLS_E_SUCCESS (0) is returned, otherwise a
+ *   negative error value.
+ *
+ * Since: 3.6.0
+ **/
+int
+gnutls_x509_crq_set_spki(gnutls_x509_crq_t crq,
+			 const gnutls_x509_spki_t spki,
+			 unsigned int flags)
+{
+	int ret;
+	gnutls_pk_algorithm_t crq_pk;
+	gnutls_x509_spki_st tpki;
+	gnutls_pk_params_st params;
+	unsigned bits;
+
+	if (crq == NULL) {
+		gnutls_assert();
+		return GNUTLS_E_INVALID_REQUEST;
+	}
+
+	ret = _gnutls_x509_crq_get_mpis(crq, &params);
+	if (ret < 0) {
+		gnutls_assert();
+		return ret;
+	}
+
+	bits = pubkey_to_bits(&params);
+	crq_pk = params.algo;
+
+	if (!_gnutls_pk_are_compat(crq_pk, spki->pk)) {
+                ret = gnutls_assert_val(GNUTLS_E_INVALID_REQUEST);
+                goto cleanup;
+	}
+
+	if (spki->pk != GNUTLS_PK_RSA_PSS) {
+		if (crq_pk == spki->pk) {
+			ret = 0;
+			goto cleanup;
+		}
+
+		gnutls_assert();
+		ret = GNUTLS_E_INVALID_REQUEST;
+		goto cleanup;
+	}
+
+	if (crq_pk == GNUTLS_PK_RSA) {
+		const mac_entry_st *me;
+
+		me = hash_to_entry(spki->rsa_pss_dig);
+		if (unlikely(me == NULL)) {
+			gnutls_assert();
+			ret = GNUTLS_E_INVALID_REQUEST;
+			goto cleanup;
+		}
+
+		memset(&tpki, 0, sizeof(gnutls_x509_spki_st));
+		tpki.pk = spki->pk;
+		tpki.rsa_pss_dig = spki->rsa_pss_dig;
+
+		/* If salt size is zero, find the optimal salt size. */
+		if (spki->salt_size == 0) {
+			tpki.salt_size =
+			    _gnutls_find_rsa_pss_salt_size(bits, me,
+							   spki->salt_size);
+		} else
+			tpki.salt_size = spki->salt_size;
+	} else if (crq_pk == GNUTLS_PK_RSA_PSS) {
+		ret = _gnutls_x509_crq_read_spki_params(crq, &tpki);
+		if (ret < 0) {
+			gnutls_assert();
+			goto cleanup;
+		}
+
+		tpki.salt_size = spki->salt_size;
+		tpki.rsa_pss_dig = spki->rsa_pss_dig;
+	}
+
+	memcpy(&params.spki, &tpki, sizeof(tpki));
+	ret = _gnutls_x509_check_pubkey_params(&params);
+	if (ret < 0) {
+		gnutls_assert();
+		goto cleanup;
+	}
+
+	ret = _gnutls_x509_write_spki_params(crq->crq,
+						"certificationRequestInfo."
+						"subjectPKInfo."
+						"algorithm",
+						&tpki);
+	if (ret < 0) {
+		gnutls_assert();
+		goto cleanup;
+	}
+
+	ret = 0;
+ cleanup:
+	gnutls_pk_params_release(&params);
+	return ret;
+}
