@@ -1,6 +1,8 @@
 /*
  * Copyright (C) 2011-2012 Free Software Foundation, Inc.
- * Author: Simon Josefsson
+ * Copyright (C) 2016-2017 Red Hat, Inc.
+ *
+ * Author: Simon Josefsson, Nikos Mavrogiannopoulos
  *
  * This file is part of GnuTLS.
  *
@@ -33,6 +35,8 @@
 #include <gnutls/ocsp.h>
 #include <auth/cert.h>
 
+#include <assert.h>
+
 typedef struct gnutls_ocsp_req_int {
 	ASN1_TYPE req;
 	unsigned init;
@@ -981,6 +985,9 @@ int gnutls_ocsp_resp_get_status(gnutls_ocsp_resp_t resp)
 		return _gnutls_asn2err(ret);
 	}
 
+	if (len != 1)
+		return gnutls_assert_val(GNUTLS_E_UNEXPECTED_PACKET);
+
 	switch (str[0]) {
 	case GNUTLS_OCSP_RESP_SUCCESSFUL:
 	case GNUTLS_OCSP_RESP_MALFORMEDREQUEST:
@@ -990,7 +997,7 @@ int gnutls_ocsp_resp_get_status(gnutls_ocsp_resp_t resp)
 	case GNUTLS_OCSP_RESP_UNAUTHORIZED:
 		break;
 	default:
-		return GNUTLS_E_UNEXPECTED_PACKET;
+		return gnutls_assert_val(GNUTLS_E_UNEXPECTED_PACKET);
 	}
 
 	return (int) str[0];
@@ -1919,28 +1926,64 @@ static gnutls_x509_crt_t find_signercert(gnutls_ocsp_resp_t resp)
 	}
 
 	for (i = 0; i < ncerts; i++) {
+		assert(certs[i] != NULL);
+		_gnutls_cert_log("checking whether signed against", certs[i]);
 		if (keyid.data != NULL) {
-			uint8_t digest[20];
+			uint8_t digest[64]; /* to support longer key IDs */
 			gnutls_datum_t spki;
+			size_t digest_size = sizeof(digest);
+			int len;
 
+			_gnutls_debug_log("checking key ID against SPK identifier\n");
+
+			/* check subject key identifier as well, some certificates
+			 * match that, but not the hash */
+			rc = gnutls_x509_crt_get_subject_key_id(certs[i], digest, &digest_size, NULL);
+			if (rc >= 0 && digest_size == keyid.size &&
+			    memcmp(keyid.data, digest, digest_size) == 0) {
+				signercert = certs[i];
+				goto quit;
+			}
+
+			_gnutls_debug_log("checking key ID against SPKI hash\n");
+
+			/* continue with checking the hash */
 			rc = _gnutls_x509_get_raw_field2(certs[i]->cert, &certs[i]->der,
 					  "tbsCertificate.subjectPublicKeyInfo.subjectPublicKey",
 					  &spki);
 			if (rc < 0 || spki.size < 6) {
+				gnutls_assert();
 				signercert = NULL;
-				goto quit;
+				continue;
 			}
 
 			/* For some reason the protocol requires we skip the
 			 * tag, length and number of unused bits.
 			 */
-			spki.data += 5;
-			spki.size -= 5;
-			rc = gnutls_hash_fast(GNUTLS_DIG_SHA1, spki.data, spki.size, digest);
+			if (spki.data[0] != 0x03) { /* bit string */
+				gnutls_assert();
+				signercert = NULL;
+				continue;
+			}
+
+			rc = asn1_get_length_der(spki.data+1, spki.size-1, &len);
+			if (rc <= 0) {
+				gnutls_assert();
+				signercert = NULL;
+				continue;
+			}
+			len += 1+1; /* skip unused bits as well */
+			if (len >= (int)spki.size) {
+				gnutls_assert();
+				signercert = NULL;
+				continue;
+			}
+
+			rc = gnutls_hash_fast(GNUTLS_DIG_SHA1, spki.data+len, spki.size-len, digest);
 			if (rc < 0) {
 				gnutls_assert();
 				signercert = NULL;
-				goto quit;
+				continue;
 			}
 
 			if ((20 == keyid.size) &&
@@ -1948,12 +1991,17 @@ static gnutls_x509_crt_t find_signercert(gnutls_ocsp_resp_t resp)
 				signercert = certs[i];
 				goto quit;
 			}
+			gnutls_assert();
 		} else {
+			_gnutls_debug_log("checking issuer DN\n");
+
+			assert(riddn.data != NULL);
 			if ((certs[i]->raw_dn.size == riddn.size)
 			    && memcmp(riddn.data, certs[i]->raw_dn.data, riddn.size) == 0) {
 				signercert = certs[i];
 				goto quit;
 			}
+			gnutls_assert();
 		}
 	}
 
@@ -2005,6 +2053,8 @@ _ocsp_resp_verify_direct(gnutls_ocsp_resp_t resp,
 		goto done;
 	}
 
+	_gnutls_cert_log("ocsp signer", signercert); \
+
 	rc = gnutls_pubkey_import_x509(pubkey, signercert, 0);
 	if (rc != GNUTLS_E_SUCCESS) {
 		gnutls_assert();
@@ -2138,6 +2188,7 @@ gnutls_ocsp_resp_verify_direct(gnutls_ocsp_resp_t resp,
 		}
 
 		if (vtmp != 0) {
+			_gnutls_reason_log("cert verification", vtmp);
 			*verify = vstatus_to_ocsp_status(vtmp);
 			gnutls_assert();
 			rc = GNUTLS_E_SUCCESS;
