@@ -52,7 +52,6 @@ struct node_st {
 	/* The trusted CRLs */
 	gnutls_x509_crl_t *crls;
 	unsigned int crl_size;
-
 };
 
 struct gnutls_x509_trust_list_iter {
@@ -368,7 +367,7 @@ advance_iter(gnutls_x509_trust_list_t list,
 	if (list->pkcs11_token != NULL) {
 		if (iter->pkcs11_list == NULL) {
 			ret = gnutls_pkcs11_obj_list_import_url2(&iter->pkcs11_list, &iter->pkcs11_size,
-			    list->pkcs11_token, (GNUTLS_PKCS11_OBJ_FLAG_PRESENT_IN_TRUSTED_MODULE|GNUTLS_PKCS11_OBJ_FLAG_CRT|GNUTLS_PKCS11_OBJ_FLAG_MARK_CA|GNUTLS_PKCS11_OBJ_FLAG_MARK_TRUSTED), 0);
+			    list->pkcs11_token, (GNUTLS_PKCS11_OBJ_FLAG_CRT|GNUTLS_PKCS11_OBJ_FLAG_MARK_CA|GNUTLS_PKCS11_OBJ_FLAG_MARK_TRUSTED), 0);
 			if (ret < 0)
 				return gnutls_assert_val(ret);
 
@@ -709,6 +708,7 @@ gnutls_x509_trust_list_add_crls(gnutls_x509_trust_list_t list,
 	unsigned x, i, j = 0;
 	unsigned int vret = 0;
 	uint32_t hash;
+	gnutls_x509_crl_t *tmp;
 
 	/* Probably we can optimize things such as removing duplicates
 	 * etc.
@@ -734,6 +734,8 @@ gnutls_x509_trust_list_add_crls(gnutls_x509_trust_list_t list,
 						   &vret);
 			if (ret < 0 || vret != 0) {
 				_gnutls_debug_log("CRL verification failed, not adding it\n");
+				if (flags & GNUTLS_TL_NO_DUPLICATES)
+					gnutls_x509_crl_deinit(crl_list[i]);
 				continue;
 			}
 		}
@@ -753,22 +755,28 @@ gnutls_x509_trust_list_add_crls(gnutls_x509_trust_list_t list,
 					} else {
 						/* The new is older, discard it */
 						gnutls_x509_crl_deinit(crl_list[i]);
-						continue;
+						goto next;
 					}
 				}
 			}
 		}
 
-		list->node[hash].crls =
-		    gnutls_realloc_fast(list->node[hash].crls,
+		tmp =
+		    gnutls_realloc(list->node[hash].crls,
 					(list->node[hash].crl_size +
 					 1) *
 					sizeof(list->node[hash].
 					       trusted_cas[0]));
-		if (list->node[hash].crls == NULL) {
+		if (tmp == NULL) {
+			ret = i;
 			gnutls_assert();
-			return i;
+			if (flags & GNUTLS_TL_NO_DUPLICATES)
+				while (i < crl_size)
+					gnutls_x509_crl_deinit(crl_list[i++]);
+			return ret;
 		}
+		list->node[hash].crls = tmp;
+
 
 		list->node[hash].crls[list->node[hash].crl_size] =
 		    crl_list[i];
@@ -964,7 +972,7 @@ int gnutls_x509_trust_list_get_issuer(gnutls_x509_trust_list_t list,
 		gnutls_datum_t der = {NULL, 0};
 		/* use the token for verification */
 		ret = gnutls_pkcs11_get_raw_issuer(list->pkcs11_token, cert, &der,
-			GNUTLS_X509_FMT_DER, GNUTLS_PKCS11_OBJ_FLAG_PRESENT_IN_TRUSTED_MODULE);
+			GNUTLS_X509_FMT_DER, 0);
 		if (ret < 0) {
 			gnutls_assert();
 			return ret;
@@ -1036,7 +1044,7 @@ int gnutls_x509_trust_list_get_issuer_by_dn(gnutls_x509_trust_list_t list,
 		gnutls_datum_t der = {NULL, 0};
 		/* use the token for verification */
 		ret = gnutls_pkcs11_get_raw_issuer_by_dn(list->pkcs11_token, dn, &der,
-			GNUTLS_X509_FMT_DER, GNUTLS_PKCS11_OBJ_FLAG_PRESENT_IN_TRUSTED_MODULE);
+			GNUTLS_X509_FMT_DER, 0);
 		if (ret < 0) {
 			gnutls_assert();
 			return ret;
@@ -1097,7 +1105,7 @@ int gnutls_x509_trust_list_get_issuer_by_subject_key_id(gnutls_x509_trust_list_t
 		gnutls_datum_t der = {NULL, 0};
 		/* use the token for verification */
 		ret = gnutls_pkcs11_get_raw_issuer_by_subject_key_id(list->pkcs11_token, dn, spki, &der,
-			GNUTLS_X509_FMT_DER, GNUTLS_PKCS11_OBJ_FLAG_PRESENT_IN_TRUSTED_MODULE);
+			GNUTLS_X509_FMT_DER, 0);
 		if (ret < 0) {
 			gnutls_assert();
 			return ret;
@@ -1175,6 +1183,15 @@ gnutls_x509_trust_list_verify_crt(gnutls_x509_trust_list_t list,
 						  NULL, 0, flags, voutput, func);
 }
 
+#define LAST_DN cert_list[cert_list_size-1]->raw_dn
+#define LAST_IDN cert_list[cert_list_size-1]->raw_issuer_dn
+/* This macro is introduced to detect a verification output
+ * which indicates an unknown signer, or a signer which uses
+ * an insecure algorithm (e.g., sha1), something that indicates
+ * a superceded signer */
+#define SIGNER_OLD_OR_UNKNOWN(output) ((output & GNUTLS_CERT_SIGNER_NOT_FOUND) || (output & GNUTLS_CERT_INSECURE_ALGORITHM))
+#define SIGNER_WAS_KNOWN(output) (!(output & GNUTLS_CERT_SIGNER_NOT_FOUND))
+
 /**
  * gnutls_x509_trust_list_verify_crt2:
  * @list: The list
@@ -1202,7 +1219,7 @@ gnutls_x509_trust_list_verify_crt(gnutls_x509_trust_list_t list,
  * flags.
  *
  * Additional verification parameters are possible via the @data types; the
- * acceptable types are %GNUTLS_DT_DNS_HOSTNAME and %GNUTLS_DT_KEY_PURPOSE_OID.
+ * acceptable types are %GNUTLS_DT_DNS_HOSTNAME, %GNUTLS_DT_IP_ADDRESS and %GNUTLS_DT_KEY_PURPOSE_OID.
  * The former accepts as data a null-terminated hostname, and the latter a null-terminated
  * object identifier (e.g., %GNUTLS_KP_TLS_WWW_SERVER).
  * If a DNS hostname is provided then this function will compare
@@ -1237,6 +1254,9 @@ gnutls_x509_trust_list_verify_crt2(gnutls_x509_trust_list_t list,
 	gnutls_x509_crt_t sorted[DEFAULT_MAX_VERIFY_DEPTH];
 	const char *hostname = NULL, *purpose = NULL, *email = NULL;
 	unsigned hostname_size = 0;
+	unsigned have_set_name = 0;
+	unsigned saved_output;
+	gnutls_datum_t ip = {NULL, 0};
 
 	if (cert_list == NULL || cert_list_size < 1)
 		return gnutls_assert_val(GNUTLS_E_INVALID_REQUEST);
@@ -1247,10 +1267,22 @@ gnutls_x509_trust_list_verify_crt2(gnutls_x509_trust_list_t list,
 			if (data[i].size > 0) {
 				hostname_size = data[i].size;
 			}
-			if (email != NULL) return gnutls_assert_val(GNUTLS_E_INVALID_REQUEST);
+
+			if (have_set_name != 0) return gnutls_assert_val(GNUTLS_E_INVALID_REQUEST);
+			have_set_name = 1;
+		} else if (data[i].type == GNUTLS_DT_IP_ADDRESS) {
+			if (data[i].size > 0) {
+				ip.data = data[i].data;
+				ip.size = data[i].size;
+			}
+
+			if (have_set_name != 0) return gnutls_assert_val(GNUTLS_E_INVALID_REQUEST);
+			have_set_name = 1;
 		} else if (data[i].type == GNUTLS_DT_RFC822NAME) {
 			email = (void*)data[i].data;
-			if (hostname != NULL) return gnutls_assert_val(GNUTLS_E_INVALID_REQUEST);
+
+			if (have_set_name != 0) return gnutls_assert_val(GNUTLS_E_INVALID_REQUEST);
+			have_set_name = 1;
 		} else if (data[i].type == GNUTLS_DT_KEY_PURPOSE_OID) {
 			purpose = (void*)data[i].data;
 		}
@@ -1299,11 +1331,9 @@ gnutls_x509_trust_list_verify_crt2(gnutls_x509_trust_list_t list,
 					    list->
 					    node[hash].trusted_ca_size,
 					    flags, purpose, func);
+	saved_output = *voutput;
 
-#define LAST_DN cert_list[cert_list_size-1]->raw_dn
-#define LAST_IDN cert_list[cert_list_size-1]->raw_issuer_dn
-
-	if ((*voutput) & GNUTLS_CERT_SIGNER_NOT_FOUND &&
+	if (SIGNER_OLD_OR_UNKNOWN(*voutput) &&
 		(LAST_DN.size != LAST_IDN.size ||
 		 memcmp(LAST_DN.data, LAST_IDN.data, LAST_IDN.size) != 0)) {
 
@@ -1315,16 +1345,25 @@ gnutls_x509_trust_list_verify_crt2(gnutls_x509_trust_list_t list,
 			  data, cert_list[cert_list_size - 1]->raw_dn.size);
 		hash %= list->size;
 
+		 _gnutls_debug_log("issuer in verification was not found or insecure; trying against trust list\n");
+
 		*voutput =
 		    _gnutls_verify_crt_status(cert_list, cert_list_size,
 					    list->node[hash].trusted_cas,
 					    list->
 					    node[hash].trusted_ca_size,
 					    flags, purpose, func);
+		if (*voutput != 0) {
+			if (SIGNER_WAS_KNOWN(saved_output))
+				*voutput = saved_output;
+			gnutls_assert();
+		}
 	}
 
+	saved_output = *voutput;
+
 #ifdef ENABLE_PKCS11
-	if ((*voutput & GNUTLS_CERT_SIGNER_NOT_FOUND) && list->pkcs11_token) {
+	if (SIGNER_OLD_OR_UNKNOWN(*voutput) && list->pkcs11_token) {
 		/* use the token for verification */
 
 		*voutput = _gnutls_pkcs11_verify_crt_status(list->pkcs11_token,
@@ -1332,6 +1371,8 @@ gnutls_x509_trust_list_verify_crt2(gnutls_x509_trust_list_t list,
 								purpose,
 								flags, func);
 		if (*voutput != 0) {
+			if (SIGNER_WAS_KNOWN(saved_output))
+				*voutput = saved_output;
 			gnutls_assert();
 		}
 	}
@@ -1349,15 +1390,28 @@ gnutls_x509_trust_list_verify_crt2(gnutls_x509_trust_list_t list,
 	if (hostname) {
 		ret =
 		    gnutls_x509_crt_check_hostname2(cert_list[0], hostname, flags);
-		if (ret == 0)
+		if (ret == 0) {
+			gnutls_assert();
 			*voutput |= GNUTLS_CERT_UNEXPECTED_OWNER|GNUTLS_CERT_INVALID;
+		}
+	}
+
+	if (ip.data) {
+		ret =
+		    gnutls_x509_crt_check_ip(cert_list[0], ip.data, ip.size, flags);
+		if (ret == 0) {
+			gnutls_assert();
+			*voutput |= GNUTLS_CERT_UNEXPECTED_OWNER|GNUTLS_CERT_INVALID;
+		}
 	}
 
 	if (email) {
 		ret =
 		    gnutls_x509_crt_check_email(cert_list[0], email, 0);
-		if (ret == 0)
+		if (ret == 0) {
+			gnutls_assert();
 			*voutput |= GNUTLS_CERT_UNEXPECTED_OWNER|GNUTLS_CERT_INVALID;
+		}
 	}
 
 	/* CRL checks follow */
