@@ -37,6 +37,7 @@
 #include <x509_int.h>
 #include <common.h>
 #include <pk.h>
+#include "supported_exts.h"
 
 /* Checks if two certs have the same name and the same key.  Return 1 on match. 
  * If @is_ca is zero then this function is identical to gnutls_x509_crt_equals()
@@ -88,6 +89,45 @@ _gnutls_check_if_same_key2(gnutls_x509_crt_t cert1,
 	return ret;
 }
 
+/* checks whether there are present unknown/unsupported critical extensions.
+ *
+ * Returns true if they are present.
+ */
+static unsigned check_for_unknown_exts(gnutls_x509_crt_t cert)
+{
+	unsigned i;
+	char oid[MAX_OID_SIZE];
+	size_t oid_size;
+	unsigned critical;
+	int ret;
+
+	for (i=0;;i++) {
+		oid_size = sizeof(oid);
+		oid[0] = 0;
+		critical = 0;
+
+		ret = gnutls_x509_crt_get_extension_info(cert, i, oid, &oid_size, &critical);
+		if (ret == GNUTLS_E_REQUESTED_DATA_NOT_AVAILABLE) {
+			return 0;
+		} else if (ret < 0) {
+			gnutls_assert();
+			/* could not decode? */
+			_gnutls_debug_log("Could not decode extension %d\n", i);
+			return 1;
+		}
+
+		if (critical == 0)
+			continue;
+
+		if (is_ext_oid_supported(oid, oid_size) == NULL) {
+			gnutls_assert();
+			_gnutls_debug_log("Unsupported critical extension: %s\n", oid);
+			return 1;
+		}
+	}
+
+	return 0;
+}
 
 /* Checks if the issuer of a certificate is a
  * Certificate Authority, or if the certificate is the same
@@ -347,18 +387,26 @@ static unsigned int check_time_status(gnutls_x509_crt_t crt, time_t now)
 	return 0;
 }
 
-unsigned _gnutls_is_broken_sig_allowed(gnutls_sign_algorithm_t sig, unsigned int flags)
+unsigned _gnutls_is_broken_sig_allowed(const gnutls_sign_entry_st *se, unsigned int flags)
 {
-	/* the first two are for backwards compatibility */
-	if ((sig == GNUTLS_SIGN_RSA_MD2)
-	    && (flags & GNUTLS_VERIFY_ALLOW_SIGN_RSA_MD2))
-		return 1;
-	if ((sig == GNUTLS_SIGN_RSA_MD5)
-	    && (flags & GNUTLS_VERIFY_ALLOW_SIGN_RSA_MD5))
-		return 1;
-	/* we no longer have individual flags - but rather a catch all */
+	gnutls_digest_algorithm_t hash;
+
+	/* we have a catch all */
 	if ((flags & GNUTLS_VERIFY_ALLOW_BROKEN) == GNUTLS_VERIFY_ALLOW_BROKEN)
 		return 1;
+
+	/* the first two are for backwards compatibility */
+	if ((se->id == GNUTLS_SIGN_RSA_MD2)
+	    && (flags & GNUTLS_VERIFY_ALLOW_SIGN_RSA_MD2))
+		return 1;
+	if ((se->id == GNUTLS_SIGN_RSA_MD5)
+	    && (flags & GNUTLS_VERIFY_ALLOW_SIGN_RSA_MD5))
+		return 1;
+
+	hash = se->hash;
+	if (hash == GNUTLS_DIG_SHA1 && (flags & GNUTLS_VERIFY_ALLOW_SIGN_WITH_SHA1))
+		return 1;
+
 	return 0;
 }
 
@@ -463,7 +511,7 @@ static unsigned is_level_acceptable(
 				return gnutls_assert_val(0);
 			}
 
-			curve = params.flags;
+			curve = params.curve;
 			gnutls_pk_params_release(&params);
 
 			if (curve != GNUTLS_ECC_CURVE_SECP256R1 &&
@@ -491,7 +539,7 @@ static unsigned is_level_acceptable(
 					return gnutls_assert_val(0);
 				}
 
-				issuer_curve = params.flags;
+				issuer_curve = params.curve;
 				gnutls_pk_params_release(&params);
 
 				if (issuer_curve != GNUTLS_ECC_CURVE_SECP256R1 &&
@@ -527,9 +575,16 @@ typedef struct verify_state_st {
 	gnutls_verify_output_function *func;
 } verify_state_st;
 
-#define MARK_INVALID(x) gnutls_assert(); \
+#define MARK_INVALID(x) { gnutls_assert(); \
 	out |= (x|GNUTLS_CERT_INVALID); \
-	result = 0
+	result = 0; }
+
+static int _gnutls_x509_verify_data(gnutls_sign_algorithm_t sign,
+				    const gnutls_datum_t * data,
+				    const gnutls_datum_t * signature,
+				    gnutls_x509_crt_t cert,
+				    gnutls_x509_crt_t issuer,
+				    unsigned vflags);
 
 /* 
  * Verifies the given certificate against a certificate list of
@@ -554,11 +609,11 @@ verify_crt(gnutls_x509_crt_t cert,
 	gnutls_datum_t cert_signed_data = { NULL, 0 };
 	gnutls_datum_t cert_signature = { NULL, 0 };
 	gnutls_x509_crt_t issuer = NULL;
-	int issuer_version, hash_algo;
+	int issuer_version;
 	unsigned result = 1;
-	const mac_entry_st * me;
 	unsigned int out = 0, usage;
 	int sigalg, ret;
+	const gnutls_sign_entry_st *se;
 
 	if (output)
 		*output = 0;
@@ -591,12 +646,14 @@ verify_crt(gnutls_x509_crt_t cert,
 
 	ret =
 	    _gnutls_x509_get_signature_algorithm(cert->cert,
-						 "signatureAlgorithm.algorithm");
+						 "signatureAlgorithm");
 	if (ret < 0) {
 		MARK_INVALID(0);
 	}
 	sigalg = ret;
 
+	se = _gnutls_sign_to_entry(sigalg);
+
 	/* issuer is not in trusted certificate
 	 * authorities.
 	 */
@@ -685,31 +742,44 @@ verify_crt(gnutls_x509_crt_t cert,
 			}
 		}
 
-		if (sigalg >= 0) {
-			hash_algo = gnutls_sign_get_hash_algorithm(sigalg);
-			me = mac_to_entry(hash_algo);
-		} else {
-			me = NULL;
-		}
-
-		if (me == NULL) {
+		if (sigalg < 0) {
 			MARK_INVALID(0);
 		} else if (cert_signed_data.data != NULL &&
-			   cert_signature.data != NULL) {
+		    cert_signature.data != NULL) {
 			ret =
-			    _gnutls_x509_verify_data(me,
+			    _gnutls_x509_verify_data(sigalg,
 						     &cert_signed_data,
 						     &cert_signature,
-						     issuer);
+						     cert,
+						     issuer, flags);
+
 			if (ret == GNUTLS_E_PK_SIG_VERIFY_FAILED) {
 				MARK_INVALID(GNUTLS_CERT_SIGNATURE_FAILURE);
+			} else if (ret == GNUTLS_E_CONSTRAINT_ERROR) {
+				MARK_INVALID(GNUTLS_CERT_SIGNER_CONSTRAINTS_FAILURE);
 			} else if (ret < 0) {
 				MARK_INVALID(0);
 			}
 		}
 	}
 
-	if (sigalg >= 0) {
+	/* we always check the issuer for unsupported critical extensions */
+	if (issuer && check_for_unknown_exts(issuer) != 0) {
+		if (!(flags & GNUTLS_VERIFY_IGNORE_UNKNOWN_CRIT_EXTENSIONS)) {
+			MARK_INVALID(GNUTLS_CERT_UNKNOWN_CRIT_EXTENSIONS);
+		}
+	}
+
+	/* we only check the end-certificate for critical extensions; that
+	 * way do not perform this check twice on the certificates when
+	 * verifying a large list */
+	if (end_cert && check_for_unknown_exts(cert) != 0) {
+		if (!(flags & GNUTLS_VERIFY_IGNORE_UNKNOWN_CRIT_EXTENSIONS)) {
+			MARK_INVALID(GNUTLS_CERT_UNKNOWN_CRIT_EXTENSIONS);
+		}
+	}
+
+	if (sigalg >= 0 && se) {
 		if (is_level_acceptable(cert, issuer, sigalg, flags) == 0) {
 			MARK_INVALID(GNUTLS_CERT_INSECURE_ALGORITHM);
 		}
@@ -718,8 +788,8 @@ verify_crt(gnutls_x509_crt_t cert,
 		 * used are secure. If the certificate is self signed it doesn't
 		 * really matter.
 		 */
-		if (gnutls_sign_is_secure(sigalg) == 0 &&
-		    _gnutls_is_broken_sig_allowed(sigalg, flags) == 0 &&
+		if (_gnutls_sign_is_secure2(se, GNUTLS_SIGN_FLAG_SECURE_FOR_CERTS) == 0 &&
+		    _gnutls_is_broken_sig_allowed(se, flags) == 0 &&
 		    is_issuer(cert, cert) == 0) {
 			MARK_INVALID(GNUTLS_CERT_INSECURE_ALGORITHM);
 		}
@@ -963,6 +1033,7 @@ cleanup:
 	return status;
 }
 
+
 #define PURPOSE_NSSGC "2.16.840.1.113730.4.1"
 #define PURPOSE_VSGC "2.16.840.1.113733.1.8.1"
 
@@ -1174,6 +1245,17 @@ _gnutls_pkcs11_verify_crt_status(const char* url,
 		goto cleanup;
 	}
 
+	/* check if the raw issuer is blacklisted (it can happen if
+	 * the issuer is both in the trusted list and the blacklisted)
+	 */
+	if (gnutls_pkcs11_crt_is_known (url, issuer,
+		GNUTLS_PKCS11_OBJ_FLAG_PRESENT_IN_TRUSTED_MODULE|
+		GNUTLS_PKCS11_OBJ_FLAG_RETRIEVE_DISTRUSTED) != 0) {
+		status |= GNUTLS_CERT_INVALID;
+		status |= GNUTLS_CERT_SIGNER_NOT_FOUND; /* if the signer is revoked - it is as if it doesn't exist */
+		goto cleanup;
+	}
+
 	/* security modules that provide trust, bundle all certificates (of all purposes)
 	 * together. In software that doesn't specify any purpose assume the default to
 	 * be www-server. */
@@ -1197,40 +1279,117 @@ cleanup:
 }
 #endif
 
+static int
+_gnutls_x509_validate_sign_params(gnutls_pk_algorithm_t pk_algorithm,
+				  ASN1_TYPE cert,
+				  const char *name,
+				  gnutls_x509_spki_st *sig_params)
+{
+	/* The signature parameter validation is only needed for RSA-PSS */
+	if (pk_algorithm == GNUTLS_PK_RSA_PSS) {
+		int result;
+		gnutls_x509_spki_st params;
+
+		result = _gnutls_x509_read_sign_params(cert, name, &params);
+		if (result < 0) {
+			/* If parameters field is absent, no parameter
+			 * validation is needed */
+			if (result != GNUTLS_E_ASN1_ELEMENT_NOT_FOUND &&
+			    result != GNUTLS_E_ASN1_VALUE_NOT_FOUND) {
+				gnutls_assert();
+				return result;
+			}
+		} else {
+			/* Check if the underlying hash algorithms are same.  */
+			if (sig_params->rsa_pss_dig != params.rsa_pss_dig) {
+				gnutls_assert();
+				return GNUTLS_E_CONSTRAINT_ERROR;
+			}
+
+			/* The salt length used to generate the
+			 * signature must be equal to or larger than
+			 * the one in the key parameter. */
+			if (sig_params->salt_size < params.salt_size) {
+				gnutls_assert();
+				return GNUTLS_E_CONSTRAINT_ERROR;
+			}
+		}
+	}
+	return 0;
+}
+
 /* verifies if the certificate is properly signed.
  * returns GNUTLS_E_PK_VERIFY_SIG_FAILED on failure and 1 on success.
  * 
  * 'data' is the signed data
  * 'signature' is the signature!
  */
-int
-_gnutls_x509_verify_data(const mac_entry_st * me,
+static int
+_gnutls_x509_verify_data(gnutls_sign_algorithm_t sign,
 			 const gnutls_datum_t * data,
 			 const gnutls_datum_t * signature,
-			 gnutls_x509_crt_t issuer)
+			 gnutls_x509_crt_t cert,
+			 gnutls_x509_crt_t issuer,
+			 unsigned vflags)
 {
-	gnutls_pk_params_st issuer_params;
+	gnutls_pk_params_st params;
+	gnutls_pk_algorithm_t issuer_pk;
 	int ret;
+	gnutls_x509_spki_st sign_params;
+	const gnutls_sign_entry_st *se;
 
 	/* Read the MPI parameters from the issuer's certificate.
 	 */
-	ret = _gnutls_x509_crt_get_mpis(issuer, &issuer_params);
+	ret = _gnutls_x509_crt_get_mpis(issuer, &params);
 	if (ret < 0) {
 		gnutls_assert();
 		return ret;
 	}
 
-	ret =
-	    pubkey_verify_data(gnutls_x509_crt_get_pk_algorithm
-			       (issuer, NULL), me, data, signature,
-			       &issuer_params);
+	issuer_pk = gnutls_x509_crt_get_pk_algorithm(issuer, NULL);
+
+	se = _gnutls_sign_to_entry(sign);
+	if (se == NULL)
+		return gnutls_assert_val(GNUTLS_E_UNSUPPORTED_SIGNATURE_ALGORITHM);
+
+	if (cert != NULL) {
+		ret = _gnutls_x509_read_sign_params(cert->cert,
+						    "signatureAlgorithm",
+						    &sign_params);
+		if (ret < 0) {
+			gnutls_assert();
+			goto cleanup;
+		}
+
+		ret = _gnutls_x509_validate_sign_params(issuer_pk,
+							issuer->cert,
+							"tbsCertificate."
+							"subjectPublicKeyInfo."
+							"algorithm",
+							&sign_params);
+		if (ret < 0) {
+			gnutls_assert();
+			goto cleanup;
+		}
+	} else {
+		memcpy(&sign_params, &params.spki,
+		       sizeof(gnutls_x509_spki_st));
+
+		sign_params.pk = se->pk;
+		if (sign_params.pk == GNUTLS_PK_RSA_PSS)
+			sign_params.rsa_pss_dig = se->hash;
+	}
+
+	ret = pubkey_verify_data(se, hash_to_entry(se->hash), data, signature, &params,
+				 &sign_params, vflags);
 	if (ret < 0) {
 		gnutls_assert();
 	}
 
+ cleanup:
 	/* release all allocated MPIs
 	 */
-	gnutls_pk_params_release(&issuer_params);
+	gnutls_pk_params_release(&params);
 
 	return ret;
 }
@@ -1401,7 +1560,7 @@ gnutls_x509_crl_verify(gnutls_x509_crl_t crl,
 	gnutls_datum_t crl_signed_data = { NULL, 0 };
 	gnutls_datum_t crl_signature = { NULL, 0 };
 	gnutls_x509_crt_t issuer = NULL;
-	int result, hash_algo;
+	int result, sigalg;
 	time_t now = gnutls_time(0);
 	unsigned int usage;
 
@@ -1431,18 +1590,16 @@ gnutls_x509_crl_verify(gnutls_x509_crl_t crl,
 		goto cleanup;
 	}
 
-	result =
+	sigalg =
 	    _gnutls_x509_get_signature_algorithm(crl->crl,
-						 "signatureAlgorithm.algorithm");
-	if (result < 0) {
+						 "signatureAlgorithm");
+	if (sigalg < 0) {
 		gnutls_assert();
 		if (verify)
 			*verify |= GNUTLS_CERT_INVALID;
 		goto cleanup;
 	}
 
-	hash_algo = gnutls_sign_get_hash_algorithm(result);
-
 	/* issuer is not in trusted certificate
 	 * authorities.
 	 */
@@ -1480,15 +1637,20 @@ gnutls_x509_crl_verify(gnutls_x509_crl_t crl,
 		}
 
 		result =
-		    _gnutls_x509_verify_data(mac_to_entry(hash_algo),
+		    _gnutls_x509_verify_data(sigalg,
 					     &crl_signed_data, &crl_signature,
-					     issuer);
+					     NULL,
+					     issuer, flags);
 		if (result == GNUTLS_E_PK_SIG_VERIFY_FAILED) {
 			gnutls_assert();
 			/* error. ignore it */
 			if (verify)
 				*verify |= GNUTLS_CERT_SIGNATURE_FAILURE;
 			result = 0;
+		} else if (result == GNUTLS_E_CONSTRAINT_ERROR) {
+			if (verify)
+				*verify |= GNUTLS_CERT_SIGNER_CONSTRAINTS_FAILURE;
+			result = 0;
 		} else if (result < 0) {
 			gnutls_assert();
 			if (verify)
@@ -1500,8 +1662,6 @@ gnutls_x509_crl_verify(gnutls_x509_crl_t crl,
 	}
 
 	{
-		int sigalg;
-
 		sigalg = gnutls_x509_crl_get_signature_algorithm(crl);
 
 		if (((sigalg == GNUTLS_SIGN_RSA_MD2) &&
