@@ -35,6 +35,7 @@
 #include <x509_b64.h>
 #include "x509_int.h"
 #include <libtasn1.h>
+#include <pk.h>
 
 static void disable_optional_stuff(gnutls_x509_crt_t cert);
 
@@ -188,7 +189,7 @@ gnutls_x509_crt_set_version(gnutls_x509_crt_t crt, unsigned int version)
 	int result;
 	unsigned char null = version;
 
-	if (crt == NULL) {
+	if (crt == NULL || version == 0 || version >= 0x80) {
 		gnutls_assert();
 		return GNUTLS_E_INVALID_REQUEST;
 	}
@@ -238,7 +239,6 @@ gnutls_x509_crt_set_key(gnutls_x509_crt_t crt, gnutls_x509_privkey_t key)
 
 	result = _gnutls_x509_encode_and_copy_PKI_params(crt->cert,
 							 "tbsCertificate.subjectPublicKeyInfo",
-							 key->pk_algorithm,
 							 &key->params);
 
 	if (result < 0) {
@@ -562,6 +562,48 @@ gnutls_x509_crt_set_key_usage(gnutls_x509_crt_t crt, unsigned int usage)
 	return 0;
 }
 
+/**
+ * gnutls_x509_crt_set_inhibit_anypolicy:
+ * @crt: a certificate of type #gnutls_x509_crt_t
+ * @skipcerts: number of certificates after which anypolicy is no longer acceptable.
+ *
+ * This function will set the Inhibit anyPolicy certificate extension.
+ *
+ * Returns: On success, %GNUTLS_E_SUCCESS (0) is returned, otherwise a
+ *   negative error value.
+ **/
+int
+gnutls_x509_crt_set_inhibit_anypolicy(gnutls_x509_crt_t crt, unsigned int skipcerts)
+{
+	int ret;
+	gnutls_datum_t der_data;
+
+	if (crt == NULL) {
+		gnutls_assert();
+		return GNUTLS_E_INVALID_REQUEST;
+	}
+
+	/* generate the extension.
+	 */
+	ret =
+	    gnutls_x509_ext_export_inhibit_anypolicy(skipcerts, &der_data);
+	if (ret < 0) {
+		gnutls_assert();
+		return ret;
+	}
+
+	ret =
+	    _gnutls_x509_crt_set_extension(crt, "2.5.29.54", &der_data, 1);
+	_gnutls_free_datum(&der_data);
+
+	if (ret < 0) {
+		gnutls_assert();
+		return ret;
+	}
+
+	return 0;
+}
+
 /**
  * gnutls_x509_crt_set_subject_alternative_name:
  * @crt: a certificate of type #gnutls_x509_crt_t
@@ -1204,6 +1246,10 @@ gnutls_x509_crt_set_expiration_time(gnutls_x509_crt_t cert,
  * with the X.509/PKIX specifications the provided @serial should be 
  * a big-endian positive number (i.e. it's leftmost bit should be zero).
  *
+ * The size of the serial is restricted to 20 bytes maximum by RFC5280.
+ * This function allows writing more than 20 bytes but the generated
+ * certificates in that case may be rejected by other implementations.
+ *
  * Returns: On success, %GNUTLS_E_SUCCESS (0) is returned, otherwise a
  *   negative error value.
  **/
@@ -1212,12 +1258,28 @@ gnutls_x509_crt_set_serial(gnutls_x509_crt_t cert, const void *serial,
 			   size_t serial_size)
 {
 	int ret;
+	unsigned all_zero, i;
+	const unsigned char *pserial = serial;
 
 	if (cert == NULL) {
 		gnutls_assert();
 		return GNUTLS_E_INVALID_REQUEST;
 	}
 
+	/* check for non-zero serial */
+	all_zero = 1;
+	for (i=0;i<serial_size;i++) {
+		if (pserial[i] != 0) {
+			all_zero = 0;
+			break;
+		}
+	}
+
+	if (all_zero) {
+		_gnutls_debug_log("error: certificate serial is zero\n");
+		return GNUTLS_E_INVALID_REQUEST;
+	}
+
 	MODIFIED(cert);
 
 	ret =
@@ -1329,16 +1391,16 @@ static void disable_optional_stuff(gnutls_x509_crt_t cert)
 	}
 
 	if (remove_issuer_unique_id)
-		asn1_write_value(cert->cert, "tbsCertificate.issuerUniqueID", NULL,
+		(void)asn1_write_value(cert->cert, "tbsCertificate.issuerUniqueID", NULL,
 				 0);
 
 	if (remove_subject_unique_id)
-		asn1_write_value(cert->cert, "tbsCertificate.subjectUniqueID",
+		(void)asn1_write_value(cert->cert, "tbsCertificate.subjectUniqueID",
 				 NULL, 0);
 
 	if (cert->use_extensions == 0) {
 		_gnutls_debug_log("Disabling X.509 extensions.\n");
-		asn1_write_value(cert->cert, "tbsCertificate.extensions",
+		(void)asn1_write_value(cert->cert, "tbsCertificate.extensions",
 				 NULL, 0);
 	}
 
@@ -1746,8 +1808,14 @@ gnutls_x509_crt_privkey_sign(gnutls_x509_crt_t crt,
 	 */
 	disable_optional_stuff(crt);
 
+	result = _gnutls_check_cert_sanity(crt);
+	if (result < 0) {
+		gnutls_assert();
+		return result;
+	}
+
 	result = _gnutls_x509_pkix_sign(crt->cert, "tbsCertificate",
-					dig, issuer, issuer_key);
+					dig, flags, issuer, issuer_key);
 	if (result < 0) {
 		gnutls_assert();
 		return result;
@@ -1921,3 +1989,118 @@ gnutls_x509_crt_set_policy(gnutls_x509_crt_t crt,
 	return ret;
 }
 
+/**
+ * gnutls_x509_crt_set_spki:
+ * @crt: a certificate of type #gnutls_x509_crt_t
+ * @spki: a SubjectPublicKeyInfo structure of type #gnutls_x509_spki_t
+ * @flags: must be zero
+ *
+ * This function will set the certificate's subject public key
+ * information explicitly. This is intended to be used in the cases
+ * where a single public key (e.g., RSA) can be used for multiple
+ * signature algorithms (RSA PKCS1-1.5, and RSA-PSS).
+ *
+ * To export the public key (i.e., the SubjectPublicKeyInfo part), check
+ * gnutls_pubkey_import_x509().
+ *
+ * Returns: On success, %GNUTLS_E_SUCCESS (0) is returned, otherwise a
+ *   negative error value.
+ *
+ * Since: 3.6.0
+ **/
+int
+gnutls_x509_crt_set_spki(gnutls_x509_crt_t crt,
+			 const gnutls_x509_spki_t spki,
+			 unsigned int flags)
+{
+	int ret;
+	gnutls_pk_algorithm_t crt_pk;
+	gnutls_x509_spki_st tpki;
+	gnutls_pk_params_st params;
+	unsigned bits;
+
+	if (crt == NULL) {
+		gnutls_assert();
+		return GNUTLS_E_INVALID_REQUEST;
+	}
+
+	ret = _gnutls_x509_crt_get_mpis(crt, &params);
+	if (ret < 0) {
+		gnutls_assert();
+		return ret;
+	}
+
+	bits = pubkey_to_bits(&params);
+	crt_pk = params.algo;
+
+	if (!_gnutls_pk_are_compat(crt_pk, spki->pk)) {
+		ret = gnutls_assert_val(GNUTLS_E_INVALID_REQUEST);
+		goto cleanup;
+	}
+
+	if (spki->pk != GNUTLS_PK_RSA_PSS) {
+		if (crt_pk == spki->pk) {
+			ret = 0;
+			goto cleanup;
+		}
+
+		gnutls_assert();
+		ret = GNUTLS_E_INVALID_REQUEST;
+		goto cleanup;
+	}
+
+	if (crt_pk == GNUTLS_PK_RSA) {
+		const mac_entry_st *me;
+
+		me = hash_to_entry(spki->rsa_pss_dig);
+		if (unlikely(me == NULL)) {
+			gnutls_assert();
+			ret = GNUTLS_E_INVALID_REQUEST;
+			goto cleanup;
+		}
+
+		memset(&tpki, 0, sizeof(gnutls_x509_spki_st));
+		tpki.pk = spki->pk;
+		tpki.rsa_pss_dig = spki->rsa_pss_dig;
+
+		/* If salt size is zero, find the optimal salt size. */
+		if (spki->salt_size == 0) {
+			tpki.salt_size =
+			    _gnutls_find_rsa_pss_salt_size(bits, me,
+							   spki->salt_size);
+		} else
+			tpki.salt_size = spki->salt_size;
+	} else if (crt_pk == GNUTLS_PK_RSA_PSS) {
+		ret = _gnutls_x509_crt_read_spki_params(crt, &tpki);
+		if (ret < 0) {
+			gnutls_assert();
+			goto cleanup;
+		}
+
+		tpki.salt_size = spki->salt_size;
+		tpki.rsa_pss_dig = spki->rsa_pss_dig;
+	}
+
+	memcpy(&params.spki, &tpki, sizeof(tpki));
+	ret = _gnutls_x509_check_pubkey_params(&params);
+	if (ret < 0) {
+		gnutls_assert();
+		goto cleanup;
+	}
+
+	MODIFIED(crt);
+
+	ret = _gnutls_x509_write_spki_params(crt->cert,
+						"tbsCertificate."
+						"subjectPublicKeyInfo.algorithm",
+						&tpki);
+	if (ret < 0) {
+		gnutls_assert();
+		goto cleanup;
+	}
+
+	ret = 0;
+ cleanup:
+	gnutls_pk_params_release(&params);
+	return ret;
+}
