@@ -1,6 +1,6 @@
 /*
  * Copyright (C) 2003-2016 Free Software Foundation, Inc.
- * Copyright (C) 2015-2016 Red Hat, Inc.
+ * Copyright (C) 2015-2017 Red Hat, Inc.
  *
  * This file is part of GnuTLS.
  *
@@ -63,7 +63,10 @@ static unsigned long file_size(FILE *fp)
 		return 0;
 
 	size = ftell(fp);
-	fseek(fp, cur, SEEK_SET);
+	if (fseek(fp, cur, SEEK_SET) == -1) {
+		fprintf(stderr, "Error reading file size\n");
+		app_exit(1);
+	}
 	return size;
 }
 
@@ -82,7 +85,7 @@ void fix_lbuffer(unsigned long size)
 
 	if (lbuffer == NULL) {
 		fprintf(stderr, "memory error");
-		exit(1);
+		app_exit(1);
 	}
 }
 
@@ -118,7 +121,7 @@ gnutls_datum_t *load_secret_key(int mand, common_info_st * info)
 	if (info->secret_key == NULL) {
 		if (mand) {
 			fprintf(stderr, "missing --secret-key\n");
-			exit(1);
+			app_exit(1);
 		} else
 			return NULL;
 	}
@@ -129,7 +132,7 @@ gnutls_datum_t *load_secret_key(int mand, common_info_st * info)
 	ret = gnutls_hex_decode(&hex_key, raw_key, &raw_key_size);
 	if (ret < 0) {
 		fprintf(stderr, "hex_decode: %s\n", gnutls_strerror(ret));
-		exit(1);
+		app_exit(1);
 	}
 
 	key.data = (void *) raw_key;
@@ -173,7 +176,7 @@ static gnutls_privkey_t _load_privkey(gnutls_datum_t * dat,
 	ret = gnutls_privkey_init(&key);
 	if (ret < 0) {
 		fprintf(stderr, "privkey_init: %s\n", gnutls_strerror(ret));
-		exit(1);
+		app_exit(1);
 	}
 
 	ret =
@@ -191,13 +194,13 @@ static gnutls_privkey_t _load_privkey(gnutls_datum_t * dat,
 		fprintf(stderr,
 			"import error: could not find a valid PEM header; "
 			"check if your key is PKCS #12 encoded\n");
-		exit(1);
+		app_exit(1);
 	}
 
 	if (ret < 0) {
-		fprintf(stderr, "importing --load-privkey: %s: %s\n",
+		fprintf(stderr, "error loading file at --load-privkey: %s: %s\n",
 			info->privkey, gnutls_strerror(ret));
-		exit(1);
+		app_exit(1);
 	}
 
 	return key;
@@ -211,14 +214,14 @@ static gnutls_privkey_t _load_url_privkey(const char *url)
 	ret = gnutls_privkey_init(&key);
 	if (ret < 0) {
 		fprintf(stderr, "privkey_init: %s\n", gnutls_strerror(ret));
-		exit(1);
+		app_exit(1);
 	}
 
 	ret = gnutls_privkey_import_url(key, url, 0);
 	if (ret < 0) {
-		fprintf(stderr, "importing key: %s: %s\n",
+		fprintf(stderr, "error importing key at %s: %s\n",
 			url, gnutls_strerror(ret));
-		exit(1);
+		app_exit(1);
 	}
 
 	return key;
@@ -234,14 +237,14 @@ static gnutls_pubkey_t _load_url_pubkey(const char *url)
 	if (ret < 0) {
 		fprintf(stderr, "Error in %s:%d: %s\n", __func__, __LINE__,
 			gnutls_strerror(ret));
-		exit(1);
+		app_exit(1);
 	}
 
 	ret = gnutls_pubkey_import_url(pubkey, url, obj_flags);
 	if (ret < 0) {
 		fprintf(stderr, "Error in %s:%d: %s: %s\n", __func__,
 			__LINE__, gnutls_strerror(ret), url);
-		exit(1);
+		app_exit(1);
 	}
 
 	return pubkey;
@@ -261,7 +264,7 @@ gnutls_privkey_t load_private_key(int mand, common_info_st * info)
 
 	if (info->privkey == NULL) {
 		fprintf(stderr, "missing --load-privkey\n");
-		exit(1);
+		app_exit(1);
 	}
 
 	if (gnutls_url_is_supported(info->privkey) != 0)
@@ -271,9 +274,9 @@ gnutls_privkey_t load_private_key(int mand, common_info_st * info)
 	dat.size = size;
 
 	if (!dat.data) {
-		fprintf(stderr, "error reading --load-privkey: %s\n",
+		fprintf(stderr, "error reading file at --load-privkey: %s\n",
 			info->privkey);
-		exit(1);
+		app_exit(1);
 	}
 
 	key = _load_privkey(&dat, info);
@@ -301,22 +304,22 @@ load_x509_private_key(int mand, common_info_st * info)
 
 	if (info->privkey == NULL) {
 		fprintf(stderr, "missing --load-privkey\n");
-		exit(1);
+		app_exit(1);
 	}
 
 	ret = gnutls_x509_privkey_init(&key);
 	if (ret < 0) {
 		fprintf(stderr, "privkey_init: %s\n", gnutls_strerror(ret));
-		exit(1);
+		app_exit(1);
 	}
 
 	dat.data = (void *) read_binary_file(info->privkey, &size);
 	dat.size = size;
 
 	if (!dat.data) {
-		fprintf(stderr, "error reading --load-privkey: %s\n",
+		fprintf(stderr, "error reading file at --load-privkey: %s\n",
 			info->privkey);
-		exit(1);
+		app_exit(1);
 	}
 
 	if (info->pkcs8) {
@@ -346,13 +349,13 @@ load_x509_private_key(int mand, common_info_st * info)
 		fprintf(stderr,
 			"import error: could not find a valid PEM header; "
 			"check if your key is PEM encoded\n");
-		exit(1);
+		app_exit(1);
 	}
 
 	if (ret < 0) {
-		fprintf(stderr, "importing --load-privkey: %s: %s\n",
+		fprintf(stderr, "error importing private key: %s: %s\n",
 			info->privkey, gnutls_strerror(ret));
-		exit(1);
+		app_exit(1);
 	}
 
 	return key;
@@ -392,6 +395,7 @@ gnutls_x509_crt_t *load_cert_list(int mand, size_t * crt_size,
 	gnutls_datum_t dat;
 	unsigned size;
 	unsigned int crt_max;
+	unsigned flags = 0;
 
 	*crt_size = 0;
 	if (info->verbose)
@@ -400,7 +404,7 @@ gnutls_x509_crt_t *load_cert_list(int mand, size_t * crt_size,
 	if (info->cert == NULL) {
 		if (mand) {
 			fprintf(stderr, "missing --load-certificate\n");
-			exit(1);
+			app_exit(1);
 		} else
 			return NULL;
 	}
@@ -408,7 +412,7 @@ gnutls_x509_crt_t *load_cert_list(int mand, size_t * crt_size,
 	fd = fopen(info->cert, "r");
 	if (fd == NULL) {
 		fprintf(stderr, "Could not open %s\n", info->cert);
-		exit(1);
+		app_exit(1);
 	}
 
 	fix_lbuffer(file_size(fd));
@@ -421,10 +425,13 @@ gnutls_x509_crt_t *load_cert_list(int mand, size_t * crt_size,
 	dat.data = (void *) lbuffer;
 	dat.size = size;
 
-	ret = gnutls_x509_crt_list_import2(&crt, &crt_max, &dat, GNUTLS_X509_FMT_PEM, 0);
+	if (info->sort_chain)
+		flags |= GNUTLS_X509_CRT_LIST_SORT;
+
+	ret = gnutls_x509_crt_list_import2(&crt, &crt_max, &dat, GNUTLS_X509_FMT_PEM, flags);
 	if (ret < 0) {
 		fprintf(stderr, "Error loading certificates: %s\n", gnutls_strerror(ret));
-		exit(1);
+		app_exit(1);
 	}
 
 	*crt_size = crt_max;
@@ -455,7 +462,7 @@ gnutls_x509_crl_t *load_crl_list(int mand, size_t * crl_size,
 	if (info->crl == NULL) {
 		if (mand) {
 			fprintf(stderr, "missing --load-crl\n");
-			exit(1);
+			app_exit(1);
 		} else
 			return NULL;
 	}
@@ -463,7 +470,7 @@ gnutls_x509_crl_t *load_crl_list(int mand, size_t * crl_size,
 	fd = fopen(info->crl, "r");
 	if (fd == NULL) {
 		fprintf(stderr, "Could not open %s\n", info->crl);
-		exit(1);
+		app_exit(1);
 	}
 
 	fix_lbuffer(file_size(fd));
@@ -477,9 +484,12 @@ gnutls_x509_crl_t *load_crl_list(int mand, size_t * crl_size,
 	dat.size = size;
 
 	ret = gnutls_x509_crl_list_import2(&crl, &crl_max, &dat, GNUTLS_X509_FMT_PEM, 0);
+	if (ret == GNUTLS_E_BASE64_DECODING_ERROR) {
+		ret = gnutls_x509_crl_list_import2(&crl, &crl_max, &dat, GNUTLS_X509_FMT_DER, 0);
+	}
 	if (ret < 0) {
 		fprintf(stderr, "Error loading CRLs: %s\n", gnutls_strerror(ret));
-		exit(1);
+		app_exit(1);
 	}
 
 	*crl_size = crl_max;
@@ -506,30 +516,30 @@ gnutls_x509_crq_t load_request(common_info_st * info)
 	ret = gnutls_x509_crq_init(&crq);
 	if (ret < 0) {
 		fprintf(stderr, "crq_init: %s\n", gnutls_strerror(ret));
-		exit(1);
+		app_exit(1);
 	}
 
 	dat.data = (void *) read_binary_file(info->request, &size);
 	dat.size = size;
 
 	if (!dat.data) {
-		fprintf(stderr, "error reading --load-request: %s\n",
+		fprintf(stderr, "error reading file at --load-request: %s\n",
 			info->request);
-		exit(1);
+		app_exit(1);
 	}
 
 	ret = gnutls_x509_crq_import(crq, &dat, info->incert_format);
 	if (ret == GNUTLS_E_BASE64_UNEXPECTED_HEADER_ERROR) {
 		fprintf(stderr,
 			"import error: could not find a valid PEM header\n");
-		exit(1);
+		app_exit(1);
 	}
 
 	free(dat.data);
 	if (ret < 0) {
-		fprintf(stderr, "importing --load-request: %s: %s\n",
+		fprintf(stderr, "error importing certificate request: %s: %s\n",
 			info->request, gnutls_strerror(ret));
-		exit(1);
+		app_exit(1);
 	}
 	return crq;
 }
@@ -544,7 +554,7 @@ gnutls_privkey_t load_ca_private_key(common_info_st * info)
 
 	if (info->ca_privkey == NULL) {
 		fprintf(stderr, "missing --load-ca-privkey\n");
-		exit(1);
+		app_exit(1);
 	}
 
 	if (gnutls_url_is_supported(info->ca_privkey) != 0)
@@ -554,9 +564,9 @@ gnutls_privkey_t load_ca_private_key(common_info_st * info)
 	dat.size = size;
 
 	if (!dat.data) {
-		fprintf(stderr, "error reading --load-ca-privkey: %s\n",
+		fprintf(stderr, "error reading file at --load-ca-privkey: %s\n",
 			info->ca_privkey);
-		exit(1);
+		app_exit(1);
 	}
 
 	key = _load_privkey(&dat, info);
@@ -581,30 +591,40 @@ gnutls_x509_crt_t load_ca_cert(unsigned mand, common_info_st * info)
 
 	if (info->ca == NULL) {
 		fprintf(stderr, "missing --load-ca-certificate\n");
-		exit(1);
+		app_exit(1);
 	}
 
 	ret = gnutls_x509_crt_init(&crt);
 	if (ret < 0) {
 		fprintf(stderr, "crt_init: %s\n", gnutls_strerror(ret));
-		exit(1);
+		app_exit(1);
+	}
+
+	if (gnutls_url_is_supported(info->ca) != 0) {
+		ret = gnutls_x509_crt_import_url(crt, info->ca, 0);
+		if (ret < 0) {
+			fprintf(stderr, "error importing CA certificate: %s: %s\n",
+				info->ca, gnutls_strerror(ret));
+			app_exit(1);
+		}
+		return crt;
 	}
 
 	dat.data = (void *) read_binary_file(info->ca, &size);
 	dat.size = size;
 
 	if (!dat.data) {
-		fprintf(stderr, "error reading --load-ca-certificate: %s\n",
+		fprintf(stderr, "error reading file at --load-ca-certificate: %s\n",
 			info->ca);
-		exit(1);
+		app_exit(1);
 	}
 
 	ret = gnutls_x509_crt_import(crt, &dat, info->incert_format);
 	free(dat.data);
 	if (ret < 0) {
-		fprintf(stderr, "importing --load-ca-certificate: %s: %s\n",
+		fprintf(stderr, "error importing CA certificate: %s: %s\n",
 			info->ca, gnutls_strerror(ret));
-		exit(1);
+		app_exit(1);
 	}
 
 	return crt;
@@ -625,7 +645,7 @@ gnutls_pubkey_t load_pubkey(int mand, common_info_st * info)
 
 	if (info->pubkey == NULL) {
 		fprintf(stderr, "missing --load-pubkey\n");
-		exit(1);
+		app_exit(1);
 	}
 
 	if (gnutls_url_is_supported(info->pubkey) != 0)
@@ -634,15 +654,15 @@ gnutls_pubkey_t load_pubkey(int mand, common_info_st * info)
 	ret = gnutls_pubkey_init(&key);
 	if (ret < 0) {
 		fprintf(stderr, "privkey_init: %s\n", gnutls_strerror(ret));
-		exit(1);
+		app_exit(1);
 	}
 
 	dat.data = (void *) read_binary_file(info->pubkey, &size);
 	dat.size = size;
 
 	if (!dat.data) {
-		fprintf(stderr, "error reading --load-pubkey: %s\n", info->pubkey);
-		exit(1);
+		fprintf(stderr, "error reading file at --load-pubkey: %s\n", info->pubkey);
+		app_exit(1);
 	}
 
 	ret = gnutls_pubkey_import(key, &dat, info->incert_format);
@@ -652,12 +672,12 @@ gnutls_pubkey_t load_pubkey(int mand, common_info_st * info)
 			fprintf(stderr,
 				"import error: could not find a valid PEM header; "
 				"check if your key has the PUBLIC KEY header\n");
-			exit(1);
+			app_exit(1);
 		}
 	} else if (ret < 0) {
-		fprintf(stderr, "importing --load-pubkey: %s: %s\n",
+		fprintf(stderr, "importing public key: %s: %s\n",
 			info->pubkey, gnutls_strerror(ret));
-		exit(1);
+		app_exit(1);
 	}
 
 	free(dat.data);
@@ -675,7 +695,7 @@ gnutls_pubkey_t load_public_key_or_import(int mand,
 	if (ret < 0) {
 		fprintf(stderr, "gnutls_pubkey_init: %s\n",
 			gnutls_strerror(ret));
-		exit(1);
+		app_exit(1);
 	}
 
 	if (!privkey || (ret = gnutls_pubkey_import_privkey(pubkey, privkey, 0, 0)) < 0) {	/* could not get (e.g. on PKCS #11 */
@@ -683,7 +703,7 @@ gnutls_pubkey_t load_public_key_or_import(int mand,
 		pubkey = load_pubkey(0, info);
 		if (pubkey == NULL && mand) {
 			fprintf(stderr, "Could not determine the public key for the operation.\nYou must specify --load-privkey or --load-pubkey if missing.\n");
-			exit(1);
+			app_exit(1);
 		}
 	}
 
@@ -747,7 +767,7 @@ gnutls_sec_param_t str_to_sec_param(const char *str)
 	} else {
 		fprintf(stderr, "Unknown security parameter string: %s\n",
 			str);
-		exit(1);
+		app_exit(1);
 	}
 
 }
@@ -856,7 +876,7 @@ const gnutls_ecc_curve_t *list, *p;
 		       gnutls_ecc_curve_get_name(*p));
 		p++;
 	}
-	exit(1);
+	app_exit(1);
 }
 
 void
@@ -871,14 +891,20 @@ print_ecc_pkey(FILE * outfile, gnutls_ecc_curve_t curve,
 		fprintf(outfile, "curve:\t%s\n",
 			gnutls_ecc_curve_get_name(curve));
 
-	if (k) {
+	if (k && k->data) {
 		print_head(outfile, "private key", k->size, cprint);
 		print_hex_datum(outfile, k, cprint);
 	}
-	print_head(outfile, "x", x->size, cprint);
-	print_hex_datum(outfile, x, cprint);
-	print_head(outfile, "y", y->size, cprint);
-	print_hex_datum(outfile, y, cprint);
+
+	if (x && x->data) {
+		print_head(outfile, "x", x->size, cprint);
+		print_hex_datum(outfile, x, cprint);
+	}
+
+	if (y && y->data) {
+		print_head(outfile, "y", y->size, cprint);
+		print_hex_datum(outfile, y, cprint);
+	}
 }
 
 
@@ -924,7 +950,7 @@ void _pubkey_info(FILE * outfile,
 	if (ret < 0) {
 		fprintf(stderr, "pubkey_print error: %s\n",
 			gnutls_strerror(ret));
-		exit(1);
+		app_exit(1);
 	}
 
 	fprintf(outfile, "%s\n", data.data);
@@ -936,7 +962,7 @@ void _pubkey_info(FILE * outfile,
 				 &size);
 	if (ret < 0) {
 		fprintf(stderr, "export error: %s\n", gnutls_strerror(ret));
-		exit(1);
+		app_exit(1);
 	}
 
 	fprintf(outfile, "\n%s\n", lbuffer);
@@ -1000,12 +1026,17 @@ void dh_info(FILE * infile, FILE * outfile, common_info_st * ci)
 
 	if (gnutls_dh_params_init(&dh_params) < 0) {
 		fprintf(stderr, "Error in dh parameter initialization\n");
-		exit(1);
+		app_exit(1);
 	}
 
 	params.data = (void *) fread_file(infile, &size);
 	params.size = size;
 
+	if (params.data == NULL) {
+		fprintf(stderr, "Could not read input\n");
+		app_exit(1);
+	}
+
 	ret =
 	    gnutls_dh_params_import_pkcs3(dh_params, &params,
 					  ci->incert_format);
@@ -1015,7 +1046,7 @@ void dh_info(FILE * infile, FILE * outfile, common_info_st * ci)
 		if (ret2 < 0) {
 			fprintf(stderr, "Error parsing dh params: %s\n",
 				gnutls_strerror(ret));
-			exit(1);
+			app_exit(1);
 		}
 	}
 
@@ -1023,7 +1054,7 @@ void dh_info(FILE * infile, FILE * outfile, common_info_st * ci)
 	if (ret < 0) {
 		fprintf(stderr, "Error exporting parameters: %s\n",
 			gnutls_strerror(ret));
-		exit(1);
+		app_exit(1);
 	}
 
 	if (ci->outcert_format == GNUTLS_X509_FMT_PEM)
@@ -1079,7 +1110,7 @@ int cipher_to_flags(const char *cipher)
 	}
 
 	fprintf(stderr, "unknown cipher %s\n", cipher);
-	exit(1);
+	app_exit(1);
 }
 
 static void privkey_info_int(FILE *outfile, common_info_st * cinfo,
@@ -1089,24 +1120,57 @@ static void privkey_info_int(FILE *outfile, common_info_st * cinfo,
 	unsigned int bits = 0;
 	size_t size;
 	const char *cprint;
+	gnutls_x509_spki_t spki;
+	gnutls_digest_algorithm_t dig;
 
 	/* Public key algorithm
 	 */
 	fprintf(outfile, "Public Key Info:\n");
-	ret = gnutls_x509_privkey_get_pk_algorithm2(key, &bits);
+	ret = gnutls_x509_spki_init(&spki);
+	if (ret < 0) {
+		fprintf(stderr, "spki_init: %s\n", gnutls_strerror(ret));
+		return;
+	}
+
 	fprintf(outfile, "\tPublic Key Algorithm: ");
 
-	key_type = ret;
+	key_type = gnutls_x509_privkey_get_pk_algorithm2(key, &bits);
 
 	cprint = gnutls_pk_algorithm_get_name(key_type);
 	fprintf(outfile, "%s\n", cprint ? cprint : "Unknown");
+
+	if (key_type == GNUTLS_PK_RSA_PSS) {
+		unsigned int salt_size;
+
+		ret = gnutls_x509_privkey_get_spki(key, spki, 0);
+		if (ret == GNUTLS_E_REQUESTED_DATA_NOT_AVAILABLE)
+			goto spki_skip;
+
+		if (ret < 0) {
+			fprintf(stderr, "spki_get: %s\n", gnutls_strerror(ret));
+			goto spki_skip;
+		}
+
+		ret = gnutls_x509_spki_get_rsa_pss_params(spki, &dig, &salt_size);
+		if (ret < 0) {
+			fprintf(stderr, "spki_get_rsa_pss_params: %s\n",
+				gnutls_strerror(ret));
+		} else {
+			fprintf(outfile, "\t\tHash Algorithm: %s\n",
+				gnutls_digest_get_name(dig));
+			fprintf(outfile, "\t\tSalt Length: %d\n", salt_size);
+		}
+	}
+
+ spki_skip:
+	gnutls_x509_spki_deinit(spki);
 	fprintf(outfile, "\tKey Security Level: %s (%u bits)\n\n",
 		gnutls_sec_param_get_name(gnutls_x509_privkey_sec_param
 					  (key)), bits);
 
 	/* Print the raw public and private keys
 	 */
-	if (key_type == GNUTLS_PK_RSA) {
+	if (GNUTLS_PK_IS_RSA(key_type)) {
 		gnutls_datum_t m, e, d, p, q, u, exp1, exp2;
 
 		ret =
@@ -1150,7 +1214,7 @@ static void privkey_info_int(FILE *outfile, common_info_st * cinfo,
 			gnutls_free(q.data);
 			gnutls_free(g.data);
 		}
-	} else if (key_type == GNUTLS_PK_EC) {
+	} else if (key_type == GNUTLS_PK_ECDSA || key_type == GNUTLS_PK_EDDSA_ED25519) {
 		gnutls_datum_t y, x, k;
 		gnutls_ecc_curve_t curve;
 
@@ -1174,15 +1238,18 @@ static void privkey_info_int(FILE *outfile, common_info_st * cinfo,
 		}
 	}
 
-	fprintf(outfile, "\n");
-
 	size = lbuffer_size;
-	ret = gnutls_x509_privkey_get_seed(key, NULL, lbuffer, &size);
+	ret = gnutls_x509_privkey_get_seed(key, &dig, lbuffer, &size);
 	if (ret >= 0) {
-		fprintf(outfile, "Seed: %s\n",
-			raw_to_string(lbuffer, size));
+		fprintf(outfile, "Validation parameters:\n");
+		fprintf(outfile, "\tHash: %s\n",
+			gnutls_digest_get_name(dig));
+		fprintf(outfile, "\tSeed: %s\n",
+			raw_to_hex(lbuffer, size));
 	}
 
+	fprintf(outfile, "\n");
+
 	size = lbuffer_size;
 	ret =
 	     gnutls_x509_privkey_get_key_id(key, GNUTLS_KEYID_USE_SHA256, lbuffer, &size);
@@ -1217,11 +1284,14 @@ print_private_key(FILE *outfile, common_info_st * cinfo, gnutls_x509_privkey_t k
 	if (!key)
 		return;
 
+	/* Only print private key parameters when an unencrypted
+	 * format is used */
+	if (cinfo->outcert_format == GNUTLS_X509_FMT_PEM)
+		privkey_info_int(outfile, cinfo, key);
+
+	switch_to_pkcs8_when_needed(cinfo, key, gnutls_x509_privkey_get_pk_algorithm(key));
+
 	if (!cinfo->pkcs8) {
-		/* Only print private key parameters when an unencrypted
-		 * format is used */
-		if (cinfo->outcert_format == GNUTLS_X509_FMT_PEM)
-			privkey_info_int(outfile, cinfo, key);
 
 		size = lbuffer_size;
 		ret = gnutls_x509_privkey_export(key, cinfo->outcert_format,
@@ -1229,24 +1299,8 @@ print_private_key(FILE *outfile, common_info_st * cinfo, gnutls_x509_privkey_t k
 		if (ret < 0) {
 			fprintf(stderr, "privkey_export: %s\n",
 				gnutls_strerror(ret));
-			exit(1);
+			app_exit(1);
 		}
-
-		if (cinfo->no_compat == 0 && gnutls_x509_privkey_get_seed(key, NULL, NULL, 0) != GNUTLS_E_INVALID_REQUEST) {
-			gnutls_x509_privkey_set_flags(key, GNUTLS_PRIVKEY_FLAG_EXPORT_COMPAT);
-
-			fwrite(lbuffer, 1, size, outfile);
-
-			size = lbuffer_size;
-			ret = gnutls_x509_privkey_export(key, cinfo->outcert_format,
-						 lbuffer, &size);
-			if (ret < 0) {
-				fprintf(stderr, "privkey_export: %s\n",
-					gnutls_strerror(ret));
-				exit(1);
-			}
-		}
-
 	} else {
 		unsigned int flags = 0;
 		const char *pass;
@@ -1262,7 +1316,7 @@ print_private_key(FILE *outfile, common_info_st * cinfo, gnutls_x509_privkey_t k
 		if (ret < 0) {
 			fprintf(stderr, "privkey_export_pkcs8: %s\n",
 				gnutls_strerror(ret));
-			exit(1);
+			app_exit(1);
 		}
 	}
 
@@ -1293,14 +1347,13 @@ int generate_prime(FILE * outfile, int how, common_info_st * info)
 	if (how != 0) {
 		if (info->provable != 0) {
 			gnutls_x509_privkey_t pkey;
-			unsigned save;
 
 			ret = gnutls_x509_privkey_init(&pkey);
 			if (ret < 0) {
 				fprintf(stderr,
 					"Error initializing key: %s\n",
 					gnutls_strerror(ret));
-				exit(1);
+				app_exit(1);
 			}
 
 			if (info->seed_size > 0) {
@@ -1308,7 +1361,7 @@ int generate_prime(FILE * outfile, int how, common_info_st * info)
 
 				if (info->seed_size < 32) {
 					fprintf(stderr, "For DH parameter generation a 32-byte seed value or larger is expected (have: %d); use -d 2 for more information.\n", (int)info->seed_size);
-					exit(1);
+					app_exit(1);
 				}
 
 				data.type = GNUTLS_KEYGEN_SEED;
@@ -1324,14 +1377,11 @@ int generate_prime(FILE * outfile, int how, common_info_st * info)
 				fprintf(stderr,
 					"Error generating DSA parameters: %s\n",
 					gnutls_strerror(ret));
-				exit(1);
+				app_exit(1);
 			}
 
 			if (info->outcert_format == GNUTLS_X509_FMT_PEM) {
-				save = info->no_compat;
-				info->no_compat = 1;
 				print_private_key(outfile, info, pkey);
-				info->no_compat = save;
 			}
 
 			ret = gnutls_dh_params_import_dsa(dh_params, pkey);
@@ -1339,7 +1389,7 @@ int generate_prime(FILE * outfile, int how, common_info_st * info)
 				fprintf(stderr,
 					"Error importing DSA parameters: %s\n",
 					gnutls_strerror(ret));
-				exit(1);
+				app_exit(1);
 			}
 
 			gnutls_x509_privkey_deinit(pkey);
@@ -1349,7 +1399,7 @@ int generate_prime(FILE * outfile, int how, common_info_st * info)
 				fprintf(stderr,
 					"Error generating parameters: %s\n",
 					gnutls_strerror(ret));
-				exit(1);
+				app_exit(1);
 			}
 		}
 
@@ -1359,12 +1409,12 @@ int generate_prime(FILE * outfile, int how, common_info_st * info)
 		if (ret < 0) {
 			fprintf(stderr, "Error exporting parameters: %s\n",
 				gnutls_strerror(ret));
-			exit(1);
+			app_exit(1);
 		}
 	} else {
 		if (info->provable != 0) {
 			fprintf(stderr, "The DH parameters obtained via this option are not provable\n");
-			exit(1);
+			app_exit(1);
 		}
 #if defined(ENABLE_DHE) || defined(ENABLE_ANON)
 		if (bits <= 2048) {
@@ -1393,7 +1443,7 @@ int generate_prime(FILE * outfile, int how, common_info_st * info)
 		if (ret < 0) {
 			fprintf(stderr, "Error exporting parameters: %s\n",
 				gnutls_strerror(ret));
-			exit(1);
+			app_exit(1);
 		}
 #elif defined(ENABLE_SRP)
 		if (bits <= 1024) {
@@ -1422,12 +1472,12 @@ int generate_prime(FILE * outfile, int how, common_info_st * info)
 		if (ret < 0) {
 			fprintf(stderr, "Error exporting parameters: %s\n",
 				gnutls_strerror(ret));
-			exit(1);
+			app_exit(1);
 		}
 #else
 		fprintf(stderr,
 			"Parameters unavailable as SRP is disabled.\n");
-		exit(1);
+		app_exit(1);
 #endif
 	}
 
@@ -1475,16 +1525,34 @@ void decode_seed(gnutls_datum_t *seed, const char *hex, unsigned hex_size)
 
 	if (seed->data == NULL) {
 		fprintf(stderr, "memory error\n");
-		exit(1);
+		app_exit(1);
 	}
 
 	seed_size = hex_size;
 	ret = gnutls_hex2bin(hex, hex_size, seed->data, &seed_size);
 	if (ret < 0) {
 		fprintf(stderr, "Could not hex decode data: %s\n", gnutls_strerror(ret));
-		exit(1);
+		app_exit(1);
 	}
 	seed->size = seed_size;
 
 	return;
 }
+
+gnutls_pk_algorithm_t figure_key_type(const char *key_type)
+{
+	if (strcasecmp(key_type, "rsa") == 0)
+		return GNUTLS_PK_RSA;
+	else if (strcasecmp(key_type, "rsa-pss") == 0)
+		return GNUTLS_PK_RSA_PSS;
+	else if (strcasecmp(key_type, "ed25519") == 0 || strcasecmp(key_type, "eddsa") == 0)
+		return GNUTLS_PK_EDDSA_ED25519;
+	else if (strcasecmp(key_type, "dsa") == 0)
+		return GNUTLS_PK_DSA;
+	else if (strcasecmp(key_type, "ecdsa") == 0 || strcasecmp(key_type, "ecc") == 0)
+		return GNUTLS_PK_ECDSA;
+	else {
+		fprintf(stderr, "unknown key type: %s\n", key_type);
+		return GNUTLS_PK_UNKNOWN;
+	}
+}
