@@ -1,5 +1,6 @@
 /*
  * Copyright (C) 2003-2012 Free Software Foundation, Inc.
+ * Copyright (C) 2017 Red Hat, Inc.
  *
  * This file is part of GnuTLS.
  *
@@ -23,6 +24,7 @@
 
 #include <gnutls/x509.h>
 #include <stdio.h>
+#include <stdbool.h>
 
 #define TYPE_CRT 1
 #define TYPE_CRQ 2
@@ -54,6 +56,11 @@ typedef struct common_info {
 	int empty_password;
 	unsigned int crq_extensions;
 	unsigned int v1_cert;
+	/* for key generation */
+	unsigned provable;
+
+	unsigned char *seed;
+	unsigned seed_size;
 
 	const char *pin;
 	const char *so_pin;
@@ -62,9 +69,50 @@ typedef struct common_info {
 	unsigned key_usage;
 
 	unsigned int batch;
+	/* when printing PKCS #11 objects, only print urls */
+	unsigned int only_urls;
 	unsigned int verbose;
+
+	unsigned rsa_pss_sign;
+	unsigned sort_chain;
 } common_info_st;
 
+static inline
+void switch_to_pkcs8_when_needed(common_info_st *cinfo, gnutls_x509_privkey_t key, unsigned key_type)
+{
+	if (cinfo->pkcs8)
+		return;
+
+	if ((key_type == GNUTLS_PK_RSA_PSS || key_type == GNUTLS_PK_EDDSA_ED25519)) {
+		if (cinfo->verbose)
+			fprintf(stderr, "Assuming --pkcs8 is given; %s private keys can only be exported in PKCS#8 format\n",
+				gnutls_pk_algorithm_get_name(key_type));
+		cinfo->pkcs8 = 1;
+		if (cinfo->password == NULL)
+			cinfo->password = "";
+	}
+
+	if (gnutls_x509_privkey_get_seed(key, NULL, NULL, 0) != GNUTLS_E_INVALID_REQUEST) {
+		if (cinfo->verbose)
+			fprintf(stderr, "Assuming --pkcs8 is given; provable private keys can only be exported in PKCS#8 format\n");
+		cinfo->pkcs8 = 1;
+		if (cinfo->password == NULL)
+			cinfo->password = "";
+	}
+}
+
+/* this must be provided by the app */
+const char *get_pass(void);
+const char *get_confirmed_pass(bool empty_ok);
+void app_exit(int val)
+#ifdef __GNUC__
+__attribute__ ((noreturn))
+#endif
+;
+int cipher_to_flags(const char *cipher);
+
+void
+print_private_key(FILE *outfile, common_info_st * cinfo, gnutls_x509_privkey_t key);
 gnutls_pubkey_t load_public_key_or_import(int mand,
 					  gnutls_privkey_t privkey,
 					  common_info_st * info);
@@ -120,4 +168,10 @@ extern unsigned long lbuffer_size;
 
 void fix_lbuffer(unsigned long);
 
+void decode_seed(gnutls_datum_t *seed, const char *hex, unsigned hex_size);
+
+#define GNUTLS_PK_IS_RSA(pk) ((pk) == GNUTLS_PK_RSA || (pk) == GNUTLS_PK_RSA_PSS)
+
+gnutls_pk_algorithm_t figure_key_type(const char *key_type);
+
 #endif
