@@ -1,5 +1,6 @@
 /*
  * Copyright (C) 2000-2012 Free Software Foundation, Inc.
+ * Copyright (C) 2017 Red Hat, Inc.
  *
  * This file is part of GnuTLS.
  *
@@ -142,6 +143,8 @@ static const TLS_TEST tls_tests[] = {
 #endif
 	{"for ephemeral Diffie-Hellman support", test_dhe, "yes", "no",
 	 "dunno"},
+	{"for RFC7919 Diffie-Hellman support", test_rfc7919, "yes", "no",
+	 "dunno"},
 	{"ephemeral Diffie-Hellman group info", test_dhe_group, NULL, "N/A",
 	 "N/A"},
 	{"for ephemeral EC Diffie-Hellman support", test_ecdhe, "yes",
@@ -285,7 +288,7 @@ int main(int argc, char **argv)
 			break;
 		}
 
-		socket_open(&hd, hostname, portname, app_proto, SOCKET_FLAG_STARTTLS, NULL, NULL);
+		socket_open(&hd, hostname, portname, app_proto, SOCKET_FLAG_STARTTLS|SOCKET_FLAG_RAW, NULL, NULL);
 		hd.verbose = verbose;
 
 		do {
@@ -296,7 +299,7 @@ int main(int argc, char **argv)
 
 			ret = tls_tests[i].func(hd.session);
 
-			if (ret != TEST_IGNORE) {
+			if (ret != TEST_IGNORE && ret != TEST_IGNORE2) {
 				printf("%58s...", tls_tests[i].test_name);
 				fflush(stdout);
 			}
