@@ -1,7 +1,7 @@
 /*
  * Copyright (C) 2000-2016 Free Software Foundation, Inc.
  * Copyright (C) 2013-2016 Nikos Mavrogiannopoulos
- * Copyright (C) 2015-2016 Red Hat, Inc.
+ * Copyright (C) 2015-2017 Red Hat, Inc.
  *
  * This file is part of GnuTLS.
  *
@@ -50,7 +50,6 @@
 #include <gnutls/abstract.h>
 #include <gnutls/dtls.h>
 #include <gnutls/x509.h>
-#include <gnutls/openpgp.h>
 #include <gnutls/pkcs11.h>
 #include <gnutls/crypto.h>
 #include <gnutls/socket.h>
@@ -79,10 +78,10 @@
 /* global stuff here */
 int resume, starttls, insecure, ranges, rehandshake, udp, mtu,
     inline_commands;
+unsigned int global_vflags = 0;
 char *hostname = NULL;
 char service[32]="";
 int record_max_size;
-int fingerprint;
 int crlf;
 int fastopen;
 unsigned int verbose = 0;
@@ -90,9 +89,6 @@ int print_cert;
 
 const char *srp_passwd = NULL;
 const char *srp_username = NULL;
-const char *pgp_keyfile = NULL;
-const char *pgp_certfile = NULL;
-const char *pgp_keyring = NULL;
 const char *x509_keyfile = NULL;
 const char *x509_certfile = NULL;
 const char *x509_cafile = NULL;
@@ -129,27 +125,6 @@ static unsigned int x509_crt_size;
 static gnutls_pcert_st x509_crt[MAX_CRT];
 static gnutls_privkey_t x509_key = NULL;
 
-static gnutls_pcert_st pgp_crt;
-static gnutls_privkey_t pgp_key = NULL;
-
-static void get_keyid(gnutls_openpgp_keyid_t keyid, const char *str)
-{
-	size_t keyid_size = GNUTLS_OPENPGP_KEYID_SIZE;
-
-	if (strlen(str) != 16) {
-		fprintf(stderr,
-			"The OpenPGP subkey ID has to be 16 hexadecimal characters.\n");
-		exit(1);
-	}
-
-	if (gnutls_hex2bin(str, strlen(str), keyid, &keyid_size) < 0) {
-		fprintf(stderr, "Error converting hex string: %s.\n", str);
-		exit(1);
-	}
-
-	return;
-}
-
 /* Load the certificate and the private key.
  */
 static void load_keys(void)
@@ -159,13 +134,16 @@ static void load_keys(void)
 	unsigned int i;
 	gnutls_datum_t data = { NULL, 0 };
 	gnutls_x509_crt_t crt_list[MAX_CRT];
-	unsigned char keyid[GNUTLS_OPENPGP_KEYID_SIZE];
 
 	if (x509_certfile != NULL && x509_keyfile != NULL) {
 #ifdef ENABLE_PKCS11
 		if (strncmp(x509_certfile, "pkcs11:", 7) == 0) {
 			crt_num = 1;
-			gnutls_x509_crt_init(&crt_list[0]);
+			ret = gnutls_x509_crt_init(&crt_list[0]);
+			if (ret < 0) {
+				fprintf(stderr, "Memory error\n");
+				exit(1);
+			}
 			gnutls_x509_crt_set_pin_function(crt_list[0],
 							 pin_callback,
 							 NULL);
@@ -280,91 +258,6 @@ static void load_keys(void)
 			"Processed %d client X.509 certificates...\n",
 			x509_crt_size);
 	}
-
-#ifdef ENABLE_OPENPGP
-	if (HAVE_OPT(PGPSUBKEY)) {
-		get_keyid(keyid, OPT_ARG(PGPSUBKEY));
-	}
-
-	if (pgp_certfile != NULL && pgp_keyfile != NULL) {
-		gnutls_openpgp_crt_t tmp_pgp_crt;
-
-		ret = gnutls_load_file(pgp_certfile, &data);
-		if (ret < 0) {
-			fprintf(stderr,
-				"*** Error loading PGP cert file.\n");
-			exit(1);
-		}
-
-		gnutls_openpgp_crt_init(&tmp_pgp_crt);
-
-		ret =
-		    gnutls_pcert_import_openpgp_raw(&pgp_crt, &data,
-						    GNUTLS_OPENPGP_FMT_BASE64,
-						    HAVE_OPT(PGPSUBKEY) ?
-						    keyid : NULL, 0);
-		if (ret < 0) {
-			fprintf(stderr,
-				"*** Error loading PGP cert file: %s\n",
-				gnutls_strerror(ret));
-			exit(1);
-		}
-
-		gnutls_free(data.data);
-
-		ret = gnutls_privkey_init(&pgp_key);
-		if (ret < 0) {
-			fprintf(stderr, "*** Error initializing key: %s\n",
-				gnutls_strerror(ret));
-			exit(1);
-		}
-
-		gnutls_privkey_set_pin_function(pgp_key, pin_callback,
-						NULL);
-
-		if (gnutls_url_is_supported(pgp_keyfile)) {
-			ret =
-			    gnutls_privkey_import_url(pgp_key, pgp_keyfile,
-						      0);
-			if (ret < 0) {
-				fprintf(stderr,
-					"*** Error loading url: %s\n",
-					gnutls_strerror(ret));
-				exit(1);
-			}
-		} else {
-			ret = gnutls_load_file(pgp_keyfile, &data);
-			if (ret < 0) {
-				fprintf(stderr,
-					"*** Error loading key file.\n");
-				exit(1);
-			}
-
-			if (HAVE_OPT(PGPSUBKEY))
-				ret =
-				    gnutls_privkey_import_openpgp_raw
-				    (pgp_key, &data, x509ctype, keyid,
-				     NULL);
-			else
-				ret =
-				    gnutls_privkey_import_openpgp_raw
-				    (pgp_key, &data, x509ctype, NULL,
-				     NULL);
-			if (ret < 0) {
-				fprintf(stderr,
-					"*** Error loading url: %s\n",
-					gnutls_strerror(ret));
-				exit(1);
-			}
-
-			gnutls_free(data.data);
-		}
-
-
-		fprintf(stdout, "Processed 1 client PGP certificate...\n");
-	}
-#endif
-
 }
 
 #define IS_NEWLINE(x) ((x[0] == '\n') || (x[0] == '\r'))
@@ -431,6 +324,7 @@ static int cert_verify_callback(gnutls_session_t session)
 	int dane = ENABLED_OPT(DANE);
 	int ca_verify = ENABLED_OPT(CA_VERIFICATION);
 	const char *txt_service;
+	gnutls_datum_t oresp;
 
 	/* On an session with TOFU the PKI/DANE verification
 	 * become advisory.
@@ -440,10 +334,26 @@ static int cert_verify_callback(gnutls_session_t session)
 		ssh = strictssh;
 	}
 
+	/* Save certificate and OCSP response */
 	if (HAVE_OPT(SAVE_CERT)) {
 		try_save_cert(session);
 	}
 
+	rc = gnutls_ocsp_status_request_get(session, &oresp);
+	if (rc < 0) {
+		oresp.data = NULL;
+		oresp.size = 0;
+	}
+
+	if (HAVE_OPT(SAVE_OCSP) && oresp.data) {
+		FILE *fp = fopen(OPT_ARG(SAVE_OCSP), "w");
+
+		if (fp != NULL) {
+			fwrite(oresp.data, 1, oresp.size, fp);
+			fclose(fp);
+		}
+	}
+
 	print_cert_info(session, verbose, print_cert);
 
 	if (ca_verify) {
@@ -641,14 +551,6 @@ cert_callback(gnutls_session_t session,
 			*pcert_length = x509_crt_size;
 			*pcert = x509_crt;
 		}
-
-	} else if (cert_type == GNUTLS_CRT_OPENPGP) {
-		if (pgp_key != NULL) {
-			*pkey = pgp_key;
-
-			*pcert_length = 1;
-			*pcert = &pgp_crt;
-		}
 	}
 
 	printf("- Successfully sent %u certificate(s) to server.\n",
@@ -729,13 +631,6 @@ gnutls_session_t init_tls_session(const char *host)
 	gnutls_certificate_set_verify_function(xcred,
 					       cert_verify_callback);
 
-	/* send the fingerprint */
-#ifdef ENABLE_OPENPGP
-	if (fingerprint != 0)
-		gnutls_openpgp_send_cert(session,
-					 GNUTLS_OPENPGP_CERT_FINGERPRINT);
-#endif
-
 	/* use the max record size extension */
 	if (record_max_size > 0 && disable_extensions == 0) {
 		if (gnutls_record_set_max_size(session, record_max_size) <
@@ -934,7 +829,8 @@ static int try_resume(socket_st * hd)
 	if (udp)
 		socket_flags |= SOCKET_FLAG_UDP;
 
-	socket_open(hd, hostname, service, OPT_ARG(STARTTLS_PROTO), socket_flags, CONNECT_MSG, &rdata);
+	socket_open(hd, hostname, service, OPT_ARG(STARTTLS_PROTO),
+		     socket_flags, CONNECT_MSG, &rdata);
 
 	printf("- Resume Handshake was completed\n");
 	if (gnutls_session_is_resumed(hd->session) != 0)
@@ -1153,14 +1049,6 @@ print_other_info(gnutls_session_t session)
 		fputs((char*)p.data, stdout);
 	}
 
-	if (HAVE_OPT(SAVE_OCSP) && oresp.data) {
-		FILE *fp = fopen(OPT_ARG(SAVE_OCSP), "w");
-
-		if (fp != NULL) {
-			fwrite(oresp.data, 1, oresp.size, fp);
-			fclose(fp);
-		}
-	}
 }
 
 static void flush_socket(socket_st *hd, unsigned ms)
@@ -1195,6 +1083,8 @@ int main(int argc, char **argv)
 	inline_cmds_st inline_cmds;
 	unsigned last_op_is_write = 0;
 	int socket_flags = 0;
+	FILE *server_fp = NULL;
+	FILE *client_fp = NULL;
 #ifndef _WIN32
 	struct sigaction new_action;
 #endif
@@ -1231,7 +1121,17 @@ int main(int argc, char **argv)
 	else if (HAVE_OPT(STARTTLS_PROTO))
 		socket_flags |= SOCKET_FLAG_STARTTLS;
 
-	socket_open(&hd, hostname, service, OPT_ARG(STARTTLS_PROTO), socket_flags, CONNECT_MSG, NULL);
+	if (HAVE_OPT(SAVE_SERVER_TRACE)) {
+		server_fp = fopen(OPT_ARG(SAVE_SERVER_TRACE), "wb");
+	}
+	if (HAVE_OPT(SAVE_CLIENT_TRACE)) {
+		client_fp = fopen(OPT_ARG(SAVE_CLIENT_TRACE), "wb");
+	}
+
+	socket_open2(&hd, hostname, service, OPT_ARG(STARTTLS_PROTO),
+		     socket_flags, CONNECT_MSG, NULL,
+		     server_fp, client_fp);
+
 	hd.verbose = verbose;
 
 	if (hd.secure) {
@@ -1568,6 +1468,10 @@ static void cmd_parser(int argc, char **argv)
 	insecure = HAVE_OPT(INSECURE);
 	ranges = HAVE_OPT(RANGES);
 
+	if (insecure || HAVE_OPT(VERIFY_ALLOW_BROKEN)) {
+		global_vflags |= GNUTLS_VERIFY_ALLOW_BROKEN;
+	}
+
 	udp = HAVE_OPT(UDP);
 	mtu = OPT_VALUE_MTU;
 
@@ -1582,8 +1486,6 @@ static void cmd_parser(int argc, char **argv)
 
 	record_max_size = OPT_VALUE_RECORDSIZE;
 
-	fingerprint = HAVE_OPT(FINGERPRINT);
-
 	if (HAVE_OPT(X509FMTDER))
 		x509ctype = GNUTLS_X509_FMT_DER;
 	else
@@ -1607,12 +1509,6 @@ static void cmd_parser(int argc, char **argv)
 	if (HAVE_OPT(X509CERTFILE))
 		x509_certfile = OPT_ARG(X509CERTFILE);
 
-	if (HAVE_OPT(PGPKEYFILE))
-		pgp_keyfile = OPT_ARG(PGPKEYFILE);
-
-	if (HAVE_OPT(PGPCERTFILE))
-		pgp_certfile = OPT_ARG(PGPCERTFILE);
-
 	if (HAVE_OPT(PSKUSERNAME))
 		psk_username = OPT_ARG(PSKUSERNAME);
 
@@ -1622,9 +1518,6 @@ static void cmd_parser(int argc, char **argv)
 	} else
 		psk_key.size = 0;
 
-	if (HAVE_OPT(PGPKEYRING))
-		pgp_keyring = OPT_ARG(PGPKEYRING);
-
 	crlf = HAVE_OPT(CRLF);
 
 #ifdef TCP_FASTOPEN
@@ -1692,7 +1585,7 @@ int do_handshake(socket_st * socket)
 
 	if (ret == 0) {
 		/* print some information */
-		print_info(socket->session, verbose, (HAVE_OPT(X509CERTFILE)||HAVE_OPT(PGPCERTFILE))?P_WAIT_FOR_CERT:0);
+		print_info(socket->session, verbose, HAVE_OPT(X509CERTFILE)?P_WAIT_FOR_CERT:0);
 		socket->secure = 1;
 	} else {
 		gnutls_alert_send_appropriate(socket->session, ret);
@@ -1740,9 +1633,9 @@ psk_callback(gnutls_session_t session, char **username,
 
 		printf("Enter PSK identity: ");
 		fflush(stdout);
-		getline(&p, &n, stdin);
+		ret = getline(&p, &n, stdin);
 
-		if (p == NULL) {
+		if (ret == -1 || p == NULL) {
 			fprintf(stderr,
 				"No username given, aborting...\n");
 			return GNUTLS_E_INSUFFICIENT_CREDENTIALS;
@@ -1788,7 +1681,11 @@ psk_callback(gnutls_session_t session, char **username,
 	if (HAVE_OPT(DEBUG)) {
 		char hexkey[41];
 		res_size = sizeof(hexkey);
-		gnutls_hex_encode(key, hexkey, &res_size);
+		ret = gnutls_hex_encode(key, hexkey, &res_size);
+		if (ret < 0) {
+			fprintf(stderr, "Error in hex encoding: %s\n", gnutls_strerror(ret));
+			exit(1);
+		}
 		fprintf(stderr, "PSK username: %s\n", *username);
 		fprintf(stderr, "PSK hint: %s\n", hint);
 		fprintf(stderr, "PSK key: %s\n", hexkey);
@@ -1827,6 +1724,8 @@ static void init_global_tls_stuff(void)
 	}
 	gnutls_certificate_set_pin_function(xcred, pin_callback, NULL);
 
+	gnutls_certificate_set_verify_flags(xcred, global_vflags);
+
 	if (x509_cafile != NULL) {
 		ret = gnutls_certificate_set_x509_trust_file(xcred,
 							     x509_cafile,
@@ -1859,19 +1758,6 @@ static void init_global_tls_stuff(void)
 
 	load_keys();
 
-#ifdef ENABLE_OPENPGP
-	if (pgp_keyring != NULL) {
-		ret =
-		    gnutls_certificate_set_openpgp_keyring_file(xcred,
-								pgp_keyring,
-								GNUTLS_OPENPGP_FMT_BASE64);
-		if (ret < 0) {
-			fprintf(stderr,
-				"Error setting the OpenPGP keyring file\n");
-		}
-	}
-#endif
-
 #ifdef ENABLE_SRP
 	if (srp_username && srp_passwd) {
 		/* SRP stuff */
@@ -1942,7 +1828,13 @@ static int cert_verify_ocsp(gnutls_session_t session)
 	for (it = 0; it < cert_list_size; it++) {
 		if (deinit_cert)
 			gnutls_x509_crt_deinit(cert);
-		gnutls_x509_crt_init(&cert);
+
+		ret = gnutls_x509_crt_init(&cert);
+		if (ret < 0) {
+			fprintf(stderr, "Memory error: %s\n", gnutls_strerror(ret));
+			goto cleanup;
+		}
+
 		deinit_cert = 1;
 		ret = gnutls_x509_crt_import(cert, &cert_list[it], GNUTLS_X509_FMT_DER);
 		if (ret < 0) {
@@ -1957,7 +1849,11 @@ static int cert_verify_ocsp(gnutls_session_t session)
 
 		ret = gnutls_certificate_get_issuer(xcred, cert, &issuer, 0);
 		if (ret < 0 && cert_list_size - it > 1) {
-			gnutls_x509_crt_init(&issuer);
+			ret = gnutls_x509_crt_init(&issuer);
+			if (ret < 0) {
+				fprintf(stderr, "Memory error: %s\n", gnutls_strerror(ret));
+				goto cleanup;
+			}
 			deinit_issuer = 1;
 			ret = gnutls_x509_crt_import(issuer, &cert_list[it + 1], GNUTLS_X509_FMT_DER);
 			if (ret < 0) {
@@ -1965,7 +1861,8 @@ static int cert_verify_ocsp(gnutls_session_t session)
 				goto cleanup;
 			}
 		} else if (ret < 0) {
-			fprintf(stderr, "Cannot find issuer\n");
+			if (it == 0)
+				fprintf(stderr, "Cannot find issuer: %s\n", gnutls_strerror(ret));
 			goto cleanup;
 		}
 
