@@ -33,7 +33,6 @@
 #include <string.h>
 #include <gnutls/gnutls.h>
 #include <gnutls/x509.h>
-#include <gnutls/openpgp.h>
 #include <gnutls/crypto.h>
 #include <time.h>
 #include <common.h>
@@ -128,7 +127,12 @@ static void print_x509_info_compact(gnutls_session_t session)
 		return;
 	}
 
-	gnutls_x509_crt_init(&crt);
+	ret = gnutls_x509_crt_init(&crt);
+	if (ret < 0) {
+		fprintf(stderr, "Memory error\n");
+		return;
+	}
+
 	ret =
 	    gnutls_x509_crt_import(crt, &cert_list[0],
 				   GNUTLS_X509_FMT_DER);
@@ -168,7 +172,12 @@ print_x509_info(gnutls_session_t session, FILE *out, int flag, int print_cert)
 	for (j = 0; j < cert_list_size; j++) {
 		gnutls_datum_t cinfo;
 
-		gnutls_x509_crt_init(&crt);
+		ret = gnutls_x509_crt_init(&crt);
+		if (ret < 0) {
+			fprintf(stderr, "Memory error\n");
+			return;
+		}
+
 		ret =
 		    gnutls_x509_crt_import(crt, &cert_list[j],
 					   GNUTLS_X509_FMT_DER);
@@ -189,147 +198,28 @@ print_x509_info(gnutls_session_t session, FILE *out, int flag, int print_cert)
 		}
 
 		if (print_cert) {
-			size_t size = 0;
-			char *p = NULL;
+			gnutls_datum_t pem;
 
 			ret =
-			    gnutls_x509_crt_export(crt,
-						   GNUTLS_X509_FMT_PEM, p,
-						   &size);
-			if (ret == GNUTLS_E_SHORT_MEMORY_BUFFER) {
-				p = malloc(size + 1);
-				if (!p) {
-					fprintf(stderr, "gnutls_malloc\n");
-					exit(1);
-				}
-
-				ret =
-				    gnutls_x509_crt_export(crt,
-							   GNUTLS_X509_FMT_PEM,
-							   p, &size);
-			}
+			    gnutls_x509_crt_export2(crt,
+						   GNUTLS_X509_FMT_PEM, &pem);
 			if (ret < 0) {
 				fprintf(stderr, "Encoding error: %s\n",
 					gnutls_strerror(ret));
 				return;
 			}
 
-			p[size] = 0;
 			fputs("\n", out);
-			fputs(p, out);
+			fputs((char*)pem.data, out);
 			fputs("\n", out);
 
-			gnutls_free(p);
+			gnutls_free(pem.data);
 		}
 
 		gnutls_x509_crt_deinit(crt);
 	}
 }
 
-#ifdef ENABLE_OPENPGP
-static void print_openpgp_info_compact(gnutls_session_t session)
-{
-
-	gnutls_openpgp_crt_t crt;
-	const gnutls_datum_t *cert_list;
-	unsigned int cert_list_size = 0;
-	int ret;
-
-	cert_list = gnutls_certificate_get_peers(session, &cert_list_size);
-
-	if (cert_list_size > 0) {
-		gnutls_datum_t cinfo;
-
-		gnutls_openpgp_crt_init(&crt);
-		ret = gnutls_openpgp_crt_import(crt, &cert_list[0],
-						GNUTLS_OPENPGP_FMT_RAW);
-		if (ret < 0) {
-			fprintf(stderr, "Decoding error: %s\n",
-				gnutls_strerror(ret));
-			return;
-		}
-
-		ret =
-		    gnutls_openpgp_crt_print(crt, GNUTLS_CRT_PRINT_COMPACT,
-					     &cinfo);
-		if (ret == 0) {
-			printf("- OpenPGP cert: %s\n", cinfo.data);
-			gnutls_free(cinfo.data);
-		}
-
-		gnutls_openpgp_crt_deinit(crt);
-	}
-}
-
-static void
-print_openpgp_info(gnutls_session_t session, FILE *out, int flag, int print_cert)
-{
-
-	gnutls_openpgp_crt_t crt;
-	const gnutls_datum_t *cert_list;
-	unsigned int cert_list_size = 0;
-	int ret;
-
-	fprintf(out, "- Certificate type: OpenPGP\n");
-
-	cert_list = gnutls_certificate_get_peers(session, &cert_list_size);
-
-	if (cert_list_size > 0) {
-		gnutls_datum_t cinfo;
-
-		gnutls_openpgp_crt_init(&crt);
-		ret = gnutls_openpgp_crt_import(crt, &cert_list[0],
-						GNUTLS_OPENPGP_FMT_RAW);
-		if (ret < 0) {
-			fprintf(stderr, "Decoding error: %s\n",
-				gnutls_strerror(ret));
-			return;
-		}
-
-		ret = gnutls_openpgp_crt_print(crt, flag, &cinfo);
-		if (ret == 0) {
-			fprintf(out, "- %s\n", cinfo.data);
-			gnutls_free(cinfo.data);
-		}
-
-		if (print_cert) {
-			size_t size = 0;
-			char *p = NULL;
-
-			ret =
-			    gnutls_openpgp_crt_export(crt,
-						      GNUTLS_OPENPGP_FMT_BASE64,
-						      p, &size);
-			if (ret == GNUTLS_E_SHORT_MEMORY_BUFFER) {
-				p = malloc(size);
-				if (!p) {
-					fprintf(stderr, "gnutls_malloc\n");
-					exit(1);
-				}
-
-				ret =
-				    gnutls_openpgp_crt_export(crt,
-							      GNUTLS_OPENPGP_FMT_BASE64,
-							      p, &size);
-			}
-			if (ret < 0) {
-				fprintf(stderr, "Encoding error: %s\n",
-					gnutls_strerror(ret));
-				return;
-			}
-
-			fputs(p, out);
-			fputs("\n", out);
-
-			gnutls_free(p);
-		}
-
-		gnutls_openpgp_crt_deinit(crt);
-	}
-}
-
-#endif
-
 /* returns false (0) if not verified, or true (1) otherwise 
  */
 int cert_verify(gnutls_session_t session, const char *hostname, const char *purpose)
@@ -390,6 +280,13 @@ static void
 print_dh_info(gnutls_session_t session, const char *str, int print)
 {
 #if defined(ENABLE_DHE) || defined(ENABLE_ANON)
+	unsigned group;
+
+	group = gnutls_group_get(session);
+	if (group != 0) {
+		return;
+	}
+
 	printf("- %sDiffie-Hellman parameters\n", str);
 	printf(" - Using prime: %d bits\n",
 	       gnutls_dh_get_prime_bits(session));
@@ -602,11 +499,6 @@ int print_info(gnutls_session_t session, int verbose, int flags)
 	tmp = SU(gnutls_mac_get_name(gnutls_mac_get(session)));
 	printf("- MAC: %s\n", tmp);
 
-	tmp =
-	    SU(gnutls_compression_get_name
-	       (gnutls_compression_get(session)));
-	printf("- Compression: %s\n", tmp);
-
 	printf("- Options:");
 	if (gnutls_session_ext_master_secret_status(session)!=0)
 		printf(" extended master secret,");
@@ -650,13 +542,10 @@ int print_info(gnutls_session_t session, int verbose, int flags)
 			for (i = 0; i < cb.size; i++)
 				printf("%02x", cb.data[i]);
 			printf("\n");
+			gnutls_free(cb.data);
 		}
 	}
 
-	/* Warning: Do not print anything more here. The 'Compression:'
-	   output MUST be the last non-verbose output.  This is used by
-	   Emacs starttls.el code. */
-
 	fflush(stdout);
 
 	return 0;
@@ -683,11 +572,6 @@ void print_cert_info2(gnutls_session_t session, int verbose, FILE *out, int prin
 	case GNUTLS_CRT_X509:
 		print_x509_info(session, out, flag, print_cert);
 		break;
-#ifdef ENABLE_OPENPGP
-	case GNUTLS_CRT_OPENPGP:
-		print_openpgp_info(session, out, flag, print_cert);
-		break;
-#endif
 	default:
 		printf("Unknown type\n");
 		break;
@@ -704,11 +588,6 @@ void print_cert_info_compact(gnutls_session_t session)
 	case GNUTLS_CRT_X509:
 		print_x509_info_compact(session);
 		break;
-#ifdef ENABLE_OPENPGP
-	case GNUTLS_CRT_OPENPGP:
-		print_openpgp_info_compact(session);
-		break;
-#endif
 	default:
 		printf("Unknown type\n");
 		break;
@@ -796,6 +675,54 @@ void print_list(const char *priorities, int verbose)
 			}
 		}
 
+		{
+			ret = gnutls_priority_cipher_list(pcache, &list);
+
+			printf("Ciphers: ");
+			if (ret == 0)
+				printf("none\n");
+			for (i = 0; i < (unsigned) ret; i++) {
+				printf("%s",
+				       gnutls_cipher_get_name(list[i]));
+				if (i + 1 != (unsigned) ret)
+					printf(", ");
+				else
+					printf("\n");
+			}
+		}
+
+		{
+			ret = gnutls_priority_mac_list(pcache, &list);
+
+			printf("MACs: ");
+			if (ret == 0)
+				printf("none\n");
+			for (i = 0; i < (unsigned) ret; i++) {
+				printf("%s",
+				       gnutls_mac_get_name(list[i]));
+				if (i + 1 != (unsigned) ret)
+					printf(", ");
+				else
+					printf("\n");
+			}
+		}
+
+		{
+			ret = gnutls_priority_kx_list(pcache, &list);
+
+			printf("Key Exchange Algorithms: ");
+			if (ret == 0)
+				printf("none\n");
+			for (i = 0; i < (unsigned) ret; i++) {
+				printf("%s",
+				       gnutls_kx_get_name(list[i]));
+				if (i + 1 != (unsigned) ret)
+					printf(", ");
+				else
+					printf("\n");
+			}
+		}
+
 		{
 			ret =
 			    gnutls_priority_compression_list(pcache,
@@ -817,14 +744,14 @@ void print_list(const char *priorities, int verbose)
 
 		{
 			ret =
-			    gnutls_priority_ecc_curve_list(pcache, &list);
+			    gnutls_priority_group_list(pcache, &list);
 
-			printf("Elliptic curves: ");
+			printf("Groups: ");
 			if (ret == 0)
 				printf("none\n");
 			for (i = 0; i < (unsigned) ret; i++) {
-				printf("CURVE-%s",
-				       gnutls_ecc_curve_get_name(list[i]));
+				printf("GROUP-%s",
+				       gnutls_group_get_name(list[i]));
 				if (i + 1 != (unsigned) ret)
 					printf(", ");
 				else
@@ -849,6 +776,7 @@ void print_list(const char *priorities, int verbose)
 			}
 		}
 
+		gnutls_priority_deinit(pcache);
 		return;
 	}
 
@@ -963,11 +891,11 @@ void print_list(const char *priorities, int verbose)
 	}
 
 	{
-		const gnutls_ecc_curve_t *p = gnutls_ecc_curve_list();
+		const gnutls_group_t *p = gnutls_group_list();
 
-		printf("Elliptic curves: ");
+		printf("Groups: ");
 		for (; *p; p++) {
-			printf("CURVE-%s", gnutls_ecc_curve_get_name(*p));
+			printf("GROUP-%s", gnutls_group_get_name(*p));
 			if (*(p + 1))
 				printf(", ");
 			else
