@@ -50,7 +50,7 @@ static char *_saved_url = NULL;
 		if (url == NULL) { \
 			fprintf(stderr, "warning: no token URL was provided for this operation; the available tokens are:\n\n"); \
 			pkcs11_token_list(out, det, info, 1); \
-			exit(1); \
+			app_exit(1); \
 		} \
 		_saved_url = (void*)url; \
 	}
@@ -82,7 +82,7 @@ pkcs11_delete(FILE * outfile, const char *url,
 		    ("Are you sure you want to delete those objects? (y/N): ",
 		     0);
 		if (ret == 0) {
-			exit(1);
+			app_exit(1);
 		}
 	}
 
@@ -90,7 +90,7 @@ pkcs11_delete(FILE * outfile, const char *url,
 	if (ret < 0) {
 		fprintf(stderr, "Error in %s:%d: %s\n", __func__, __LINE__,
 			gnutls_strerror(ret));
-		exit(1);
+		app_exit(1);
 	}
 
 	fprintf(outfile, "\n%d objects deleted\n", ret);
@@ -144,12 +144,12 @@ pkcs11_list(FILE * outfile, const char *url, int type, unsigned int flags,
 	if (ret < 0) {
 		fprintf(stderr, "Error in crt_list_import (1): %s\n",
 			gnutls_strerror(ret));
-		exit(1);
+		app_exit(1);
 	}
 
 	if (crt_list_size == 0) {
 		fprintf(stderr, "No matching objects found\n");
-		exit(2);
+		app_exit(2);
 	}
 
 	for (i = 0; i < crt_list_size; i++) {
@@ -163,7 +163,7 @@ pkcs11_list(FILE * outfile, const char *url, int type, unsigned int flags,
 		if (ret < 0) {
 			fprintf(stderr, "Error in %s:%d: %s\n", __func__,
 				__LINE__, gnutls_strerror(ret));
-			exit(1);
+			app_exit(1);
 		}
 
 		if (info->only_urls) {
@@ -187,7 +187,7 @@ pkcs11_list(FILE * outfile, const char *url, int type, unsigned int flags,
 		if (ret < 0) {
 			fprintf(stderr, "Error in %s:%d: %s\n", __func__,
 				__LINE__, gnutls_strerror(ret));
-			exit(1);
+			app_exit(1);
 		}
 		fprintf(outfile, "\tLabel: %s\n", buf);
 
@@ -196,7 +196,7 @@ pkcs11_list(FILE * outfile, const char *url, int type, unsigned int flags,
 		if (ret < 0) {
 			fprintf(stderr, "Error in %s:%d: %s\n", __func__,
 				__LINE__, gnutls_strerror(ret));
-			exit(1);
+			app_exit(1);
 		}
 		str = gnutls_pkcs11_obj_flags_get_str(oflags);
 		if (str != NULL) {
@@ -212,7 +212,7 @@ pkcs11_list(FILE * outfile, const char *url, int type, unsigned int flags,
 		if (ret < 0) {
 			fprintf(stderr, "Error in %s:%d: %s\n", __func__,
 				__LINE__, gnutls_strerror(ret));
-			exit(1);
+			app_exit(1);
 		}
 		fprintf(outfile, "\tID: %s\n", buf);
 
@@ -274,46 +274,46 @@ pkcs11_test_sign(FILE * outfile, const char *url, unsigned int flags,
 	if (ret < 0) {
 		fprintf(stderr, "Error in %s:%d: %s\n", __func__,
 			__LINE__, gnutls_strerror(ret));
-		exit(1);
+		app_exit(1);
 	}
 
 	ret = gnutls_pubkey_init(&pubkey);
 	if (ret < 0) {
 		fprintf(stderr, "Error in %s:%d: %s\n", __func__,
 			__LINE__, gnutls_strerror(ret));
-		exit(1);
+		app_exit(1);
 	}
 
 	ret = gnutls_privkey_import_url(privkey, url, flags);
 	if (ret < 0) {
 		fprintf(stderr, "Cannot import private key: %s\n",
 			gnutls_strerror(ret));
-		exit(1);
+		app_exit(1);
 	}
 
 	ret = gnutls_pubkey_import_privkey(pubkey, privkey, GNUTLS_KEY_DIGITAL_SIGNATURE, flags);
 	if (ret < 0) {
 		fprintf(stderr, "Cannot import public key: %s\n",
 			gnutls_strerror(ret));
-		exit(1);
+		app_exit(1);
 	}
 
-	ret = gnutls_privkey_sign_data(privkey, GNUTLS_DIG_SHA1, 0, &data, &sig);
+	ret = gnutls_privkey_sign_data(privkey, GNUTLS_DIG_SHA256, 0, &data, &sig);
 	if (ret < 0) {
 		fprintf(stderr, "Cannot sign data: %s\n",
 			gnutls_strerror(ret));
-		exit(1);
+		app_exit(1);
 	}
 
 	pk = gnutls_pubkey_get_pk_algorithm(pubkey, NULL);
 
 	fprintf(stderr, "Verifying against private key parameters... ");
-	ret = gnutls_pubkey_verify_data2(pubkey, gnutls_pk_to_sign(pk, GNUTLS_DIG_SHA1),
+	ret = gnutls_pubkey_verify_data2(pubkey, gnutls_pk_to_sign(pk, GNUTLS_DIG_SHA256),
 		0, &data, &sig);
 	if (ret < 0) {
 		fprintf(stderr, "Cannot verify signed data: %s\n",
 			gnutls_strerror(ret));
-		exit(1);
+		app_exit(1);
 	}
 
 	fprintf(stderr, "ok\n");
@@ -324,7 +324,7 @@ pkcs11_test_sign(FILE * outfile, const char *url, unsigned int flags,
 	if (ret < 0) {
 		fprintf(stderr, "Error in %s:%d: %s\n", __func__,
 			__LINE__, gnutls_strerror(ret));
-		exit(1);
+		app_exit(1);
 	}
 
 	ret = gnutls_pubkey_import_url(pubkey, url, flags);
@@ -332,17 +332,17 @@ pkcs11_test_sign(FILE * outfile, const char *url, unsigned int flags,
 		fprintf(stderr, "Cannot find a corresponding public key object in token: %s\n",
 			gnutls_strerror(ret));
 		if (ret == GNUTLS_E_REQUESTED_DATA_NOT_AVAILABLE)
-			exit(0);
-		exit(1);
+			app_exit(0);
+		app_exit(1);
 	}
 
 	fprintf(stderr, "Verifying against public key in the token... ");
-	ret = gnutls_pubkey_verify_data2(pubkey, gnutls_pk_to_sign(pk, GNUTLS_DIG_SHA1),
+	ret = gnutls_pubkey_verify_data2(pubkey, gnutls_pk_to_sign(pk, GNUTLS_DIG_SHA256),
 		0, &data, &sig);
 	if (ret < 0) {
 		fprintf(stderr, "Cannot verify signed data: %s\n",
 			gnutls_strerror(ret));
-		exit(1);
+		app_exit(1);
 	}
 
 	fprintf(stderr, "ok\n");
@@ -370,21 +370,21 @@ pkcs11_export(FILE * outfile, const char *url, unsigned int flags,
 	if (ret < 0) {
 		fprintf(stderr, "Error in %s:%d: %s\n", __func__, __LINE__,
 			gnutls_strerror(ret));
-		exit(1);
+		app_exit(1);
 	}
 
 	ret = gnutls_pkcs11_obj_import_url(obj, url, obj_flags);
 	if (ret < 0) {
 		fprintf(stderr, "Error in %s:%d: %s\n", __func__, __LINE__,
 			gnutls_strerror(ret));
-		exit(1);
+		app_exit(1);
 	}
 
 	ret = gnutls_pkcs11_obj_export3(obj, info->outcert_format, &t);
 	if (ret < 0) {
 		fprintf(stderr, "Error in %s:%d: %s\n", __func__,
 				__LINE__, gnutls_strerror(ret));
-		exit(1);
+		app_exit(1);
 	}
 
 	fwrite(t.data, 1, t.size, outfile);
@@ -417,14 +417,14 @@ pkcs11_export_chain(FILE * outfile, const char *url, unsigned int flags,
 	if (ret < 0) {
 		fprintf(stderr, "Error in %s:%d: %s\n", __func__, __LINE__,
 			gnutls_strerror(ret));
-		exit(1);
+		app_exit(1);
 	}
 
 	ret = gnutls_pkcs11_obj_import_url(obj, url, obj_flags);
 	if (ret < 0) {
 		fprintf(stderr, "Error in %s:%d: %s\n", __func__, __LINE__,
 			gnutls_strerror(ret));
-		exit(1);
+		app_exit(1);
 	}
 
 	/* make a crt */
@@ -432,21 +432,21 @@ pkcs11_export_chain(FILE * outfile, const char *url, unsigned int flags,
 	if (ret < 0) {
 		fprintf(stderr, "Error in %s:%d: %s\n", __func__, __LINE__,
 			gnutls_strerror(ret));
-		exit(1);
+		app_exit(1);
 	}
 
 	ret = gnutls_x509_crt_import_pkcs11(xcrt, obj);
 	if (ret < 0) {
 		fprintf(stderr, "Error in %s:%d: %s\n", __func__,
 				__LINE__, gnutls_strerror(ret));
-		exit(1);
+		app_exit(1);
 	}
 
 	ret = gnutls_pkcs11_obj_export3(obj, GNUTLS_X509_FMT_PEM, &t);
 	if (ret < 0) {
 		fprintf(stderr, "Error in %s:%d: %s\n", __func__,
 				__LINE__, gnutls_strerror(ret));
-		exit(1);
+		app_exit(1);
 	}
 	fwrite(t.data, 1, t.size, outfile);
 	fputs("\n\n", outfile);
@@ -461,7 +461,7 @@ pkcs11_export_chain(FILE * outfile, const char *url, unsigned int flags,
 		if (ret < 0) {
 			fprintf(stderr, "Error in %s:%d: %s\n", __func__,
 					__LINE__, gnutls_strerror(ret));
-			exit(1);
+			app_exit(1);
 		}
 
 		fwrite(t.data, 1, t.size, outfile);
@@ -473,14 +473,14 @@ pkcs11_export_chain(FILE * outfile, const char *url, unsigned int flags,
 		if (ret < 0) {
 			fprintf(stderr, "Error in %s:%d: %s\n", __func__,
 					__LINE__, gnutls_strerror(ret));
-			exit(1);
+			app_exit(1);
 		}
 
 		ret = gnutls_x509_crt_import(xcrt, &t, GNUTLS_X509_FMT_PEM);
 		if (ret < 0) {
 			fprintf(stderr, "Error in %s:%d: %s\n", __func__,
 					__LINE__, gnutls_strerror(ret));
-			exit(1);
+			app_exit(1);
 		}
 
 		gnutls_free(t.data);
@@ -565,7 +565,7 @@ pkcs11_token_list(FILE * outfile, unsigned int detailed,
 		if (ret < 0) {
 			fprintf(stderr, "Error in %s:%d: %s\n", __func__,
 				__LINE__, gnutls_strerror(ret));
-			exit(1);
+			app_exit(1);
 		}
 
 		if (brief != 0) {
@@ -583,7 +583,7 @@ pkcs11_token_list(FILE * outfile, unsigned int detailed,
 		if (ret < 0) {
 			fprintf(stderr, "Error in %s:%d: %s\n", __func__,
 				__LINE__, gnutls_strerror(ret));
-			exit(1);
+			app_exit(1);
 		}
 
 		fprintf(outfile, "\tLabel: %s\n", buf);
@@ -604,7 +604,7 @@ pkcs11_token_list(FILE * outfile, unsigned int detailed,
 		if (ret < 0) {
 			fprintf(stderr, "Error in %s:%d: %s\n", __func__,
 				__LINE__, gnutls_strerror(ret));
-			exit(1);
+			app_exit(1);
 		}
 
 		fprintf(outfile, "\tManufacturer: %s\n", buf);
@@ -617,7 +617,7 @@ pkcs11_token_list(FILE * outfile, unsigned int detailed,
 		if (ret < 0) {
 			fprintf(stderr, "Error in %s:%d: %s\n", __func__,
 				__LINE__, gnutls_strerror(ret));
-			exit(1);
+			app_exit(1);
 		}
 
 		fprintf(outfile, "\tModel: %s\n", buf);
@@ -630,7 +630,7 @@ pkcs11_token_list(FILE * outfile, unsigned int detailed,
 		if (ret < 0) {
 			fprintf(stderr, "Error in %s:%d: %s\n", __func__,
 				__LINE__, gnutls_strerror(ret));
-			exit(1);
+			app_exit(1);
 		}
 
 		fprintf(outfile, "\tSerial: %s\n", buf);
@@ -668,7 +668,7 @@ static void find_same_pubkey_with_id(const char *url, gnutls_x509_crt_t crt, gnu
 	ret = gnutls_pubkey_init(&pubkey);
 	if (ret < 0) {
 		fprintf(stderr, "memory error\n");
-		exit(1);
+		app_exit(1);
 	}
 
 	ret = gnutls_pubkey_import_x509(pubkey, crt, 0);
@@ -691,7 +691,7 @@ static void find_same_pubkey_with_id(const char *url, gnutls_x509_crt_t crt, gnu
 	if (ret < 0) {
 		fprintf(stderr, "Error in obj_list_import (1): %s\n",
 			gnutls_strerror(ret));
-		exit(1);
+		app_exit(1);
 	}
 
 	if (obj_list_size == 0)
@@ -730,14 +730,14 @@ static void find_same_pubkey_with_id(const char *url, gnutls_x509_crt_t crt, gnu
 			if (ret < 0) {
 				fprintf(stderr, "Error in %s:%d: %s\n", __func__,
 					__LINE__, gnutls_strerror(ret));
-				exit(1);
+				app_exit(1);
 			}
 
 			cid->data = gnutls_malloc(size);
 			cid->size = size;
 			if (cid->data == NULL) {
 				fprintf(stderr, "memory error\n");
-				exit(1);
+				app_exit(1);
 			}
 
 			memcpy(cid->data, buf, size);
@@ -772,7 +772,7 @@ static void find_same_privkey_with_id(const char *url, gnutls_x509_crt_t crt, gn
 	ret = gnutls_pubkey_init(&pubkey);
 	if (ret < 0) {
 		fprintf(stderr, "memory error\n");
-		exit(1);
+		app_exit(1);
 	}
 
 	ret = gnutls_pubkey_import_x509(pubkey, crt, 0);
@@ -795,7 +795,7 @@ static void find_same_privkey_with_id(const char *url, gnutls_x509_crt_t crt, gn
 	if (ret < 0) {
 		fprintf(stderr, "Error in obj_list_import (1): %s\n",
 			gnutls_strerror(ret));
-		exit(1);
+		app_exit(1);
 	}
 
 	if (obj_list_size == 0)
@@ -822,7 +822,7 @@ static void find_same_privkey_with_id(const char *url, gnutls_x509_crt_t crt, gn
 		ret = gnutls_privkey_init(&privkey);
 		if (ret < 0) {
 			fprintf(stderr, "memory error\n");
-			exit(1);
+			app_exit(1);
 		}
 
 		ret = gnutls_privkey_import_url(privkey, purl, 0);
@@ -839,7 +839,7 @@ static void find_same_privkey_with_id(const char *url, gnutls_x509_crt_t crt, gn
 		ret = gnutls_pubkey_init(&pubkey);
 		if (ret < 0) {
 			fprintf(stderr, "memory error\n");
-			exit(1);
+			app_exit(1);
 		}
 
 		ret = gnutls_pubkey_import_privkey(pubkey, privkey, 0, 0);
@@ -866,14 +866,14 @@ static void find_same_privkey_with_id(const char *url, gnutls_x509_crt_t crt, gn
 			if (ret < 0) {
 				fprintf(stderr, "Error in %s:%d: %s\n", __func__,
 					__LINE__, gnutls_strerror(ret));
-				exit(1);
+				app_exit(1);
 			}
 
 			cid->data = gnutls_malloc(size);
 			cid->size = size;
 			if (cid->data == NULL) {
 				fprintf(stderr, "memory error\n");
-				exit(1);
+				app_exit(1);
 			}
 
 			memcpy(cid->data, buf, size);
@@ -921,7 +921,7 @@ pkcs11_write(FILE * outfile, const char *url, const char *label,
 		ret = gnutls_hex2bin(id, strlen(id), raw_id, &raw_id_size);
 		if (ret < 0) {
 			fprintf(stderr, "Error converting hex: %s\n", gnutls_strerror(ret));
-			exit(1);
+			app_exit(1);
 		}
 		cid.data = raw_id;
 		cid.size = raw_id_size;
@@ -937,7 +937,7 @@ pkcs11_write(FILE * outfile, const char *url, const char *label,
 		if (ret < 0) {
 			fprintf(stderr, "Error in %s:%d: %s\n", __func__,
 				__LINE__, gnutls_strerror(ret));
-			exit(1);
+			app_exit(1);
 		}
 	}
 
@@ -959,7 +959,7 @@ pkcs11_write(FILE * outfile, const char *url, const char *label,
 				ret = gnutls_hex_encode2(&cid, &hex);
 				if (ret < 0) {
 					fprintf(stderr, "Error converting hex: %s\n", gnutls_strerror(ret));
-					exit(1);
+					app_exit(1);
 				}
 				fprintf(stderr, "note: will re-use ID %s from corresponding public key\n", hex.data);
 				gnutls_free(hex.data);
@@ -970,7 +970,7 @@ pkcs11_write(FILE * outfile, const char *url, const char *label,
 					ret = gnutls_hex_encode2(&cid, &hex);
 					if (ret < 0) {
 						fprintf(stderr, "Error converting hex: %s\n", gnutls_strerror(ret));
-						exit(1);
+						app_exit(1);
 					}
 					fprintf(stderr, "note: will re-use ID %s from corresponding private key\n", hex.data);
 					gnutls_free(hex.data);
@@ -985,7 +985,7 @@ pkcs11_write(FILE * outfile, const char *url, const char *label,
 			     (flags & GNUTLS_PKCS11_OBJ_FLAG_MARK_TRUSTED)) && 
 			    (flags & GNUTLS_PKCS11_OBJ_FLAG_LOGIN_SO) == 0)
 				fprintf(stderr, "note: some tokens may require security officer login for this operation\n");
-			exit(1);
+			app_exit(1);
 		}
 
 		gnutls_x509_crt_get_key_usage(xcrt, &key_usage, NULL);
@@ -1002,7 +1002,7 @@ pkcs11_write(FILE * outfile, const char *url, const char *label,
 		if (ret < 0) {
 			fprintf(stderr, "Error in %s:%d: %s\n", __func__,
 				__LINE__, gnutls_strerror(ret));
-			exit(1);
+			app_exit(1);
 		}
 		gnutls_x509_privkey_deinit(xkey);
 	}
@@ -1016,7 +1016,7 @@ pkcs11_write(FILE * outfile, const char *url, const char *label,
 		if (ret < 0) {
 			fprintf(stderr, "Error in %s:%d: %s\n", __func__,
 				__LINE__, gnutls_strerror(ret));
-			exit(1);
+			app_exit(1);
 		}
 		gnutls_pubkey_deinit(xpubkey);
 	}
@@ -1024,7 +1024,7 @@ pkcs11_write(FILE * outfile, const char *url, const char *label,
 	if (xkey == NULL && xcrt == NULL && secret_key == NULL && xpubkey == NULL) {
 		fprintf(stderr,
 			"You must use --load-privkey, --load-certificate, --load-pubkey or --secret-key to load the file to be copied\n");
-		exit(1);
+		app_exit(1);
 	}
 
 	UNFIX;
@@ -1053,7 +1053,7 @@ pkcs11_generate(FILE * outfile, const char *url, gnutls_pk_algorithm_t pk,
 		ret = gnutls_hex2bin(id, strlen(id), raw_id, &raw_id_size);
 		if (ret < 0) {
 			fprintf(stderr, "Error converting hex: %s\n", gnutls_strerror(ret));
-			exit(1);
+			app_exit(1);
 		}
 		cid.data = raw_id;
 		cid.size = raw_id_size;
@@ -1067,20 +1067,18 @@ pkcs11_generate(FILE * outfile, const char *url, gnutls_pk_algorithm_t pk,
 		label = read_str("warning: Label was not specified.\nLabel: ");
 	}
 
-	fprintf(stderr, "Generating an %s key...\n", gnutls_pk_get_name(pk));
-
 	ret =
 	    gnutls_pkcs11_privkey_generate3(url, pk, bits, label, &cid,
 					    GNUTLS_X509_FMT_PEM, &pubkey,
 					    info->key_usage,
-					    flags|GNUTLS_PKCS11_OBJ_FLAG_MARK_SENSITIVE);
+					    flags);
 	if (ret < 0) {
 		fprintf(stderr, "Error in %s:%d: %s\n", __func__, __LINE__,
 			gnutls_strerror(ret));
 		if (bits != 1024 && pk == GNUTLS_PK_RSA)
 			fprintf(stderr,
 				"note: several smart cards do not support arbitrary size keys; try --bits 1024 or 2048.\n");
-		exit(1);
+		app_exit(1);
 	}
 
 	fwrite(pubkey.data, 1, pubkey.size, outfile);
@@ -1111,14 +1109,14 @@ pkcs11_export_pubkey(FILE * outfile, const char *url, int detailed, unsigned int
 	if (ret < 0) {
 		fprintf(stderr, "Error in %s:%d: %s\n", __func__, __LINE__,
 			gnutls_strerror(ret));
-		exit(1);
+		app_exit(1);
 	}
 
 	ret = gnutls_pkcs11_privkey_import_url(pkey, url, 0);
 	if (ret < 0) {
 		fprintf(stderr, "Error in %s:%d: %s\n", __func__, __LINE__,
 			gnutls_strerror(ret));
-		exit(1);
+		app_exit(1);
 	}
 
 	ret =
@@ -1128,7 +1126,7 @@ pkcs11_export_pubkey(FILE * outfile, const char *url, int detailed, unsigned int
 	if (ret < 0) {
 		fprintf(stderr, "Error in %s:%d: %s\n", __func__, __LINE__,
 			gnutls_strerror(ret));
-		exit(1);
+		app_exit(1);
 	}
 	gnutls_pkcs11_privkey_deinit(pkey);
 
@@ -1151,12 +1149,12 @@ pkcs11_init(FILE * outfile, const char *url, const char *label,
 
 	if (url == NULL) {
 		fprintf(stderr, "error: no token URL given to initialize!\n");
-		exit(1);
+		app_exit(1);
 	}
 
 	if (label == NULL) {
 		fprintf(stderr, "error: no label provided for token initialization!\n");
-		exit(1);
+		app_exit(1);
 	}
 
 	if (info->so_pin != NULL)
@@ -1166,11 +1164,11 @@ pkcs11_init(FILE * outfile, const char *url, const char *label,
 		if (pin == NULL && info->batch == 0)
 			pin = getpass("Enter Security Officer's PIN: ");
 		if (pin == NULL)
-			exit(1);
+			app_exit(1);
 	}
 
 	if (strlen(pin) >= sizeof(so_pin) || pin[0] == '\n')
-		exit(1);
+		app_exit(1);
 
 	strcpy(so_pin, pin);
 
@@ -1179,7 +1177,7 @@ pkcs11_init(FILE * outfile, const char *url, const char *label,
 	if (ret < 0) {
 		fprintf(stderr, "\nError in %s:%d: %s\n", __func__, __LINE__,
 			gnutls_strerror(ret));
-		exit(1);
+		app_exit(1);
 	}
 	fprintf(stderr, "done\n");
 
@@ -1198,7 +1196,7 @@ pkcs11_set_pin(FILE * outfile, const char *url, common_info_st * info, unsigned
 
 	if (url == NULL) {
 		fprintf(stderr, "error: no token URL given to initialize!\n");
-		exit(1);
+		app_exit(1);
 	}
 
 	fprintf(stderr, "Setting token's user PIN...\n");
@@ -1211,7 +1209,7 @@ pkcs11_set_pin(FILE * outfile, const char *url, common_info_st * info, unsigned
 			if (pin == NULL && info->batch == 0)
 				pin = getpass("Enter Administrators's new PIN: ");
 			if (pin == NULL)
-				exit(1);
+				app_exit(1);
 		}
 	} else {
 		if (info->pin != NULL) {
@@ -1221,235 +1219,24 @@ pkcs11_set_pin(FILE * outfile, const char *url, common_info_st * info, unsigned
 			if (pin == NULL && info->batch == 0)
 				pin = getpass("Enter User's new PIN: ");
 			if (pin == NULL)
-				exit(1);
+				app_exit(1);
 		}
 	}
 
 	if (pin == NULL || pin[0] == '\n')
-		exit(1);
+		app_exit(1);
 
 	ret = gnutls_pkcs11_token_set_pin(url, NULL, pin, (so!=0)?GNUTLS_PIN_SO:GNUTLS_PIN_USER);
 	if (ret < 0) {
 		fprintf(stderr, "Error in %s:%d: %s\n", __func__, __LINE__,
 			gnutls_strerror(ret));
-		exit(1);
+		app_exit(1);
 	}
 
 	return;
 }
 
-const char *mech_list[] = {
-	[0] = "CKM_RSA_PKCS_KEY_PAIR_GEN",
-	[1] = "CKM_RSA_PKCS",
-	[2] = "CKM_RSA_9796",
-	[3] = "CKM_RSA_X_509",
-	[4] = "CKM_MD2_RSA_PKCS",
-	[5] = "CKM_MD5_RSA_PKCS",
-	[6] = "CKM_SHA1_RSA_PKCS",
-	[7] = "CKM_RIPEMD128_RSA_PKCS",
-	[8] = "CKM_RIPEMD160_RSA_PKCS",
-	[9] = "CKM_RSA_PKCS_OAEP",
-	[0xa] = "CKM_RSA_X9_31_KEY_PAIR_GEN",
-	[0xb] = "CKM_RSA_X9_31",
-	[0xc] = "CKM_SHA1_RSA_X9_31",
-	[0xd] = "CKM_RSA_PKCS_PSS",
-	[0xe] = "CKM_SHA1_RSA_PKCS_PSS",
-	[0x10] = "CKM_DSA_KEY_PAIR_GEN",
-	[0x11] = "CKM_DSA",
-	[0x12] = "CKM_DSA_SHA1",
-	[0x20] = "CKM_DH_PKCS_KEY_PAIR_GEN",
-	[0x21] = "CKM_DH_PKCS_DERIVE",
-	[0x30] = "CKM_X9_42_DH_KEY_PAIR_GEN",
-	[0x31] = "CKM_X9_42_DH_DERIVE",
-	[0x32] = "CKM_X9_42_DH_HYBRID_DERIVE",
-	[0x33] = "CKM_X9_42_MQV_DERIVE",
-	[0x40] = "CKM_SHA256_RSA_PKCS",
-	[0x41] = "CKM_SHA384_RSA_PKCS",
-	[0x42] = "CKM_SHA512_RSA_PKCS",
-	[0x43] = "CKM_SHA256_RSA_PKCS_PSS",
-	[0x44] = "CKM_SHA384_RSA_PKCS_PSS",
-	[0x45] = "CKM_SHA512_RSA_PKCS_PSS",
-	[0x100] = "CKM_RC2_KEY_GEN",
-	[0x101] = "CKM_RC2_ECB",
-	[0x102] = "CKM_RC2_CBC",
-	[0x103] = "CKM_RC2_MAC",
-	[0x104] = "CKM_RC2_MAC_GENERAL",
-	[0x105] = "CKM_RC2_CBC_PAD",
-	[0x110] = "CKM_RC4_KEY_GEN",
-	[0x111] = "CKM_RC4",
-	[0x120] = "CKM_DES_KEY_GEN",
-	[0x121] = "CKM_DES_ECB",
-	[0x122] = "CKM_DES_CBC",
-	[0x123] = "CKM_DES_MAC",
-	[0x124] = "CKM_DES_MAC_GENERAL",
-	[0x125] = "CKM_DES_CBC_PAD",
-	[0x130] = "CKM_DES2_KEY_GEN",
-	[0x131] = "CKM_DES3_KEY_GEN",
-	[0x132] = "CKM_DES3_ECB",
-	[0x133] = "CKM_DES3_CBC",
-	[0x134] = "CKM_DES3_MAC",
-	[0x135] = "CKM_DES3_MAC_GENERAL",
-	[0x136] = "CKM_DES3_CBC_PAD",
-	[0x140] = "CKM_CDMF_KEY_GEN",
-	[0x141] = "CKM_CDMF_ECB",
-	[0x142] = "CKM_CDMF_CBC",
-	[0x143] = "CKM_CDMF_MAC",
-	[0x144] = "CKM_CDMF_MAC_GENERAL",
-	[0x145] = "CKM_CDMF_CBC_PAD",
-	[0x200] = "CKM_MD2",
-	[0x201] = "CKM_MD2_HMAC",
-	[0x202] = "CKM_MD2_HMAC_GENERAL",
-	[0x210] = "CKM_MD5",
-	[0x211] = "CKM_MD5_HMAC",
-	[0x212] = "CKM_MD5_HMAC_GENERAL",
-	[0x220] = "CKM_SHA_1",
-	[0x221] = "CKM_SHA_1_HMAC",
-	[0x222] = "CKM_SHA_1_HMAC_GENERAL",
-	[0x230] = "CKM_RIPEMD128",
-	[0x231] = "CKM_RIPEMD128_HMAC",
-	[0x232] = "CKM_RIPEMD128_HMAC_GENERAL",
-	[0x240] = "CKM_RIPEMD160",
-	[0x241] = "CKM_RIPEMD160_HMAC",
-	[0x242] = "CKM_RIPEMD160_HMAC_GENERAL",
-	[0x250] = "CKM_SHA256",
-	[0x251] = "CKM_SHA256_HMAC",
-	[0x252] = "CKM_SHA256_HMAC_GENERAL",
-	[0x260] = "CKM_SHA384",
-	[0x261] = "CKM_SHA384_HMAC",
-	[0x262] = "CKM_SHA384_HMAC_GENERAL",
-	[0x270] = "CKM_SHA512",
-	[0x271] = "CKM_SHA512_HMAC",
-	[0x272] = "CKM_SHA512_HMAC_GENERAL",
-	[0x300] = "CKM_CAST_KEY_GEN",
-	[0x301] = "CKM_CAST_ECB",
-	[0x302] = "CKM_CAST_CBC",
-	[0x303] = "CKM_CAST_MAC",
-	[0x304] = "CKM_CAST_MAC_GENERAL",
-	[0x305] = "CKM_CAST_CBC_PAD",
-	[0x310] = "CKM_CAST3_KEY_GEN",
-	[0x311] = "CKM_CAST3_ECB",
-	[0x312] = "CKM_CAST3_CBC",
-	[0x313] = "CKM_CAST3_MAC",
-	[0x314] = "CKM_CAST3_MAC_GENERAL",
-	[0x315] = "CKM_CAST3_CBC_PAD",
-	[0x320] = "CKM_CAST128_KEY_GEN",
-	[0x321] = "CKM_CAST128_ECB",
-	[0x322] = "CKM_CAST128_CBC",
-	[0x323] = "CKM_CAST128_MAC",
-	[0x324] = "CKM_CAST128_MAC_GENERAL",
-	[0x325] = "CKM_CAST128_CBC_PAD",
-	[0x330] = "CKM_RC5_KEY_GEN",
-	[0x331] = "CKM_RC5_ECB",
-	[0x332] = "CKM_RC5_CBC",
-	[0x333] = "CKM_RC5_MAC",
-	[0x334] = "CKM_RC5_MAC_GENERAL",
-	[0x335] = "CKM_RC5_CBC_PAD",
-	[0x340] = "CKM_IDEA_KEY_GEN",
-	[0x341] = "CKM_IDEA_ECB",
-	[0x342] = "CKM_IDEA_CBC",
-	[0x343] = "CKM_IDEA_MAC",
-	[0x344] = "CKM_IDEA_MAC_GENERAL",
-	[0x345] = "CKM_IDEA_CBC_PAD",
-	[0x350] = "CKM_GENERIC_SECRET_KEY_GEN",
-	[0x360] = "CKM_CONCATENATE_BASE_AND_KEY",
-	[0x362] = "CKM_CONCATENATE_BASE_AND_DATA",
-	[0x363] = "CKM_CONCATENATE_DATA_AND_BASE",
-	[0x364] = "CKM_XOR_BASE_AND_DATA",
-	[0x365] = "CKM_EXTRACT_KEY_FROM_KEY",
-	[0x370] = "CKM_SSL3_PRE_MASTER_KEY_GEN",
-	[0x371] = "CKM_SSL3_MASTER_KEY_DERIVE",
-	[0x372] = "CKM_SSL3_KEY_AND_MAC_DERIVE",
-	[0x373] = "CKM_SSL3_MASTER_KEY_DERIVE_DH",
-	[0x374] = "CKM_TLS_PRE_MASTER_KEY_GEN",
-	[0x375] = "CKM_TLS_MASTER_KEY_DERIVE",
-	[0x376] = "CKM_TLS_KEY_AND_MAC_DERIVE",
-	[0x377] = "CKM_TLS_MASTER_KEY_DERIVE_DH",
-	[0x380] = "CKM_SSL3_MD5_MAC",
-	[0x381] = "CKM_SSL3_SHA1_MAC",
-	[0x390] = "CKM_MD5_KEY_DERIVATION",
-	[0x391] = "CKM_MD2_KEY_DERIVATION",
-	[0x392] = "CKM_SHA1_KEY_DERIVATION",
-	[0x3a0] = "CKM_PBE_MD2_DES_CBC",
-	[0x3a1] = "CKM_PBE_MD5_DES_CBC",
-	[0x3a2] = "CKM_PBE_MD5_CAST_CBC",
-	[0x3a3] = "CKM_PBE_MD5_CAST3_CBC",
-	[0x3a4] = "CKM_PBE_MD5_CAST128_CBC",
-	[0x3a5] = "CKM_PBE_SHA1_CAST128_CBC",
-	[0x3a6] = "CKM_PBE_SHA1_RC4_128",
-	[0x3a7] = "CKM_PBE_SHA1_RC4_40",
-	[0x3a8] = "CKM_PBE_SHA1_DES3_EDE_CBC",
-	[0x3a9] = "CKM_PBE_SHA1_DES2_EDE_CBC",
-	[0x3aa] = "CKM_PBE_SHA1_RC2_128_CBC",
-	[0x3ab] = "CKM_PBE_SHA1_RC2_40_CBC",
-	[0x3b0] = "CKM_PKCS5_PBKD2",
-	[0x3c0] = "CKM_PBA_SHA1_WITH_SHA1_HMAC",
-	[0x400] = "CKM_KEY_WRAP_LYNKS",
-	[0x401] = "CKM_KEY_WRAP_SET_OAEP",
-	[0x1000] = "CKM_SKIPJACK_KEY_GEN",
-	[0x1001] = "CKM_SKIPJACK_ECB64",
-	[0x1002] = "CKM_SKIPJACK_CBC64",
-	[0x1003] = "CKM_SKIPJACK_OFB64",
-	[0x1004] = "CKM_SKIPJACK_CFB64",
-	[0x1005] = "CKM_SKIPJACK_CFB32",
-	[0x1006] = "CKM_SKIPJACK_CFB16",
-	[0x1007] = "CKM_SKIPJACK_CFB8",
-	[0x1008] = "CKM_SKIPJACK_WRAP",
-	[0x1009] = "CKM_SKIPJACK_PRIVATE_WRAP",
-	[0x100a] = "CKM_SKIPJACK_RELAYX",
-	[0x1010] = "CKM_KEA_KEY_PAIR_GEN",
-	[0x1011] = "CKM_KEA_KEY_DERIVE",
-	[0x1020] = "CKM_FORTEZZA_TIMESTAMP",
-	[0x1030] = "CKM_BATON_KEY_GEN",
-	[0x1031] = "CKM_BATON_ECB128",
-	[0x1032] = "CKM_BATON_ECB96",
-	[0x1033] = "CKM_BATON_CBC128",
-	[0x1034] = "CKM_BATON_COUNTER",
-	[0x1035] = "CKM_BATON_SHUFFLE",
-	[0x1036] = "CKM_BATON_WRAP",
-	[0x1040] = "CKM_ECDSA_KEY_PAIR_GEN",
-	[0x1041] = "CKM_ECDSA",
-	[0x1042] = "CKM_ECDSA_SHA1",
-	[0x1050] = "CKM_ECDH1_DERIVE",
-	[0x1051] = "CKM_ECDH1_COFACTOR_DERIVE",
-	[0x1052] = "CKM_ECMQV_DERIVE",
-	[0x1060] = "CKM_JUNIPER_KEY_GEN",
-	[0x1061] = "CKM_JUNIPER_ECB128",
-	[0x1062] = "CKM_JUNIPER_CBC128",
-	[0x1063] = "CKM_JUNIPER_COUNTER",
-	[0x1064] = "CKM_JUNIPER_SHUFFLE",
-	[0x1065] = "CKM_JUNIPER_WRAP",
-	[0x1070] = "CKM_FASTHASH",
-	[0x1080] = "CKM_AES_KEY_GEN",
-	[0x1081] = "CKM_AES_ECB",
-	[0x1082] = "CKM_AES_CBC",
-	[0x1083] = "CKM_AES_MAC",
-	[0x1084] = "CKM_AES_MAC_GENERAL",
-	[0x1085] = "CKM_AES_CBC_PAD",
-	[0x2000] = "CKM_DSA_PARAMETER_GEN",
-	[0x2001] = "CKM_DH_PKCS_PARAMETER_GEN",
-	[0x2002] = "CKM_X9_42_DH_PARAMETER_GEN",
-	[0x1200] = "CKM_GOSTR3410_KEY_PAIR_GEN",
-	[0x1201] = "CKM_GOSTR3410",
-	[0x1202] = "CKM_GOSTR3410_WITH_GOSTR3411",
-	[0x1203] = "CKM_GOSTR3410_KEY_WRAP",
-	[0x1204] = "CKM_GOSTR3410_DERIVE",
-	[0x1210] = "CKM_GOSTR3411",
-	[0x1211] = "CKM_GOSTR3411_HMAC",
-	[0x255] = "CKM_SHA224",
-	[0x256] = "CKM_SHA224_HMAC",
-	[0x257] = "CKM_SHA224_HMAC_GENERAL",
-	[0x46] = "CKM_SHA224_RSA_PKCS",
-	[0x47] = "CKM_SHA224_RSA_PKCS_PSS",
-	[0x396] = "CKM_SHA224_KEY_DERIVATION",
-	[0x550] = "CKM_CAMELLIA_KEY_GEN",
-	[0x551] = "CKM_CAMELLIA_ECB",
-	[0x552] = "CKM_CAMELLIA_CBC",
-	[0x553] = "CKM_CAMELLIA_MAC",
-	[0x554] = "CKM_CAMELLIA_MAC_GENERAL",
-	[0x555] = "CKM_CAMELLIA_CBC_PAD",
-	[0x556] = "CKM_CAMELLIA_ECB_ENCRYPT_DATA",
-	[0x557] = "CKM_CAMELLIA_CBC_ENCRYPT_DATA"
-};
+#include "mech-list.h"
 
 void
 pkcs11_mechanism_list(FILE * outfile, const char *url, unsigned int flags,
@@ -1500,14 +1287,14 @@ pkcs11_get_random(FILE * outfile, const char *url, unsigned bytes,
 	output = malloc(bytes);
 	if (output == NULL) {
 		fprintf(stderr, "Memory error\n");
-		exit(1);
+		app_exit(1);
 	}
 
 	ret = gnutls_pkcs11_token_get_random(url, output, bytes);
 	if (ret < 0) {
 		fprintf(stderr, "gnutls_pkcs11_token_get_random: %s\n",
 			gnutls_strerror(ret));
-		exit(1);
+		app_exit(1);
 	}
 
 	fwrite(output, 1, bytes, outfile);
@@ -1533,14 +1320,14 @@ void pkcs11_set_val(FILE * outfile, const char *url, int detailed,
 	if (ret < 0) {
 		fprintf(stderr, "Error in %s:%d: %s\n", __func__, __LINE__,
 			gnutls_strerror(ret));
-		exit(1);
+		app_exit(1);
 	}
 
 	ret = gnutls_pkcs11_obj_import_url(obj, url, flags);
 	if (ret < 0) {
 		fprintf(stderr, "Error in %s:%d: %s\n", __func__, __LINE__,
 			gnutls_strerror(ret));
-		exit(1);
+		app_exit(1);
 	}
 
 	ret =
@@ -1548,7 +1335,7 @@ void pkcs11_set_val(FILE * outfile, const char *url, int detailed,
 	if (ret < 0) {
 		fprintf(stderr, "Error in %s:%d: %s\n", __func__, __LINE__,
 			gnutls_strerror(ret));
-		exit(1);
+		app_exit(1);
 	}
 	gnutls_pkcs11_obj_deinit(obj);
 
