@@ -1,6 +1,7 @@
 /*
  * Copyright (C) 2004-2012 Free Software Foundation, Inc.
  * Copyright (C) 2001,2002 Paul Sheer
+ * Copyright (C) 2016-2017 Red Hat, Inc.
  * Portions Copyright (C) 2002,2003 Nikos Mavrogiannopoulos
  *
  * This file is part of GnuTLS.
@@ -34,7 +35,6 @@
 #include <string.h>
 #include <gnutls/gnutls.h>
 #include <gnutls/dtls.h>
-#include <gnutls/openpgp.h>
 #include <sys/time.h>
 #include <sys/select.h>
 #include <fcntl.h>
@@ -67,15 +67,10 @@ int disable_client_cert;
 const char *psk_passwd = NULL;
 const char *srp_passwd = NULL;
 const char *srp_passwd_conf = NULL;
-const char *pgp_keyring = NULL;
-const char *pgp_keyfile = NULL;
-const char *pgp_certfile = NULL;
-const char *x509_keyfile = NULL;
-const char *x509_certfile = NULL;
-const char *x509_dsakeyfile = NULL;
-const char *x509_dsacertfile = NULL;
-const char *x509_ecckeyfile = NULL;
-const char *x509_ecccertfile = NULL;
+const char **x509_keyfile = NULL;
+const char **x509_certfile = NULL;
+unsigned x509_certfile_size = 0;
+unsigned x509_keyfile_size = 0;
 const char *x509_cafile = NULL;
 const char *dh_params_file = NULL;
 const char *x509_crlfile = NULL;
@@ -84,6 +79,9 @@ const char *status_response_ocsp = NULL;
 const char *sni_hostname = NULL;
 int sni_hostname_fatal = 0;
 
+const char **alpn_protos = NULL;
+unsigned alpn_protos_size = 0;
+
 gnutls_datum_t session_ticket_key;
 static void tcp_server(const char *name, int port);
 
@@ -272,7 +270,7 @@ int ret;
 		if (!require_cert && gnutls_certificate_get_peers(session, &size) == NULL)
 			return 0;
 
-		if (require_cert || ENABLED_OPT(VERIFY_CLIENT_CERT)) {
+		if (ENABLED_OPT(VERIFY_CLIENT_CERT)) {
 			if (cert_verify(session, NULL, NULL) == 0) {
 				do {
 					ret = gnutls_alert_send(session, GNUTLS_AL_FATAL, GNUTLS_A_ACCESS_DENIED);
@@ -365,11 +363,15 @@ end:
 	return ret;
 }
 
+#define MAX_ALPN_PROTOCOLS 16
 gnutls_session_t initialize_session(int dtls)
 {
 	gnutls_session_t session;
 	int ret;
+	unsigned i;
 	const char *err;
+	gnutls_datum_t alpn[MAX_ALPN_PROTOCOLS];
+	unsigned alpn_size;
 
 	if (priorities == NULL)
 		priorities = "NORMAL";
@@ -408,6 +410,18 @@ gnutls_session_t initialize_session(int dtls)
 		exit(1);
 	}
 
+	alpn_size = MIN(MAX_ALPN_PROTOCOLS,alpn_protos_size);
+	for (i=0;i<alpn_size;i++) {
+		alpn[i].data = (void*)alpn_protos[i];
+		alpn[i].size = strlen(alpn_protos[i]);
+	}
+
+	ret = gnutls_alpn_set_protocols(session, alpn, alpn_size, HAVE_OPT(ALPN_FATAL)?GNUTLS_ALPN_MANDATORY:0);
+	if (ret < 0) {
+		fprintf(stderr, "Error setting ALPN protocols: %s\n", gnutls_strerror(ret));
+		exit(1);
+	}
+
 	gnutls_credentials_set(session, GNUTLS_CRD_ANON, dh_cred);
 
 	if (srp_cred != NULL)
@@ -509,7 +523,7 @@ static char *peer_print_info(gnutls_session_t session, int *ret_length,
 		    gnutls_certificate_get_peers(session, &cert_list_size);
 
 		for (i = 0; i < cert_list_size; i++) {
-			gnutls_x509_crt_t cert;
+			gnutls_x509_crt_t cert = NULL;
 			gnutls_datum_t info;
 
 			if (gnutls_x509_crt_init(&cert) == 0 &&
@@ -539,6 +553,7 @@ static char *peer_print_info(gnutls_session_t session, int *ret_length,
 				crtinfo[ncrtinfo] = '\0';
 				gnutls_free(info.data);
 			}
+			gnutls_x509_crt_deinit(cert);
 		}
 	}
 
@@ -821,7 +836,7 @@ int listen_socket(const char *name, int listen_port, int socktype)
 			yes = 1;
 			/* avoid listen on ipv6 addresses failing
 			 * because already listening on ipv4 addresses: */
-			setsockopt(s, IPPROTO_IPV6, IPV6_V6ONLY,
+			(void)setsockopt(s, IPPROTO_IPV6, IPV6_V6ONLY,
 				   (const void *) &yes, sizeof(yes));
 		}
 #endif
@@ -890,7 +905,7 @@ static void strip(char *data)
 
 	for (i = 0; i < len; i++) {
 		if (data[i] == '\r' && data[i + 1] == '\n'
-		    && data[i + 1] == 0) {
+		    && data[i + 2] == 0) {
 			data[i] = '\n';
 			data[i + 1] = 0;
 			break;
@@ -1074,81 +1089,24 @@ int main(int argc, char **argv)
 			printf("Processed %d CRL(s).\n", ret);
 		}
 	}
-#ifdef ENABLE_OPENPGP
-	if (pgp_keyring != NULL) {
-		ret =
-		    gnutls_certificate_set_openpgp_keyring_file(cert_cred,
-								pgp_keyring,
-								GNUTLS_OPENPGP_FMT_BASE64);
-		if (ret < 0) {
-			fprintf(stderr,
-				"Error setting the OpenPGP keyring file\n");
-			GERR(ret);
+
+	if (x509_certfile_size > 0 && x509_keyfile_size > 0) {
+		unsigned i;
+
+		for (i = 0; i < x509_certfile_size; i++) {
+			ret = gnutls_certificate_set_x509_key_file
+			    (cert_cred, x509_certfile[i], x509_keyfile[i], x509ctype);
+			if (ret < 0) {
+				fprintf(stderr,
+						"Error reading '%s' or '%s'\n",
+						x509_certfile[i], x509_keyfile[i]);
+				GERR(ret);
+				exit(1);
+			} else
+				cert_set = 1;
 		}
 	}
 
-	if (pgp_certfile != NULL && pgp_keyfile != NULL) {
-		if (HAVE_OPT(PGPSUBKEY))
-			ret = gnutls_certificate_set_openpgp_key_file2
-			    (cert_cred, pgp_certfile, pgp_keyfile,
-			     OPT_ARG(PGPSUBKEY),
-			     GNUTLS_OPENPGP_FMT_BASE64);
-		else
-			ret = gnutls_certificate_set_openpgp_key_file
-			    (cert_cred, pgp_certfile, pgp_keyfile,
-			     GNUTLS_OPENPGP_FMT_BASE64);
-
-		if (ret < 0) {
-			fprintf(stderr,
-				"Error[%d] while reading the OpenPGP key pair ('%s', '%s')\n",
-				ret, pgp_certfile, pgp_keyfile);
-			GERR(ret);
-		} else
-			cert_set = 1;
-	}
-#endif
-
-	if (x509_certfile != NULL && x509_keyfile != NULL) {
-		ret = gnutls_certificate_set_x509_key_file
-		    (cert_cred, x509_certfile, x509_keyfile, x509ctype);
-		if (ret < 0) {
-			fprintf(stderr,
-				"Error reading '%s' or '%s'\n",
-				x509_certfile, x509_keyfile);
-			GERR(ret);
-			exit(1);
-		} else
-			cert_set = 1;
-	}
-
-	if (x509_dsacertfile != NULL && x509_dsakeyfile != NULL) {
-		ret = gnutls_certificate_set_x509_key_file
-		    (cert_cred, x509_dsacertfile, x509_dsakeyfile,
-		     x509ctype);
-		if (ret < 0) {
-			fprintf(stderr,
-				"Error reading '%s' or '%s'\n",
-				x509_dsacertfile, x509_dsakeyfile);
-			GERR(ret);
-			exit(1);
-		} else
-			cert_set = 1;
-	}
-
-	if (x509_ecccertfile != NULL && x509_ecckeyfile != NULL) {
-		ret = gnutls_certificate_set_x509_key_file
-		    (cert_cred, x509_ecccertfile, x509_ecckeyfile,
-		     x509ctype);
-		if (ret < 0) {
-			fprintf(stderr,
-				"Error reading '%s' or '%s'\n",
-				x509_ecccertfile, x509_ecckeyfile);
-			GERR(ret);
-			exit(1);
-		} else
-			cert_set = 1;
-	}
-
 	if (cert_set == 0) {
 		fprintf(stderr,
 			"Warning: no private key and certificate pairs were set.\n");
@@ -1285,8 +1243,7 @@ static void retry_handshake(listener_item *j)
 	} else if (r < 0) {
 		j->http_state = HTTP_STATE_CLOSING;
 		check_alert(j->tls_session, r);
-		fprintf(stderr, "Error in handshake\n");
-		GERR(r);
+		fprintf(stderr, "Error in handshake: %s\n", gnutls_strerror(r));
 
 		do {
 			ret = gnutls_alert_send_appropriate(j->tls_session, r);
@@ -1325,7 +1282,7 @@ int r, ret;
 		do {
 			ret = gnutls_alert_send_appropriate(j->tls_session, r);
 		} while (ret == GNUTLS_E_AGAIN || ret == GNUTLS_E_INTERRUPTED);
-		GERR(r);
+		fprintf(stderr, "Error in rehandshake: %s\n", gnutls_strerror(r));
 		j->http_state = HTTP_STATE_CLOSING;
 	} else {
 		j->http_state = HTTP_STATE_REQUEST;
@@ -1458,7 +1415,7 @@ static void tcp_server(const char *name, int port)
 
 			if (FD_ISSET(j->fd, &rd) && !j->listen_socket) {
 /* read partial GET request */
-				char buf[1024];
+				char buf[16*1024];
 				int r;
 
 				if (j->handshake_ok == 0) {
@@ -1469,7 +1426,7 @@ static void tcp_server(const char *name, int port)
 					r = gnutls_record_recv(j->
 							       tls_session,
 							       buf,
-							       MIN(1024,
+							       MIN(sizeof(buf),
 								   SMALL_READ_TEST));
 					if (r == GNUTLS_E_INTERRUPTED || r == GNUTLS_E_AGAIN) {
 						/* do nothing */
@@ -1677,35 +1634,31 @@ static void cmd_parser(int argc, char **argv)
 	if (HAVE_OPT(DHPARAMS))
 		dh_params_file = OPT_ARG(DHPARAMS);
 
-	if (HAVE_OPT(X509KEYFILE))
-		x509_keyfile = OPT_ARG(X509KEYFILE);
-	if (HAVE_OPT(X509CERTFILE))
-		x509_certfile = OPT_ARG(X509CERTFILE);
+	if (HAVE_OPT(ALPN)) {
+		alpn_protos = STACKLST_OPT(ALPN);
+		alpn_protos_size = STACKCT_OPT(ALPN);
+	}
 
-	if (HAVE_OPT(X509DSAKEYFILE))
-		x509_dsakeyfile = OPT_ARG(X509DSAKEYFILE);
-	if (HAVE_OPT(X509DSACERTFILE))
-		x509_dsacertfile = OPT_ARG(X509DSACERTFILE);
+	if (HAVE_OPT(X509KEYFILE)) {
+		x509_keyfile = STACKLST_OPT(X509KEYFILE);
+		x509_keyfile_size = STACKCT_OPT(X509KEYFILE);
+	}
 
+	if (HAVE_OPT(X509CERTFILE)) {
+		x509_certfile = STACKLST_OPT(X509CERTFILE);
+		x509_certfile_size = STACKCT_OPT(X509CERTFILE);
+	}
 
-	if (HAVE_OPT(X509ECCKEYFILE))
-		x509_ecckeyfile = OPT_ARG(X509ECCKEYFILE);
-	if (HAVE_OPT(X509ECCCERTFILE))
-		x509_ecccertfile = OPT_ARG(X509ECCCERTFILE);
+	if (x509_certfile_size != x509_keyfile_size) {
+		fprintf(stderr, "The certificate number provided (%u) doesn't match the keys (%u)\n",
+			x509_certfile_size, x509_keyfile_size);
+	}
 
 	if (HAVE_OPT(X509CAFILE))
 		x509_cafile = OPT_ARG(X509CAFILE);
 	if (HAVE_OPT(X509CRLFILE))
 		x509_crlfile = OPT_ARG(X509CRLFILE);
 
-	if (HAVE_OPT(PGPKEYFILE))
-		pgp_keyfile = OPT_ARG(PGPKEYFILE);
-	if (HAVE_OPT(PGPCERTFILE))
-		pgp_certfile = OPT_ARG(PGPCERTFILE);
-
-	if (HAVE_OPT(PGPKEYRING))
-		pgp_keyring = OPT_ARG(PGPKEYRING);
-
 	if (HAVE_OPT(SRPPASSWD))
 		srp_passwd = OPT_ARG(SRPPASSWD);
 	if (HAVE_OPT(SRPPASSWDCONF))
