@@ -43,13 +43,6 @@
 #include <c-ctype.h>
 #include "sockets.h"
 
-#ifdef HAVE_LIBIDN2
-#include <idn2.h>
-#elif defined HAVE_LIBIDN
-#include <idna.h>
-#include <idn-free.h>
-#endif
-
 #define MAX_BUF 4096
 
 /* Functions to manipulate sockets
@@ -236,10 +229,10 @@ socket_starttls(socket_st * socket)
 			printf("Negotiating SMTP STARTTLS\n");
 
 		wait_for_text(socket, "220 ", 4);
-		snprintf(buf, sizeof(buf), "EHLO %s\n", socket->hostname);
+		snprintf(buf, sizeof(buf), "EHLO %s\r\n", socket->hostname);
 		send_line(socket, buf);
 		wait_for_text(socket, "250 ", 4);
-		send_line(socket, "STARTTLS\n");
+		send_line(socket, "STARTTLS\r\n");
 		wait_for_text(socket, "220 ", 4);
 	} else if (strcasecmp(socket->app_proto, "imap") == 0 || strcasecmp(socket->app_proto, "imap2") == 0) {
 		if (socket->verbose)
@@ -272,6 +265,44 @@ socket_starttls(socket_st * socket)
 		wait_for_text(socket, "211 ", 4);
 		send_line(socket, "AUTH TLS\r\n");
 		wait_for_text(socket, "234", 3);
+	} else if (strcasecmp(socket->app_proto, "lmtp") == 0) {
+		if (socket->verbose)
+			printf("Negotiating LMTP STARTTLS\n");
+
+		wait_for_text(socket, "220 ", 4);
+		snprintf(buf, sizeof(buf), "LHLO %s\r\n", socket->hostname);
+		send_line(socket, buf);
+		wait_for_text(socket, "250 ", 4);
+		send_line(socket, "STARTTLS\r\n");
+		wait_for_text(socket, "220 ", 4);
+	} else if (strcasecmp(socket->app_proto, "pop3") == 0) {
+		if (socket->verbose)
+			printf("Negotiating POP3 STARTTLS\n");
+
+		wait_for_text(socket, "+OK", 3);
+		send_line(socket, "STLS\r\n");
+		wait_for_text(socket, "+OK", 3);
+	} else if (strcasecmp(socket->app_proto, "nntp") == 0) {
+		if (socket->verbose)
+			printf("Negotiating NNTP STARTTLS\n");
+
+		wait_for_text(socket, "200 ", 4);
+		send_line(socket, "STARTTLS\r\n");
+		wait_for_text(socket, "382 ", 4);
+	} else if (strcasecmp(socket->app_proto, "sieve") == 0) {
+		if (socket->verbose)
+			printf("Negotiating Sieve STARTTLS\n");
+
+		wait_for_text(socket, "OK ", 3);
+		send_line(socket, "STARTTLS\r\n");
+		wait_for_text(socket, "OK ", 3);
+	} else if (strcasecmp(socket->app_proto, "postgres") == 0 || strcasecmp(socket->app_proto, "postgresql") == 0) {
+		if (socket->verbose)
+			printf("Negotiating PostgreSQL STARTTLS\n");
+
+#define POSTGRES_STR "\x00\x00\x00\x08\x04\xD2\x16\x2F"
+		send(socket->fd, POSTGRES_STR, sizeof(POSTGRES_STR)-1, 0);
+		wait_for_text(socket, NULL, 0);
 	} else {
 		if (!c_isdigit(socket->app_proto[0])) {
 			static int warned = 0;
@@ -298,8 +329,9 @@ starttls_proto_to_port(const char *app_proto)
 
 	s = getservbyname(app_proto, NULL);
 	if (s != NULL) {
-		return s->s_port;
+		return ntohs(s->s_port);
 	}
+
 	endservent();
 
 	return 443;
@@ -355,6 +387,11 @@ void socket_bye(socket_st * socket, unsigned polite)
 	gnutls_free(socket->rdata.data);
 	socket->rdata.data = NULL;
 
+	if (socket->server_trace)
+		fclose(socket->server_trace);
+	if (socket->client_trace)
+		fclose(socket->client_trace);
+
 	socket->fd = -1;
 	socket->secure = 0;
 }
@@ -377,10 +414,45 @@ void canonicalize_host(char *hostname, char *service, unsigned service_size)
 	snprintf(service, service_size, "%s", p+1);
 }
 
+static ssize_t
+wrap_pull(gnutls_transport_ptr_t ptr, void *data, size_t len)
+{
+	socket_st *hd = ptr;
+	ssize_t r;
+
+	r = recv(hd->fd, data, len, 0);
+	if (r > 0 && hd->server_trace) {
+		fwrite(data, 1, len, hd->server_trace);
+	}
+	return r;
+}
+
+static ssize_t
+wrap_push(gnutls_transport_ptr_t ptr, const void *data, size_t len)
+{
+	socket_st *hd = ptr;
+
+	if (hd->client_trace) {
+		fwrite(data, 1, len, hd->client_trace);
+	}
+
+	return send(hd->fd, data, len, 0);
+}
+
+/* inline is used to avoid a gcc warning if used in mini-eagain */
+inline static int wrap_pull_timeout_func(gnutls_transport_ptr_t ptr,
+					   unsigned int ms)
+{
+	socket_st *hd = ptr;
+
+	return gnutls_system_recv_timeout((gnutls_transport_ptr_t)(long)hd->fd, ms);
+}
+
 
 void
-socket_open(socket_st * hd, const char *hostname, const char *service,
-	    const char *app_proto, int flags, const char *msg, gnutls_datum_t *rdata)
+socket_open2(socket_st * hd, const char *hostname, const char *service,
+	    const char *app_proto, int flags, const char *msg, gnutls_datum_t *rdata,
+	    FILE *server_trace, FILE *client_trace)
 {
 	struct addrinfo hints, *res, *ptr;
 	int sd, err = 0;
@@ -491,7 +563,16 @@ socket_open(socket_st * hd, const char *hostname, const char *service,
 				gnutls_session_set_data(hd->session, hd->rdata.data, hd->rdata.size);
 			}
 
-			gnutls_transport_set_int(hd->session, sd);
+			if (server_trace)
+				hd->server_trace = server_trace;
+
+			if (client_trace)
+				hd->client_trace = client_trace;
+
+			gnutls_transport_set_push_function(hd->session, wrap_push);
+			gnutls_transport_set_pull_function(hd->session, wrap_pull);
+			gnutls_transport_set_pull_timeout_function(hd->session, wrap_pull_timeout_func);
+			gnutls_transport_set_ptr(hd->session, hd);
 		}
 
 		if (!(flags & SOCKET_FLAG_RAW) && !(flags & SOCKET_FLAG_SKIP_INIT)) {
