@@ -114,11 +114,13 @@ static int generate_create_conf(const char *tpasswd_conf)
 
 		if (gnutls_srp_base64_encode_alloc(&n, &str_n) < 0) {
 			fprintf(stderr, "Could not encode\n");
+			fclose(fd);
 			return -1;
 		}
 
 		if (gnutls_srp_base64_encode_alloc(&g, &str_g) < 0) {
 			fprintf(stderr, "Could not encode\n");
+			fclose(fd);
 			return -1;
 		}
 
@@ -572,7 +574,7 @@ crypt_int(const char *username, const char *passwd, int salt_size,
 		if (fd == NULL) {
 			fprintf(stderr, "Cannot open '%s' for write\n",
 				tpasswd);
-			remove(tmpname);
+			(void)remove(tmpname);
 			return -1;
 		}
 
@@ -580,7 +582,7 @@ crypt_int(const char *username, const char *passwd, int salt_size,
 		if (fd2 == NULL) {
 			fprintf(stderr, "Cannot open '%s' for read\n",
 				tmpname);
-			remove(tmpname);
+			(void)remove(tmpname);
 			return -1;
 		}
 
@@ -613,7 +615,7 @@ crypt_int(const char *username, const char *passwd, int salt_size,
 		fclose(fd);
 		fclose(fd2);
 
-		remove(tmpname);
+		(void)remove(tmpname);
 
 	}
 
