@@ -202,6 +202,28 @@ test_code_t test_dhe(gnutls_session_t session)
 	return TEST_IGNORE;
 }
 
+test_code_t test_rfc7919(gnutls_session_t session)
+{
+#ifdef ENABLE_DHE
+	int ret;
+
+	sprintf(prio_str, INIT_STR
+		ALL_CIPHERS ":" ALL_COMP ":" ALL_CERTTYPES ":%s:" ALL_MACS
+		":+DHE-RSA:+DHE-DSS:+GROUP-ALL:%s", protocol_str, rest);
+	_gnutls_priority_set_direct(session, prio_str);
+
+	gnutls_credentials_set(session, GNUTLS_CRD_CERTIFICATE, xcred);
+
+	ret = do_handshake(session);
+
+	if (ret != TEST_FAILED && (gnutls_session_get_flags(session) & GNUTLS_SFLAGS_RFC7919))
+		return TEST_SUCCEED;
+	else
+		return TEST_FAILED;
+#endif
+	return TEST_IGNORE;
+}
+
 test_code_t test_ecdhe(gnutls_session_t session)
 {
 	int ret;
@@ -296,7 +318,7 @@ test_code_t test_rfc7507(gnutls_session_t session)
 
 	ret = do_handshake(session);
 	if (ret < 0)
-		return TEST_IGNORE;
+		return TEST_IGNORE2;
 
 	if (handshake_output < 0)
 		return TEST_SUCCEED;
@@ -433,7 +455,7 @@ test_code_t test_dhe_group(gnutls_session_t session)
 	const char *print;
 	FILE *fp;
 
-	remove("debug-dh.out");
+	(void)remove("debug-dh.out");
 
 	if (verbose == 0 || pubkey.data == NULL)
 		return TEST_IGNORE;
@@ -654,11 +676,15 @@ test_code_t test_camellia_cbc(gnutls_session_t session)
 {
 	int ret;
 
+	if (gnutls_fips140_mode_enabled())
+		return TEST_IGNORE;
+
 	sprintf(prio_str,
 		INIT_STR "+CAMELLIA-128-CBC:" ALL_COMP ":" ALL_CERTTYPES
 		":%s:" ALL_MACS ":" ALL_KX ":%s", protocol_str, rest);
 	_gnutls_priority_set_direct(session, prio_str);
 
+
 	gnutls_credentials_set(session, GNUTLS_CRD_CERTIFICATE, xcred);
 
 	ret = do_handshake(session);
@@ -669,6 +695,9 @@ test_code_t test_camellia_gcm(gnutls_session_t session)
 {
 	int ret;
 
+	if (gnutls_fips140_mode_enabled())
+		return TEST_IGNORE;
+
 	sprintf(prio_str,
 		INIT_STR "+CAMELLIA-128-GCM:" ALL_COMP ":" ALL_CERTTYPES
 		":%s:" ALL_MACS ":" ALL_KX ":%s", protocol_str, rest);
@@ -724,6 +753,9 @@ test_code_t test_md5(gnutls_session_t session)
 {
 	int ret;
 
+	if (gnutls_fips140_mode_enabled())
+		return TEST_IGNORE;
+
 	sprintf(prio_str,
 		INIT_STR "+AES-128-CBC:" ALL_CIPHERS ":" ALL_COMP ":"
 		ALL_CERTTYPES ":%s:+MD5:" ALL_KX ":%s", protocol_str,
@@ -801,6 +833,9 @@ test_code_t test_arcfour(gnutls_session_t session)
 {
 	int ret;
 
+	if (gnutls_fips140_mode_enabled())
+		return TEST_IGNORE;
+
 	sprintf(prio_str,
 		INIT_STR "+ARCFOUR-128:" ALL_COMP ":" ALL_CERTTYPES ":%s:"
 		ALL_MACS ":" ALL_KX ":%s", protocol_str, rest);
@@ -815,6 +850,9 @@ test_code_t test_chacha20(gnutls_session_t session)
 {
 	int ret;
 
+	if (gnutls_fips140_mode_enabled())
+		return TEST_IGNORE;
+
 	sprintf(prio_str,
 		INIT_STR "+CHACHA20-POLY1305:" ALL_COMP ":" ALL_CERTTYPES ":%s:"
 		ALL_MACS ":" ALL_KX ":%s", protocol_str, rest);
@@ -1304,7 +1342,7 @@ test_code_t test_certificate(gnutls_session_t session)
 	int ret;
 	FILE *fp;
 
-	remove("debug-certs.out");
+	(void)remove("debug-certs.out");
 
 	if (verbose == 0)
 		return TEST_IGNORE;
@@ -1354,12 +1392,12 @@ test_code_t test_chain_order(gnutls_session_t session)
 		return ret;
 
 	if (gnutls_certificate_type_get(session) != GNUTLS_CRT_X509)
-		return TEST_IGNORE;
+		return TEST_IGNORE2;
 
 	cert_list = gnutls_certificate_get_peers(session, &cert_list_size);
 	if (cert_list_size == 0) {
 		ext_text = "No certificates found!";
-		return TEST_IGNORE;
+		return TEST_IGNORE2;
 	}
 
 	if (cert_list_size == 1)
@@ -1455,7 +1493,7 @@ test_code_t test_server_cas(gnutls_session_t session)
 {
 	int ret;
 
-	remove("debug-cas.out");
+	(void)remove("debug-cas.out");
 	if (verbose == 0)
 		return TEST_IGNORE;
 
