@@ -31,9 +31,22 @@ echo "$CFLAGS"|grep sanitize && exit 77
 
 ${PKGCONFIG} --version >/dev/null || exit 77
 
+${PKGCONFIG} --libs nettle
+if test $? != 0;then
+	echo "Nettle was not found in pkg-config"
+	exit 77
+fi
+
 PKG_CONFIG_PATH=${top_builddir}/lib
 export PKG_CONFIG_PATH
 
+OTHER=$(${PKGCONFIG} --libs --static libidn)
+OTHER="${OTHER} $(${PKGCONFIG} --libs --static p11-kit-1)"
+if test -n "${OTHER}" && test "${OTHER#*-R}" != "$OTHER";then
+	echo "Found invalid string in libidn flags: ${OTHER}"
+	exit 77
+fi
+
 set -e
 
 cat >$TMPFILE <<__EOF__
